/*
 *	Intel CPU Microcode Update driver for Linux
 *
 *	Copyright (C) 2000 Tigran Aivazian
 *
 *	This driver allows to upgrade microcode on Intel processors
 *	belonging to IA-32 family - PentiumPro, Pentium II, 
 *	Pentium III, Xeon, Pentium 4, etc.
 *
 *	Reference: Section 8.10 of Volume III, Intel Pentium 4 Manual, 
 *	Order Number 245472 or free download from:
 *		
 *	http://developer.intel.com/design/pentium4/manuals/245472.htm
 *
 *	For more information, go to http://www.urbanmyth.org/microcode
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	1.0	16 Feb 2000, Tigran Aivazian <tigran@sco.com>
 *		Initial release.
 *	1.01	18 Feb 2000, Tigran Aivazian <tigran@sco.com>
 *		Added read() support + cleanups.
 *	1.02	21 Feb 2000, Tigran Aivazian <tigran@sco.com>
 *		Added 'device trimming' support. open(O_WRONLY) zeroes
 *		and frees the saved copy of applied microcode.
 *	1.03	29 Feb 2000, Tigran Aivazian <tigran@sco.com>
 *		Made to use devfs (/dev/cpu/microcode) + cleanups.
 *	1.04	06 Jun 2000, Simon Trimmer <simon@veritas.com>
 *		Added misc device support (now uses both devfs and misc).
 *		Added MICROCODE_IOCFREE ioctl to clear memory.
 *	1.05	09 Jun 2000, Simon Trimmer <simon@veritas.com>
 *		Messages for error cases (non intel & no suitable microcode).
 *	1.06	03 Aug 2000, Tigran Aivazian <tigran@veritas.com>
 *		Removed ->release(). Removed exclusive open and status bitmap.
 *		Added microcode_rwsem to serialize read()/write()/ioctl().
 *		Removed global kernel lock usage.
 *	1.07	07 Sep 2000, Tigran Aivazian <tigran@veritas.com>
 *		Write 0 to 0x8B msr and then cpuid before reading revision,
 *		so that it works even if there were no update done by the
 *		BIOS. Otherwise, reading from 0x8B gives junk (which happened
 *		to be 0 on my machine which is why it worked even when I
 *		disabled update by the BIOS)
 *		Thanks to Eric W. Biederman <ebiederman@lnxi.com> for the fix.
 *	1.08	11 Dec 2000, Richard Schaal <richard.schaal@intel.com> and
 *			     Tigran Aivazian <tigran@veritas.com>
 *		Intel Pentium 4 processor support and bugfixes.
 *	1.09	30 Oct 2001, Tigran Aivazian <tigran@veritas.com>
 *		Bugfix for HT (Hyper-Threading) enabled processors
 *		whereby processor resources are shared by all logical processors
 *		in a single CPU package.
 *	1.10	28 Feb 2002 Asit K Mallick <asit.k.mallick@intel.com> and
 *		Tigran Aivazian <tigran@veritas.com>,
 *		Serialize updates as required on HT processors due to speculative
 *		nature of implementation.
 *	1.11	22 Mar 2002 Tigran Aivazian <tigran@veritas.com>
 *		Fix the panic when writing zero-length microcode chunk.
 *	1.12	19 May 2003 Nitin Kamble <nitin.a.kamble@intel.com>, Jun Nakajima<jun.nakajima@intel.com>
 *		Support for microcode updates in the new fromat.
 */

#include <linux/init.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/miscdevice.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/spinlock.h>

#include <asm/msr.h>
#include <asm/uaccess.h>
#include <asm/processor.h>


#define get_totalsize(mc) \
	(mc->totalsize ? mc->totalsize : 2048)

#define get_datasize(mc) \
	(mc->datasize ? mc->datasize : 2000)

#define get_next_microcode(mc) \
	(microcode_t *)((_get_next_microcode(mc) >= \
	(char *)microcode + microcode_size) ? \
	NULL : _get_next_microcode(mc))

#define _get_next_microcode(mc) \
	 ((char *) mc + get_totalsize(mc)) 

#define get_extended_sigtable(mc) \
	((get_totalsize(mc) - get_datasize(mc) == 48) ? \
	 NULL : (struct extended_sigtable *)((char *) & mc->bits +\
		 get_datasize(mc))) 

#define sigmatch(a, b, c, d) ((a == b) && (c & d))

#define exttable_size(et) (et->count * 12 + 20)

#define MC_HEADER_SIZE	48	/* bytes */
#define DWSIZE	(sizeof(u32))

static spinlock_t microcode_update_lock = SPIN_LOCK_UNLOCKED;

#define MICROCODE_VERSION 	"1.12"

MODULE_DESCRIPTION("Intel CPU (IA-32) microcode update driver");
MODULE_AUTHOR("Tigran Aivazian <tigran@veritas.com>");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;

#define MICRO_DEBUG 0

#if MICRO_DEBUG
#define printf(x...) printk(KERN_INFO x)
#else
#define printf(x...)
#endif

/* VFS interface */
static int microcode_open(struct inode *, struct file *);
static ssize_t microcode_read(struct file *, char *, size_t, loff_t *);
static ssize_t microcode_write(struct file *, const char *, size_t, loff_t *);
static int microcode_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static int do_microcode_update(void);
static void do_update_one(void *);
static void update_microcode(microcode_t *, int sig, int pf, int cksum);

/* read()/write()/ioctl() are serialized on this */
static DECLARE_RWSEM(microcode_rwsem);

static struct microcode *microcode; /* microcode data received from user */
static unsigned int microcode_size;  /* number of bytes in microcode */
static char *mc_applied;            /* applied microcode blocks */
static unsigned int mc_fsize;       /* file size of /dev/cpu/microcode */

/* we share file_operations between misc and devfs mechanisms */
static struct file_operations microcode_fops = {
	owner:		THIS_MODULE,
	read:		microcode_read,
	write:		microcode_write,
	ioctl:		microcode_ioctl,
	open:		microcode_open,
};

static struct miscdevice microcode_dev = {
	minor: MICROCODE_MINOR,
	name:	"microcode",
	fops:	&microcode_fops,
};

static devfs_handle_t devfs_handle;

static int __init microcode_init(void)
{
	int error;

	error = misc_register(&microcode_dev);
	if (error)
		printk(KERN_WARNING 
			"microcode: can't misc_register on minor=%d\n",
			MICROCODE_MINOR);

	devfs_handle = devfs_register(NULL, "cpu/microcode",
			DEVFS_FL_DEFAULT, 0, 0, S_IFREG | S_IRUSR | S_IWUSR, 
			&microcode_fops, NULL);
	if (devfs_handle == NULL && error) {
		printk(KERN_ERR "microcode: failed to devfs_register()\n");
		goto out;
	}
	error = 0;
	printk(KERN_INFO 
		"IA-32 Microcode Update Driver: v%s <tigran@veritas.com>\n", 
		MICROCODE_VERSION);

out:
	return error;
}

static void __exit microcode_exit(void)
{
	misc_deregister(&microcode_dev);
	devfs_unregister(devfs_handle);
	if (mc_applied)
		kfree(mc_applied);
	printk(KERN_INFO "IA-32 Microcode Update Driver v%s unregistered\n", 
			MICROCODE_VERSION);
}

module_init(microcode_init)
module_exit(microcode_exit)

static int microcode_open(struct inode *unused1, struct file *unused2)
{
	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

/*
 * update_req[cpu].err is set to 1 if update failed on 'cpu', 0 otherwise
 * if err==0, update_req[cpu].mc points to the applied block of microcode update
 */
struct update_req {
	int err;
	microcode_t * mc;
} update_req[NR_CPUS];

static char *ucode_buf[NR_CPUS];    /* microcode data buffer */

#define IS_UCODE_NOT_ALIGNED(ucode_addr) ((unsigned long)(ucode_addr) & 0xfUL)

static int get_max_unaligned_ucode(microcode_t *microcode)
{
	microcode_t 	*next;
	int 		cur_ucode_len;
	int 		max_ucode_len = 0;
	int		all_ucode_aligned = 1;

	next = microcode;
	do {
		cur_ucode_len = get_datasize(next);
		if (cur_ucode_len > max_ucode_len)
			max_ucode_len = cur_ucode_len;

		if (IS_UCODE_NOT_ALIGNED(&(next->bits)))
			all_ucode_aligned = 0;
	} while((next = get_next_microcode(next)));

	if (all_ucode_aligned)
		max_ucode_len = 0;
	return max_ucode_len;
}

static int do_microcode_update(void)
{
	int i, error = 0, err;
	int max_ucode_len;

	max_ucode_len = get_max_unaligned_ucode(microcode);
	if (max_ucode_len) {
		for (i = 0; i < smp_num_cpus; i++) {
			ucode_buf[i] = vmalloc(max_ucode_len);
			if (ucode_buf[i] == NULL) {
				int j;
				printk(KERN_ERR "microcode: CPU%d aborting, "
					"can not allocate memory \n", i);
				for (j = 0; j < i; j++)
					vfree(ucode_buf[j]);
				return EIO;
			}
		}
	}
	if (smp_call_function(do_update_one, NULL, 1, 1) != 0) {
		printk(KERN_ERR "microcode: IPI timeout, giving up\n");
		return -EIO;
	}

	do_update_one(NULL);

	if (max_ucode_len) {
		for (i = 0; i < smp_num_cpus; i++)
			vfree(ucode_buf[i]);
	}

	for (mc_fsize=0, i=0; i<smp_num_cpus; i++) {
		err = update_req[i].err;
		error += err;
		if(!err && update_req[i].mc) 
			mc_fsize += get_totalsize(update_req[i].mc);
	}

	mc_applied = kmalloc(mc_fsize, GFP_KERNEL);
	if (!mc_applied) {
		printk(KERN_ERR "microcode: out of memory for saved microcode\n");
		mc_fsize = 0;
		return error;
	}

	for (mc_fsize=0, i=0; i<smp_num_cpus; i++) {
		microcode_t * mc = update_req[i].mc;
		if (!update_req[i].err && mc) {
			memcpy(&mc_applied[mc_fsize], mc, get_totalsize(mc)); 
			mc_fsize += get_totalsize(mc);
		}
	}
	return error;
}

static void do_update_one(void *unused)
{
	int cpu_num = smp_processor_id();
	struct cpuinfo_x86 *c = cpu_data + cpu_num;
	struct update_req *req = update_req + cpu_num;
	unsigned int pf = 0, val[2], sig;
	microcode_t * next;
	
	req->err = 1; /* assume update will fail on this cpu */

	if (c->x86_vendor != X86_VENDOR_INTEL || c->x86 < 6 ||
		test_bit(X86_FEATURE_IA64, &c->x86_capability)){
		printk(KERN_ERR "microcode: CPU%d not a capable Intel processor\n", cpu_num);
		return;
	}

	sig = c->x86_mask + (c->x86_model<<4) + (c->x86<<8);

	if ((c->x86_model >= 5) || (c->x86 > 6)) {
		/* get processor flags from MSR 0x17 */
		rdmsr(MSR_IA32_PLATFORM_ID, val[0], val[1]);
		pf = 1 << ((val[1] >> 18) & 7);
	}

	next = microcode;
	do {
		struct extended_sigtable * extsigtable;
		printf("microcode: next=%p\n", next);
		if (next->ldrver == 1 && next->hdrver == 1) {
			if (sigmatch(next->sig, sig, next->pf, pf)) {
				update_microcode(next, next->sig, next->pf, next->cksum);
				return;
			}
			extsigtable = get_extended_sigtable(next);
			
			if (extsigtable) {
				int sum=extsigtable->count;
				int i=exttable_size(extsigtable)/DWSIZE;
				while(--i) sum += ((int *)extsigtable)[i];
				if (sum) {
					printk(KERN_WARNING "microcode: cpu %d aborting,"
							" bad extended signature table checksum\n", cpu_num);
				}
				else for (i=0; i<extsigtable->count; i++){
					struct extended_signature * extsig;
					extsig = &extsigtable->sigs[i];
					if (sigmatch(extsig->sig, sig, extsig->pf, pf)) {
						update_microcode(next, extsig->sig, extsig->pf, extsig->cksum);
						return;
					}
				}
			}
		}
	} while((next = get_next_microcode(next)));

	printk(KERN_ERR
	       "microcode: CPU%d no microcode found! (sig=%x, pflags=%d)\n", 
	       cpu_num, sig, pf);
}

static void update_microcode(microcode_t * mc, int sig, int pf, int cksum)
{
	int i, sum = 0;
	unsigned long flags;
	unsigned int val[2], rev;
	int cpu_num = smp_processor_id();
	struct update_req *req = update_req + cpu_num;
	char *mc_data;

	printf("Microcode\n");
	printf("   Header Revision %d\n",mc->hdrver);
	printf("   Loader Revision %x\n",mc->ldrver);
	printf("   Date %x/%x/%x\n",
		((mc->date >> 24 ) & 0xff),
		((mc->date >> 16 ) & 0xff),
		(mc->date & 0xFFFF));
	printf("   Type %x Family %x Model %x Stepping %x\n",
		((sig >> 12) & 0x3),
		((sig >> 8) & 0xf),
		((sig >> 4) & 0xf),
		((sig & 0xf)));
	printf("   Processor Flags %x\n\n",pf);
	printf("   Checksum %x\n",cksum);

	req->mc = mc;

	/* serialize access to update decision */
	spin_lock_irqsave(&microcode_update_lock, flags);          

	/* trick, to work even if there was no prior update by the BIOS */
	wrmsr(MSR_IA32_UCODE_REV, 0, 0);
	__asm__ __volatile__ ("cpuid" : : : "ax", "bx", "cx", "dx");

	/* get current (on-cpu) revision into rev (ignore val[0]) */
	rdmsr(MSR_IA32_UCODE_REV, val[0], rev);
	
	if (mc->rev < rev) {
		printk(KERN_ERR 
		       "microcode: CPU%d not 'upgrading' to earlier revision"
		       " %d (current=%d)\n", cpu_num, mc->rev, rev);
		goto out_unlock;
	} else if (mc->rev == rev) {
		/* notify the caller of success on this cpu */
		req->err = 0;
		printk(KERN_ERR 
			"microcode: CPU%d already at revision"
			" %d (current=%d)\n", cpu_num, mc->rev, rev);
		goto out_unlock;
	}

	/* Verify the checksum */
	i = (MC_HEADER_SIZE + get_datasize(mc)) / DWSIZE;
	while(--i) sum += ((int *)mc)[i];
	sum += ((int *)mc)[0];

	sum += (sig + pf + cksum) - (mc->sig + mc->pf + mc->cksum);

	if (sum != 0) {
		printk(KERN_ERR "microcode: CPU%d aborting, "
		       "bad checksum\n", cpu_num);
		goto out_unlock;
	}

	if (IS_UCODE_NOT_ALIGNED(&mc->bits)) {
		mc_data = ucode_buf[cpu_num];
		memcpy(mc_data, &mc->bits, get_datasize(mc));
	} else
		mc_data = (char *)&mc->bits;

	/* write microcode via MSR 0x79 */
	wrmsr(MSR_IA32_UCODE_WRITE, (unsigned int)(mc_data), 0);
	wrmsr(MSR_IA32_UCODE_REV, 0, 0);

	__asm__ __volatile__ ("cpuid" : : : "ax", "bx", "cx", "dx");
	/* get the current revision from MSR 0x8B */
	rdmsr(MSR_IA32_UCODE_REV, val[0], val[1]);

	/* notify the caller of success on this cpu */
	req->err = 0;
	printk(KERN_INFO "microcode: CPU%d updated from revision "
	       "%u to %u, date=%08x\n", 
	       cpu_num, rev, val[1], mc->date);
out_unlock:
	spin_unlock_irqrestore(&microcode_update_lock, flags);
	return;
}
	


static ssize_t microcode_read(struct file *file, char *buf, size_t len, loff_t *ppos)
{
	ssize_t ret = 0;

	down_read(&microcode_rwsem);
	if (*ppos >= mc_fsize)
		goto out;
	if (*ppos + len > mc_fsize)
		len = mc_fsize - *ppos;
	ret = -EFAULT;
	if (copy_to_user(buf, mc_applied + *ppos, len))
		goto out;
	*ppos += len;
	ret = len;
out:
	up_read(&microcode_rwsem);
	return ret;
}

static ssize_t microcode_write(struct file *file, const char *buf, size_t len, loff_t *ppos)
{
	ssize_t ret;

	if (len <= sizeof(struct microcode)) {
		printk(KERN_ERR "microcode: not enough data\n"); 
		return -EINVAL;
	}
	if ((len >> PAGE_SHIFT) > num_physpages) {
		printk(KERN_ERR "microcode: too much data (max %ld pages)\n", num_physpages);
		return -EINVAL;
	}
	down_write(&microcode_rwsem);
	if (mc_applied) {
		kfree(mc_applied);
		mc_applied = NULL;
		mc_fsize = 0;
	}
	
	microcode = vmalloc(len);
	if (!microcode) {
		ret = -ENOMEM;
		goto out_unlock;
	}
	microcode_size = len;

	if (copy_from_user(microcode, buf, len)) {
		ret = -EFAULT;
		goto out_free;
	}

	if(do_microcode_update()) {
		ret = -EIO;
	} else 
		ret = (ssize_t)len;

out_free:
	devfs_set_file_size(devfs_handle, mc_fsize);
	vfree(microcode);
out_unlock:
	up_write(&microcode_rwsem);
	return ret;
}

static int microcode_ioctl(struct inode *inode, struct file *file, 
		unsigned int cmd, unsigned long arg)
{
	switch(cmd) {
		case MICROCODE_IOCFREE:
			down_write(&microcode_rwsem);
			if (mc_applied) {
				devfs_set_file_size(devfs_handle, 0);
				kfree(mc_applied);
				mc_applied = NULL;
				printk(KERN_INFO "microcode: freed %d bytes\n", mc_fsize);
				mc_fsize = 0;
				up_write(&microcode_rwsem);
				return 0;
			}
			up_write(&microcode_rwsem);
			return -ENODATA;

		default:
			printk(KERN_ERR "microcode: unknown ioctl cmd=%d\n", cmd);
			return -EINVAL;
	}
	return -EINVAL;
}
