/*
 *  c 2003 PPC 64 Team, IBM Corp
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 * Advanced RTAS Interface for PPC64  
 * John Rose <johnrose@austin.ibm.com>
 *
 * Various applications on PPC64 require user-space access to a set of RTAS 
 * calls.  This module uses /proc/ppc64/rtas/ to provide an interface through
 * which these calls can be made.  The following list describes the
 * new proc entries and the RTAS calls they implement:
 * File					RTAS call
 * /proc/ppc64/rtas/cfg_connector	ibm,configure-connector
 * /proc/ppc64/rtas/get_power		get-power-level
 * /proc/ppc64/rtas/set_power		set-power-level
 * /proc/ppc64/rtas/get_sensor		get-sensor-state
 * /proc/ppc64/rtas/set_indicator 	set-indicator
 * /proc/ppc64/rtas/get_sysparm 	ibm,get-system-parameter
 * /proc/ppc64/rtas/eeh_set_option	ibm,set-eeh-option
 * /proc/ppc64/rtas/eeh_set_slot_reset	ibm,set-slot-reset
 * /proc/ppc64/rtas/eeh_reset_state	ibm,read-slot-reset-state
 */

#include <linux/module.h>
#include <linux/init.h>
#include <asm/proc_fs.h>
#include <asm/rtas.h>
#include <asm/page.h>
#include <asm/uaccess.h>
#include <asm/delay.h>

#define MODULE_VERSION "1.0"
#define MODULE_NAME "rtas_user"

#define GET_POWER_NAME   "get_power"
#define SET_POWER_NAME   "set_power"
#define CFG_CONN_NAME    "cfg_connector"
#define GET_SENSOR_NAME  "get_sensor"
#define SET_IND_NAME     "set_indicator"
#define GET_SYSPARM_NAME "get_sysparm"
#define EEH_SET_OPTION   "eeh_set_option"
#define EEH_SLOT_RESET   "eeh_set_slot_reset"
#define EEH_RESET_STATE  "eeh_reset_state"

extern int eeh_rtas_set_option(ulong, ulong, ulong, int);
extern int eeh_rtas_set_slot_reset(ulong, ulong, ulong, int);
extern int eeh_rtas_reset_state(ulong, ulong, ulong, ulong*);

static int ppc_rtas_common_open(struct inode *inode, struct file *file);
static int ppc_rtas_common_release(struct inode *inode, struct file *file);
static ssize_t ppc_rtas_common_read(struct file * file, char * buf,
		size_t count, loff_t *ppos);
static int copy_to_pde_buf(struct file *file, const char *user_buf, 
		size_t count, char **pde_buf);
static void free_pde_data(struct file *file);

static ssize_t ppc_rtas_get_power_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t ppc_rtas_set_power_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t ppc_rtas_cfg_conn_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t ppc_rtas_get_sensor_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t ppc_rtas_set_ind_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t ppc_rtas_get_sysparm_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t rtas_eeh_set_option_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t rtas_eeh_set_slot_reset_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static ssize_t rtas_eeh_reset_state_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos);
static int rtas_cfg_connector(char *workarea);

static void create_user_proc_entry(const char *name, 
		struct proc_dir_entry *dir, struct file_operations *ops);

struct file_operations ppc_rtas_get_power_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_get_power_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_set_power_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_set_power_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_cfg_conn_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_cfg_conn_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_get_sensor_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_get_sensor_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_set_ind_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_set_ind_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_get_sysparm_operations = {
	.open = 	ppc_rtas_common_open,
	.read = 	ppc_rtas_common_read,
	.write = 	ppc_rtas_get_sysparm_write,
	.release =	ppc_rtas_common_release
};
struct file_operations ppc_rtas_eeh_set_option_operations = {
       .open =         ppc_rtas_common_open,
       .read =         ppc_rtas_common_read,
       .write =        rtas_eeh_set_option_write,
       .release =      ppc_rtas_common_release
};
struct file_operations ppc_rtas_eeh_set_slot_reset_operations = {
       .open =         ppc_rtas_common_open,
       .read =         ppc_rtas_common_read,
       .write =        rtas_eeh_set_slot_reset_write,
       .release =      ppc_rtas_common_release
};
struct file_operations ppc_rtas_eeh_reset_state_operations = {
       .open =         ppc_rtas_common_open,
       .read =         ppc_rtas_common_read,
       .write =        rtas_eeh_reset_state_write,
       .release =      ppc_rtas_common_release
};

static char cfg_extentarea[RTAS_DATA_BUF_SIZE] __page_aligned;
static spinlock_t rtas_file_open_lock = SPIN_LOCK_UNLOCKED;

static int ppc_rtas_common_open(struct inode *inode, struct file *file)
{
	struct proc_dir_entry *dp = file->f_dentry->d_inode->u.generic_ip;
	
	/* Enforce exclusive open with use count of PDE */
	spin_lock(&rtas_file_open_lock);
	if (atomic_read(&dp->count) > 1) {
		spin_unlock(&rtas_file_open_lock);
		return -EAGAIN;
	}

	atomic_inc(&dp->count);
	spin_unlock(&rtas_file_open_lock);
	
	dp->data = NULL;
	
	return 0;
}

static int ppc_rtas_common_release(struct inode *inode, struct file *file)
{
	struct proc_dir_entry *dp = file->f_dentry->d_inode->u.generic_ip;
	
	free_pde_data(file);
	atomic_dec(&dp->count);

	return 0;
}

/* The dp->data buffer contains the size of the RTAS call structure 
 * in the first 32 bits, followed by the structure itself.
 */
static ssize_t ppc_rtas_common_read(struct file *file, char *buf, 
		size_t count, loff_t *ppos)
{
	struct proc_dir_entry *dp = file->f_dentry->d_inode->u.generic_ip;
	unsigned int copy_size;
	char *pde_buf;
	
	if (verify_area(VERIFY_WRITE, buf, count)) 
		return -EFAULT;
	
	/* In the data buffer of the PDE, the first 32 bits contains the
	 * size of the rest of the buffer. */
	copy_size = 0;
	if (dp->data) {
		copy_size = *(unsigned int *)dp->data;

		if (*ppos >= copy_size)
			return 0;
		else if (*ppos >= 0) 
			copy_size -= *ppos;	

		if (copy_size > count)
			copy_size = count;
			
		pde_buf = (char *) dp->data + sizeof(unsigned int);	
		if (copy_to_user(buf, pde_buf + *ppos, copy_size) != 0) 
			return -EFAULT;
	
		*ppos += copy_size;
	}
	
	return copy_size;
}

static int copy_to_pde_buf(struct file *file, const char *user_buf, 
		size_t count, char **pde_buf)
{
	struct proc_dir_entry *dp = file->f_dentry->d_inode->u.generic_ip;

	if (verify_area(VERIFY_READ, user_buf, count)) 
		return -EFAULT;

	free_pde_data(file);
	
	dp->data = kmalloc(sizeof(unsigned int) + count, GFP_KERNEL);
	if (!dp->data)
		return -ENOMEM;
	
	/* In the data buffer of the PDE, the first 32 bits contains the
	 * size of the rest of the buffer. */
	*(unsigned int *)dp->data = count;	

	*pde_buf = (char *) dp->data + sizeof(unsigned int);
	if (copy_from_user(*pde_buf, user_buf, count) != 0) {
		free_pde_data(file);
		return -EFAULT;
	}

	return 0;
}

static void free_pde_data(struct file *file)
{
	struct proc_dir_entry *dp = file->f_dentry->d_inode->u.generic_ip;

	if (dp->data) {
		kfree(dp->data);
		dp->data = NULL;
	}
}

static ssize_t ppc_rtas_get_power_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_get_power_level *gpl;
	int rc;
	
	if (count != sizeof(struct rtas_get_power_level))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &gpl);
	if (rc != 0) 
		return rc;
		
	gpl->status = rtas_get_power_level(gpl->powerdomain, &gpl->level);
	return count;
}

static ssize_t ppc_rtas_set_power_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_set_power_level *spl;
	int rc;
	
	if (count != sizeof(struct rtas_set_power_level))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &spl);
	if (rc != 0) 
		return rc;
		
	spl->status = rtas_set_power_level(spl->powerdomain, spl->level, 
			 	&spl->setlevel);
	return count;	
}

static ssize_t ppc_rtas_cfg_conn_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_cfg_connector *cc;
	int rc;
	
	if (count != sizeof(struct rtas_cfg_connector))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &cc);
	if (rc != 0) 
		return rc;
		
	cc->status = rtas_cfg_connector(&cc->workarea[0]);
	return count;	
}

static ssize_t ppc_rtas_get_sensor_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_get_sensor *gs;
	int rc;
	
	if (count != sizeof(struct rtas_get_sensor))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &gs);
	if (rc != 0) 
		return rc;
		
	gs->status = rtas_get_sensor(gs->sensor, gs->index, &gs->state);
	return count;	
}

static ssize_t ppc_rtas_set_ind_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_set_indicator *si;
	int rc;
	
	if (count != sizeof(struct rtas_set_indicator))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &si);
	if (rc != 0) 
		return rc;
		
	si->status = rtas_set_indicator(si->indicator, si->index, 
				si->new_value);
	return count;	
}

static ssize_t ppc_rtas_get_sysparm_write(struct file * file, const char * buf,
		size_t count, loff_t *ppos)
{
	struct rtas_get_sysparm *gs;
	int rc;
	
	if (count != sizeof(struct rtas_get_sysparm))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **) &gs);
	if (rc != 0) 
		return rc;

	if (gs->length > RTAS_DATA_BUF_SIZE) { 
 		free_pde_data(file);
		return -EINVAL;
	}
		
	gs->status = rtas_get_sysparm(gs->parameter, gs->data, gs->length);
	
	return count;	
}

static ssize_t rtas_eeh_set_option_write(struct file *file, const char *buf,
		size_t count, loff_t *ppos)
{
	struct rtas_eeh_set_options	*opts;
	int	rc;

	if (count != sizeof(struct rtas_eeh_set_options))
		return -EINVAL;

	rc = copy_to_pde_buf(file, buf, count, (char **)&opts);
 	if (rc != 0)
		return rc;

	opts->status = eeh_rtas_set_option(opts->config_addr, opts->buid_hi,
					   opts->buid_lo, opts->function);

	return count;
}

static ssize_t rtas_eeh_set_slot_reset_write(struct file *file, const char *buf,
		size_t count, loff_t *ppos)
{
	struct rtas_eeh_set_options	*opts;
	int	rc;

	if (count != sizeof(struct rtas_eeh_set_options))
		return -EINVAL;
 
 	rc = copy_to_pde_buf(file, buf, count, (char **)&opts);
 	if (rc != 0)
		return rc;
 
 	opts->status = eeh_rtas_set_slot_reset(opts->config_addr, opts->buid_hi,
 					       opts->buid_lo, opts->function);
 
 	return count;
}
 
static ssize_t rtas_eeh_reset_state_write(struct file *file, const char *buf,
 		size_t count, loff_t *ppos)
{
 	struct rtas_eeh_reset_state *reset_state;
 	ulong	rets[2];
 	int	rc;
 
 	if (count != sizeof(struct rtas_eeh_reset_state)) {
 	     	printk("EEH_RSRS: invalid read size\n");
 		return -EINVAL;
 	}
 
 	rc = copy_to_pde_buf(file, buf, count, (char **)&reset_state);
 	if (rc != 0)
 		return rc;
 
 	reset_state->status = eeh_rtas_reset_state(reset_state->config_addr, 
 						   reset_state->buid_hi,
 					           reset_state->buid_lo, rets);
 
	reset_state->reset_state = rets[0];
	reset_state->capabilities = rets[1];
	return count;
}

static int rtas_cfg_connector(char *workarea)
{
	int token = rtas_token("ibm,configure-connector");
	unsigned int wait_time;
	unsigned long extent_arg;
	int rc;
	
	if (token == RTAS_UNKNOWN_SERVICE)
		return RTAS_UNKNOWN_OP;
	
	rc = 0;
	while (1) {
		if (rc == RTAS_MEM) 
			extent_arg = __pa(cfg_extentarea);
	 	else
			extent_arg = 0;
			
		spin_lock(&rtas_data_buf_lock);
		memcpy(rtas_data_buf, workarea, RTAS_DATA_BUF_SIZE);
		rc = (int) rtas_call(token, 2, 1, NULL, __pa(rtas_data_buf), 
					extent_arg);
		memcpy(workarea, rtas_data_buf, RTAS_DATA_BUF_SIZE);
		spin_unlock(&rtas_data_buf_lock);
		
		if (rc == RTAS_BUSY) 
			udelay(1);	
		else if (rtas_is_extended_busy(rc)) {
			wait_time = rtas_extended_busy_delay_time(rc);
			udelay(wait_time * 1000);
		}
		else if (rc != RTAS_MEM)
			break;
	}

	return rc;
}

static void create_user_proc_entry(const char *name, 
		struct proc_dir_entry *dir, struct file_operations *ops)
{
	struct proc_dir_entry *entry;

	entry = create_proc_entry(name, S_IWUSR|S_IRUSR, dir);
	if (entry) {
		entry->proc_fops = ops;		
		entry->owner = THIS_MODULE;
	}
}

int __init rtas_user_init(void)
{

	if (!rtas_proc_dir) {
		printk(KERN_WARNING "rtas proc dir does not already exist");
		return -ENOENT;
	}
	
	create_user_proc_entry(GET_POWER_NAME, rtas_proc_dir, 
			&ppc_rtas_get_power_operations);
	create_user_proc_entry(SET_POWER_NAME, rtas_proc_dir, 
			&ppc_rtas_set_power_operations);
	create_user_proc_entry(CFG_CONN_NAME, rtas_proc_dir, 
			&ppc_rtas_cfg_conn_operations);
	create_user_proc_entry(GET_SENSOR_NAME, rtas_proc_dir,
			&ppc_rtas_get_sensor_operations);
	create_user_proc_entry(SET_IND_NAME, rtas_proc_dir,
			&ppc_rtas_set_ind_operations);
	create_user_proc_entry(GET_SYSPARM_NAME, rtas_proc_dir,
			&ppc_rtas_get_sysparm_operations);
	create_user_proc_entry(EEH_SET_OPTION, rtas_proc_dir,
			&ppc_rtas_eeh_set_option_operations);
	create_user_proc_entry(EEH_SLOT_RESET, rtas_proc_dir,
			&ppc_rtas_eeh_set_slot_reset_operations);
	create_user_proc_entry(EEH_RESET_STATE, rtas_proc_dir,
			&ppc_rtas_eeh_reset_state_operations);

	return 0;
}

void __exit rtas_user_cleanup(void)
{
	if (!rtas_proc_dir)
		return;

	remove_proc_entry(GET_POWER_NAME, rtas_proc_dir);
	remove_proc_entry(SET_POWER_NAME, rtas_proc_dir);
	remove_proc_entry(CFG_CONN_NAME, rtas_proc_dir);
	remove_proc_entry(GET_SENSOR_NAME, rtas_proc_dir);
	remove_proc_entry(SET_IND_NAME, rtas_proc_dir);
	remove_proc_entry(GET_SYSPARM_NAME, rtas_proc_dir);
	remove_proc_entry(EEH_SET_OPTION, rtas_proc_dir);
	remove_proc_entry(EEH_SLOT_RESET, rtas_proc_dir);
	remove_proc_entry(EEH_RESET_STATE, rtas_proc_dir);	
}

module_init(rtas_user_init);
module_exit(rtas_user_cleanup);
MODULE_LICENSE("GPL");
