#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h> 
#include <linux/file.h> 
#include <linux/utime.h>
#include <linux/times.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/aio.h>
#include <linux/aio_abi.h>
#include <asm/mman.h>
#include <asm/types.h>
#include <asm/uaccess.h>

struct timespec32 { 
	u32 tv_sec;
	u32 tv_nsec;
};

#define u32_to_ptr(x) ((void *)(u64)(x))


int sys_sched_getaffinity(pid_t pid, unsigned int len, unsigned long *new_mask_ptr); 
int sys_sched_setaffinity(pid_t pid, unsigned int len, unsigned long *new_mask_ptr); 

/* only works on LE */
long sys32_sched_setaffinity(pid_t pid, unsigned int len,
			    unsigned int *new_mask_ptr)
{
	mm_segment_t oldfs = get_fs(); 
	unsigned long mask; 
	int err;
	if (get_user(mask, new_mask_ptr)) 
		return -EFAULT;	
	set_fs(KERNEL_DS); 
	err = sys_sched_setaffinity(pid,sizeof(mask),&mask); 	
	set_fs(oldfs); 
	return err;
}

/* only works on LE */ 
long sys32_sched_getaffinity(pid_t pid, unsigned int len,
			    unsigned int *new_mask_ptr)
{
	mm_segment_t oldfs = get_fs(); 
	unsigned long mask; 
	int err;
	mask = 0; 
	set_fs(KERNEL_DS); 
	err = sys_sched_getaffinity(pid,sizeof(mask),&mask); 	
	set_fs(oldfs); 
	if (err > 0) 
		err = put_user((u32)mask, new_mask_ptr); 
	return err;
}

extern int sys_futex(unsigned long uaddr, int op, int val, struct timespec *t); 

asmlinkage long
sys32_futex(unsigned long uaddr, int op, int val, struct timespec32 *utime32)
{
	struct timespec t;
	mm_segment_t oldfs = get_fs(); 
	int err;

	if (utime32) {
		if (verify_area(VERIFY_READ, utime32, sizeof(*utime32)))
			return -EFAULT;

		if (__get_user(t.tv_sec, &utime32->tv_sec) ||
		    __get_user(t.tv_nsec, &utime32->tv_nsec))
			return -EFAULT;
		
	}

	/* the set_fs is safe because futex doesn't use the seg limit 
	   for valid page checking of uaddr. */ 
	set_fs(KERNEL_DS); 
	err = sys_futex(uaddr, op, val, &t);
	set_fs(oldfs); 
	return err; 
}

extern long sys_io_setup(unsigned nr_reqs, aio_context_t *ctx);

long sys32_io_setup(unsigned nr_reqs, u32 *ctx32p)
{ 
	long ret; 
	aio_context_t ctx64;
	mm_segment_t oldfs = get_fs(); 	
	set_fs(KERNEL_DS); 
	ret = sys_io_setup(nr_reqs, &ctx64); 
	set_fs(oldfs); 
	/* truncating is ok because it's a user address */
	if (!ret) 
		ret = put_user((u32)ctx64, ctx32p);
	return ret;
} 

asmlinkage long sys32_io_submit(aio_context_t ctx_id, int nr,
			      u32 *iocbpp)
{
	struct kioctx *ctx;
	long ret = 0;
	int i;
	
	if (unlikely(nr < 0))
		return -EINVAL;

	if (unlikely(!access_ok(VERIFY_READ, iocbpp, (nr*sizeof(*iocbpp)))))
		return -EFAULT;

	ctx = lookup_ioctx(ctx_id);
	if (unlikely(!ctx)) {
		pr_debug("EINVAL: io_submit: invalid context id\n");
		return -EINVAL;
	}

	for (i=0; i<nr; i++) {
		u32 p32;
		struct iocb *user_iocb, tmp;

		if (unlikely(__get_user(p32, iocbpp + i))) {
			ret = -EFAULT;
			break;
		}
		user_iocb = u32_to_ptr(p32);

		if (unlikely(copy_from_user(&tmp, user_iocb, sizeof(tmp)))) {
			ret = -EFAULT;
			break;
		}

		ret = io_submit_one(ctx, user_iocb, &tmp);
		if (ret)
			break;
	}

	put_ioctx(ctx);
	return i ? i : ret;
}

extern asmlinkage long sys_io_getevents(aio_context_t ctx_id,
					  long min_nr,
					  long nr,
					  struct io_event *events,
					  struct timespec *timeout);

asmlinkage long sys32_io_getevents(aio_context_t ctx_id,
				 unsigned long min_nr,
				 unsigned long nr,
				 struct io_event *events,
				 struct timespec32 *timeout)
{ 	
	long ret;
	mm_segment_t oldfs; 
	struct timespec t;
	/* Harden against bogus ptrace */
	if (nr >= 0xffffffff || 
	    !access_ok(VERIFY_WRITE, events, nr * sizeof(struct io_event)))
		return -EFAULT;
	if (timeout && (get_user(t.tv_sec, &timeout->tv_sec) ||
			get_user(t.tv_nsec, &timeout->tv_nsec)))
		return -EFAULT; 
	oldfs = get_fs();
	set_fs(KERNEL_DS); 
	ret = sys_io_getevents(ctx_id,min_nr,nr,events,timeout ? &t : NULL); 
	set_fs(oldfs); 

	if (timeout) {
		if (verify_area(VERIFY_WRITE, timeout, sizeof(struct timespec32)) ||
		    __put_user (t.tv_sec, &timeout->tv_sec) ||
	    	    __put_user (t.tv_nsec, &timeout->tv_nsec))
			return -EFAULT;
	}
	return ret;
} 
