/*
 * C-Brick Serial Port (and console) driver for SGI Altix machines.
 *
 * This driver is NOT suitable for talking to the l1-controller for
 * anything other than 'console activities' --- please use the l1
 * driver for that.
 *
 *
 * Copyright (C) 2000-2003 Silicon Graphics, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 * Mountain View, CA  94043, or:
 *
 * http://www.sgi.com
 *
 * For further information regarding this notice, see:
 *
 * http://oss.sgi.com/projects/GenInfo/NoticeExplan
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/module.h>
#ifdef CONFIG_MAGIC_SYSRQ
#include <linux/sysrq.h>
#endif
#include <linux/circ_buf.h>
#include <linux/serial_reg.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/sn/sn_sal.h>
#include <asm/sn/nodepda.h>
#include <asm/sn/simulator.h>
#include <asm/sn/sn2/intr.h>
#include <asm/sn/sn2/sn_private.h>

#if defined(CONFIG_SGI_L1_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
static char sysrq_serial_str[] = "\eSYS";
static char *sysrq_serial_ptr = sysrq_serial_str;
static unsigned long sysrq_requested;
#endif	/* CONFIG_SGI_L1_SERIAL_CONSOLE && CONFIG_MAGIC_SYSRQ */

static char *serial_revdate = "2003-08-15";

static struct tty_struct *sn_sal_tty;
static struct tq_struct tq;
static struct circ_buf xmit;
static struct timer_list sn_sal_timer_list;
static int sn_sal_event; /* event type for task queue */
static int sn_sal_refcount;
static char *sal_tmp_buffer;
static int sn_sal_irq;
static spinlock_t sn_sal_lock = SPIN_LOCK_UNLOCKED;
static int tx_count;
static int rx_count;
static int sn_sal_basic_init;
static int sn_sal_write(struct tty_struct *tty, int from_user,
			const unsigned char *buf, int count);

static struct tty_struct *sn_sal_table;
static struct termios *sn_sal_termios;
static struct termios *sn_sal_termios_locked;

static DECLARE_TASK_QUEUE(tq_serial);

/* driver subtype - what does this mean? */
#define SN_SAL_SUBTYPE 1

/* minor device number */
#define SN_SAL_MINOR 64

/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

/* number of characters we can transmit to the SAL console at a time */
#define SN_SAL_MAX_CHARS 120

/* event types for our task queue -- so far just one */
#define SN_SAL_EVENT_WRITE_WAKEUP	0

#define CONSOLE_RESTART 1

static unsigned long interrupt_timeout;

extern u64 master_node_bedrock_address;
extern void early_sn_setup(void);

struct sn_sal_ops_struct {
	int (*puts)(const char *s, int len);
	int (*getc)(void);
	int (*input_pending)(void);
	int (*console_write)(struct tty_struct *tty, int from_user,
			const unsigned char *buf, int count);
};

static struct sn_sal_ops_struct *sn_sal_ops;

/* routines for running the console in polling mode */

static int poll_puts(const char *s, int len)
{
	return ia64_sn_console_putb(s, len);
}

static int poll_getc(void)
{
	int ch;
	/* XXX chadt what does the return value mean */
	ia64_sn_console_getc(&ch);
	return ch;
}

static int poll_input_pending(void)
{
	int status, input;

	status = ia64_sn_console_check(&input);
	if (!status && input)
		return SAL_CONSOLE_INTR_RECV;
	else
		return 0;
}

static int
poll_console_write(struct tty_struct *tty, int from_user,
		   const unsigned char *buf, int count)
{
	return ia64_sn_console_putb(buf, count);
}


static struct sn_sal_ops_struct poll_ops = {
	.puts = poll_puts,
	.getc = poll_getc,
	.input_pending = poll_input_pending,
	.console_write = poll_console_write
};


/* routines for running the console on the simulator */

static int sim_puts(const char *str, int count)
{
	int counter = count;

#ifdef FLAG_DIRECT_CONSOLE_WRITES
	/* This is an easy way to pre-pend the output to know whether the output
	 * was done via sal or directly */
	writeb('[', master_node_bedrock_address + (UART_TX << 3));
	writeb('+', master_node_bedrock_address + (UART_TX << 3));
	writeb(']', master_node_bedrock_address + (UART_TX << 3));
	writeb(' ', master_node_bedrock_address + (UART_TX << 3));
#endif	/* FLAG_DIRECT_CONSOLE_WRITES */
	while (counter > 0) {
		writeb(*str, master_node_bedrock_address + (UART_TX << 3));
		counter--;
		str++;
	}

	return count;
}

static int sim_getc(void)
{
	return readb(master_node_bedrock_address + (UART_RX << 3));
}

static int sim_input_pending(void)
{
	int status;
	status = readb(master_node_bedrock_address + (UART_LSR << 3));
	if (status & UART_LSR_DR)
		return SAL_CONSOLE_INTR_RECV;
	else
		return 0;
}

static int
sim_console_write(struct tty_struct *tty, int from_user,
		  const unsigned char *buf, int count)
{
	return sim_puts(buf, count);
}

static struct sn_sal_ops_struct sim_ops = {
	.puts = sim_puts,
	.getc = sim_getc,
	.input_pending = sim_input_pending,
	.console_write = sim_console_write
};


/* routines for an interrupt driven console (normal) */
static int intr_puts(const char *str, int len)
{
	return ia64_sn_console_xmit_chars(str, len);
}

static int intr_getc(void)
{
	return ia64_sn_console_readc();
}

static int intr_input_pending(void)
{
	return ia64_sn_console_intr_status() & SAL_CONSOLE_INTR_RECV;
}

static struct sn_sal_ops_struct intr_ops = {
	.puts = intr_puts,
	.getc = intr_getc,
	.input_pending = intr_input_pending,
	.console_write = sn_sal_write
};


/*********************************************************************
 * Interrupt handling routines.
 */

static inline void
sn_sal_sched_event(int event)
{
	sn_sal_event |= (1 << event);
	queue_task(&tq, &tq_serial);
	mark_bh(SERIAL_BH);
}

static int
receive_chars(struct tty_struct *tty, struct pt_regs *regs)
{
	int ch;
	int status;

	while ((status = sn_sal_ops->input_pending())) {
		ch = sn_sal_ops->getc();
		if (ch < 0) {
			printk(KERN_ERR "sn_serial: An error occured while "
			       "obtaining data from console (0x%0x)\n", ch);
			break;
		}
#if defined(CONFIG_SGI_L1_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
		if (sysrq_requested) {
			unsigned long sysrq_timeout = sysrq_requested + HZ*5;
			sysrq_requested = 0;
			if (ch && time_before(jiffies, sysrq_timeout)) {
				handle_sysrq(ch, regs, NULL, NULL);
				goto ignore_char;
			}
		}
		if (ch == *sysrq_serial_ptr) {
			if (!(*++sysrq_serial_ptr)) {
				sysrq_requested = jiffies;
				sysrq_serial_ptr = sysrq_serial_str;
			}
		} else
			sysrq_serial_ptr = sysrq_serial_str;
#endif	/* CONFIG_SGI_L1_SERIAL_CONSOLE && CONFIG_MAGIC_SYSRQ */

		if (tty) {
			*tty->flip.char_buf_ptr = ch;
			tty->flip.char_buf_ptr++;
			tty->flip.count++;
		}
		rx_count++;

#if defined(CONFIG_SGI_L1_SERIAL_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
	ignore_char:
#endif
	}

	if (tty)
		tty_flip_buffer_push(tty);

	return status;
}


static void
transmit_chars(void)
{
	int xmit_count, tail, head, loops, ii;
	int result;
	unsigned long flags;
	char *start;

	if (xmit.head == xmit.tail)
		return;

	/* Make sure no one gets here until we have drained the buffer */
	spin_lock_irqsave(&sn_sal_lock, flags);

	head = xmit.head;
	tail = xmit.tail;
	start = &xmit.buf[tail];

	/* twice around gets the tail to the end of the buffer and then to the head, if needed */
	loops = (head < tail) ? 2 : 1;

	for (ii = 0; ii < loops; ii++) {
		xmit_count = (head < tail) ? (SERIAL_XMIT_SIZE - tail) : (head - tail);

		if (xmit_count > 0) {
			result = sn_sal_ops->puts(start, xmit_count);
			if (result > 0) {
				xmit_count -= result;
				tx_count += result;
				tail += result;
				tail &= SERIAL_XMIT_SIZE - 1;
				xmit.tail = tail;
				start = &xmit.buf[tail];
			}
		}
	}
	spin_unlock_irqrestore(&sn_sal_lock, flags);

	if (CIRC_CNT(xmit.head, xmit.tail, SERIAL_XMIT_SIZE) < WAKEUP_CHARS) {
		/* There's few enough characters left in the xmit buffer
		 * that we could stand for the upper layer to send us some
		 * more.  Ask for it.
		 */
		sn_sal_sched_event(SN_SAL_EVENT_WRITE_WAKEUP);
	}
}


static void
sn_sal_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int status;

	if (irq) {
		status = ia64_sn_console_intr_status();

		if (status & SAL_CONSOLE_INTR_RECV) {
			status = receive_chars(sn_sal_tty, regs);
		}

		if (status & SAL_CONSOLE_INTR_XMIT) {
			transmit_chars();
		}
	} else {
		/* Polling */
		int input;

		if (IS_RUNNING_ON_SIMULATOR()) {
			input = sim_input_pending();
			if (input)
				status = receive_chars(sn_sal_tty, regs);
		} else {
			status = ia64_sn_console_check(&input);
			if (!status && input)
				status = receive_chars(sn_sal_tty, regs);
		}
		transmit_chars();
	}
}


static void
sn_sal_connect_interrupt(void)
{
	cpuid_t intr_cpuid;
	unsigned int intr_cpuloc;
	nasid_t console_nasid;
	unsigned int console_irq;
	int result;

	/* if it is an old prom, run in poll mode */
	if ((sn_sal_rev_major() <= 1) && (sn_sal_rev_minor() <= 3)) {
		/* before version 1.06 doesn't work */
		printk(KERN_INFO "========== OLD prom version %x.%x - running in polled mode\n",
		       sn_sal_rev_major(), sn_sal_rev_minor());
		return;
	}

	console_nasid = ia64_sn_get_console_nasid();
	intr_cpuid = NODEPDA(NASID_TO_COMPACT_NODEID(console_nasid))
		->node_first_cpu;
	intr_cpuloc = cpu_physical_id(intr_cpuid);
	console_irq = CPU_VECTOR_TO_IRQ(intr_cpuloc, SGI_UART_VECTOR);

	result = intr_connect_level(intr_cpuid, SGI_UART_VECTOR,
				    0 /*not used*/, 0 /*not used*/);
	if (result != SGI_UART_VECTOR) {
		return;
	}

	result = request_irq(console_irq, sn_sal_interrupt,
			     SA_INTERRUPT,  "SAL console driver", &sn_sal_tty);
	if (result >= 0) {
		/* sn_sal_irq is a global variable.  When it's set to a non-zero
		 * value, the console driver stops polling for input (since interrupts
		 * should now be enabled).
		 */
		sn_sal_irq = console_irq;

		/* xmit interrupts are enabled by SAL as necessary */
	} else
		printk(KERN_INFO "sn_serial: Console proceeding in polled mode\n");
}
    

/*
 * End of the interrupt routines.
 *********************************************************************/


/*********************************************************************
 * From drivers/char/serial.c:
 *
 * "This routine is used to handle the "bottom half" processing for
 * the serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * [sn_sal_interrupt()] has returned, BUT WITH INTERRUPTS TURNED ON.
 * This is where time-consuming activities which can not be done in
 * the interrupt driver proper are done; the interrupt driver
 * schedules them using [sn_sal_sched_event()], and they get done
 * here."
 */
static void
do_sn_sal_bh(void)
{
	run_task_queue(&tq_serial);
}

static void
do_softint(void *data)
{
	struct tty_struct *tty;

	if (!(tty = sn_sal_tty))
		return;

	if (test_and_clear_bit(SN_SAL_EVENT_WRITE_WAKEUP, &sn_sal_event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
		{
			(tty->ldisc.write_wakeup)(tty);
		}
		wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
		wake_up_interruptible(&tty->poll_wait);
#endif
	}
}


/*
 * This function handles polled mode.
 */
static void
sn_sal_timer(unsigned long dummy)
{
	if (sn_sal_irq == 0) {
		sn_sal_interrupt(0, NULL, NULL);
	}
#if CONSOLE_RESTART
	else {
		sn_sal_interrupt(1, NULL, NULL);
	}
#endif
	mod_timer(&sn_sal_timer_list, jiffies + interrupt_timeout);
}

/*
 * End of "sofware interrupt" routines.
 *********************************************************************/


/*********************************************************************
 * User-level console routines
 */

static int
sn_sal_open(struct tty_struct *tty, struct file *filp)
{
	if (!sn_sal_tty)
		sn_sal_tty = tty;

	if (!sn_sal_irq || CONSOLE_RESTART)
		mod_timer(&sn_sal_timer_list, jiffies + interrupt_timeout);

	return 0;
}


/* We're keeping all our resources.  We're keeping interrupts turned
 * on.  Maybe just let the tty layer finish its stuff...? GMSH
 */
static void
sn_sal_close(struct tty_struct *tty, struct file * filp)
{
	mb();
	if (tty->count == 1) {
		tty->closing = 1;
		if (tty->driver.flush_buffer)
			tty->driver.flush_buffer(tty);
		if (tty->ldisc.flush_buffer)
			tty->ldisc.flush_buffer(tty);
		tty->closing = 0;
		sn_sal_tty = NULL;
	}
}


static int
sn_sal_write(struct tty_struct *tty, int from_user,
	     const unsigned char *buf, int count)
{
	int c, ret = 0;
	unsigned long flags;

	if (/*!tty || */ !xmit.buf) {
		return 0;
	}

	if (from_user) {
		if (!sal_tmp_buffer)
			return 0;
		while (1) {
			int c1;
			c = CIRC_SPACE_TO_END(xmit.head, xmit.tail,
					      SERIAL_XMIT_SIZE);

			if (count < c)
				c = count;
			if (c <= 0)
				break;

			c -= copy_from_user(sal_tmp_buffer, buf, c);
			if (!c) {
				if (!ret)
					ret = -EFAULT;
				break;
			}

			/* Turn off interrupts and see if the xmit buffer has
			 * moved since the last time we looked.
			 */
			spin_lock_irqsave(&sn_sal_lock, flags);
			c1 = CIRC_SPACE_TO_END(xmit.head, xmit.tail,
					       SERIAL_XMIT_SIZE);

			if (c1 < c)
				c = c1;

			memcpy(xmit.buf + xmit.head, sal_tmp_buffer, c);
			xmit.head = ((xmit.head + c) & (SERIAL_XMIT_SIZE - 1));
			spin_unlock_irqrestore(&sn_sal_lock, flags);

			buf += c;
			count -= c;
			ret += c;
		}
	} else {
		/* The buffer passed in isn't coming from userland,
		 * so cut out the middleman (sal_tmp_buffer).
		 */
		spin_lock_irqsave(&sn_sal_lock, flags);
		while (1) {
			c = CIRC_SPACE_TO_END(xmit.head, xmit.tail,
					      SERIAL_XMIT_SIZE);

			if (count < c)
				c = count;
			if (c <= 0) {
				break;
			}
			memcpy(xmit.buf + xmit.head, buf, c);
			xmit.head = ((xmit.head + c) & (SERIAL_XMIT_SIZE - 1));
			buf += c;
			count -= c;
			ret += c;
		}
		spin_unlock_irqrestore(&sn_sal_lock, flags);
	}

	if ((xmit.head != xmit.tail) /* && !tty->stopped && !tty->hw_stopped */)
		transmit_chars();
    
	return ret;
}


static void
sn_sal_put_char(struct tty_struct *tty, unsigned char ch)
{
	unsigned long flags;

	if (!tty || !xmit.buf)
		return;

	spin_lock_irqsave(&sn_sal_lock, flags);
	if (CIRC_SPACE(xmit.head, xmit.tail, SERIAL_XMIT_SIZE) != 0) {
		xmit.buf[xmit.head] = ch;
		xmit.head = (xmit.head + 1) & (SERIAL_XMIT_SIZE-1);
	}
	spin_unlock_irqrestore(&sn_sal_lock, flags);
}


static void
sn_sal_flush_chars(struct tty_struct *tty)
{
	if (CIRC_CNT(xmit.head, xmit.tail, SERIAL_XMIT_SIZE))
		transmit_chars();
}


static int
sn_sal_write_room(struct tty_struct *tty)
{
	return CIRC_SPACE(xmit.head, xmit.tail, SERIAL_XMIT_SIZE);
}


static int
sn_sal_chars_in_buffer(struct tty_struct *tty)
{
	return CIRC_CNT(xmit.head, xmit.tail, SERIAL_XMIT_SIZE);
}


static int
sn_sal_ioctl(struct tty_struct *tty, struct file *file,
	       unsigned int cmd, unsigned long arg)
{
	/* nothing supported*/
	return -ENOIOCTLCMD;
}


static void
sn_sal_flush_buffer(struct tty_struct *tty)
{
	unsigned long flags;

	/* drop everything */
	spin_lock_irqsave(&sn_sal_lock, flags);
	xmit.head = xmit.tail = 0;
	spin_unlock_irqrestore(&sn_sal_lock, flags);

	/* wake up tty level */
	wake_up_interruptible(&tty->write_wait);
#ifdef SERIAL_HAVE_POLL_WAIT
	wake_up_interruptible(&tty->poll_wait);
#endif
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}


static void
sn_sal_hangup(struct tty_struct *tty)
{
	sn_sal_flush_buffer(tty);
}


static void
sn_sal_wait_until_sent(struct tty_struct *tty, int timeout)
{
	/* this is SAL's problem */
}


/*
 * sn_sal_read_proc
 *
 * Console /proc interface
 */

static int
sn_sal_read_proc(char *page, char **start, off_t off, int count,
		 int *eof, void *data)
{
	int len = 0;
	off_t	begin = 0;
	extern nasid_t get_console_nasid(void);

	len += sprintf(page, "sn_serial: revision:%s nasid:%d irq:%d tx:%d rx:%d\n",
		       serial_revdate, get_console_nasid(), sn_sal_irq,
		       tx_count, rx_count);
	*eof = 1;

	if (off >= len+begin)
		return 0;
	*start = page + (off-begin);

	return (count < begin+len-off) ? count : begin+len-off;
}


static struct tty_driver sn_sal_driver = {
	.magic = TTY_DRIVER_MAGIC,
	.driver_name = "sn_serial",
#if defined(CONFIG_DEVFS_FS)
	.name = "tts/%d",
#else
	.name = "ttyS",
#endif
	.major = TTY_MAJOR,
	.minor_start = SN_SAL_MINOR,
	.num = 1,
	.type = TTY_DRIVER_TYPE_SERIAL,
	.subtype = SN_SAL_SUBTYPE,
	.flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS,
	.refcount = &sn_sal_refcount,
	.table = &sn_sal_table,
	.termios = &sn_sal_termios,
	.termios_locked = &sn_sal_termios_locked,

	.open = sn_sal_open,
	.close = sn_sal_close,
	.write = sn_sal_write,
	.put_char = sn_sal_put_char,
	.flush_chars = sn_sal_flush_chars,
	.flush_buffer = sn_sal_flush_buffer,
	.write_room = sn_sal_write_room,
	.chars_in_buffer = sn_sal_chars_in_buffer,
/*
	.ioctl = sn_sal_ioctl,
*/
	.hangup = sn_sal_hangup,
	.wait_until_sent = sn_sal_wait_until_sent,
	.read_proc = sn_sal_read_proc,
};

/* sn_sal_init wishlist:
 * - allocate sal_tmp_buffer
 * - fix up the tty_driver struct
 * - turn on receive interrupts
 * - do any termios twiddling once and for all
 */

/*
 * Boot-time initialization code
 */
static int __init
sn_sal_init(void)
{
	int retval = 0;

	if (!ia64_platform_is("sn2"))
		return -ENODEV;

	if (!sn_sal_basic_init) {
		sn_sal_basic_init = 1;
		early_sn_setup();
	}

	/* Initialize the tty_driver structure */
	init_bh(SERIAL_BH, do_sn_sal_bh);

	sn_sal_driver.init_termios = tty_std_termios;
	sn_sal_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;

	/* initialize xmit */
	xmit.buf = (char *)get_zeroed_page(GFP_KERNEL);
	if (!xmit.buf) {
		printk(KERN_ERR "sn_serial: Unable to allocate xmit buffer\n");
		retval = -ENOMEM;
		goto error_out;
	}

	xmit.head = 0;
	xmit.tail = 0;

	/* allocate a temporary buffer for copying data from user land */
	sal_tmp_buffer = (char *)get_zeroed_page(GFP_KERNEL);
	if (!sal_tmp_buffer) {
		printk(KERN_ERR "sn_serial: unable to allocate temp buffer\n");
		retval = -ENOMEM;
		goto free_xmit;
	}

	/* task queue initialization */
	tq.routine = do_softint;
	tq.data = NULL;

	if (IS_RUNNING_ON_SIMULATOR())
		sn_sal_ops = &sim_ops;
	else {
		/* turn on interrupts */
		sn_sal_connect_interrupt();
		if (sn_sal_irq) {
			unsigned long flags;
			spin_lock_irqsave(&sn_sal_lock, flags);
			sn_sal_ops = &intr_ops;
			spin_unlock_irqrestore(&sn_sal_lock, flags);
			ia64_sn_console_intr_enable(SAL_CONSOLE_INTR_RECV);
		} else
			sn_sal_ops = &poll_ops;
	}

	/*
	 * Without interrupts, we have to rely on a timer to poll the
	 * SAL console driver.
	 */
	if (!sn_sal_irq || CONSOLE_RESTART) {
		init_timer(&sn_sal_timer_list);
		sn_sal_timer_list.function = sn_sal_timer;

		if (IS_RUNNING_ON_SIMULATOR())
			interrupt_timeout = 6;
		else if (sn_sal_irq && CONSOLE_RESTART)
			interrupt_timeout = 20*HZ;
		else
			interrupt_timeout = 2*HZ/100;

		mod_timer(&sn_sal_timer_list, jiffies + interrupt_timeout);
	}

	if ((retval = tty_register_driver(&sn_sal_driver))) {
		printk(KERN_ERR "sn_serial: Unable to register tty driver\n");
		retval = -ENODEV;
		goto free_tmp;
	}

	tty_register_devfs(&sn_sal_driver, 0, sn_sal_driver.minor_start);

	return 0;

 free_tmp:
	free_page((unsigned long)sal_tmp_buffer);
 free_xmit:
	free_page((unsigned long)xmit.buf);
 error_out:
	return retval;
}


static void __exit
sn_sal_fini(void)
{
	unsigned long flags;
	int e;

	del_timer_sync(&sn_sal_timer_list);
	spin_lock_irqsave(&sn_sal_lock, flags);
	remove_bh(SERIAL_BH);

	if ((e = tty_unregister_driver(&sn_sal_driver)))
		printk("sn_serial: failed to unregister driver (%d)\n", e);

	spin_unlock_irqrestore(&sn_sal_lock, flags);

	if (sal_tmp_buffer) {
		unsigned long pg = (unsigned long) sal_tmp_buffer;
		sal_tmp_buffer = NULL;
		free_page(pg);
	}
}

module_init(sn_sal_init);
module_exit(sn_sal_fini);

/*
 * End of user-level console routines.
 *********************************************************************/


/*********************************************************************
 * Kernel console definitions
 */

#ifdef CONFIG_SGI_L1_SERIAL_CONSOLE
/*
 * Print a string to the SAL console.  The console_lock must be held
 * when we get here.
 */
static void
sn_sal_console_write(struct console *co, const char *s, unsigned count)
{
	sn_sal_ops->console_write(NULL, 0, s, count);
}

#if defined(CONFIG_IA64_EARLY_PRINTK_SGI_SN)
void
early_printk_sn_sal(const char *s, unsigned int count)
{
	if (!sn_sal_basic_init) {
		early_sn_setup();
		if (IS_RUNNING_ON_SIMULATOR())
			sn_sal_ops = &sim_ops;
		else
			sn_sal_ops = &poll_ops;
		sn_sal_basic_init = 1;
	}

	sn_sal_ops->puts(s, count);
}
#endif	/* CONFIG_IA64_EARLY_PRINTK_SGI_SN */

static kdev_t
sn_sal_console_device(struct console *c)
{
	return MKDEV(TTY_MAJOR, 64 + c->index);
}

static int __init
sn_sal_console_setup(struct console *co, char *options)
{
	return 0;
}


static struct console sal_console = {
	.name = "ttyS",
	.write = sn_sal_console_write,
	.device = sn_sal_console_device,
	.setup = sn_sal_console_setup,
	.flags = CON_PRINTBUFFER,
	.index = -1
};

/*
 * End of kernel console definitions.
 *********************************************************************/


void __init
sn_sal_serial_console_init(void)
{
	if (ia64_platform_is("sn2")) {
		early_sn_setup();
		if (IS_RUNNING_ON_SIMULATOR())
			sn_sal_ops = &sim_ops;
		else
			sn_sal_ops = &poll_ops;
		sn_sal_basic_init = 1;
		mb();
		register_console(&sal_console);
	}
}

#endif /* CONFIG_SGI_L1_SERIAL_CONSOLE */
