/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/dprobes.h>
#include <linux/hook.h>

void dprobes_hooks_asm_code_start(void) { }

USE_HOOK(DO_BREAK);
USE_HOOK(DO_ILLOP);
USE_HOOK(DO_DEBUG);
USE_HOOK(MATH_STATE_RESTORE);
USE_HOOK(DO_PAGE_FAULT);
USE_HOOK(DPROBES_CALLK);


static int do_break_exit(struct hook *h, unsigned long break_num, struct pt_regs * regs)
{
	if (break_num == DP_BREAK_NUM) {
		if (dp_trap(regs, DP_PROBE_BREAKPOINT))
			return HOOK_RETURN;
	}
	return HOOK_CONTINUE;
}

static int do_illop_exit(struct hook *h, struct pt_regs * regs)
{
	dp_handle_fault(regs);
	return HOOK_CONTINUE;
}

static int do_debug_exit(struct hook *h, unsigned long vector, struct pt_regs * regs)
{
	switch (vector) {
			case 29:
			  if (dp_trap(regs, DP_PROBE_WATCHPOINT))
					return HOOK_RETURN;
			  break;
			case 36:
			  if (dp_ss_fault(regs)) {
					/* only dprobes is single stepping across
					   the probe point so we just return */
					return HOOK_RETURN;
			  }
			  /* somebody else single stepping across the probe
				 point so we continue normal processing */
			  break;
			default:
			  dp_handle_fault(regs);
			  break;
	}
	return HOOK_CONTINUE;
}

static int do_page_fault_exit(struct hook *h, struct pt_regs * regs)
{
	if (dp_pf(regs))
		return HOOK_RETURN;
	return HOOK_CONTINUE;
}


struct hook_rec do_illop_hook_rec;
struct hook_rec do_debug_hook_rec;
struct hook_rec do_break_hook_rec;
struct hook_rec do_page_fault_hook_rec;

int register_asm_dprobes_hooks(void)
{
	int retval;
	
	do_illop_hook_rec.hook_exit = do_illop_exit;
	if ((retval = hook_exit_register(&DO_ILLOP, &do_illop_hook_rec)))
		goto err;

	do_debug_hook_rec.hook_exit = do_debug_exit;
	if ((retval = hook_exit_register(&DO_DEBUG, &do_debug_hook_rec)))
		goto err1;

	do_break_hook_rec.hook_exit = do_break_exit;
	if ((retval = hook_exit_register(&DO_BREAK, &do_break_hook_rec)))
		goto err2;

	do_page_fault_hook_rec.hook_exit = do_page_fault_exit;
	if ((retval = hook_exit_register(&DO_PAGE_FAULT, &do_page_fault_hook_rec)))
		goto err3;
	return 0;
	
err3:	hook_exit_deregister(&do_break_hook_rec);
err2:	hook_exit_deregister(&do_debug_hook_rec);
err1:	hook_exit_deregister(&do_illop_hook_rec);
err:	return retval;

}

int deregister_asm_dprobes_hooks(void)
{
	hook_exit_deregister(&do_page_fault_hook_rec);
	hook_exit_deregister(&do_break_hook_rec);
	hook_exit_deregister(&do_debug_hook_rec);
	hook_exit_deregister(&do_illop_hook_rec);
	return 0;
}

int arm_all_asm_dprobes_hooks(void)
{
	hook_exit_arm(&do_illop_hook_rec);
	hook_exit_arm(&do_debug_hook_rec);
	hook_exit_arm(&do_break_hook_rec);
	hook_exit_arm(&do_page_fault_hook_rec);
	
	return 0;
}

int disarm_all_asm_dprobes_hooks(void)
{
	hook_exit_disarm(&do_page_fault_hook_rec);
	hook_exit_disarm(&do_break_hook_rec);
	hook_exit_disarm(&do_debug_hook_rec);
	hook_exit_disarm(&do_illop_hook_rec);
	return 0;
}

void dprobes_hooks_asm_code_end(void) { }
