/*
 *  linux/drivers/s390/misc/z90main.c
 *
 *  z90crypt 1.1.1
 *
 *  Copyright (C)  2001-2002 IBM Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * The functions:
 *     init_module
 *     cleanup_module
 *
 */

#include <linux/interrupt.h>   // for tasklets
#include <linux/spinlock.h>
#include <linux/config.h>      // to determine whether 64-bit
#include <linux/random.h>      // get_random_bytes
#include <linux/delay.h>       // udelay
#include <linux/stddef.h>      // NULL
#include <linux/errno.h>       // some error numbers
#include <linux/slab.h>        // kmalloc
#include <linux/random.h>      // get_random_bytes
#include <asm/uaccess.h>       // copy_from_user
#include <linux/ioctl.h>       // to make z90crypt.h work properly
#include <linux/sched.h>       // for wait_event
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/list.h>
#include <linux/poll.h>
#include "z90crypt.h"
#include "z90common.h"

#define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0) /* version < 2.4 */
#  error "This kernel is too old: not supported"
#endif
#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0) /* version > 2.5 */
#  error "This kernel is too recent: not supported by this file"
#endif
 
#ifdef Z90CRYPT_DEBUG
/*
 * If we're debugging, don't update the module usage count.
 *
 * Instead define them as no-ops.
 */
#undef MOD_INC_USE_COUNT
#define MOD_INC_USE_COUNT       do { } while (0)
#undef MOD_DEC_USE_COUNT
#define MOD_DEC_USE_COUNT       do { } while (0)
#endif

#include <linux/errno.h>   /* error codes */
#include <linux/proc_fs.h>

/****************************************************************************
 * Ioctl definitions  See z90crypt.h                                        *
 ****************************************************************************/
/*****************************************************************************
* Compiled defaults that may be modified in one of several ways:             *
* 1) at compile time:                                                        *
*    + in this file                                                          *
*    + on the command line                                                   *
*    + in the Makefile                                                       *
* 2) on the insmod command line                                              *
*****************************************************************************/

/*
 * Use a dynamic major device number by default
 *
 * You can specify a different one at compile time.
 */
#ifndef Z90CRYPT_MAJOR
#define Z90CRYPT_MAJOR  0
#endif

/*
 * This is the default domain index.
 *
 * You can specify a different one at compile time or on the insmod command
 * line.
 */
#ifndef DOMAIN_INDEX
#define DOMAIN_INDEX   -1
#endif

/*
 * This is the name under which the device is registered in /proc/modules.
 */
#define REG_NAME        "z90crypt"

/*****************************************************************************
* Compiled defaults that should not be changed unless you are quite sure     *
* you know what you are doing.                                               *
*****************************************************************************/

/*
 * errno definitions
 */
#define ENOBUFF        129     /* filp->private_data->...>work_elem_p->buffer
                                  is NULL                                   */
#define EWORKPEND      130     /* user issues ioctl while another pending   */
#define ERELEASED      131     /* user released while ioctl pending         */
#define EQUIESCE       132     /* z90crypt quiescing (no more work allowed) */
#define ETIMEOUT       133     /* request timed out                         */
#define EUNKNOWN       134     /* some unrecognized error occured           */
#define EGETBUFF       135     // Error getting buffer

/*
 * Cleanup should run every CLEANUPTIME seconds and should clean up requests
 * older than CLEANUPTIME seconds in the past.
 */
#ifndef CLEANUPTIME
#define CLEANUPTIME 15
#endif

/*
 * Config should run every CONFIGTIME seconds
 */
#ifndef CONFIGTIME
#define CONFIGTIME 30
#endif

// The first execution of the config task should take place
// immediately after initialization

#ifndef INITIAL_CONFIGTIME
#define INITIAL_CONFIGTIME 1
#endif

/*
 * Reader should run every READERTIME milliseconds
 */
#ifndef READERTIME
#define READERTIME 2
#endif

/*
 * Macros to split minor device numbers into two parts
 */
#define TYPE(dev)   (MINOR(dev) >> 6)   /* high 2 bits */
#define NUM(dev)    (MINOR(dev) & 0x3f) /* low  6 bits */
/*
 * Macros
 *
 * turn long device array index into device pointer
 * turn short device array index into long device array index
 * turn short device array index into device pointer
 */
#define LONG2DEVPTR(ndx) (z90crypt.z90c_device_p[(ndx)])
#define SHRT2LONG(ndx)\
                     (z90crypt.z90c_overall_device_x.z90c_device_index[(ndx)])
#define SHRT2DEVPTR(ndx) LONG2DEVPTR(SHRT2LONG(ndx))
/*
 * bits in work_element_t.status[0]
 *
 * high -----------------------> low
 * _________________________________
 * | 7 | 6 | 5 | 4 | 3 | 2 | 1 | 0 |
 * ---------------------------------
 *
 * The default status is 0x00
 */
#define STAT_DEFAULT  0x00 /* STAT_ROUTED bit is off,            */

#define STAT_ROUTED   0x80 /* bit 7: requests get routed to specific device
                                  otherwise, device is determined each write*/
#define STAT_FAILED   0x40 /* bit 6: this bit is set if the request failed
                                     before being sent to the hardware.     */
#define STAT_WRITTEN  0x30 /* bits 5,4: work to be done, not sent to device */
#define STAT_QUEUED   0x20 /* bits 5,4: work has been sent to a device      */
#define STAT_READPEND 0x10 /* bits 5,4: work done, we're returning data now */
#define STAT_NOWORK   0x00 /* bits off: no work on any queue                */


// Audit Trail.  Progress of a Work element
// audit[0]:  Unless noted otherwise, these bits are all set by the process
#define FP_COPYFROM   0x80 // Caller's buffer has been copied to work element
#define FP_BUFFREQ    0x40 // Low Level buffer requested
#define FP_BUFFGOT    0x20 // Low Level buffer obtained
#define FP_SENT       0x10 // Work element sent to a crypto device
                           // Above may be set by process or by reader task
#define FP_PENDING    0x08 // Work element placed on pending queue
                           // Above may be set by process or by reader task
#define FP_REQUEST    0x04 // Work element placed on request queue
#define FP_ASLEEP     0x02 // Work element about to sleep
#define FP_AWAKE      0x01 // Work element has been awakened
// audit[1]:  These bits are set by the reader task and/or the cleanup task
#define FP_NOTPENDING 0x80 // Work element removed from pending queue
#define FP_AWAKENING  0x40 // Caller about to be awakened
#define FP_TIMEDOUT   0x20 // Caller timed out
#define FP_RESPSIZESET 0x10 // Response size copied to work element
#define FP_RESPADDRCOPIED 0x08 // Response address copied to work element
#define FP_RESPBUFFCOPIED 0x04 // Response buffer copied to work element
#define FP_REMREQUEST 0x02 // Work element removed from request queue
#define FP_SIGNALED   0x01 // Work element was awakened by a signal
// audit[2]:  Reserved

/*
 * masks of bits in work_element_t.status[0]
 */
#define STAT_RDWRMASK 0x30 /* mask to determine caller's current state      */
#define CHECK_RDWRMASK(statbyte) ((statbyte) & STAT_RDWRMASK)
#define SET_RDWRMASK(statbyte, newval) \
  {(statbyte) &= ~STAT_RDWRMASK; (statbyte) |= newval;}
/*
 * Constants to indicate the state of the file handle for
 * this device in private_data.status
 */
#define STAT_OPEN 0
#define STAT_CLOSED 1

/*
 * PID() expands to the process ID of the current process
 */
#define PID() (current->pid)

//
// BUFFERSIZE is the maximum size request buffer passed in an ioctl call
//
#define BUFFERSIZE 128

/*
 * lowlevel definitions
 */

/*-------------------------------------------------------------------*/
/* Selected Constants.  The number of APs and the number of devices  */
/*-------------------------------------------------------------------*/
#ifndef Z90CRYPT_NUM_APS
#define Z90CRYPT_NUM_APS 64
#endif
#ifndef Z90CRYPT_NUM_CCFS
#define Z90CRYPT_NUM_CCFS 2
#endif
#ifndef Z90CRYPT_NUM_DEVS
#define Z90CRYPT_NUM_DEVS Z90CRYPT_NUM_APS
#endif
#ifndef Z90CRYPT_NUM_TYPES
#define Z90CRYPT_NUM_TYPES 2
#endif
#ifndef DEFAULT_DEVICE_TYPE
#define DEFAULT_DEVICE_TYPE 0
#endif
#ifndef MAX_RESPONSE_SIZE
#define MAX_RESPONSE_SIZE 0x00000710
#endif

/*-------------------------------------------------------------------*/
/* z90c_status_str                                                   */
/*                                                                   */
/* A count and status-byte mask                                      */
/*                                                                   */
/*-------------------------------------------------------------------*/
typedef struct z90c_status_str {
  int z90c_st_count;             // number of enabled bits in mask
  int z90c_disabled_count;       // number of disabled devices
  UCHAR z90c_st_mask[Z90CRYPT_NUM_APS]; //***
} Z90C_STATUS_STR;

/*-------------------------------------------------------------------*/
/* The array of device indexes is a mechanism for fast indexing into */
/* a long (and sparse) array.  For instance, if APs 3, 9 and 47 are  */
/* installed, z90CDeviceIndex[0] is 3, z90CDeviceIndex[1] is 9, and  */
/* z90CDeviceIndex[2] is 47.                                         */
/*-------------------------------------------------------------------*/
typedef struct z90c_device_x {
  int  z90c_device_index[Z90CRYPT_NUM_DEVS];
} Z90C_DEVICE_X;

/*-------------------------------------------------------------------*/
/* z90crypt is the topmost data structure in the hierarchy.          */
/*                                                                   */
/*-------------------------------------------------------------------*/
typedef struct z90crypt {
  int   z90c_max_count;            // Number of possible crypto devices
  struct z90c_status_str z90c_mask; // Hardware Status array for all APs
  int   z90c_q_depth_array[Z90CRYPT_NUM_DEVS]; // queue depths
  struct z90c_device_x z90c_overall_device_x; // array device indexes
  struct z90c_device *z90c_device_p[Z90CRYPT_NUM_DEVS]; // Addr array
  UCHAR z90c_rsvd[8];              // Reserved
  UCHAR z90c_terminating;          // 1:  terminating
  BOOL  z90c_domain_established;   // TRUE:  domain has been found
  UCHAR z90c_stat[1];              // Reserved for add'l status info
  int   z90c_cdx;                  // Crypto Domain Index
  int   z90c_len;                  // Length of this data structure
  struct z90c_hdware_block * z90c_hdware_info;
} Z90CRYPT;

/*-------------------------------------------------------------------*/
/* z90c_hdware_block                                                 */
/*   There's a status_str and a device_index_array for each          */
/*   device type.  When created, this block will contain:            */
/*                                                                   */
/*     +----------------------...-----------------+                  */
/*     |   mask of available devices              |                  */
/*     +----+-------------------------------------+                  */
/*     |   2|                                                        */
/*     +----+-----------------...-----------------+                  */
/*     |   mask of available Leeds2s              |                  */
/*     +----------------------...-----------------+                  */
/*     |   mask of available LeedsLites           |                  */
/*     +----------------------...-----------------+                  */
/*     |   index array of available LeedsLites    |                  */
/*     +----------------------...-----------------+                  */
/*     |   index array of available Leeds2s       |                  */
/*     +----------------------...-----------------+                  */
/*     |   mask of device types                   |                  */
/*     +----------------------...-----------------+                  */
/*                                                                   */
/*-------------------------------------------------------------------*/
typedef struct z90c_hdware_block {
  struct z90c_status_str z90c_hdware_mask;
  int z90c_nr_types;
  struct z90c_status_str z90c_type_mask[Z90CRYPT_NUM_TYPES];
  struct z90c_device_x z90c_type_x_addr[Z90CRYPT_NUM_TYPES];
  UCHAR  z90c_device_type_array[Z90CRYPT_NUM_APS]; //***
} Z90C_HDWARE_BLOCK;

/*-------------------------------------------------------------------*/
/* The queue depth array has the maximum number of elements in the   */
/* work queue for a given device.                                    */
/*-------------------------------------------------------------------*/
typedef struct z90c_q_depth_array {
  int z90c_queue_depth[Z90CRYPT_NUM_DEVS];
} Z90C_Q_DEPTH_ARRAY;

/*-------------------------------------------------------------------*/
/* All devices are arranged in a single array: 64 APs (Leeds2 and/or */
/* LeedsLites).                                                      */
/*-------------------------------------------------------------------*/
typedef struct z90c_device {
  CDEVICE_TYPE z90c_dev_type;      // 0 (LEEDSLITE) or 1 (LEEDS2)
  DEVSTAT z90c_dev_stat;           // current device status
  int    z90c_dev_self_x;          // Index in array
  int    z90c_dev_len;             // Length of this struct
  BOOL   z90c_disabled;            // Set when device is in error
  char   z90c_dev_sn[8];           // Reserved
  int    z90c_dev_cdx;             // Crypto Domain Index
  int    z90c_dev_q_depth;         // q depth
  UCHAR *z90c_dev_resp_p;          // Response buffer address
  int    z90c_dev_resp_l;          // Response Buffer length
  UCHAR  z90c_dev_stat_info[4];    // Reserved for additional stat info
  int    z90c_dev_caller_count;    // Number of callers
  struct list_head z90c_dev_caller_list; // List of callers 
} Z90C_DEVICE;

/*-------------------------------------------------------------------*/
/* z90c_caller                                                       */
/*                                                                   */
/* An array of these structures is pointed to from z90c_dev_callr    */
/* The length of the array depends on the device type. For APs       */
/* (Leeds2 and LeedsLite) there are 8.                               */
/*                                                                   */
/* The z90c_request is permanently associated with this descriptor.  */
/* The z90c_caller_buf is allocated to the user at OPEN.  At  WRITE, */
/* it contains the request; at READ, the response.   The function    */
/* sendToCryptoDevice will convert the request to device-dependent   */
/* form and use the caller's OPEN-allocated buffer for the response. */
/*                                                                   */
/*-------------------------------------------------------------------*/
typedef struct z90c_buffer {
  int z90c_buff_len;
  UCHAR * z90c_buff;
} Z90C_BUFFER;

typedef struct z90c_caller {
  int    z90c_caller_buf_l;        // length of original request
  UCHAR *z90c_caller_buf_p;        /* Original request on WRITE      */
  int    z90c_caller_dev_dep_req_l; // length of device dependent requ
  UCHAR *z90c_caller_dev_dep_req_p; // Device dependent form of requ
  UCHAR  z90c_caller_status[4];    //
  UCHAR  z90c_caller_id[8];        // program supplied message id
  int    z90c_caller_func;         // mod_expo or crt        
  HWRD   z90c_caller_function;     // decrypt or encrypt
  UCHAR  z90c_pad_rule[8];         // padding rule
  struct z90c_device *z90c_caller_dev_p; // pointer to a Z90C_DEVICE
  struct list_head z90c_caller_liste; // list element
  long   z90c_caller_sen_time;     // time at which req was sent 
  long   z90c_caller_rec_time;     // time at which resp was received
  UCHAR  z90c_caller_dev_dep_req[MAX_RESPONSE_SIZE]; // response buffer
} Z90C_CALLER;
/*-------------------------------------------------------------------*/
/* A generic descriptor is returned by receiveFromCryptoDevice.      */
/* In most cases, it's a caller descriptor, but in case of error it  */
/* may be a device descriptor.                                       */
/*-------------------------------------------------------------------*/
typedef union z90c_descriptor {
  struct z90c_caller clr;
  struct z90c_device dev;
} Z90C_DESCRIPTOR;

/*-------------------------------------------------------------------*/
/* Function Codes.  Because the IOCTL buffer contains no indicator   */
/* of what sort of request this is, a function code is a necessary   */
/* input to sendToCryptoDevice.                                      */
/*                                                                   */
/* These will be found in z90crypt.h                                 */
/*                                                                   */
/*-------------------------------------------------------------------*/

int createZ90crypt (int * cdx_array);
/*********************************************************************/
/*                                                                   */
/*         input:  cdx                --  crypto domain index        */
/*                                                                   */
/* implied input:  z90crypt                                          */
/*                                                                   */
/* At start-up, z90crypt calls this method.                          */
/*                                                                   */
/* For example,                                                      */
/*                                                                   */
/*  int rv;                                                          */
/*  Z90CRYPT the_z90crypt;                                           */
/*  int crypto_domain_index=0;                                       */
/*                                                                   */
/*  rv = createZ90crypt (&crypto_domain_index);                      */
/*                                                                   */
/*********************************************************************/

int refreshZ90crypt (int * cdx_p);
/*********************************************************************/
/*                                                                   */
/*        inputs:  cdx_p    --   address of space to return cdx      */
/*                                                                   */
/* implied input:  z90crypt                                          */
/*                                                                   */
/* At start-up, and periodically thereafter, z90crypt calls this     */
/* method to revise the device status mask and device address array. */
/*                                                                   */
/*********************************************************************/

int findCryptoDevices(Z90C_STATUS_STR *  deviceMask);
/*********************************************************************/
/*                                                                   */
/*         inputs:  none                                             */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  deviceMask                                       */
/*                                                                   */
/* The function returns the number and hardware status of any online */
/* crypto devices.  It revises the queue depth array.  It's called   */
/* by refreshZ90crypt.                                               */
/*                                                                   */
/*********************************************************************/

int createCryptoDevice(int index);
/*********************************************************************/
/*                                                                   */
/*         inputs:  index                                            */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  deviceType                                       */
/*                                                                   */
/* implied output:  the z90crypt  the device descriptor is           */
/*                                allocated and pointed to from      */
/*                                z90crypt.z90c_device_p[index]      */
/*                                                                   */
/* After calling findCryptoDevices, refreshZ90crypt calls this       */
/* function for any device that has just come online.  In this       */
/* release, the returned deviceType will always be 0 (LeedsLite).    */
/*                                                                   */
/*********************************************************************/

int destroyCryptoDevice(int index);
/*********************************************************************/
/*                                                                   */
/*         inputs:  index                                            */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  none                                             */
/*                                                                   */
/* implied output:  the z90crypt  the device descriptor is           */
/*                                unallocated and the pointer        */
/*                                z90crypt.z90c_device_p[index]      */
/*                                is nullified.                      */
/*                                                                   */
/*                                the device status becomes DEV_GONE.*/
/*                                                                   */
/*                                the device queue-depth is zeroed.  */
/*                                                                   */
/* After calling findCryptoDevices, z90crypt calls this function for */
/* any device that is no longer attached.                            */
/*                                                                   */
/*********************************************************************/

int getCryptoRequestBuffer(UCHAR * psmid,
                     int func,
                     int buff_len,
                     UCHAR * buff_ptr,
                     CDEVICE_TYPE * devType_p,
                     unsigned char * reqBuffp);
/*********************************************************************/
/*                                                                   */
/*         inputs:  psmid -- an 8 byte identifier                    */
/*                                                                   */
/*                  func -- ICARSAMODEXPO or ICARSACRT as defined in */
/*                          z90crypt.h                               */
/*                                                                   */
/*                  buff_len -- length of caller's buffer            */
/*                                                                   */
/*                  buff_ptr -- address of caller's buffer           */
/*                                                                   */
/*                  *devType_p -- 0 (LeedsLite) or 1 (Leeds2)  or -1 */
/*                                (no preference)                    */
/*                                                                   */
/*                  *devNr_p   -- a device number from 0 to 63 or -1 */
/*                                (no preference)                    */
/*                                                                   */
/*                  *reqBuffp  -- address of a pointer to the        */
/*                                request buffer                     */
/*                                                                   */
/* The caller's buffer should contain an instance of either of two   */
/* types in z90crypt.h:                                              */
/*                                                                   */
/* 1. ica_rsa_modexpo_crt_t                                          */
/*                                                                   */
/*    char         *inputdata;                                       */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *bp_key;                                          */
/*    char         *bq_key;                                          */
/*    char         *np_prime;                                        */
/*    char         *nq_prime;                                        */
/*    char         *u_mult_inv;                                      */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*  inputdatalength is even and inputdata contains no leading zeroes */
/*                                                                   */
/*  bp_keylength = np_primelength = u_mult_inv_length =              */
/*                                             8 + inputdatalength/2 */
/*  and                                                              */
/*                                                                   */
/*  bq_keylength = nq_primelength = inputdatalength/2                */
/*                                                                   */
/* 2. ica_rsa_modexpo_t                                              */
/*                                                                   */
/*    char         *inputdata                                        */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *b_key;                                           */
/*    char         *n_modulus;                                       */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*         b_keylength = n_moduluslength = inputdatalength           */
/*                                                                   */
/*                                                                   */
/* Note that in both types, all key parts are right justified in     */
/* their fields, padded on the left with zeroes.                     */
/*                                                                   */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  *devType_p -- 0 (LeedsLite) or 1 (Leeds2) to     */
/*                                which the request was sent.        */
/*                                                                   */
/*                  *devNr_p -- a device number from 0 to 63 to      */
/*                              which the request was sent.          */
/*                                                                   */
/*                                                                   */
/* implied output:  None.                                            */
/*                                                                   */
/*********************************************************************/

int sendToCryptoDevice(UCHAR * psmid,
                     int func,
                     int buff_len,
                     UCHAR * buff_ptr,
                     CDEVICE_TYPE devType,
                     int * devNr_p,
                     unsigned char * reqBuff);
/*********************************************************************/
/*                                                                   */
/*         inputs:  psmid -- an 8 byte identifier                    */
/*                                                                   */
/*                  func -- ICARSAMODEXPO or ICARSACRT as defined in */
/*                          z90crypt.h                               */
/*                                                                   */
/*                  buff_len -- length of caller's buffer            */
/*                                                                   */
/*                  buff_ptr -- address of caller's buffer           */
/*                                                                   */
/*                  *devType_p -- 0 (LeedsLite) or 1 (Leeds2)  or -1 */
/*                                (no preference)                    */
/*                                                                   */
/*                  *devNr_p   -- a device number from 0 to 63 or -1 */
/*                                (no preference)                    */
/*                                                                   */
/* The caller's buffer should contain an instance of either of two   */
/* types in z90crypt.h:                                              */
/*                                                                   */
/* 1. ica_rsa_modexpo_crt_t                                          */
/*                                                                   */
/*    char         *inputdata;                                       */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *bp_key;                                          */
/*    char         *bq_key;                                          */
/*    char         *np_prime;                                        */
/*    char         *nq_prime;                                        */
/*    char         *u_mult_inv;                                      */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*  inputdatalength is even and inputdata contains no leading zeroes */
/*                                                                   */
/*  bp_keylength = np_primelength = u_mult_inv_length =              */
/*                                             8 + inputdatalength/2 */
/*  and                                                              */
/*                                                                   */
/*  bq_keylength = nq_primelength = inputdatalength/2                */
/*                                                                   */
/* 2. ica_rsa_modexpo_t                                              */
/*                                                                   */
/*    char         *inputdata                                        */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *b_key;                                           */
/*    char         *n_modulus;                                       */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*         b_keylength = n_moduluslength = inputdatalength           */
/*                                                                   */
/*                                                                   */
/* Note that in both types, all key parts are right justified in     */
/* their fields, padded on the left with zeroes.                     */
/*                                                                   */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  *devType_p -- 0 (LeedsLite) or 1 (Leeds2) to     */
/*                                which the request was sent.        */
/*                                                                   */
/*                  *devNr_p -- a device number from 0 to 63 to      */
/*                              which the request was sent.          */
/*                                                                   */
/*                                                                   */
/* implied output:  None.                                            */
/*                                                                   */
/*********************************************************************/

int receiveFromCryptoDevice(int index,
                            UCHAR * psmid,
                            int * buff_len_p,
                            UCHAR * buff,
                            UCHAR ** dest);              // @rhbA
/*********************************************************************/
/*                                                                   */
/*         inputs:  index -- index of the device from which          */
/*                           the work element is to be received      */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  psmid     -- 8 byte identifier                   */
/*                                                                   */
/*                  *buff_len_p -- length of buffer                  */
/*                                                                   */
/*                  buff      -- holding buffer in kernel space      */
/*                               for output                          */
/*                                                                   */
/*                  *dest     -- address of output buffer in user    */
/*                               space                               */
/*                                                                   */
/*                               When control returns, the caller's  */
/*                               results will have been copied to    */
/*                               buff, the length of the result      */
/*                               will be stored in *buff_len_p, and  */
/*                               the user-space destination address  */
/*                               will be in *dest.                   */
/*                                                                   */
/*********************************************************************/

int queryZ90crypt(Z90C_STATUS_STR * z90cryptStatus);

/*********************************************************************/
/*                                                                   */
/*         inputs:  none                                             */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*         output:  z90cryptStatus                                   */
/*                                                                   */
/* The output is a count/mask pair.                                  */
/*                                                                   */
/*********************************************************************/

DEVSTAT queryCryptoDevice(int index);
/*********************************************************************/
/*                                                                   */
/*         inputs:  an index                                         */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*         output:  DEV_GONE, DEV_ONLINE, aut ceterum.               */
/*                                                                   */
/*                                                                   */
/*********************************************************************/

int destroyZ90crypt();
/*********************************************************************/
/*                                                                   */
/*  implied input:  the z90crypt                                     */
/*                                                                   */
/*         output:  none                                             */
/*                                                                   */
/* All work will be drained.  All structures pointed to from the     */
/* z90crypt will be unallocated.                                     */
/*                                                                   */
/*********************************************************************/

int refreshIndexArray(struct z90c_status_str *status_str,
                      struct z90c_device_x *index_array);

int selectDevice (short func, CDEVICE_TYPE *devType_p , int *device_nr_p);

int selectDeviceType (short func, CDEVICE_TYPE *devType_p);

int getTestMsg(UCHAR *, UINT *);
int testReply(UCHAR *);
int probeDeviceType(Z90C_DEVICE *);
int probeCryptoDomain();
int getZ90cryptStatus(ica_z90_status * sp, int sl);
int isPKCS1_2padded(unsigned char *, int);
int isPkcs1_1Padded(unsigned char *, int);

/*
 * global pointer to the main data structure (defined in main.c)
 * global boolean indicating the device is shutting down
 * global pointer to the proc_dir_entry struct
 * global pointer to the head of the request queue and its semaphore
 * global pointer to the head of the pending queue and its semaphore
 * global pointer to the head of the status struct and its semaphore
 * global timer struct for the reader task scheduling
 * global timer struct for the config task scheduling
 * global timer struct for the cleanup task scheduling
 */

//
// MD5 definitions, structs, macros
// 

//
// The md5 context is used to compute pseudo-random numbers
//
typedef struct md5_ctx {
  unsigned int i[2];               // number of bits mod 2^64
  unsigned int buf[4];             // scratch buffer
  unsigned char in[64];            // input buffer
  unsigned char digest[16];        // output digest
} MD5_CTX;

/******************************************************************************/
/* Rotate a word (32 bits) left by a specified number of bits.  The 32-bit    */
/* number invalue is circularly rotated left by num_bits bit positions.  The  */
/* result is returned as the function result.                                 */
/*----------------------------------------------------------------------------*/
#define rotate_left(Data, bit_cnt)              \
        (Data = ( (Data << bit_cnt) |           \
                  (Data >> (32 - bit_cnt)) ) )

/******************************************************************************/
/* Implement the MD5 algorithm's "F" function.  This function performs a      */
/* transform on three input words, designated x, y, and z, producing a        */
/* single word output value.  The transform is:                               */
/*                                                                            */
/*    output = ( x AND y ) OR ( (NOT x) AND z )                               */
/*----------------------------------------------------------------------------*/
#define F( x, y, z) ( ( x & y ) | ( (~x) & z) )

/******************************************************************************/
/* Implement the MD5 algorithm's "G" function.  This function performs a      */
/* transform on three input words, designated x, y, and z, producing a        */
/* single word output value.  The transform is:                               */
/*                                                                            */
/*    output = ( x AND z ) OR ( y AND (NOT z) )                               */
/*----------------------------------------------------------------------------*/
#define G( x, y, z) ( ( x & z ) | ( y & (~z) ) )

/******************************************************************************/
/* Implement the MD5 algorithm's "H" function.  This function performs a      */
/* transform on three input words, designated x, y, and z, producing a        */
/* single word output value.  The transform is:                               */
/*                                                                            */
/*    output = ( x XOR y XOR z )                                              */
/*----------------------------------------------------------------------------*/
#define H( x, y, z) ( x ^ y ^ z )

/******************************************************************************/
/* Implement the MD5 algorithm's "I" function.  This function performs a      */
/* transform on three input words, designated x, y, and z, producing a        */
/* single word output value.  The transform is:                               */
/*                                                                            */
/*    output = ( y XOR ( x OR (NOT z) ) )                                     */
/*----------------------------------------------------------------------------*/
#define I( x, y, z) ( y ^ ( x | (~z) ) )
#define MD5_HASH_SIZE 16
#define MD5_INPUT_SIZE 64
void md5_init(MD5_CTX * pCtx);
void md5_transform(unsigned int *, unsigned int *);
int md5_update(MD5_CTX *pCtx,
               unsigned char * indata,
               unsigned int indata_len);
int md5_final(MD5_CTX * pCtx,
              unsigned char * otdata,
              unsigned int * pOtdata_len);
int md5_hash(MD5_CTX * pCtx,
             unsigned char * indata,
             unsigned int indata_len,
             unsigned char * otdata,
             unsigned int * pOtdata_len);

/****************************************************************************
 * structure prototypes                                                     *
 ****************************************************************************/

/*
 * filp->private_data points to this structure
 * work_element.opener points back to this structure
 *
 * We use it to keep a point to the device handling our request
 * as well as some record-keeping information about this request
 */

typedef struct private_data {
  pid_t               opener_pid;  // process id of opening process
  int                 type;
  int                 num;
  unsigned char random_seed[16];
  struct work_element *priv_we_p;  // address of work element
  struct md5_ctx      priv_md5;    // imbedded context;
  unsigned char       status;      // 0: open  1: closed
} private_data_t;         
  
  
// filp->private_data->free_stack points to the following structure
// work_element.next_free points to the following structure
 
typedef struct work_element
{
  struct private_data  *opener_p;   // private data of opening pid       
  pid_t                pid;         // pid of this work element
  int                  devindex;    // index of device processing this w_e
                                    // (If request did not specify device,
                                    // -1 until placed onto a queue)     
  CDEVICE_TYPE         devtype;     // device type
  struct list_head     liste;       // used for requestq and pendingq
  char          buffer[BUFFERSIZE]; // local copy of user request      
  int                  buffsize;    // size of the buffer for the request 
  char      respbuff[RESPBUFFSIZE]; // response buffer
  int                  respsize;    // size of the response buffer
  char                 *respaddr;   // address of response in user space
  unsigned int         funccode;    // function code of request
  wait_queue_head_t    waitq;       // waitq for this w_e
  unsigned long        requestsent; // time at which the request was sent
  atomic_t             alarmrung;   // wake-up signal
  unsigned char        caller_id[8];// pid, followed by counter, for this w_e
  unsigned char        status[1];   // bits to indicate status of the request
  unsigned char        audit[3];    // record of work element's progress
  unsigned char      * requestptr;  // address of request buffer
  int                  returncode;  // return code of request
} work_element_t;

/****************************************************************************
 * function prototypes                                                      *
 ****************************************************************************/
int z90crypt_open(struct inode *, struct file *);
int z90crypt_release(struct inode *, struct file *);

ssize_t z90crypt_read(struct file *, char *, size_t, loff_t *);
ssize_t z90crypt_write(struct file *, const char *, size_t, loff_t *);

int z90crypt_ioctl(struct inode *, struct file *, unsigned int,unsigned long);
loff_t z90crypt_llseek(struct file *, loff_t off, int);
unsigned int z90crypt_poll(struct file *, poll_table *);
int z90crypt_rsa(private_data_t *, pid_t, unsigned int, unsigned long);
int z90crypt_prepare(work_element_t *, unsigned int, const char *);
int z90crypt_send(work_element_t *, const char *); 
int z90crypt_process_results(work_element_t *, char *);

int z90crypt_schedule_reader_timer(void);
void z90crypt_schedule_reader_task(unsigned long);
int z90crypt_schedule_config_task(unsigned int);
int z90crypt_schedule_initial_config_task(void);
int z90crypt_schedule_cleanup_task(void);

int allocate_work_element(work_element_t **,private_data_t *, pid_t);
int create_work_element(work_element_t **, pid_t);
int init_work_element(work_element_t *, private_data_t *, pid_t);
int free_work_element(work_element_t *);
void destroy_private_data(private_data_t *);
void destroy_work_element(work_element_t *);

int z90crypt_status(char *, char **, off_t, int, int *, void*);
int sprinthx(unsigned char *, unsigned char *, unsigned char *, unsigned int);
int sprintrw(unsigned char *, unsigned char *, unsigned int);
int sprintcl(unsigned char *, unsigned char *, unsigned int);


/*--------------------------------------------------------------------------*/
/* Storage allocated at initialization and used throughout the life of      */
/* this insmod                                                              */
/*--------------------------------------------------------------------------*/
int z90crypt_major = Z90CRYPT_MAJOR;
int domain         = DOMAIN_INDEX; /* this is the default; can be changed
                                      on loading of module by specifying
                                      "domain=xx" on the command line       */
Z90CRYPT z90crypt;                 /* initialized by createZ90crypt call    */
BOOL quiesce_z90crypt;             /* bool indicating device shutting down  */
spinlock_t queuespinlock;
struct list_head request_list;
int requestqCount;
struct list_head pending_list;
int pendingqCount;

void (z90crypt_reader_task)(unsigned long);
struct tasklet_struct reader_tasklet; /* struct for reader task             */
struct timer_list reader_timer;    /* struct for delayed reader_tasklet     */
struct timer_list config_timer;    /* struct for scheduling config task     */
struct timer_list cleanup_timer;   /* struct for scheduling cleanup task    */
atomic_t totalOpen;
atomic_t z90crypt_step;            /* this is the second half of the id field
                                      in the Z90C_CALLER struct.  It is
                                      incremented on each ioctl.             */
struct file_operations z90crypt_fops =
{
  owner:     THIS_MODULE,
  llseek:    z90crypt_llseek,
  read:      z90crypt_read,
  write:     z90crypt_write,
  poll:      z90crypt_poll,
  ioctl:     z90crypt_ioctl,
  open:      z90crypt_open,
  release:   z90crypt_release
};
/*---------------------------------------------------------------------------*/
/* Documentation values.                                                     */
/*---------------------------------------------------------------------------*/
MODULE_AUTHOR("IBM z900 Linux Crypto Team: " \
	      "Robert H. Burroughs and Eric D. Rossman");
MODULE_DESCRIPTION("z90crypt 1.1.1, Copyright 2001, 2002 IBM Corporation");
MODULE_PARM(domain, "i");
MODULE_PARM_DESC(domain, "domain index for device");
MODULE_LICENSE("GPL");

/*---------------------------------------------------------------------------*/
/* The module initialization code.                                           */
/*---------------------------------------------------------------------------*/
int
init_module(void)
{
  int result,   /* return code */
      nresult;

  EXPORT_NO_SYMBOLS;
  PDEBUG("init_module -> PID %d\n", PID());

  //
  // Register the major (or get a dynamic one).
  //
  result = register_chrdev(z90crypt_major, REG_NAME, &z90crypt_fops);
  if (result < 0)
  {
    PRINTKW("register_chrdev (major %d) failed with %d.\n",
            z90crypt_major, result);
    return result;
  }

  PDEBUG("Registered " DEV_NAME " with result %d\n", result);

  /* If z90crypt_major was 0, set it to the dynamically allocated one.  */
  if (z90crypt_major == 0)
    z90crypt_major = result;

  /* allocate the base structure, and initialize it */
  result = createZ90crypt(&domain);
  if (result == Z90C_AMBIGUOUS_DOMAIN) {
    PRINTK("More than one domain defined to this LPAR\n");
  }

  if (result == Z90C_INCORRECT_DOMAIN) {
    PRINTK("Specified domain doesn't match assigned domain\n");
  }

  if (result != OK) {
    PRINTKW("createZ90crypt (domain index %d) failed with %d.\n",
            domain, result);
    result = -ENOMEM;
    goto init_module_cleanup;
  }

  if (result == OK) {
    PRINTK("Using domain index 0x%02X\n", domain);
    PDEBUG("createZ90crypt (domain index %d) successful.\n", domain);
  }
  else {
    PRINTK("No devices at startup\n");
  }

  /*
   * There may be some setting up of semaphores for device types as disabling
   * entire device types may be implemented in the future.
   */

  /*
   * Initialize the config lock, the request queue, pending queue, reader timer,
   * timer, configuration timer, cleanup timer, and boolean indicating that
   * the device is quiescing.
   */
  spin_lock_init(&queuespinlock);

  INIT_LIST_HEAD(&pending_list);
  pendingqCount=0;

  INIT_LIST_HEAD(&request_list);
  requestqCount=0;

  init_timer(&reader_timer);
  tasklet_init(&reader_tasklet,
               z90crypt_reader_task,
               (unsigned long) NULL);
  init_timer(&config_timer);
  init_timer(&cleanup_timer);
  quiesce_z90crypt = FALSE;
  atomic_set(&totalOpen,0);
  atomic_set(&z90crypt_step,0);

  //
  // Set up the cleanup task.
  //
  if ((result = z90crypt_schedule_cleanup_task()))
  {
    //
    // Currently, there is nothing that could go wrong, but in the future,
    // there may be.  This is a catch-all.
    //
    PRINTK("schedule_cleanup_task failed with %d.\n", result);
  }
  else
  {
    PDEBUG("schedule_cleanup_task successful.\n");
  }

  //
  // Set up the configuration task.
  //
  if ((result = z90crypt_schedule_config_task(INITIAL_CONFIGTIME)))
  {
    //
    // Currently, there is nothing that could go wrong, but in the future,
    // there may be.  This is a catch-all.
    //
    PRINTK("schedule_config_task failed with %d.\n", result);
  }
  else
  {
    PDEBUG("schedule_config_task successful.\n");
  }

  //
  // Set up the proc file system
  //
  create_proc_read_entry("driver/z90crypt",
                         0,
                         NULL,
                         z90crypt_status,
                         NULL);

  return 0; // succeed

init_module_cleanup:
  if ((nresult = unregister_chrdev(z90crypt_major, REG_NAME)))
  {
    PRINTK("unregister_chrdev failed with %d.\n", nresult);
  }
  else
  {
    PDEBUG("unregister_chrdev successful.\n");
  }

  return result;
}

/*--------------------------------------------------------------------------*/
/* The module termination code                                              */
/*--------------------------------------------------------------------------*/
void
cleanup_module(void)
{
  int nresult;

  PDEBUG("cleanup_module -> PID %d\n", PID());
  //
  // Remove proc fs code
  //
  remove_proc_entry("driver/z90crypt", NULL);

  /*
   * Unregister the driver.
   */
  if ((nresult = unregister_chrdev(z90crypt_major, REG_NAME)))
  {
    PRINTK("unregister_chrdev failed with %d.\n", nresult);
  }
  else
  {
    PDEBUG("unregister_chrdev successful.\n");
  }

  /*
   * Remove the reader task, configuration task, and cleanup task from the
   * timer queue.
   */
  tasklet_kill(&reader_tasklet);
  del_timer(&reader_timer);
  del_timer(&config_timer);
  del_timer(&cleanup_timer);

  destroyZ90crypt();
}




/*----------------------------------------------------------------------------*/
/* Top Half Functions                                                         */
/*                                                                            */
/* The I/O functions:                                                         */
/*     z90crypt_open                                                          */
/*     z90crypt_close                                                         */
/*     z90crypt_read                                                          */
/*     z90crypt_write                                                         */
/*     z90crypt_llseek                                                        */
/*     z90crypt_poll                                                          */
/*     z90crypt_ioctl                                                         */
/*                                                                            */
/*     z90crypt_status           // proc-fs routine                           */
/*                                                                            */
/* Helper functions:                                                          */
/*     z90crypt_rsa                                                           */
/*       z90crypt_prepare                                                     */
/*       z90crypt_send                                                        */
/*       z90crypt_process_results                                             */
/*                                                                            */
/*       sprinthx                                                             */
/*       sprintrw                                                             */
/*       sprintcl                                                             */
/*                                                                            */
/* the fops struct is built here                                              */
/*                                                                            */
/* read() will return a string of random bytes.                               */
/* write() will fail.                                                         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/* Statics needed by MD5 hashing used for pseudo-random number generation.    */
/*                                                                            */
/* Define PADDING.  PADDING is used if the input data to md5_hash doesn't     */
/* number exactly 64 bytes.                                                   */
/*                                                                            */
/* Define the MD5 "T[]" table.  This table consists of 64 4-byte (MD5_word)   */
/* entries, designated T[1] through T[64].  (Note that this is different from */
/* the way C will index them, from 0..63 instead of 1..64.)                   */
/*                                                                            */
/* If the index of each entry is i, where i ranges from 1 to 64, then the     */
/* value in each entry is given by the following formula.                     */
/*                                                                            */
/*      T[i] = int ( 4294967296 * abs ( sin ( i ) ) )                         */
/*                                                                            */
/* where the function sin(i) expects i in radians, and the function int(x)    */
/* returns the integer portion of a floating point number x.                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/

static unsigned char PADDING[64] = {
  0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static unsigned int T[64] = {
                   0xD76AA478,         /* T[01]                           */
                   0xE8C7B756,         /* T[02]                           */
                   0x242070DB,         /* T[03]                           */
                   0xC1BDCEEE,         /* T[04]                           */
                   0xF57C0FAF,         /* T[05]                           */
                   0x4787C62A,         /* T[06]                           */
                   0xA8304613,         /* T[07]                           */
                   0xFD469501,         /* T[08]                           */
                   0x698098D8,         /* T[09]                           */
                   0x8B44F7AF,         /* T[10]                           */
                   0xFFFF5BB1,         /* T[11]                           */
                   0x895CD7BE,         /* T[12]                           */
                   0x6B901122,         /* T[13]                           */
                   0xFD987193,         /* T[14]                           */
                   0xA679438E,         /* T[15]                           */
                   0x49B40821,         /* T[16]                           */
                   0xF61E2562,         /* T[17]                           */
                   0xC040B340,         /* T[18]                           */
                   0x265E5A51,         /* T[19]                           */
                   0xE9B6C7AA,         /* T[20]                           */
                   0xD62F105D,         /* T[21]                           */
                   0x02441453,         /* T[22]                           */
                   0xD8A1E681,         /* T[23]                           */
                   0xE7D3FBC8,         /* T[24]                           */
                   0x21E1CDE6,         /* T[25]                           */
                   0xC33707D6,         /* T[26]                           */
                   0xF4D50D87,         /* T[27]                           */
                   0x455A14ED,         /* T[28]                           */
                   0xA9E3E905,         /* T[29]                           */
                   0xFCEFA3F8,         /* T[30]                           */
                   0x676F02D9,         /* T[31]                           */
                   0x8D2A4C8A,         /* T[32]                           */
                   0xFFFA3942,         /* T[33]                           */
                   0x8771F681,         /* T[34]                           */
                   0x6D9D6122,         /* T[35]                           */
                   0xFDE5380C,         /* T[36]                           */
                   0xA4BEEA44,         /* T[37]                           */
                   0x4BDECFA9,         /* T[38]                           */
                   0xF6BB4B60,         /* T[39]                           */
                   0xBEBFBC70,         /* T[40]                           */
                   0x289B7EC6,         /* T[41]                           */
                   0xEAA127FA,         /* T[42]                           */
                   0xD4EF3085,         /* T[43]                           */
                   0x04881D05,         /* T[44]                           */
                   0xD9D4D039,         /* T[45]                           */
                   0xE6DB99E5,         /* T[46]                           */
                   0x1FA27CF8,         /* T[47]                           */
                   0xC4AC5665,         /* T[48]                           */
                   0xF4292244,         /* T[49]                           */
                   0x432AFF97,         /* T[50]                           */
                   0xAB9423A7,         /* T[51]                           */
                   0xFC93A039,         /* T[52]                           */
                   0x655B59C3,         /* T[53]                           */
                   0x8F0CCC92,         /* T[54]                           */
                   0xFFEFF47D,         /* T[55]                           */
                   0x85845DD1,         /* T[56]                           */
                   0x6FA87E4F,         /* T[57]                           */
                   0xFE2CE6E0,         /* T[58]                           */
                   0xA3014314,         /* T[59]                           */
                   0x4E0811A1,         /* T[60]                           */
                   0xF7537E82,         /* T[61]                           */
                   0xBD3AF235,         /* T[62]                           */
                   0x2AD7D2BB,         /* T[63]                           */
                   0xEB86D391 };       /* T[64]                           */

/*
 * The different file operations
 */
//
// z90crypt_status -- returns z90crypt status as /proc/driver/z90crypt
//
//
int z90crypt_status(char * respbuff, char **start, off_t offset, int count,
                    int * eof, void * data)
{
  int ret = 0;
  ica_z90_status * stat;
  int len = 0;

  // respbuff is a page.  Use the right half for a work area
  stat = (ica_z90_status *)(respbuff+2000);

  ret = getZ90cryptStatus(stat, 2000);

  if (ret == 0) {
    stat->requestqWaitCount = requestqCount;
    stat->pendingqWaitCount = pendingqCount;
    stat->totalOpenCount = atomic_read(&totalOpen);
    stat->cryptoDomain = z90crypt.z90c_cdx;

    len += sprintf (respbuff+len,"\nz90crypt version: %1d.%1d.%1d\n",
                    z90crypt_VERSION,
                    z90crypt_RELEASE,
                    z90crypt_VARIANT);
    len += sprintf (respbuff+len,"Cryptographic domain: %i\n",stat->cryptoDomain);
    len += sprintf (respbuff+len,"Total device count: %i\n",stat->totalcount);
    len += sprintf (respbuff+len,"PCICA count: %i\n",stat->leedslitecount);
    len += sprintf (respbuff+len,"PCICC count: %i\n",stat->leeds2count);
    len += sprintf (respbuff+len,"requestq count: %i\n",stat->requestqWaitCount);
    len += sprintf (respbuff+len,"pendingq count: %i\n",stat->pendingqWaitCount);
    len += sprintf (respbuff+len,"Total open handles: %i\n\n",stat->totalOpenCount);

    len += sprinthx("Mask of online devices: 01 means PCICA, 02 means PCICC",
           respbuff+len, stat->status,sizeof(stat->status));
    len += sprinthx("Mask of waiting work element counts",
           respbuff+len, stat->qdepth, sizeof(stat->qdepth));

  }
  else {
    len += sprintf(respbuff,"\nStatus cannot be obtained.\n");
  }

  *eof = 1;
  memset(stat, 0, sizeof(ica_z90_status));
  return len;

} // end z90crypt_status

/*
 * z90crypt_open
 */
int
z90crypt_open (struct inode *inode, struct file *filp)
{
  int         type = TYPE(inode->i_rdev),
              num = NUM(inode->i_rdev);
  private_data_t *private_data_p = NULL;
  pid_t       pid;            /* PID of the current task                    */

  /*
   * Are we quiescing?
   */
  if (quiesce_z90crypt)
    return -EQUIESCE;

  /*
   * set up needed variables
   */
  pid = PID();

  PDEBUG("open -> PID %d (filp %p), devtype %d, devnum %d\n",
         pid, filp, type, num);

  /* deal with the "type" of device */
  switch(type)
  {
    case 0: /* non-routed request; num should be zero */
      if (num)
        return -ENODEV;
      break;

    case 1: /* routed request to an AP; num is relevant */
      /* don't do anything */
      break;

      /*
       * We do not support CCFs (1/8/2001)
       */
    case 2: /* routed request to a CCF; low bit of num is relevant */
//    if (num & 0xFE)
        return -ENODEV;
      break;

    default: /* there is no type 3 */
      PDEBUG("PID %d failed to open device type %d, device num %d.\n",
             pid, type, num);
      return -ENODEV;
  }

  /*
   * filp->private_data points to the device data
   *
   * Allocate and initialize the device data
   */
  if (!(private_data_p = kmalloc(sizeof(private_data_t),GFP_KERNEL)))
    return -ENOMEM;
  memset((void *)private_data_p,0,sizeof(private_data_t));
  private_data_p->status = STAT_OPEN;
  private_data_p->opener_pid = pid;
  private_data_p->type = type;
  private_data_p->num = num;

  filp->private_data = private_data_p;

  atomic_inc(&totalOpen);

  return 0;

} /* End of z90crypt_open */


/*
 * z90crypt_release
 */
int
z90crypt_release (struct inode *inode, struct file *filp)
{
  pid_t       pid;            /* PID of the current task                    */
  private_data_t *private_data_p; /* private data for this opener           */
  work_element_t   *work_elem_p;  /* any work element associate w  opener   */
  int        rv = 0;          // return value from free_work_element

  /*
   * set up needed variables
   */
  private_data_p = filp->private_data;
  pid = PID();

  PDEBUG("release -> PID %d (filp %p)\n", pid, filp);

  /*
   * Any outstanding work covered by this handle will complete, except
   * work left by the closing thread itself.
   */
  private_data_p->status = STAT_CLOSED;
  work_elem_p = private_data_p->priv_we_p;

  if (work_elem_p != NULL) {
    if (CHECK_RDWRMASK(work_elem_p->status[0]))
    {
      /*
       * Mark the request as failed, wake up the caller, and clean up.
       */
      PDEBUG("PID %d tried to release filp %p, which has status 0x%02X.  "
             "Failing outstanding request and waking caller.\n",
             pid, filp, work_elem_p->status[0]);
      work_elem_p->returncode = -ERELEASED;
      work_elem_p->status[0] |= STAT_FAILED;
      work_elem_p->audit[1] |= FP_AWAKENING;
      atomic_set(&work_elem_p->alarmrung,1);
      wake_up(&work_elem_p->waitq);
    } // end if (CHECK_...
    /*
     * If we are shutting down normally (not STAT_FAILED/retcode -ERELEASED)
     * Get rid of work element
     */
    if (!((work_elem_p->status[0] & STAT_FAILED) &&
          (work_elem_p->returncode == -ERELEASED)))
      rv = free_work_element(work_elem_p);

  } // endif this work belongs to closer

  /*
   * Clear the private data pointer in filp.
   * If the private_data struct hasn't been freed, it will be
   * by the cleanup task when the last work element has been
   * destroyed.
   */
  filp->private_data = NULL;

  // If the private data structure wasn't destroyed by the
  // last free_work_elem call, see whether it can be destroyed now
  if (rv == 0) {
    destroy_private_data(private_data_p);
  }

  atomic_dec(&totalOpen);

  return 0;
} /* End of z90crypt_release */


/*
 * Data management
 *
 * read() just returns a string of random bytes.  Since we have no way
 * to generate these in an actual device, we just get_random_bytes for
 * a seed and then hash the seed iteratively until we've generated the
 * requisite number of bytes.
 *
 */
ssize_t
z90crypt_read (struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
  pid_t   pid;                 /* PID of the current task                   */
  ssize_t retval = 0;          /* count (if successful), error code (if not)*/
  private_data_t *private_data_p;
  unsigned char * temp_buff = NULL;
  unsigned int temp_len = MD5_HASH_SIZE;
  unsigned char * src = NULL;
  int i,j;
  int out16 = 0;
  int counter = 0;
  int page_count = 0;
  int page_residue = 0;
  unsigned char * tgt = NULL;

  /*
   * Are we quiescing?
   */
  if (quiesce_z90crypt)
    return -EQUIESCE;

  /*
   * set up needed variables
   */
  pid = PID();
  PDEBUG("read -> filp %p (PID %d)\n", filp, pid);
  counter = (int) count;
  if (counter < 0) {
    PRINTK("Requested random byte count is negative!: %i\n", (int) count);
    return -EINVAL;
  }

  if (counter == 0)
    return 0;

  private_data_p = filp->private_data;

  if ((retval = create_work_element((work_element_t **)&temp_buff, pid)) != 0)
    return (retval);

  if (temp_buff == NULL)
    return -ENOMEM;

  if (private_data_p->random_seed[0] == 0) {
    get_random_bytes(private_data_p->random_seed,
                     sizeof(private_data_p->random_seed));
    md5_init(&(private_data_p->priv_md5));
  }

  for (i=0; i<8; i ++)
    memcpy(temp_buff+i*sizeof(private_data_p->random_seed),
           private_data_p->random_seed,
           sizeof(private_data_p->random_seed));

  page_count = ((int)count)/4000;
  page_residue = ((int)count)%4000;
  tgt = buf;

  for (j=0; j<page_count; j++) {
    counter = 4000;
    do {
      // Prime the output buffer
      temp_len = MD5_HASH_SIZE;
      if ((retval = md5_hash(&(private_data_p->priv_md5),
                             temp_buff,
                             64,
                             temp_buff+64,
                             &temp_len)) != 0) {
        retval = -ENOMSG;
        break;
      }

      if (counter <= MD5_HASH_SIZE) {
        break;
      }

      out16 = counter/MD5_HASH_SIZE;

      // Hash the hashes
      src = temp_buff+MD5_HASH_SIZE;
      for (i = 0; i < out16; i++) {
        temp_len = MD5_HASH_SIZE;
        if ((retval = md5_hash(&(private_data_p->priv_md5),
                               src,
                               64,
                               src+64,
                               &temp_len)) != 0) {
          retval = -ENOMSG;
          break;
        }
        src += MD5_HASH_SIZE;
      }

    } while (0);

    if (retval == 0) {
      if ((retval = copy_to_user(tgt, temp_buff+64, counter)) != 0) {
        retval = -EFAULT;
      }
    }

    if (retval == 0) {
      tgt += 4000;
      memcpy (temp_buff, temp_buff+(4000-64), 64);
    }
    else {
      break;
    }

  } // end for

  if (retval == 0 && page_residue != 0) {
    counter = page_residue;
    do {
      // Prime the output buffer
      temp_len = MD5_HASH_SIZE;
      if ((retval = md5_hash(&(private_data_p->priv_md5),
                             temp_buff,
                             64,
                             temp_buff+64,
                             &temp_len)) != 0) {
        retval = -ENOMSG;
        break;
      }

      if (counter <= MD5_HASH_SIZE) {
        break;
      }

      out16 = counter/MD5_HASH_SIZE;

      // Hash the hashes
      src = temp_buff+MD5_HASH_SIZE;
      for (i = 0; i < out16; i++) {
        temp_len = MD5_HASH_SIZE;
        if ((retval = md5_hash(&(private_data_p->priv_md5),
                               src,
                               64,
                               src+64,
                               &temp_len)) != 0) {
          retval = -ENOMSG;
          break;
        }
        src += MD5_HASH_SIZE;
      }

    } while (0);

    if (retval == 0) {
      if ((retval = copy_to_user(tgt, temp_buff+64, counter)) != 0) {
        retval = -EFAULT;
      }
    }

  } // end if

  if (retval < 0)
    memset(private_data_p->random_seed,0,sizeof(private_data_p->random_seed));
  else {
    memcpy(private_data_p->random_seed,
           temp_buff+64,
           sizeof(private_data_p->random_seed));
    retval = (int)count;
  }

  destroy_work_element((work_element_t * )temp_buff);

  return retval;

} // end z90crypt_read

/*
 * Currently, write() does nothing other than return -EPERM (operation not
 * permitted).
 */
ssize_t
z90crypt_write (struct file *filp, const char *buf, size_t count,
                loff_t *f_pos)
{
  pid_t   pid;                 /* PID of the current task                   */
  ssize_t retval = 0;          /* count (if successful), error code (if not)*/

  /*
   * set up needed variables
   */
  pid = PID();

  PDEBUG("write -> filp %p (PID %d)\n", filp, pid);

  retval = -EPERM;

//write_cleanup:

  return retval;
}


/*
 * The "extended" operations -- llseek, poll, and ioctl
 */
loff_t
z90crypt_llseek (struct file *filp, loff_t off, int whence)
{
  pid_t   pid;                 /* PID of the current task                   */
  ssize_t retval = 0;          /* count (if successful), error code (if not)*/

  /*
   * set up needed variables
   */
  pid = PID();

  PDEBUG("llseek -> filp %p (PID %d)\n", filp, pid);

  /*
   * This device is unseekable.
   */
  retval = -ESPIPE;

//llseek_cleanup:

  return retval;
}

/*
 * Polling the device:
 * - You can always read.
 * - You can never write.
 *
 * There is no need to issue poll_wait().  The device is always ready.
 */
unsigned int
z90crypt_poll (struct file *filp, poll_table *wait)
{
  pid_t   pid;                 /* PID of the current task                   */
  unsigned int mask = 0;

  /*
   * set up needed variables
   */
  pid = PID();

  PDEBUG("poll -> filp %p (PID %d)\n", filp, pid);

  mask |= POLLIN | POLLRDNORM;  /* readable */

  return mask;
}


/*
 * The ioctl() implementation
 *
 * ioctl is the primary entry point for this device.
 */
int
z90crypt_ioctl (struct inode *inode, struct file *filp, unsigned int cmd,
                unsigned long arg)
{
  int       size = _IOC_SIZE(cmd), /* size bitfield in cmd                  */
            ret = 0,               /* return code from various calls        */
            i,                     // loop counter
            loopLim;               // loop limit
  work_element_t *work_elem_p;     /* pointer to work element struct        */
  pid_t     pid;                   /* PID of the current task               */
  private_data_t * private_data_p;

  /*
   * set up needed variables
   */

  private_data_p = filp->private_data;
  pid = PID();

  PDEBUG("ioctl -> filp %p (PID %d), cmd 0x%08X\n", filp, pid, cmd);
  PDEBUG("cmd 0x%08X: dir %s, size 0x%04X, type 0x%02X, nr 0x%02X\n",
         cmd,
         !_IOC_DIR(cmd) ? "NO"
           : ((_IOC_DIR(cmd) == (_IOC_READ|_IOC_WRITE)) ? "RW"
             : ((_IOC_DIR(cmd) == _IOC_READ) ? "RD"
               : "WR")),
         _IOC_SIZE(cmd), _IOC_TYPE(cmd), _IOC_NR(cmd));

  /*
   * extract the type bitfield, and don't decode wrong cmds:
   * return ENOTTY (inappropriate ioctl) before access_ok()
   */
  if (_IOC_TYPE(cmd) != ICA_IOCTL_MAGIC)
  {
    PRINTK("cmd 0x%08X contains a bad type\n", cmd);
    return -ENOTTY;
  }

  /*
   * the direction is a bitmask, and VERIFY_WRITE catches R/W
   * transfers. 'type' is user-oriented, while
   * access_ok is kernel-oriented, so the concept of "read" and
   * "write" is reversed
   */
  if (_IOC_DIR(cmd) & _IOC_READ)
    if (!access_ok(VERIFY_WRITE, (void *)arg, size))
      return -EFAULT;
  if (_IOC_DIR(cmd) & _IOC_WRITE)
    if (!access_ok(VERIFY_READ,  (void *)arg, size))
      return -EFAULT;

  switch(cmd)
  {
    //
    // Request for status.
    //
    // arg is pointer to user's buffer.
    //
    // Status is returned in raw binary format.
    //
    case ICAZ90STATUS:
      PDEBUG("in ioctl got ICAZ90STATUS!\n");

      if ((ret = allocate_work_element(&work_elem_p,
                                       private_data_p, pid)) != 0)
        break;

      if ((ret = copy_from_user(work_elem_p->respbuff,
                                (void *)arg,
                                sizeof(ica_z90_status))) != 0){
        free_work_element(work_elem_p);
        ret = -EFAULT;
        break;
      }

      if ((ret = getZ90cryptStatus(
                (ica_z90_status *)work_elem_p->respbuff,
                              RESPBUFFSIZE)) != 0) {
        free_work_element(work_elem_p);
        break;
      }

      ((ica_z90_status *)(work_elem_p->respbuff))->requestqWaitCount =
                                              requestqCount;
      ((ica_z90_status *)(work_elem_p->respbuff))->pendingqWaitCount =
                                              pendingqCount;
      ((ica_z90_status *)(work_elem_p->respbuff))->totalOpenCount =
                                              atomic_read(&totalOpen);
      ((ica_z90_status *)(work_elem_p->respbuff))->cryptoDomain =
                                              z90crypt.z90c_cdx;
      if ((ret = copy_to_user((void *)arg,
                              work_elem_p->respbuff,
                              sizeof(ica_z90_status))) != 0){
        ret = -EFAULT;
      }

      free_work_element(work_elem_p);

      break;

    //
    // Request for device shutdown.
    //
    case ICAZ90QUIESCE:

      PRINTK("QUIESCING device under directions from PID %d.\n", pid);
      if(current->euid != 0)
        ret = -EACCES;
      else
        quiesce_z90crypt = TRUE;
      break;

    //
    // Request for device reset
    //
    case ICAZ90HARDRESET:
#ifdef Z90CRYPT_DEBUG
    #ifdef MODULE
        PRINTK("RESETING device under directions from PID %d.\n", pid);
        if(current->euid != 0)
            ret = -EACCES;
        else
            atomic_set(&__this_module.uc.usecount, 1);
        break;
    #else
        ret = -EPERM;	
        break;
    #endif // MODULE
#else
    ret = -ENOTTY;
    break;
#endif	// Z90CRYPT_DEBUG

    //
    // Request for RSA Clear-Key Decrypt
    //
    // arg is pointer to structure passed by caller.
    //
    case ICARSAMODEXPO:
    case ICARSACRT:

      if (quiesce_z90crypt)
        ret = -EQUIESCE;
      else {
        loopLim = z90crypt.z90c_hdware_info->z90c_hdware_mask.z90c_st_count;
        for (i = 0; i < loopLim; i++) {
          ret = z90crypt_rsa(private_data_p, pid, cmd, arg);
          if (ret != -ERESTARTSYS)
            break;
        }
      if (ret == -ERESTARTSYS)
        ret = -ENODEV;
      }

      break;

    default:  // user passed an invalid IOCTL number
      PDEBUG("cmd 0x%08X contains an invalid ioctl code\n", cmd);
      ret = -ENOTTY;

  } // end switch(cmd)

  return ret;
} // end z90crypt_ioctl

int allocate_work_element(work_element_t ** work_elem_pp,
                          struct private_data * priv_data_p,
                          pid_t pid)
{
  int rv = 0;
  work_element_t * we_p = NULL;

  if ((rv = create_work_element(&we_p, pid)) != 0)
    return (rv);

  init_work_element(we_p, priv_data_p, pid);

  *work_elem_pp = we_p;

  return (rv);

} // end allocate_work_element


int create_work_element(work_element_t ** work_elem_pp, pid_t pid)
{
  work_element_t * work_elem_p = NULL;

  // A page is 4K bytes.  Currently the maximum size needed is
  // sizeof(work_element_t) + sizeof(Z90C_CALLER) = 2336 bytes
  if (!(work_elem_p = (work_element_t *)get_zeroed_page(GFP_KERNEL)))
    return -ENOMEM;

  *work_elem_pp = work_elem_p;

  return 0;

} // end create_work_element


int init_work_element(work_element_t * work_elem_p,
                      struct private_data * priv_data_p,
                      pid_t pid)
{

  int step;

  work_elem_p->requestptr =
             (unsigned char *)work_elem_p + sizeof(work_element_t);

  // Come up with a unique id for this caller.

  step = atomic_inc_return(&z90crypt_step);
  memcpy (work_elem_p->caller_id+0, (void *) &pid,  sizeof(pid));
  memcpy (work_elem_p->caller_id+4, (void *) &step, sizeof(step));
  work_elem_p->pid = pid;
  work_elem_p->opener_p = priv_data_p;

  // Set the status byte to the defaults
  //   not STAT_ROUTED
  //   not STAT_FAILED
  //   STAT_NOWORK

  work_elem_p->status[0] = STAT_DEFAULT;

  // Set the audit trail
  work_elem_p->audit[0] = 0;
  work_elem_p->audit[1] = 0;

  // Initialize response size to none
  work_elem_p->respsize = 0;

  // Initialize return code
  work_elem_p->returncode = OK;

  // Initialize device information
  work_elem_p->devindex = -1; // sendToCrypto selects the device
  work_elem_p->devtype = -1;  // getCryptoBuffer selects the type

  // Set up the wait queue for this work element.
  atomic_set(&work_elem_p->alarmrung,0);
  init_waitqueue_head(&work_elem_p->waitq);

  // Not in the request list or the pending list yet
  INIT_LIST_HEAD(&(work_elem_p->liste));

  return 0;

} // end init_work_element

int free_work_element(work_element_t * work_elem_p)
{

  int rv = 0;
  //
  // Notify lowlevel code, in case this caller has resources allocated
  // there
  //

  // Put the work element back
  destroy_work_element(work_elem_p);

  return (rv);
} // end free_work_element

void destroy_private_data(private_data_t * private_data_p)
{
  memset(private_data_p, 0, sizeof(private_data_t));
  kfree(private_data_p);
} // end destroy_private_data

void destroy_work_element(work_element_t *work_elem_p)
{
  free_page((long)work_elem_p);
} // end destroy_work_element

/*-------------------------------------------------------------------*/
/*                                                                   */
/* z90crypt_rsa:  This is the driver's reason for being; it performs */
/*                rsa encryptions and decryptions.                   */
/*                                                                   */
/* 1.  Call allocate_work_element to get storage                     */
/*                                                                   */
/* 2.  Call z90crypt_prepare to copy input data to the work element  */
/*     and prepare the data for submission to a crypto device        */
/*                                                                   */
/* 3.  Enqueue the work element to the list of work elements         */
/*     waiting for any crypto device                                 */
/*                                                                   */
/* 4.  Sleep while awaiting results                                  */
/*                                                                   */
/* 5.  When awakened:                                                */
/*       Copy results to user storage                                */
/*       Give back the work element                                  */
/*                                                                   */
/* 6.  Return to caller                                              */
/*                                                                   */
/*-------------------------------------------------------------------*/
int z90crypt_rsa(private_data_t *private_data_p, pid_t pid,
                 unsigned int cmd, unsigned long arg)
{
  int rv = 0;
  int keep_work_element = 0;
  work_element_t * work_elem_p;
  struct list_head * lptr = NULL;  // pointer to list element in workelement
  struct list_head * hptr = NULL;  // pointer to list head

  rv = allocate_work_element(&work_elem_p, private_data_p, pid);
  if (rv != 0) {
    return (rv);
  }

  do {
    //
    // Prepare the work element for crypto operations
    //
    if ((rv = z90crypt_prepare(work_elem_p, cmd, (const char *)arg))) {
      break;
    }

    //
    // Place the work on a device or on the request queue
    //
    if ((rv = z90crypt_send(work_elem_p, (const char *)arg)))
    {
      PDEBUG("PID %d trying to write: error %d detected in "
             "send.\n",
              pid, rv);
      break;
    }

    //
    // Await results
    //
    work_elem_p->audit[0] |= FP_ASLEEP;
    wait_event(work_elem_p->waitq,
               atomic_read(&work_elem_p->alarmrung));
    //
    // Put the results into user space
    //
    work_elem_p->audit[0] |= FP_AWAKE;

    rv = work_elem_p->returncode;

    if (rv == OK) {
      rv = z90crypt_process_results(work_elem_p, (char *)arg);
    }

    // Leave

  } while(0);

  //
  // Cleanup.  In all cases except status STAT_QUEUED, free the
  //           work element
  //
  if ((work_elem_p->status[0] & STAT_FAILED))
  {
    //
    // If we timed out the device
    //
    // If we released the device and the request is on the request queue
    //   (STAT_WRITTEN), remove it from the request queue
    //
    // If we released the device and the request is no longer on any queue
    // (STAT_READPEND)
    //
    // If we released the device and the request is on the pending queue
    // (STAT_PENDING), leave it alone, and let the reader_task clean it up.
    //
    // If we didn't release the device and we didn't timeout the request, mark
    //   it as not failed.
    //
    switch (rv)
    {
      // EINVAL *after* receive is almost always a padding error issued
      // by a leeds2 card.  Since leedslite doesn't check padding, and since
      // results should be consistent when this device driver is used by
      // libica, we convert this return value to -EGETBUFF.  This triggers
      // a resort to software in libica, which should produce results identical
      // to what a leedslite would have done.  Mind, this is almost certainly
      // an error condition, so all we're guaranteeing here is the *same*
      // error condition no matter which card is in use.  Things will have
      // gone very slowly, but this is an error path after all.
      case -EINVAL:
        if (work_elem_p->devtype == LEEDS2)
          rv = -EGETBUFF;
        break;

      // ETIMEOUT can happen because a device has gone offline.  In
      // that case, the work should be retried if other devices are
      // available.
      case -ETIMEOUT:
        /*
        if ((work_elem_p->devindex == -1) ||
            (z90crypt.z90c_mask.z90c_st_mask[work_elem_p->devindex] == DEV_GONE)) {
        */
          if (z90crypt.z90c_mask.z90c_st_count > 0)
            rv = -ERESTARTSYS;
          else
            rv = -ENODEV;
        /*
        }
        else {
          PRINTK("Timeout:  device index: %d\n",work_elem_p->devindex);
          PRINTK("Timeout:  device count: %d\n",z90crypt.z90c_mask.z90c_st_count);
          break;
        }
        */

      case -ERESTARTSYS:
        // fall through to clean up request queue
      case -ERELEASED:
        switch(CHECK_RDWRMASK(work_elem_p->status[0]))
        {
          case STAT_WRITTEN:    // Get the request off the requestq or pendingq

            spin_lock_irq(&queuespinlock);
            hptr = &request_list;
            list_for_each(lptr, hptr) {
              if (lptr == &(work_elem_p->liste)) {
                list_del(lptr);
                --requestqCount;
                break;
              }
            }
            hptr = &pending_list;
            list_for_each(lptr, hptr) {
              if (lptr == &(work_elem_p->liste)) {
                list_del(lptr);
                --pendingqCount;
                break;
              }
            }
            spin_unlock_irq(&queuespinlock);

            break;

          case STAT_READPEND:
          case STAT_NOWORK:     // This should never happen
            break;

          case STAT_QUEUED:
            //
            // Request is still on the hardware.  Let reader_task clean up
            //
            keep_work_element = 1;
            break;

          default:
            break;
        } // end switch(status)
        break;

      default:
        work_elem_p->status[0] ^= STAT_FAILED; // remove failed-bit
        break;

    } // end switch(rv)

  } // end if STAT_FAILED

  //
  // In all cases except STAT_QUEUED (see above), free the work element
  //
  if (!(keep_work_element))
    free_work_element(work_elem_p);

  return rv;

} // end z90crypt_rsa

//
// Prepare copies the user's data to kernel space and
// formats it for dispatch to a crypto device
//
int
z90crypt_prepare(work_element_t * work_elem_p,
                 unsigned int funccode, const char *caller_buf)
{
  int       ret = 0;           // zero (if successful) or error code

  //
  // Set the correct size for the structure used
  //
  if (funccode == ICARSAMODEXPO)
    work_elem_p->buffsize = sizeof(ica_rsa_modexpo_t);
  else
    work_elem_p->buffsize = sizeof(ica_rsa_modexpo_crt_t);

  //
  // Copy the structure from the user.
  //
  if (copy_from_user (work_elem_p->buffer, caller_buf, work_elem_p->buffsize))
  {
    ret = -EFAULT;
    goto prepare_cleanup;
  }
  work_elem_p->audit[0] |= FP_COPYFROM;

  //
  // Set the function code in the work element.
  //
  SET_RDWRMASK(work_elem_p->status[0], STAT_WRITTEN);
  work_elem_p->funccode = funccode;

  //
  // Get a request buffer.  The lowlevel code will analyze the
  // request and return an error if the request can't be satisfied
  //
  work_elem_p->devtype = -1;   // no particular type can be specified
  work_elem_p->audit[0] |= FP_BUFFREQ;
  ret = getCryptoRequestBuffer(work_elem_p->caller_id,
                               work_elem_p->funccode,
                               work_elem_p->buffsize,
                               work_elem_p->buffer,
                               &work_elem_p->devtype,
                               work_elem_p->requestptr);
  switch(ret) {
    case 0:
      work_elem_p->audit[0] |= FP_BUFFGOT;
      break;
    case SEN_USER_ERROR:
      ret = -EINVAL;
      break;
    case SEN_QUEUE_FULL:
      ret = 0;
      break;
    case SEN_RELEASED:
      ret = -EFAULT;
      break;
    case REC_NO_RESPONSE:
      ret = -ENODEV;
    case SEN_NOT_AVAIL:
    default:
      PRINTK("Error from getCryptoRequestBuffer: %d\n",ret);
      ret = -EGETBUFF;
  }

prepare_cleanup:
  if (CHECK_RDWRMASK(work_elem_p->status[0]) ==STAT_WRITTEN)
    SET_RDWRMASK(work_elem_p->status[0], STAT_DEFAULT);
  work_elem_p->devindex = -1;  // only dev *type* is important
  return ret;
} // end z90crypt_prepare

//
// Enqueue puts the user's data on one of two queues:
//   the pending queue if the send was successful
//   the request queie if the send failed because device full or busy
//
int
z90crypt_send(work_element_t * we_p, const char *buf)
{
  pid_t     pid;              /* PID of the current task                    */
  int       rv = 0;           /* zero (if successful) or error code (if not)*/

  pid = PID();
  PDEBUG("send -> PID %d\n", pid);

  //
  // User is not allowed to have any outstanding requests.
  //
  if (CHECK_RDWRMASK(we_p->status[0]) != STAT_NOWORK)
  {
    PDEBUG("PID %d tried to send more work but has outstanding "
           "work.\n",
           pid);
    return -EWORKPEND;
  }
  // Reset device number
  we_p->devindex = -1;

  // Serialize on the request and pending queues
  spin_lock_irq(&queuespinlock);

  // Send the work to a crypto device
  rv = sendToCryptoDevice(we_p->caller_id,
                             we_p->funccode,
                             we_p->buffsize,
                             we_p->buffer,
                             we_p->devtype,
                             &we_p->devindex,
                             we_p->requestptr);
  // Check the return value
  switch (rv) {
    case OK:
      // Time stamp
      we_p->requestsent = jiffies;      // keep track of age on queue
      // footprint
      we_p->audit[0] |= FP_SENT;    
      // Put on the pending queue
      list_add_tail(&we_p->liste, &pending_list);
      ++pendingqCount;
      we_p->audit[0] |= FP_PENDING;
      break;

    case SEN_BUSY:
    case SEN_QUEUE_FULL:
      rv = 0;            // Fake it, so the caller will sleep
      // reset device number
      we_p->devindex = -1;             // any device will do
      // Time stamp
      we_p->requestsent = jiffies;     // keep track of age on queue
      // Put on the request queue
      list_add_tail(&we_p->liste, &request_list);
      ++requestqCount;
      we_p->audit[0] |= FP_REQUEST;
      break;

    case SEN_RETRY:
      rv = -ERESTARTSYS;
      break;

    case SEN_NOT_AVAIL:  // From here down the request is rejected
      PRINTK("*** No devices available.\n");
      rv = we_p->returncode = -ENODEV;
      we_p->status[0] |= STAT_FAILED;
      break;
    case REC_OPERAND_INVALID:
    case REC_OPERAND_SIZE:
    case REC_EVEN_MODULUS:
    case REC_INVALID_PADDING:
      rv = we_p->returncode = -EINVAL;
      we_p->status[0] |= STAT_FAILED;
      break;
    default:
      // Propagate return code
      we_p->returncode = rv;
      // Mark this element a failure
      we_p->status[0] |= STAT_FAILED;
      break;
  } // end switch (rv)

  // Mark the status
  if (rv != -ERESTARTSYS)
    SET_RDWRMASK(we_p->status[0], STAT_WRITTEN);

  // Unserialize
  spin_unlock_irq(&queuespinlock);

  // If this work element is on a device or the request queue, get it off
  if (rv == OK)
    tasklet_schedule(&reader_tasklet);

  return rv;
} // end z90crypt_send

//
// process_results copies the user's work from kernel space.
//
int
z90crypt_process_results (work_element_t * work_elem_p, char *buf)
{
  pid_t     pid;              /* PID of the current task                    */
  int       retval = 0;       /* zero (if successful) or error code (if not)*/

  /*
   * set up needed variables
   */
  pid = PID();

  PDEBUG("process_results -> work_elem_p %p (PID %d)\n", work_elem_p, pid);

  /*
   * Change the RDWRMASK to STAT_READPEND before we do anything.
   */
  SET_RDWRMASK(work_elem_p->status[0], STAT_READPEND);

  /*
   * Check that the buffer holding the reply is not NULL.
   */
  if (!work_elem_p->buffer)
  {
    PRINTK("work_elem_p %p (PID %d) claims STAT_READPEND: buffer is NULL.\n",
           work_elem_p, pid);
    /*
     * This is BAD!  We'll clean up before we tell the user that things
     * went south.
     */
    retval = -ENOBUFF;
    goto process_results_cleanup;
  }

  if ((retval =
        copy_to_user (buf, work_elem_p->buffer, work_elem_p->buffsize))!=0)
  {
    PDEBUG("copy_to_user failed:  rv = %d\n",retval);
    retval = -EFAULT;
    goto process_results_cleanup;
  }

  retval = work_elem_p->returncode;
  if (retval)
    goto process_results_cleanup;

  /* here copy from work_elem_p->respbuff to user space */  // @rhbA
  if (work_elem_p->respsize)
  {
    if (copy_to_user (work_elem_p->respaddr,
        work_elem_p->respbuff,
        work_elem_p->respsize))
    {
      retval = -EFAULT;
      goto process_results_cleanup;
    }
  }

process_results_cleanup:
  /*
   * Change the RDWRMASK to STAT_NOWORK before we go home.
   *
   * Either we copied data to the user or failed with an unrecoverable error.
   */

  SET_RDWRMASK(work_elem_p->status[0], STAT_NOWORK);

  return retval;
} // end z90crypt_process_results

int sprintcl(unsigned char * outaddr,
             unsigned char * addr,
             unsigned int len)
{
  int hl = 0;
  int i;

  for (i=0;i<len;i++) {
    hl += sprintf(outaddr+hl,"%02x",(unsigned int) addr[i]);
  }
  hl += sprintf(outaddr+hl," ");

  return hl;
}

int sprintrw(unsigned char * outaddr,
             unsigned char * addr,
             unsigned int len)
{
  int hl = 0;
  int inl = 0;
  int c, cx;

  hl += sprintf(outaddr,"    ");                         // left margin

  for (c=0;c<len/8;c++) {
    hl += sprintcl(outaddr+hl, addr+inl, 8);
    inl += 8;
  }

  cx = len%8;
  if (cx) {
    hl += sprintcl(outaddr+hl, addr+inl, cx);
    inl += cx;
  }

  hl += sprintf(outaddr+hl,"\n");

  return hl;
}

int sprinthx(unsigned char * title,
             unsigned char * outaddr,
             unsigned char * addr,
             unsigned int len)
{
  int hl = 0;
  int inl = 0;
  int r,rx;

  hl += sprintf(outaddr,"\n%s\n",title);

  for (r=0; r<len/32; r++) {
    hl += sprintrw(outaddr+hl, addr+inl, 32);
    inl += 32;
  }
  rx = len%32;
  if (rx) {
    hl += sprintrw(outaddr+hl, addr+inl, rx);
    inl += rx;
  }

  hl += sprintf(outaddr+hl,"\n");

  return hl;
}

void md5_init( MD5_CTX *context )
{
  context->i[0] = context->i[1] = 0;

  // Load magic initialization constants.
  //
  context->buf[0] = 0x67452301;
  context->buf[1] = 0xefcdab89;
  context->buf[2] = 0x98badcfe;
  context->buf[3] = 0x10325476;
} // end md5_init

int md5_hash(
              MD5_CTX  * ctx,
              unsigned char * in_data,
              unsigned int    in_data_len,
              unsigned char * out_data,
              unsigned int * out_data_len )
{
   int rc = 0;

   rc = md5_update(ctx, in_data, in_data_len);
   if (rc != 0){
      return rc;
   }

   return md5_final(ctx, out_data, out_data_len);
} // end md5_hash

int md5_update( MD5_CTX  * context,
            unsigned char * in_data,
            unsigned int in_data_len )
{
  unsigned int in[16];
  int      mdi;
  unsigned int i, ii;

  // compute number of bytes mod 64
  //
  mdi = ((context->i[0] >> 3) & 0x3F);

  // update number of bits
  //
  if ((context->i[0] + (in_data_len << 3)) < context->i[0])
    context->i[1]++;

  context->i[0] += (in_data_len << 3);
  context->i[1] += (in_data_len >> 29);

  while (in_data_len--) {
    // add new character to buffer, increment mdi
    //
    context->in[mdi++] = *in_data++;

    // transform if necessary
    //
    if (mdi == 0x40) {
      for (i = 0, ii = 0; i < 16; i++, ii += 4)
        in[i] = ((context->in[ii+3]) << 24) |
                ((context->in[ii+2]) << 16) |
                ((context->in[ii+1]) << 8) |
                 (context->in[ii]);
      md5_transform (context->buf, in);
      mdi = 0;
    }
  }

  return 0;

} // end md5_update


int md5_final( MD5_CTX *context,
               unsigned char *out_data,
               unsigned int * out_data_len )
{
  unsigned int in[16];
  int       mdi;
  unsigned int i, ii;
  unsigned int padLen;

  // save number of bits
  //
  in[14] = context->i[0];
  in[15] = context->i[1];

  // compute number of bytes mod 64
  //
  mdi = ((context->i[0] >> 3) & 0x3F);

  // pad out to 56 mod 64
  //
  padLen = (mdi < 56) ? (56 - mdi) : (120 - mdi);
  md5_update( context, PADDING, padLen );

  // append length in bits and transform
  //
  for (i = 0, ii = 0; i < 14; i++, ii += 4)
    in[i] = ((context->in[ii+3]) << 24) |
            ((context->in[ii+2]) << 16) |
            ((context->in[ii+1]) << 8) |
            (context->in[ii]);

  // store buffer in digest
  //
  for (i = 0, ii = 0; i < 4; i++, ii += 4) {
    context->digest[ii  ] = (unsigned char) (context->buf[i] & 0xFF);
    context->digest[ii+1] = (unsigned char) ((context->buf[i] >>  8) & 0xFF);
    context->digest[ii+2] = (unsigned char)((context->buf[i] >> 16) & 0xFF);
    context->digest[ii+3] = (unsigned char)((context->buf[i] >> 24) & 0xFF);
  }

  memcpy( out_data, context->digest, MD5_HASH_SIZE );
  *out_data_len = MD5_HASH_SIZE;

  return 0;
} // end md5_final

// Basic MD5 step. Transform buf based on in.
//
void md5_transform(unsigned int *buf, unsigned int *in )
{
   unsigned int AA = 0x00000000;          /* Temp. save areas for A, B, C, D */
   unsigned int BB = 0x00000000;          /* Temp. save areas for A, B, C, D */
   unsigned int CC = 0x00000000;          /* Temp. save areas for A, B, C, D */
   unsigned int DD = 0x00000000;          /* Temp. save areas for A, B, C, D */

   /*-------------------------------------------------------------------------*/
   /*                                                                         */
   /* Macro ROUND_FCN defines the common function that is performed           */
   /* throughout the MD5 round.  Parameters are:                              */
   /*                                                                         */
   /*   - The name of the function to be performed on the data for this       */
   /*     round.  There are four logical functions, F, G, H, and I, and they  */
   /*     are each used throughout the algorithm.                             */
   /*                                                                         */
   /*   - a, b, c, and d are the four md5-word parameters to the functions    */
   /*     F, G, H, and I.  They are replaced with varying permutations of the */
   /*     accumulated hash values in A, B, C, and D.                          */
   /*                                                                         */
   /*   - x_index is an index into the in[] array, where in[] is the input    */
   /*     block of message text.                                              */
   /*                                                                         */
   /*   - t_index is an index into the T[] array, a set of constants.         */
   /*                                                                         */
   /*   - rotate_cnt is the number of bits the result must be rotated.        */
   /*                                                                         */
   /*-------------------------------------------------------------------------*/

   #define ROUND_FCN(FCN, a, b, c, d, x_index, rotate_cnt, t_index) \
    { a += FCN(b,c,d) + in[x_index] + T[t_index-1];           \
      rotate_left( a, rotate_cnt);                           \
      a += b;                                                \
    }

   /* Save the MD buffer in the temporary locations AA-DD.                    */

   AA = buf[0];
   BB = buf[1];
   CC = buf[2];
   DD = buf[3];

   /*==================================================================*/
   /*                                                                  */
   /* Process the four rounds for each 16-word block.                  */
   /*                                                                  */
   /* The function for each of these has the form:                     */
   /*                                                                  */
   /*    a = b + (( a + fcn( b, c, d ) + in[k] + T[i] ) <<< s )        */
   /*                                                                  */
   /* for a function fcn() which can be F, G, H, or I, and for input   */
   /* values a, b, c, d, k, i, and s.  Array T is the array of         */
   /* constants, computed from the sin() function.  Array in is the    */
   /* current input block.  Value s is the number of bits to rotate    */
   /* left, where <<< represents a 32-bit left rotation.               */
   /*                                                                  */
   /* The definitions of these functions are taken directly from the   */
   /* definition of MD5 in RFC 1321.                                   */
   /*                                                                  */
   /*==================================================================*/

   /*------------------------------------------------------------------*/
   /*                                                                  */
   /* Round 1                                                          */
   /*                                                                  */
   /*------------------------------------------------------------------*/
   ROUND_FCN(F, buf[0], buf[1], buf[2], buf[3],  0,  7,  1);
   ROUND_FCN(F, buf[3], buf[0], buf[1], buf[2],  1, 12,  2);
   ROUND_FCN(F, buf[2], buf[3], buf[0], buf[1],  2, 17,  3);
   ROUND_FCN(F, buf[1], buf[2], buf[3], buf[0],  3, 22,  4);
   ROUND_FCN(F, buf[0], buf[1], buf[2], buf[3],  4,  7,  5);
   ROUND_FCN(F, buf[3], buf[0], buf[1], buf[2],  5, 12,  6);
   ROUND_FCN(F, buf[2], buf[3], buf[0], buf[1],  6, 17,  7);
   ROUND_FCN(F, buf[1], buf[2], buf[3], buf[0],  7, 22,  8);
   ROUND_FCN(F, buf[0], buf[1], buf[2], buf[3],  8,  7,  9);
   ROUND_FCN(F, buf[3], buf[0], buf[1], buf[2],  9, 12, 10);
   ROUND_FCN(F, buf[2], buf[3], buf[0], buf[1], 10, 17, 11);
   ROUND_FCN(F, buf[1], buf[2], buf[3], buf[0], 11, 22, 12);
   ROUND_FCN(F, buf[0], buf[1], buf[2], buf[3], 12,  7, 13);
   ROUND_FCN(F, buf[3], buf[0], buf[1], buf[2], 13, 12, 14);
   ROUND_FCN(F, buf[2], buf[3], buf[0], buf[1], 14, 17, 15);
   ROUND_FCN(F, buf[1], buf[2], buf[3], buf[0], 15, 22, 16);

   /*------------------------------------------------------------------*/
   /*                                                                  */
   /* Round 2                                                          */
   /*                                                                  */
   /*------------------------------------------------------------------*/

   ROUND_FCN(G, buf[0], buf[1], buf[2], buf[3],  1,  5, 17);
   ROUND_FCN(G, buf[3], buf[0], buf[1], buf[2],  6,  9, 18);
   ROUND_FCN(G, buf[2], buf[3], buf[0], buf[1], 11, 14, 19);
   ROUND_FCN(G, buf[1], buf[2], buf[3], buf[0],  0, 20, 20);
   ROUND_FCN(G, buf[0], buf[1], buf[2], buf[3],  5,  5, 21);
   ROUND_FCN(G, buf[3], buf[0], buf[1], buf[2], 10,  9, 22);
   ROUND_FCN(G, buf[2], buf[3], buf[0], buf[1], 15, 14, 23);
   ROUND_FCN(G, buf[1], buf[2], buf[3], buf[0],  4, 20, 24);
   ROUND_FCN(G, buf[0], buf[1], buf[2], buf[3],  9,  5, 25);
   ROUND_FCN(G, buf[3], buf[0], buf[1], buf[2], 14,  9, 26);
   ROUND_FCN(G, buf[2], buf[3], buf[0], buf[1],  3, 14, 27);
   ROUND_FCN(G, buf[1], buf[2], buf[3], buf[0],  8, 20, 28);
   ROUND_FCN(G, buf[0], buf[1], buf[2], buf[3], 13,  5, 29);
   ROUND_FCN(G, buf[3], buf[0], buf[1], buf[2],  2,  9, 30);
   ROUND_FCN(G, buf[2], buf[3], buf[0], buf[1],  7, 14, 31);
   ROUND_FCN(G, buf[1], buf[2], buf[3], buf[0], 12, 20, 32);

   /*------------------------------------------------------------------*/
   /*                                                                  */
   /* Round 3                                                          */
   /*                                                                  */
   /*------------------------------------------------------------------*/
   ROUND_FCN(H, buf[0], buf[1], buf[2], buf[3],  5,  4, 33);
   ROUND_FCN(H, buf[3], buf[0], buf[1], buf[2],  8, 11, 34);
   ROUND_FCN(H, buf[2], buf[3], buf[0], buf[1], 11, 16, 35);
   ROUND_FCN(H, buf[1], buf[2], buf[3], buf[0], 14, 23, 36);
   ROUND_FCN(H, buf[0], buf[1], buf[2], buf[3],  1,  4, 37);
   ROUND_FCN(H, buf[3], buf[0], buf[1], buf[2],  4, 11, 38);
   ROUND_FCN(H, buf[2], buf[3], buf[0], buf[1],  7, 16, 39);
   ROUND_FCN(H, buf[1], buf[2], buf[3], buf[0], 10, 23, 40);
   ROUND_FCN(H, buf[0], buf[1], buf[2], buf[3], 13,  4, 41);
   ROUND_FCN(H, buf[3], buf[0], buf[1], buf[2],  0, 11, 42);
   ROUND_FCN(H, buf[2], buf[3], buf[0], buf[1],  3, 16, 43);
   ROUND_FCN(H, buf[1], buf[2], buf[3], buf[0],  6, 23, 44);
   ROUND_FCN(H, buf[0], buf[1], buf[2], buf[3],  9,  4, 45);
   ROUND_FCN(H, buf[3], buf[0], buf[1], buf[2], 12, 11, 46);
   ROUND_FCN(H, buf[2], buf[3], buf[0], buf[1], 15, 16, 47);
   ROUND_FCN(H, buf[1], buf[2], buf[3], buf[0],  2, 23, 48);

   /*------------------------------------------------------------------*/
   /*                                                                  */
   /* Round 4                                                          */
   /*                                                                  */
   /*------------------------------------------------------------------*/
   ROUND_FCN(I, buf[0], buf[1], buf[2], buf[3],  0,  6, 49);
   ROUND_FCN(I, buf[3], buf[0], buf[1], buf[2],  7, 10, 50);
   ROUND_FCN(I, buf[2], buf[3], buf[0], buf[1], 14, 15, 51);
   ROUND_FCN(I, buf[1], buf[2], buf[3], buf[0],  5, 21, 52);
   ROUND_FCN(I, buf[0], buf[1], buf[2], buf[3], 12,  6, 53);
   ROUND_FCN(I, buf[3], buf[0], buf[1], buf[2],  3, 10, 54);
   ROUND_FCN(I, buf[2], buf[3], buf[0], buf[1], 10, 15, 55);
   ROUND_FCN(I, buf[1], buf[2], buf[3], buf[0],  1, 21, 56);
   ROUND_FCN(I, buf[0], buf[1], buf[2], buf[3],  8,  6, 57);
   ROUND_FCN(I, buf[3], buf[0], buf[1], buf[2], 15, 10, 58);
   ROUND_FCN(I, buf[2], buf[3], buf[0], buf[1],  6, 15, 59);
   ROUND_FCN(I, buf[1], buf[2], buf[3], buf[0], 13, 21, 60);
   ROUND_FCN(I, buf[0], buf[1], buf[2], buf[3],  4,  6, 61);
   ROUND_FCN(I, buf[3], buf[0], buf[1], buf[2], 11, 10, 62);
   ROUND_FCN(I, buf[2], buf[3], buf[0], buf[1],  2, 15, 63);
   ROUND_FCN(I, buf[1], buf[2], buf[3], buf[0],  9, 21, 64);

   /* Add to each MD buffer variable the value it had before this block was   */
   /* started.                                                                */

   buf[0] += AA;
   buf[1] += BB;
   buf[2] += CC;
   buf[3] += DD;

   /* Undefine the ROUND_FCN macro we used in this function.                  */
   #undef ROUND_FCN

} // end md5_transform


/*--------------------------------------------------------------------------*/
/* Bottom Half Functions                                                    */
/*                                                                          */
/* The task functions:                                                      */
/*     z90crypt_reader_task                                                 */
/*     z90crypt_config_task                                                 */
/*     z90crypt_cleanup_task                                                */
/*                                                                          */
/* Helper functions:                                                        */
/*     z90crypt_schedule_reader_timer                                       */
/*     z90crypt_schedule_reader_task                                        */
/*     z90crypt_schedule_config_task                                        */
/*     z90crypt_schedule_cleanup_task                                       */
/*                                                                          */
/*--------------------------------------------------------------------------*/

void
z90crypt_reader_task (unsigned long ptr)
{
  int       workavail,
            remaining,          /* number of unfinished requests            */
            index,
            rc, rv,
            buff_len;           /* Response length                          */
  UCHAR     psmid[8],
            buff[256],          /* Response buffer                          */
            *respaddr;          /* Response address in user space           */
  BOOL      remove_from_queue;  //
  work_element_t *pq_p;         /* used to walk the pending queue           */
  work_element_t *rq_p;         /* from the request queue                   */
  struct list_head * lptr;      // pointer to list struct in work element
  struct list_head * tptr;      // temp pointer to list struct in work element
  struct list_head * hptr;      // pointer to list head

  workavail = 2;
  remaining = 0;
  buff_len = 0;
  memset(buff, 0x00, sizeof(buff));
  respaddr = NULL;
  remove_from_queue = FALSE;
  pq_p = NULL;
  rq_p = NULL;
  PDEBUG("reader_task -> jiffies %ld\n", jiffies);

  remaining = 0;

  while (workavail)
  {
    /*
     * Decrement workavail by 1.
     *
     * We need to have nothing on a queue twice in a row to stop looking, so
     * we decrement it each time before we scan, and we will reset it to 2 if
     * we find anything on the queues.
     *
     * Reset unfinished work count to zero.  If we leave the while loop after
     * checking each device, and this count is still zero, there is no work
     * still to be finished and thus, no reason to reschedule the reader_task.
     */
    workavail--;
    rc = rv = 0;

    // Serialize
    spin_lock_irq(&queuespinlock);

    for (index = 0; index < z90crypt.z90c_mask.z90c_st_count; index++)
    {
      /*
       * Dequeue finished work from hardware.
       *
       * Dequeue once from each device in round robin.  It seems like a fairer
       * algorithm than checking one device over and over again until it
       * returns empty and then going on to the next device.
       */


      // Initialize switches, return codes
      remove_from_queue = FALSE;
      rc = rv = OK;

      PDEBUG("About to receive.\n");

      rc = receiveFromCryptoDevice (SHRT2LONG(index),
                                    psmid,
                                    &buff_len,
                                    buff,
                                    &respaddr);
      PDEBUG("Dequeued: rc = %d.\n",
             rc);

      switch (rc)
      {
        case OK:
          remove_from_queue = TRUE;
          break;

        case REC_BUSY:
        case REC_NO_WORK:
        case REC_EMPTY:
        case REC_RETRY_THIS_DEVICE:
        case REC_FATAL_ERROR: // There is no work returned.  Move on.
          remove_from_queue = FALSE;
          break;

        case REC_NO_RESPONSE:
          remove_from_queue = FALSE;
          workavail = 0;
          break;

        case REC_OPERAND_INVALID:
        case REC_OPERAND_SIZE:
        case REC_EVEN_MODULUS:
        case REC_INVALID_PADDING:
          remove_from_queue = TRUE;
          break;

        default:
          PRINTK("Error %d detected in reader_task from device %d.\n",
                  rc,SHRT2LONG(index));
          rc = REC_NO_RESPONSE;
          remove_from_queue = FALSE;
          workavail = 0;
          break;

      } // end switch

      if (remove_from_queue == TRUE) {

        if (rc != REC_NO_RESPONSE) {
          // Now there's room for more work on the hardware--put some there
          rq_p = NULL;
          if (!(list_empty(&request_list))) {
            lptr = request_list.next;
            // Remove from queue
            list_del(lptr);
            requestqCount--;
            rq_p = list_entry(lptr, struct work_element, liste);
            rq_p->audit[1] |= FP_REMREQUEST;
            if (rq_p->devtype == SHRT2DEVPTR(index)->z90c_dev_type) {
              // Send to device
              rq_p->devindex = SHRT2LONG(index); // send to *this* device
              rv = sendToCryptoDevice(rq_p->caller_id,
                                      rq_p->funccode,
                                      rq_p->buffsize,
                                      rq_p->buffer,
                                      rq_p->devtype,
                                      &rq_p->devindex,
                                      rq_p->requestptr);
              if (rv == OK) {
                // Time stamp
                rq_p->requestsent = jiffies; // keep track of age on *this *queue
                rq_p->audit[0] |= FP_SENT;    
                // Put on the pending queue
                list_add_tail(lptr, &pending_list);
                ++pendingqCount;
                rq_p->audit[0] |= FP_PENDING;
              } // end if rv is OK
              else {
                // Propagate return code
                switch (rv) {
                  case SEN_OPERAND_INVALID:
                  case REC_OPERAND_SIZE:
                  case REC_EVEN_MODULUS:
                  case SEN_PADDING_ERROR:
                    rq_p->returncode = -EINVAL;
                    break;
                  case SEN_NOT_AVAIL:
                  case SEN_RETRY:
                  case REC_NO_RESPONSE:
                  default:
                    if (z90crypt.z90c_mask.z90c_st_count > 1)
                      rq_p->returncode = -ERESTARTSYS;
                    else
                      rq_p->returncode = -ENODEV;
                    break;
                } // end switch on rv
                // Mark this element a failure
                rq_p->status[0] |= STAT_FAILED;
                // Give the caller the bad news
                rq_p->audit[1] |= FP_AWAKENING;
                atomic_set(&rq_p->alarmrung,1);
                wake_up(&rq_p->waitq);
              } // end if rv isn't OK
            } // end if device types match
            else {
              if (z90crypt.z90c_mask.z90c_st_count > 1)
                rq_p->returncode = -ERESTARTSYS;
              else
                rq_p->returncode = -ENODEV;
              // Mark this element a failure
              rq_p->status[0] |= STAT_FAILED;
              // Give the caller the bad news
              rq_p->audit[1] |= FP_AWAKENING;
              atomic_set(&rq_p->alarmrung,1);
              wake_up(&rq_p->waitq);
            } // end else device types don't match
          } // end if request list not empty

            // Since rc wasn't REC_NO_RESPONSE, reset workavail to 2 more tries.
            // We have to go two rounds without work to give up.
            workavail = 2;

        } // end if rc isn't REC_NO_RESPONSE

        /*
         * Wake up task waiting for this data, so that it can take care of
         * it.
         */

        // Search for this work element on the pending queue
        pq_p = NULL;
        hptr = &pending_list;
        list_for_each_safe(lptr,tptr,hptr) {
          pq_p = list_entry(lptr, struct work_element, liste);
          if (!(memcmp(pq_p->caller_id, psmid, sizeof(pq_p->caller_id)))){
            list_del(lptr);
            --pendingqCount;
            pq_p->audit[1] |= FP_NOTPENDING;
            break;
          } // end if the psmid's match
          else {
            pq_p = NULL;
          } // end else the psmid's don't match
        }

        if(!pq_p) // if Not Found
        {
          PRINTK("device %d has work but no caller "
                 "exists on pending queue\n",
                 SHRT2LONG(index));
        } // end Not Found
        else  { // Found
          if (rc == OK) {
            pq_p->respsize = buff_len;
            pq_p->audit[1] |= FP_RESPSIZESET;
            if (buff_len)
            {
              pq_p->respaddr = respaddr;
              pq_p->audit[1] |= FP_RESPADDRCOPIED;
              memcpy(pq_p->respbuff,
                     buff,
                     buff_len);
              pq_p->audit[1] |= FP_RESPBUFFCOPIED;
            } // end if buff_len non zero
          } // end if rc is OK
          else {
            switch (rc) {
              case REC_OPERAND_INVALID:
              case REC_OPERAND_SIZE:
              case REC_EVEN_MODULUS:
              case REC_INVALID_PADDING:
                PDEBUG(" -EINVAL after application error %d\n",rc);
                pq_p->returncode = -EINVAL;
                pq_p->status[0] |= STAT_FAILED;
                break;
              case REC_NO_RESPONSE:
              default:
                if (z90crypt.z90c_mask.z90c_st_count > 1)
                  pq_p->returncode = -ERESTARTSYS;
                else
                  pq_p->returncode = -ENODEV;
                PRINTK("In tasks, after hw failed.  rc = %d\n", pq_p->returncode);
                pq_p->status[0] |= STAT_FAILED;
                break;
            } // end switch on rc
          } // end else rc isn't OK
          if ((pq_p->status[0] != STAT_FAILED ||
               pq_p->returncode != -ERELEASED)) {
            pq_p->audit[1] |= FP_AWAKENING;
            atomic_set(&pq_p->alarmrung,1);
            wake_up(&pq_p->waitq);
          }
        } // end else Found
      } // end if remove_from_queue = TRUE )

      /*
       * Update the count of remaining work on this queue.
       */
      if (rc == REC_FATAL_ERROR)
        remaining = 0;
      else
        if (rc != REC_NO_RESPONSE)
          remaining += SHRT2DEVPTR(index)->z90c_dev_caller_count;

    } // end for each device

    // Unserialize
    spin_unlock_irq(&queuespinlock);

  } // end while workavail

  //
  // Reschedule self, if there is still work on hardware.
  //
  if (remaining)
  {
    //
    // Schedule the reader_timer, if the previous step was successful.
    //
    spin_lock_irq(&queuespinlock);
    z90crypt_schedule_reader_timer();
    spin_unlock_irq(&queuespinlock);
  } // end if there's remaining work to do
} // end z90crypt_reader_task

void
z90crypt_config_task (unsigned long ptr)
{
  int rc = 0;

  PDEBUG("config_task -> jiffies %ld\n", jiffies);

  if ((rc = refreshZ90crypt(&z90crypt.z90c_cdx)))
  {
    // Complain loudly.  This is show-stoppingly bad.
    PRINTK("Error %d detected in refreshZ90crypt.\n", rc);
  }
  // If return was fatal, don't bother reconfiguring
  if (rc != TSQ_FATAL_ERROR && rc != RSQ_FATAL_ERROR) {
    // Reschedule self.
    if ((rc = z90crypt_schedule_config_task(CONFIGTIME)))
    {
      // Currently, there is nothing that could go wrong, but in the future,
      // there may be.  This is a catch-all.
      //
      PRINTK("Error %d detected in schedule_config_task.\n", rc);
    }
    else
    {
      PDEBUG("schedule_config_task successful.\n");
    }
  }
}

void
z90crypt_cleanup_task (unsigned long ptr)
{
  work_element_t *pq_p;         /* used to walk the pending queue           */
  struct list_head *lptr = NULL;// used to walk the queues
  struct list_head *tptr = NULL;// temp for running queues
  struct list_head *hptr = NULL;// pointer to the queue head
  int       rc;
  int deviceCount = 0;
  long      timelimit;          /* time at which no task on the pending
                                   queue should have been started before
                                   (saves reading a volatile variable every
                                   time we check a task on the pending queue
                                   with no loss of accuracy                 */

  PDEBUG("cleanup_task -> jiffies %ld\n", jiffies);

  //
  // Check whether all the crypto devices are gone.  That would mean
  // that every waiting process--waiting on either the request queue
  // or the pending queue--will have to be posted and notified that
  // its work isn't going to get done.
  //

  spin_lock_irq(&queuespinlock);

  deviceCount = z90crypt.z90c_mask.z90c_st_count;

  if (deviceCount <= 0) {

    // Post everybody on the pending queue
    hptr = &pending_list;
    list_for_each_safe(lptr, tptr, hptr) {
      pq_p = list_entry(lptr, struct work_element, liste);
      pq_p->returncode = -ENODEV;      // set return code and status
      pq_p->status[0] |= STAT_FAILED;
      list_del(lptr);
      pendingqCount--;
      pq_p->audit[1] |= FP_NOTPENDING;
      pq_p->audit[1] |= FP_AWAKENING;
      atomic_set(&pq_p->alarmrung,1);
      wake_up(&pq_p->waitq);
    } // end list_for_each

    // Post everybody on the request queue
    hptr = &request_list;
    list_for_each_safe(lptr, tptr, hptr) {
      pq_p = list_entry(lptr, struct work_element, liste);
      pq_p->returncode = -ENODEV;      // set return code and status
      pq_p->status[0] |= STAT_FAILED;
      list_del(lptr);
      requestqCount--;
      pq_p->audit[1] |= FP_REMREQUEST;
      pq_p->audit[1] |= FP_AWAKENING;
      atomic_set(&pq_p->alarmrung,1);
      wake_up(&pq_p->waitq);
    } // end list_for_each

  } // end if deviceCount is zero
  else {
    /*
     * Grab the earliest time at which a request should have been started
     * (CLEANUPTIME seconds ago: HZ = 1 second)
     *
     * Tasks started before this will be set to STAT_FAILED (return code
     * ETIMEOUT) and awoken to return to their caller.
     */

    timelimit = jiffies - (CLEANUPTIME * HZ);

    /*
     * We need to hold the pendingq semaphore so that no work gets added while
     * we are cleaning up.
     */

    /*
     * Walk the pending queue and clean up requests more than CLEANUPTIME
     * seconds old.  The list is in strict chronological order, so a
     * work element younger than any that's not too old is not too old
     */

    hptr = &pending_list;
    list_for_each_safe(lptr, tptr, hptr) {
      pq_p = list_entry(lptr, struct work_element, liste);
      if (pq_p->requestsent < timelimit) { // too old!
        pq_p->returncode = -ETIMEOUT;      // set return code and status
        pq_p->status[0] |= STAT_FAILED;
        list_del(lptr);
        pendingqCount--;
        pq_p->audit[1] |= FP_TIMEDOUT;
        pq_p->audit[1] |= FP_NOTPENDING;
        pq_p->audit[1] |= FP_AWAKENING;
        atomic_set(&pq_p->alarmrung,1);
        wake_up(&pq_p->waitq);
      } // end too old
      else {
        break;
      } // end else not too old
    } // end list_for_each

    //
    // If pending count is zero, items left on the request queue may
    // never be processed.
    //
    if (pendingqCount <= 0) {
      hptr = &request_list;
      list_for_each_safe(lptr, tptr, hptr) {
        pq_p = list_entry(lptr, struct work_element, liste);
        if (pq_p->requestsent < timelimit) { // too old!
          pq_p->returncode = -ETIMEOUT;      // set return code and status
          pq_p->status[0] |= STAT_FAILED;
          list_del(lptr);
          requestqCount--;
          pq_p->audit[1] |= FP_TIMEDOUT;
          pq_p->audit[1] |= FP_REMREQUEST;
          pq_p->audit[1] |= FP_AWAKENING;
          atomic_set(&pq_p->alarmrung,1);
          wake_up(&pq_p->waitq);
        } // end too old
        else {
          break;
        } // end else not too old
      } // end list_for_each
    } // end if pendingqCount 0
  } // end else deviceCount isn't zero

  // Unserialize
  spin_unlock_irq(&queuespinlock);

  /*
   * Reschedule self.
   */
  if ((rc = z90crypt_schedule_cleanup_task()))
  {
    /*
     * Currently, there is nothing that could go wrong, but in the future,
     * there may be.  This is a catch-all.
     */
    PRINTK("Error %d detected in schedule_cleanup_task.\n", rc);
  }
  else
  {
//  PDEBUG("schedule_cleanup_task successful.\n");
  }

}

/*
 * Helper functions
 */
int
z90crypt_schedule_reader_timer (void)
{
  //PDEBUG("schedule_reader_timer -> jiffies %ld\n", jiffies);
  do {
    // if this timer is already scheduled, don't schedule it again
    if (timer_pending(&reader_timer)) {
      break;
    }

    //
    // Initialize the timer structure and add the timer.
    //
    // HZ jiffies = 1 second.
    //
    // We will execute the reader task every READERTIME milliseconds.
    //
    reader_timer.function = z90crypt_schedule_reader_task;
    reader_timer.data = (unsigned long) NULL;
    reader_timer.expires = jiffies + (READERTIME * HZ / 1000);
    add_timer(&reader_timer);
  } while(0);

  //
  // In the future, there may be errors to report.
  //
  return 0;
}

void
z90crypt_schedule_reader_task (unsigned long ptr)
{
  tasklet_schedule(&reader_tasklet);
}

int
z90crypt_schedule_config_task (unsigned int expiration)
{
  PDEBUG("schedule_config_task -> jiffies %ld\n", jiffies);

  /*
   * Initialize the timer structure and add the timer.
   * We will execute the config task every CONFIGTIME seconds.
   */
  do {
    if (timer_pending(&config_timer))
        break;
    config_timer.function = z90crypt_config_task;
    config_timer.data = (unsigned long) NULL;
    config_timer.expires = jiffies + (expiration * HZ);
    add_timer(&config_timer);
  } while(0);

  /*
   * In the future, there may be errors to report.
   */
  return 0;
}

int
z90crypt_schedule_cleanup_task (void)
{
//PDEBUG("schedule_cleanup_task -> jiffies %ld\n", jiffies);

  /*
   * Initialize the timer structure and add the timer.
   *
   * HZ jiffies = 1 second.
   *
   * We will execute the cleanup task every CLEANUPTIME seconds.
   */
  do {
    if (timer_pending(&cleanup_timer))
      break;
    cleanup_timer.function = z90crypt_cleanup_task;
    cleanup_timer.data = (unsigned long) NULL;
    cleanup_timer.expires = jiffies + (CLEANUPTIME * HZ);
    add_timer(&cleanup_timer);
  } while(0);

  /*
   * In the future, there may be errors to report.
   */
  return 0;
}


/*-------------------------------------------------------------------*/
/*                                                                   */
/* Lowlevel Functions:                                               */
/*                                                                   */
/*   createZ90crypt:  creates and initializes basic data structures  */
/*                                                                   */
/*   refreshZ90crypt:  re-initializes basic data structures          */
/*                                                                   */
/*   findCryptoDevices:  returns a count and mask of hardware status */
/*                                                                   */
/*   createCryptoDevice:  builds the descriptor for a device         */
/*                                                                   */
/*   destroyCryptoDevice:  unallocates the descriptor for a device   */
/*                                                                   */
/*   sendToCryptoDevice:  sends work to a device                     */
/*                                                                   */
/*   receiveFromCryptoDevice:  solicits completed work from a device */
/*                                                                   */
/*   queryCryptoDevice:  returns status information about the device */
/*                                                                   */
/*   destroyZ90crypt:  drains all work, unallocates structs          */
/*                                                                   */
/*   getZ90cryptStatus:  returns status of crypto devices, amount    */
/*                       of work on each.                            */
/*                                                                   */
/*-------------------------------------------------------------------*/


/*-------------------------------------------------------------------*/
/* Buffer size for receiving responses.  The maximum Response Size   */
/* is actually the maximum request size, since in an error condition */
/* the request itself may be returned unchanged.                     */
/*-------------------------------------------------------------------*/
#define MAX_RESPONSE_SIZE 0x00000710 

static UCHAR NULL_psmid[8] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

/*-------------------------------------------------------------------*/
/* MIN_MOD_SIZE is a leeds2 limit, but it's not unreasonable to      */
/* apply it to both device types.  MAX_MOD_SIZE is a hard limit      */
/* for both cards.                                                   */
/*-------------------------------------------------------------------*/
#define MIN_MOD_SIZE 64                         // bytes             
#define MAX_LEEDS2_MOD_SIZE 128                 // bytes
#define MAX_CCF_MOD_SIZE 128                    // bytes
#define MAX_MOD_SIZE 256	

/*-------------------------------------------------------------------*/
/* Used in device configuration functions                            */
/*-------------------------------------------------------------------*/
#define MAX_RESET 90

/*-------------------------------------------------------------------*/
/* Function prototypes for lowlevel routines                         */ 
/*-------------------------------------------------------------------*/
DEVSTAT queryOnLine (int deviceNr, int cdx, int resetNr, int *q_depth);
/*********************************************************************/
/*                                                                   */
/*           inputs:  deviceNr -- a device number                    */
/*                    cdx      -- the crypto domain index            */
/*                    resetNr  -- internal.  Should be 45            */
/*                                                                   */
/*    implied input:  none                                           */
/*                                                                   */
/*          outputs:  q_depth  -- maximum number of outstanding      */
/*                                requests for this device           */
/*                                                                   */
/*     return value:  one of {HD_NOT_THERE, HD_BUSY, HD_ONLINE}      */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*********************************************************************/

DEVSTAT resetDevice (int deviceNr, int cdx, int resetNr);
/*********************************************************************/
/*                                                                   */
/*           inputs:  deviceNr -- a device number                    */
/*                    cdx      -- the crypto domain index            */
/*                    resetNr  -- internal.  Should be 45            */
/*                                                                   */
/*    implied input:  none                                           */
/*                                                                   */
/*          outputs:  no output parameters                           */
/*                                                                   */
/*     return value:  one of {DEV_ONLINE or DEV_BUSY}                */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*********************************************************************/

DEVSTAT sendToAP(int dev_nr,
                 int cdx,
                 int msg_len,
                 UCHAR * msg_ext_p);
/*********************************************************************/
/*                                                                   */
/*           inputs:  dev_nr   -- an AP number                       */
/*                    cdx      -- a crypto domain index              */
/*                    msg_len  -- length of input message            */
/*                    msg_ext_p-- a message of type depending on     */
/*                                the device type to be equeued to   */
/*                                the device, *preceded* by a 12     */
/*                                byte prefix.                       */
/*                                                                   */
/*                                Note that msg_len above does *not* */
/*                                include the 12 byte prefix.        */
/*                                                                   */
/*     hidden input:  psmid    -- 8-byte id string (bytes 5-12       */
/*                                of the of the 12 byte prefix).     */
/*                                                                   */
/*    implied input:  none                                           */
/*                                                                   */
/*          outputs:  no output parameters                           */
/*                                                                   */
/*     return value:  a device status                                */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*********************************************************************/

DEVSTAT receiveFromAP(int dev_nr,
                      int cdx,
                      int resp_l,
                      unsigned char * resp,
                      unsigned char * psmid);
/*********************************************************************/
/*                                                                   */
/*           inputs:  dev_nr   -- an AP number                       */
/*                    cdx      -- a crypto domain index              */
/*                    resp_l   -- length of output message buffer    */
/*                                                                   */
/*          outputs:  resp     -- output message                     */
/*                    psmid    -- 8-byte id string (bytes 4-12       */
/*                                of the of the 12 byte prefix.      */
/*                                                                   */
/*     return value:  a device status                                */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*********************************************************************/

int convertRequest (UCHAR * buffer,
                    int func,
                    short function,
                    UCHAR * pad_rule,
                    int cdx,
                    CDEVICE_TYPE dev_type,
                    int * msg_l_p,
                    UCHAR * msg_p);

int convertResponse (UCHAR * response,
                     CDEVICE_TYPE dev_type,
                     int icaMsg_l,
                     UCHAR * buffer,
                     short function,
                     UCHAR * pad_rule,
                     int *respbufflen,
                     UCHAR * respbuff);

int createZ90crypt (int * cdx_p)
/*********************************************************************/
/*                                                                   */
/* implied input:  z90crypt                                          */
/*                                                                   */
/*         input:  *cdx_p             --  -1 or the domain specified */
/*                                        by caller                  */
/*                                                                   */
/*        output:  *cdx_p             --  crypto domain index        */
/*                                                                   */
/* At start-up, z90crypt calls this method.                          */
/*                                                                   */
/*********************************************************************/
{
  int rv = OK;
  Z90C_HDWARE_BLOCK * hdware_blk_p = NULL;

  memset(&z90crypt,0,sizeof(struct z90crypt));
  z90crypt.z90c_domain_established = FALSE;
  z90crypt.z90c_len = sizeof(struct z90crypt);
  z90crypt.z90c_max_count = Z90CRYPT_NUM_DEVS;

  do {
    if ((hdware_blk_p = (Z90C_HDWARE_BLOCK *)
        kmalloc(sizeof(struct z90c_hdware_block),GFP_ATOMIC)) == NULL){
      PDEBUG("kmalloc for hardware block failed\n");
      rv = ENOMEM;
      break;
    }
    memset(hdware_blk_p,0,sizeof(struct z90c_hdware_block));

    z90crypt.z90c_hdware_info = hdware_blk_p;

    hdware_blk_p->z90c_nr_types = Z90CRYPT_NUM_TYPES;

  } while(0);

  return (rv);

}; // end createZ90crypt

int refreshZ90crypt(int * cdx_p)
/*********************************************************************/
/*                                                                   */
/*        inputs:  no explicit inputs                                */
/*                                                                   */
/* implied input:  z90crypt                                          */
/*                                                                   */
/* This function, called by the config task at startup and           */
/* periodically thereafter, returns the status of each device.       */
/*                                                                   */
/* Ensures the z90crypt has been initialized.                        */
/*                                                                   */
/* If no devices have been found yet, probes for crypto domain.      */
/*                                                                   */
/* Calls findCryptoDevices.  Compares results with previous results; */
/*                                                                   */
/* If there hasn't been any change:                                  */
/*   returns 0                                                       */
/*                                                                   */
/* If there's been any change:                                       */
/*                                                                   */
/* a.  Serializes by getting the queuespinlock                       */
/*                                                                   */
/* b.  For each device:                                              */
/*                                                                   */
/*     1) if the device was not HD_ONLINE but is now,                */
/*          calls createCryptoDevice                                 */
/*     2) if the device was HD_ONLINE but isn't any longer,          */
/*          calls destroyCryptoDevice                                */
/*                                                                   */
/* c.  copies the new hardware status mask into                      */
/*     z90crypt.z90c_hdware_info->z90c_hdware_mask.                  */
/*                                                                   */
/* d.  Refreshes the short index arrays for:                         */
/*     1) the overall device index array z90crypt.z90c_mask          */
/*     2) the leedslite device index array                           */
/*          z90crypt.z90c_hdware_info->z90c_type_mask[0]             */
/*     3) the leeds2 device index array                              */
/*          z90crypt.z90c_hdware_info->z90c_type_mask[1]             */
/*                                                                   */
/*     Note that all three of these masks will have been modified    */
/*     by createCryptoDevice and/or destroyCryptoDevice              */
/*                                                                   */
/* e.  Unserializes by giving back the queuespinlock                 */
/*                                                                   */
/* f.  Returns 0                                                     */
/*                                                                   */
/*********************************************************************/
{
  int i;
  int rv = OK;
  Z90C_STATUS_STR local_mask;
  Z90C_DEVICE * devPtr;
  int j;
  int indx;
  UCHAR oldStat;
  UCHAR newStat;
  BOOL return_unchanged = TRUE;

  if (z90crypt.z90c_len != sizeof(struct z90crypt))
    return ENOTINIT;

  // If the terminating flag is on, don't refresh
  if (z90crypt.z90c_terminating == 1)
    return TSQ_FATAL_ERROR;

  // If no devices have been found yet, first probe for domain
  if (z90crypt.z90c_hdware_info->z90c_hdware_mask.z90c_st_count == 0 &&
      z90crypt.z90c_domain_established == FALSE)
    rv = probeCryptoDomain(cdx_p);

  if (z90crypt.z90c_terminating == 1)
    return TSQ_FATAL_ERROR;

  if (rv == 0) {
    z90crypt.z90c_cdx = *cdx_p;
    z90crypt.z90c_domain_established = TRUE;
  }
  else {
    PRINTK("Probe Domain returned %d\n",rv);
    return (rv);
  }

  if ((rv = findCryptoDevices(&local_mask)) != OK) {
    PRINTK("Find Crypto Devices returned %d\n",rv);
    return (rv);
  }
  if ((rv = memcmp((void *)&local_mask,
                   (void *)&(z90crypt.z90c_hdware_info->z90c_hdware_mask),
                   sizeof(struct z90c_status_str)))== OK) { // no change
    return_unchanged = TRUE;
    for (i=0;i<Z90CRYPT_NUM_TYPES;i++) {

      // Check for disabled cards.  If any device is marked disabled,
      // destroy it.

      for (j = 0; j < z90crypt.z90c_hdware_info->z90c_type_mask[i].z90c_st_count;j++)      {
        indx = z90crypt.z90c_hdware_info->z90c_type_x_addr[i].z90c_device_index[j];
        devPtr = z90crypt.z90c_device_p[indx];
        if (devPtr != NULL && devPtr->z90c_disabled == TRUE) {
          local_mask.z90c_st_mask[indx] = HD_NOT_THERE;
          return_unchanged = FALSE;
        } // end if
      } // end for j ...
    } // end for i ...

    if (return_unchanged == TRUE)
      return OK;
  }

  rv = OK;

  spin_lock_irq(&queuespinlock);

  for (i =  0; i < z90crypt.z90c_max_count; i++) {
    oldStat = z90crypt.z90c_hdware_info->z90c_hdware_mask.z90c_st_mask[i];
    newStat = local_mask.z90c_st_mask[i];
    if (oldStat == HD_ONLINE && newStat != HD_ONLINE) {
      destroyCryptoDevice(i);
    }
    else if (oldStat != HD_ONLINE && newStat == HD_ONLINE) {
      rv = createCryptoDevice(i);
      if (rv >= REC_FATAL_ERROR)
        break;
      if (rv != OK) {
        local_mask.z90c_st_mask[i] = HD_NOT_THERE;
      }
    }
  }

  if (rv >= REC_FATAL_ERROR)
    return rv;

  memcpy((void *)&z90crypt.z90c_hdware_info->z90c_hdware_mask,
         (void *)&local_mask,
         sizeof(struct z90c_status_str));

  refreshIndexArray(&z90crypt.z90c_mask,
                    &z90crypt.z90c_overall_device_x);

  for (i=0;i<Z90CRYPT_NUM_TYPES;i++) {
    refreshIndexArray(
                &(z90crypt.z90c_hdware_info->z90c_type_mask[i]),
                &(z90crypt.z90c_hdware_info->z90c_type_x_addr[i]));
  } // end for i

  spin_unlock_irq(&queuespinlock);

  return (rv);

}; // end refreshZ90crypt

int findCryptoDevices(Z90C_STATUS_STR * deviceMask)
/*********************************************************************/
/*                                                                   */
/*         inputs:  none                                             */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  deviceMask                                       */
/*                                                                   */
/*        process:  1.  Zeroes the count                             */
/*                  2.  For each possible device number | cdx        */
/*                        calls queryDevStat to get the status       */
/*                        if the device is online or active, records */
/*                        the queue depth in the queue depth array.  */
/*                        Note that access to this array is *not*    */
/*                        serialized.                                */
/*                  3.  Returns the deviceMask                       */
/*                                                                   */
/*  This function is called by refreshZ90crypt.                      */
/*                                                                   */
/*  Note that this process is completely non-intrusive, therefore    */
/*  the configlock need not be held by a process calling this        */
/*  function.                                                        */
/*                                                                   */
/*********************************************************************/
{
  int rv= OK;
  int i;                      // loop counter
  int q_depth;                // number of work elements at one time
  HDSTAT hd_stat;             // device status

  deviceMask->z90c_st_count = 0;

  for (i=0;i<z90crypt.z90c_max_count;i++) {
    hd_stat = queryOnLine(i, z90crypt.z90c_cdx, MAX_RESET, &q_depth);
    if (hd_stat == HD_TSQ_EXCEPTION) {
      rv = TSQ_FATAL_ERROR;
      z90crypt.z90c_terminating = 1;
      PRINTKC("Exception during probe for crypto devices\n");
      break;
      }
    deviceMask->z90c_st_mask[i] = hd_stat;
    if (hd_stat == HD_ONLINE) {
      PDEBUG("Got an online crypto!: %d\n",i);
      PDEBUG("Got a queue depth of %d\n",q_depth);
      deviceMask->z90c_st_count++;
      if (q_depth <= 0) {
        rv = TSQ_FATAL_ERROR;
        break;
      }
      z90crypt.z90c_q_depth_array[i] = q_depth;
    } // end if hd_stat = hd_online
  } // end for(i...

  return (rv);
} // end findCryptoDevices

int probeCryptoDomain(int * cdx_p)
/*********************************************************************/
/*                                                                   */
/*         inputs:  none                                             */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  the crypto domain                                */
/*                                                                   */
/*        process:  1.  Zeroes the count                             */
/*                  2.  For each possible device number | cdx        */
/*                        calls queryDevStat to get the status       */
/*                        if the device is online or active.         */
/*                  3.  Returns the domain in which the above        */
/*                      condition was encountered.                   */
/*                                                                   */
/*  This function is called by refreshZ90crypt.                      */
/*                                                                   */
/*  Note that this process is completely non-intrusive, therefore    */
/*  the  configlock need not be held by a process calling this       */
/*  function.                                                        */
/*                                                                   */
/*********************************************************************/
{
  int rv = 0;
  int correct_cdx_found = 0;
  int i,j,k;                  // loop counters
  HDSTAT hd_stat;             // device status
  int cdx_array[16];          // array of found cdx's
  int q_depth = 0;

  k = 0;
  do {
  for (i=0;i<z90crypt.z90c_max_count;i++) {
    hd_stat = HD_NOT_THERE;
    for (j=0;j<15;cdx_array[j++]=-1);
    k = 0;
    for (j=0;j<15;j++){
      hd_stat = queryOnLine(i, j, MAX_RESET, &q_depth);
      if (hd_stat == HD_TSQ_EXCEPTION) {
        z90crypt.z90c_terminating = 1;
        PRINTKC("Exception during probe for crypto domain\n");
        break;
      }
      if (hd_stat == HD_ONLINE) {
        cdx_array[k++] = j;
        if (*cdx_p == j) {
          correct_cdx_found  = 1;
          break;
        }
      }
    } // end for(j...

    if (correct_cdx_found == 1 || k != 0)
      break;

    if (z90crypt.z90c_terminating)
      break;

  } // end for(i...

  if (z90crypt.z90c_terminating)
    return (TSQ_FATAL_ERROR);

  if (correct_cdx_found) {
    rv = 0;
    break;
  }

  switch(k) {
    case 0:
      *cdx_p = 0;
      rv = 1;
      break;

    case 1:
      if (*cdx_p == -1 || z90crypt.z90c_domain_established == FALSE) {
        *cdx_p = cdx_array[0];
        rv = 0;
      }
      else {
        if (*cdx_p != cdx_array[0]) {
          PRINTK("Specified cdx: %d.  Assigned cdx: %d\n",*cdx_p,cdx_array[0]);
          rv = Z90C_INCORRECT_DOMAIN;
        }
        else
          rv = 0;
      }
      break;

    default:
      rv = Z90C_AMBIGUOUS_DOMAIN;
      break;
  } // end switch
  } while(0);

  return (rv);
} // end probeCryptoDomain

int refreshIndexArray(struct z90c_status_str *status_str,
                      struct z90c_device_x *index_array)
/*********************************************************************/
/*                                                                   */
/*           input:  status_str -- a count and mask of device status */
/*                                                                   */
/*   implied input:  none                                            */
/*                                                                   */
/*          output:  index_array -- a count and array of indexes     */
/*                                                                   */
/*                   returns count                                   */
/*                                                                   */
/*********************************************************************/
{
  int i = -1;
  int count = 0;
  DEVSTAT stat;

  do {
    stat = status_str->z90c_st_mask[++i];
    if (stat == DEV_ONLINE)
      index_array->z90c_device_index[count++] = i;
  } while (i<Z90CRYPT_NUM_DEVS && count < status_str->z90c_st_count);

  return count;

} // end refreshIndexArray

int createCryptoDevice(int index)
/*********************************************************************/
/*                                                                   */
/*         inputs:  index                                            */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/* implied output:  the z90crypt  the device descriptor is           */
/*                                allocated and pointed to from      */
/*                                z90crypt.z90c_device_p[index]      */
/*                                                                   */
/* During refreshZ90crypt, the driver calls this function for any    */
/* device that has just come on line.  Note that this function       */
/* assumes the  configlock is held exclusive by the caller.          */
/*                                                                   */
/* The last step in this process is to invoke init_semaphore         */
/*********************************************************************/
{
  int rv = OK;
  Z90C_DEVICE * dev_ptr;
  struct z90c_status_str * type_str_p;
  int devstat = DEV_GONE;
  CDEVICE_TYPE deviceType = LEEDSLITE;
  int total_size;

  if((dev_ptr = z90crypt.z90c_device_p[index]) == NULL) {
    total_size = sizeof(struct z90c_device) +
                 z90crypt.z90c_q_depth_array[index] * sizeof(int);

    if ((dev_ptr =
            (Z90C_DEVICE *)kmalloc(total_size,GFP_ATOMIC))==NULL){
      PRINTK("Could not malloc device %i\n",index);
      return ENOMEM;
    }

    memset(dev_ptr,0,total_size);

    if ((dev_ptr->z90c_dev_resp_p=kmalloc(MAX_RESPONSE_SIZE,GFP_ATOMIC))
                                                                  ==NULL){
      kfree(dev_ptr);
      PRINTK("Could not malloc device %i receiving buffer\n",index);
      return ENOMEM;
    }
    dev_ptr-> z90c_dev_resp_l = MAX_RESPONSE_SIZE;
    INIT_LIST_HEAD(&(dev_ptr->z90c_dev_caller_list));
  }

  devstat = resetDevice(index, z90crypt.z90c_cdx, MAX_RESET);
  if (devstat == DEV_RSQ_EXCEPTION) {
    PRINTK("Exception during reset device %d\n",index);
    kfree(dev_ptr->z90c_dev_resp_p);
    kfree(dev_ptr);
    return RSQ_FATAL_ERROR;
  }

  if (devstat == DEV_ONLINE) {
    dev_ptr->z90c_dev_self_x = index;
    rv = probeDeviceType(dev_ptr);
    if (rv != OK) {
      kfree (dev_ptr->z90c_dev_resp_p);
      kfree (dev_ptr);
      return rv;
    }
    deviceType = dev_ptr->z90c_dev_type;
    z90crypt.z90c_hdware_info->z90c_device_type_array[index] =
                                           dev_ptr->z90c_dev_type + 1;
  }

  dev_ptr-> z90c_dev_len = sizeof(struct z90c_device);
  dev_ptr-> z90c_dev_q_depth = z90crypt.z90c_q_depth_array[index];
  dev_ptr-> z90c_dev_stat = devstat;
  dev_ptr-> z90c_disabled = FALSE;
  z90crypt.z90c_device_p[index] = dev_ptr;

  if (devstat == DEV_ONLINE) {

    if (z90crypt.z90c_mask.z90c_st_mask[index] != DEV_ONLINE) {
      z90crypt.z90c_mask.z90c_st_mask[index] = DEV_ONLINE;
      z90crypt.z90c_mask.z90c_st_count++;
    }

    type_str_p = &(z90crypt.z90c_hdware_info->z90c_type_mask[deviceType]);
    if (type_str_p->z90c_st_mask[index] != DEV_ONLINE) {
      type_str_p->z90c_st_mask[index] = DEV_ONLINE;
      type_str_p->z90c_st_count++;
    }
  }

  return OK;
} // end createCryptoDevice

int destroyCryptoDevice(int index)
/*********************************************************************/
/*                                                                   */
/*         inputs:  index                                            */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  none                                             */
/*                                                                   */
/* implied output:  the z90crypt  the device descriptor is           */
/*                                unallocated and the pointer        */
/*                                z90crypt.z90c_device_p[index]      */
/*                                is nullified.                      */
/*                                                                   */
/* During refreshZ90crypt, the driver calls this function for any    */
/* device that has just gone off line.  Note that this function      */
/* assumes the  configlock is held exclusive by the caller.          */
/*                                                                   */
/*********************************************************************/
{
  Z90C_DEVICE * dev_ptr;
  CDEVICE_TYPE t = -1;
  BOOL disabledFlag = FALSE;

  dev_ptr = z90crypt.z90c_device_p[index];

  // remember device type; get rid of device struct
  if (dev_ptr != NULL) {
    disabledFlag = dev_ptr->z90c_disabled;
    t = dev_ptr->z90c_dev_type;
    if (dev_ptr->z90c_dev_resp_p != NULL)
      kfree(dev_ptr->z90c_dev_resp_p);
    kfree(dev_ptr);
  }

  // remove the address of the struct from the array of device struct addresses
  z90crypt.z90c_device_p[index] = (Z90C_DEVICE *) NULL;

  // if the type is valid, remove the device from the type_mask
  if (t != -1) {
    if (z90crypt.z90c_hdware_info->z90c_type_mask[t].z90c_st_mask[index] != 0x00) {
      z90crypt.z90c_hdware_info->z90c_type_mask[t].z90c_st_mask[index] = 0x00;
      z90crypt.z90c_hdware_info->z90c_type_mask[t].z90c_st_count--;
      if (disabledFlag == TRUE)
        z90crypt.z90c_hdware_info->z90c_type_mask[t].z90c_disabled_count--;
    }
  }

  // Get the device out of the overall status array
  if (z90crypt.z90c_mask.z90c_st_mask[index] != DEV_GONE) {
    z90crypt.z90c_mask.z90c_st_mask[index] = DEV_GONE;
    z90crypt.z90c_mask.z90c_st_count--;
  }

  // Change the device type to null (00)
  z90crypt.z90c_hdware_info->z90c_device_type_array[index] = 0;

  return OK;
} // end destroyCryptoDevice


int buildCaller(char * psmid,
                int func,
                short function,
                UCHAR * pad_rule,
                int buff_len,
                char * buff_ptr,
                CDEVICE_TYPE dev_type,
                Z90C_CALLER * caller_p)
/*********************************************************************/
/*  builds struct z90c_caller, converts message from generic format  */
/*  to device-dependent format.                                      */
/*                                                                   */
/*         inputs:  psmid -- an 8 byte identifier                    */
/*                                                                   */
/*                  func -- ICARSAMODEXPO or ICARSACRT as defined    */
/*                          in z90crypt.h                            */
/*                                                                   */
/*                  buff_len -- length of caller's buffer            */
/*                                                                   */
/*                  buff_ptr -- address of caller's buffer           */
/*                                                                   */
/*                  dev_type  -- a device type                       */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  caller_p -- contents of a struct z90c_caller     */
/*                                                                   */
/*                  returns rv:                                      */
/*                      0 -- all went well                           */
/*                      4 --                                         */
/*                     12 -- device not available                    */
/*                                                                   */
/*        process:  1.  If something's wrong with the device type,   */
/*                      returns rv = device not available.           */
/*                                                                   */
/*                  2.  Fills in the input caller struct with        */
/*                      the correct information, returns rv = ok     */
/*                                                                   */
/*********************************************************************/
{
  int rv = OK;
  UCHAR * temp_up = NULL;

  if (dev_type != LEEDS2 && dev_type != LEEDSLITE) {
    return (SEN_NOT_AVAIL);
  }

  memcpy(caller_p->z90c_caller_id,psmid,sizeof(caller_p->z90c_caller_id));
  caller_p->z90c_caller_func = func;
  caller_p->z90c_caller_function = function;
  caller_p->z90c_caller_dev_dep_req_p = caller_p->z90c_caller_dev_dep_req;
  caller_p->z90c_caller_dev_dep_req_l = MAX_RESPONSE_SIZE; // re-init buffer len
  memcpy(caller_p->z90c_pad_rule, pad_rule, sizeof(caller_p->z90c_pad_rule));
  caller_p->z90c_caller_buf_p = buff_ptr;
  INIT_LIST_HEAD(&(caller_p->z90c_caller_liste));

  if ((rv = convertRequest(buff_ptr,
                           func,
                           function,
                           pad_rule,
                           z90crypt.z90c_cdx,
                           dev_type,
                           &(caller_p->z90c_caller_dev_dep_req_l),
                           caller_p->z90c_caller_dev_dep_req_p))!=OK) {
    return (rv);
  }

  temp_up = &(caller_p->z90c_caller_dev_dep_req_p[4]);
  memcpy(temp_up, psmid, 8);

  return rv;

} // end buildCaller

int unbuildCaller(Z90C_DEVICE * device_p,
                  Z90C_CALLER * caller_p)
/*********************************************************************/
/*  cleans up caller struct                                          */
/*  pushes the caller struct on the device's caller stack            */
/*                                                                   */
/*         inputs:  device_p  -- address of a device descriptor      */
/*                                                                   */
/*                  caller_p -- address of a struct z90c_caller      */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*                                                                   */
/*                  returns rv:                                      */
/*                      8 -- queue empty                             */
/*                                                                   */
/*         process: 1.  clean up caller struct                       */
/*                                                                   */
/*                  2.  push the caller struct on the device caller  */
/*                      stack                                        */
/*                                                                   */
/*********************************************************************/
{
  if (caller_p != NULL) {
    memset(caller_p->z90c_caller_id,0,sizeof(caller_p->z90c_caller_id));
    caller_p->z90c_caller_func = 0;
    caller_p->z90c_caller_function = 0;
    caller_p->z90c_caller_dev_p = NULL;
    memset(caller_p->z90c_pad_rule, 0, sizeof(caller_p->z90c_pad_rule));
  }

  return 0;

} // end unbuildCaller

int removeDevice(Z90C_DEVICE * device_p)
/*********************************************************************/
/*                                                                   */
/*         inputs:  device_p  -- address of a device descriptor      */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*                  returns rv = 0                                   */
/*                                                                   */
/*         process: 2.  get the device semaphore                     */
/*                                                                   */
/*                  3.  call removeDeviceFromType                    */
/*                                                                   */
/*                  4.  remove the device from the mask/index array  */
/*                                                                   */
/*                  5.  give back the device semaphore               */
/*                                                                   */
/*                  6.  remove the device                            */
/*                                                                   */
/*                  8.  return rv = 0                                */
/*                                                                   */
/*********************************************************************/
{
  int rv = OK;
  int indx;
  CDEVICE_TYPE devTp = -1;

  if ((device_p != NULL) &&
      (device_p->z90c_disabled == FALSE)) {
    device_p->z90c_disabled = TRUE;
    indx = device_p->z90c_dev_self_x;
    devTp = device_p->z90c_dev_type;
    z90crypt.z90c_hdware_info->z90c_type_mask[devTp].z90c_disabled_count++;
  }

  return (rv);

} // end removeDevice

int getCryptoRequestBuffer(UCHAR * psmid,
                     int func,
                     int buff_len,
                     UCHAR * buff_ptr,
                     CDEVICE_TYPE * devType_p,
                     unsigned char * caller_p)
/*********************************************************************/
/*                                                                   */
/*         inputs:  psmid -- an 8 byte identifier                    */
/*                                                                   */
/*                  func -- ICARSAMODEXPO or ICARSACRT as defined in */
/*                          z90crypt.h                               */
/*                                                                   */
/*                  buff_len -- length of caller's buffer            */
/*                                                                   */
/*                  buff_ptr -- address of caller's buffer           */
/*                                                                   */
/*                  *devType_p -- 0 (LeedsLite) or 1 (Leeds2)  or -1 */
/*                                (no preference)                    */
/*                                                                   */
/*                  reqBuffp  -- address of the request buffer       */
/*                                                                   */
/* The caller's buffer should contain an instance of either of two   */
/* types in z90crypt.h:                                              */
/*                                                                   */
/* 1. ica_rsa_modexpo_crt_t                                          */
/*                                                                   */
/*    char         *inputdata;                                       */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *bp_key;                                          */
/*    char         *bq_key;                                          */
/*    char         *np_prime;                                        */
/*    char         *nq_prime;                                        */
/*    char         *u_mult_inv;                                      */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*  inputdatalength is even and inputdata contains no leading zeroes */
/*                                                                   */
/*  bp_keylength = np_primelength = u_mult_inv_length =              */
/*                                             8 + inputdatalength/2 */
/*  and                                                              */
/*                                                                   */
/*  bq_keylength = nq_primelength = inputdatalength/2                */
/*                                                                   */
/* 2. ica_rsa_modexpo_t                                              */
/*                                                                   */
/*    char         *inputdata                                        */
/*    unsigned int  inputdatalength;                                 */
/*    char         *outputdata;                                      */
/*    unsigned int  outputdatalength;                                */
/*    char         *b_key;                                           */
/*    char         *n_modulus;                                       */
/*                                                                   */
/* Note: it is assumed that                                          */
/*                                                                   */
/*         b_keylength = n_moduluslength = inputdatalength           */
/*                                                                   */
/*                                                                   */
/* Note that in both types, all key parts are right justified in     */
/* their fields, padded on the left with zeroes.                     */
/*                                                                   */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  *devType_p -- 0 (LeedsLite) or 1 (Leeds2) to     */
/*                                which the request was sent.        */
/*                                                                   */
/*                                                                   */
/* implied output:  None.                                            */
/*                                                                   */
/*********************************************************************/
{
  int rv = 0;
  int parmLen = 0;
  ica_rsa_modexpo_t * mex_p;
  ica_rsa_modexpo_crt_t * crt_p;
  ext_input_t * inp_p;
  UCHAR temp_buffer[256];
  short function;
  UCHAR pad_rule[8];

  PDEBUG("device type input to getCryptoRequestBuffer: %i\n",*devType_p);

  // Refuse request if z90c_terminating flag is on
  if (z90crypt.z90c_terminating == 1){
    return (REC_NO_RESPONSE);
  }

  /*
   * The psmid can't be NULL or nothing works at all.
   *
   * The buffer can't be NULL or nothing works at all.
   *
   * Input data length has to be even, at least 64 and at most 256.
   *
   * Output data length will be checked before any store, must be
   * at least 16.
   */

  if ((rv=memcmp(psmid,NULL_psmid,8))==0)
    return (SEN_FATAL_ERROR);

  rv = 0;

  if (buff_ptr == NULL) {
    PRINTK("Buffer pointer Null in getCryptoRequestBuff\n");
    return (SEN_USER_ERROR);
  }

  mex_p = (ica_rsa_modexpo_t *)buff_ptr;
  parmLen = mex_p->inputdatalength;
  PDEBUG("send: parmLen at top = %d\n",parmLen);
  PDEBUG("send: buffLen at top = %d\n",buff_len);
  if ((parmLen != 2*(parmLen/2)) ||
      (parmLen > MAX_MOD_SIZE)) {
    return (SEN_USER_ERROR);
  }

  if (mex_p->outputdatalength < sizeof(ext_input_t)) {
    return (SEN_USER_ERROR);
  }

  /*
   * Function must be ICARSAMODEXPO or ICARSACRT
   *
   * Buffer length has to be consistent with function
   *
   * None of the key or data address can be NULL
   */

  switch(func) {
    case ICARSAMODEXPO:
      if (buff_len < sizeof(ica_rsa_modexpo_t)) {
        rv = SEN_USER_ERROR;
        break;
      }

      if ((mex_p->inputdata==NULL) ||
          (mex_p->outputdata==NULL) ||
          (mex_p->b_key == NULL) ||
          (mex_p->n_modulus == NULL) ||
          (mex_p->inputdatalength <= 0) ||
          (mex_p->inputdatalength > 256)) {
        rv = SEN_USER_ERROR;
      }

      break;

    case ICARSACRT:
      if (buff_len < sizeof(ica_rsa_modexpo_crt_t)) {
        rv = SEN_USER_ERROR;
        break;
      }
      crt_p = (ica_rsa_modexpo_crt_t *)buff_ptr;
      if ((crt_p->inputdata == NULL) ||
          (crt_p->outputdata == NULL) ||
          (crt_p->bp_key == NULL) ||
          (crt_p->bq_key == NULL) ||
          (crt_p->np_prime == NULL) ||
          (crt_p->nq_prime == NULL) ||
          (crt_p->u_mult_inv == NULL) ||
          (crt_p->inputdatalength <= 0) ||
          (crt_p->inputdatalength > 256)) {
        rv = SEN_USER_ERROR;
      }

      break;

    default:
      rv = SEN_USER_ERROR;
  }
  if (rv != OK)
    return (rv);

  /*
   * Device type must be 0, 1 or -1
   */

  if ((*devType_p != LEEDSLITE) &&
      (*devType_p != LEEDS2) &&
      (*devType_p != -1)) {
    return (SEN_USER_ERROR);
  }

  // Device type must be leedslite if this is a modexpo call and
  // modulus size (inputdatalength) exceeds 128

  if (*devType_p == -1 &&
      ((func == ICARSAMODEXPO && parmLen > 128) ||
       (parmLen < MIN_MOD_SIZE)))
    *devType_p = LEEDSLITE;

  // caller_p had better not be null
  if (caller_p == NULL) {
    return (SEN_USER_ERROR);
  }

  if ((rv=copy_from_user(temp_buffer,
                         mex_p->outputdata,
                         sizeof(ext_input_t))) != 0)
    return (SEN_RELEASED);

  inp_p = (ext_input_t *) temp_buffer;
  function = inp_p -> functioncode;

  memset(temp_buffer,0,sizeof(temp_buffer));
  if ((rv=copy_from_user(temp_buffer,
                         mex_p->inputdata,
                         mex_p->inputdatalength)) != 0)
    return (SEN_RELEASED);

  if ((func == ICARSAMODEXPO) && (isPKCS1_2padded(temp_buffer,
                                         mex_p->inputdatalength)))
    function = PCI_FUNC_KEY_ENCRYPT;
  else
    function = PCI_FUNC_KEY_DECRYPT;

  if (function != PCI_FUNC_KEY_ENCRYPT)
    function = PCI_FUNC_KEY_DECRYPT;
  // memcpy (pad_rule, PKCS_PAD_RULE, sizeof(PKCS_PAD_RULE));

  PDEBUG("function: %04x\n",function);

  // Device type must be leedslite if the input is PKCS#1.1 padded
  if ((*devType_p == -1) &&
      (isPkcs1_1Padded(temp_buffer, mex_p->inputdatalength)))
    *devType_p = LEEDSLITE;

  do {
    rv = selectDeviceType(function,
                          devType_p);
    PDEBUG("device type after Select: %i\n",*devType_p);
    if (rv < 0){
      rv = SEN_NOT_AVAIL;
      break;
    }

    if ((rv = buildCaller(psmid,
                          func,
                          function,
                          pad_rule,
                          buff_len,
                          buff_ptr,
                          *devType_p,
                          (Z90C_CALLER *)caller_p)) != OK) {
      break;
    }
  } while (0);

  return (rv);
} // end getCryptoRequestBuffer

int sendToCryptoDevice(UCHAR * psmid,
                     int func,
                     int buff_len,
                     UCHAR * buff_ptr,
                     CDEVICE_TYPE devType,
                     int * devNr_p,
                     unsigned char * reqBuff)
/*********************************************************************/
/*                                                                   */
/*         inputs:  caller_p -- pointer to a caller struct           */
/*                                                                   */
/*         process: 2.  Time stamps the request                      */
/*                                                                   */
/*                  3.  Invokes sendToAP                             */
/*                                                                   */
/*                  4.  An exception may be taken during             */
/*                      send, in which case DEV_SEN_EXCEPTION        */
/*                      will be returned.  In that case, return      */
/*                      SEN_FATAL_ERROR.                             */
/*                                                                   */
/*********************************************************************/
{
  int rv = 0;
  int dv = DEV_ONLINE;
  Z90C_CALLER * caller_p;
  Z90C_DEVICE * device_p;
  int dev_nr;

  if (!(reqBuff))
    return (SEN_FATAL_ERROR);

  caller_p = (Z90C_CALLER *) reqBuff;

  caller_p->z90c_caller_sen_time = jiffies;

  dev_nr = *devNr_p;

  if ((rv = selectDevice(caller_p->z90c_caller_function,&devType,&dev_nr))==-1) {
    if (z90crypt.z90c_hdware_info->z90c_hdware_mask.z90c_st_count != 0) {
      return SEN_RETRY;
    }
    else {
      return SEN_NOT_AVAIL;
    }
  }

  *devNr_p = dev_nr;

  device_p = z90crypt.z90c_device_p[dev_nr];

  if (device_p->z90c_dev_type != devType)
    return SEN_RETRY;

  if (device_p == NULL)
    return SEN_NOT_AVAIL;

  if (device_p->z90c_dev_caller_count > device_p->z90c_dev_q_depth) {
    return SEN_QUEUE_FULL;
  }

  do {
    PDEBUG("device number prior to send:  %i\n",dev_nr);
    if ((dv = sendToAP(dev_nr,
                       z90crypt.z90c_cdx,
                       caller_p->z90c_caller_dev_dep_req_l,
                       caller_p->z90c_caller_dev_dep_req_p)) != OK){
      if (dv==DEV_SEN_EXCEPTION) {
        rv = SEN_FATAL_ERROR;
        z90crypt.z90c_terminating = 1;
        PRINTKC("Exception during send to device %i\n",dev_nr);
        break;
      }
      PDEBUG("return value from sendToAP:  %i\n",rv);
      switch(dv) {
        case DEV_GONE:
          PRINTK("Device %d not available\n", dev_nr);
          removeDevice(device_p);
          rv = SEN_NOT_AVAIL;
          break;
        case DEV_ONLINE:
          rv = OK;
          break;
        case DEV_EMPTY:
          rv = SEN_NOT_AVAIL;
          break;
        case DEV_NO_WORK:
          rv = SEN_FATAL_ERROR;
          break;
        case DEV_BAD_MESSAGE:
          rv = SEN_USER_ERROR;
          break;
        case DEV_QUEUE_FULL:
          rv = SEN_QUEUE_FULL;
          break;
        default:
          break;
      }
      break;
    }
  } while(0);

  // Get the caller on the list
  if (rv == OK) {
    list_add_tail(&(caller_p->z90c_caller_liste),
                  &(device_p->z90c_dev_caller_list));
    ++(device_p->z90c_dev_caller_count);
  }

  return (rv);
} // end sendToCryptoDevice

int receiveFromCryptoDevice(int index,
                            UCHAR * psmid,
                            int * buff_len_p,
                            UCHAR * buff,
                            UCHAR ** dest_p_p)
/*********************************************************************/
/*                                                                   */
/*         inputs:  index -- index of the device from which          */
/*                           the work element is to be received      */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*        outputs:  psmid     -- 8 byte identifier                   */
/*                                                                   */
/*                  *buff_len_p -- length of output                  */
/*                                                                   */
/*                  buff      -- output from receive                 */
/*                                                                   */
/*                  *dest_p_p -- user buffer that is the ultimate    */
/*                               destination of the output.          */
/*                                                                   */
/*                               When control returns, the caller's  */
/*                               results will have been copied to    */
/*                               buff, and the length of the result  */
/*                               will be stored in *buff_len_p.      */
/*                               *dest_p_p will hold the address     */
/*                               into which the output should be     */
/*                               copied from buff.                   */
/*                                                                   */
/*                               In the event of serious error, of   */
/*                               course, outputdata and              */
/*                               outputdatalength will be unchanged. */
/*                                                                   */
/*        process:  1.  Invoke receiveFromAP with input index,       */
/*                      z90crypt.z90c_cdx, the output buffer         */
/*                      pointed to from the z90c_device(index) and   */
/*                      response struct imbedded in                  */
/*                      z90c_device(index).                          */
/*                                                                   */
/*                  2.  If the receiveFromAP returns DEV_ONLINE      */
/*                      (thus indicating success):                   */
/*                                                                   */
/*                      a.  find the caller in the queue pointed to  */
/*                          from the z90c_device by matching the     */
/*                          psmid returned by receiveFromAP to the   */
/*                          the psmid's in the device's callerQueue. */
/*                          If the psmid can't be matched, return    */
/*                          REC_USER_GONE.                           */
/*                                                                   */
/*                      b.  call convertResponse to extract the      */
/*                          reply code and (if the reply code        */
/*                          indicates success) the result of the     */
/*                          operation from the response.             */
/*                                                                   */
/*                          A non-success reply code extracted by    */
/*                          convertResponse will indicate one of     */
/*                          four conditions:                         */
/*                          1) wrong message type.  In that          */
/*                             case, convertResponse will return     */
/*                             BAD_MESSAGE_DRIVER.  This routine     */
/*                             will return REC_FATAL_ERROR.          */
/*                          2) incorrect message part.  In that      */
/*                             case, convertResponse will return     */
/*                             BAD_MESSAGE_USER.  This routine will  */
/*                             return REC_USER_ERROR.                */
/*                          3) incorrect message syntax.  In this    */
/*                             case, convertResponse will return     */
/*                             BAD_MESSAGE_DRIVER.  This routine     */
/*                             will return REC_FATAL_ERROR.          */
/*                          4) storage access error.  In that case,  */
/*                             the REC_RELEASED will be returned.    */
/*                             This routine will simply propagate    */
/*                             the REC_RELEASED error code.          */
/*                                                                   */
/*                  3.  If receiveFromAP returns DEV_DECONFIGURED,   */
/*                      DEV_CHECKSTOPPED or DEV_GONE, call           */
/*                      removeDevice and return REC_NO_RESPONSE.     */
/*                                                                   */
/*                  4.  If receiveFromAP returns DEV_EMPTY, return   */
/*                      REC_EMPTY.                                   */
/*                                                                   */
/*                  5.  If receiveFromAP returns DEV_BAD_MESSAGE,    */
/*                      return REC_NO_RESPONSE.  In that case, the   */
/*                      device is considered in error.               */
/*                                                                   */
/*                  6.  If receiveFromAP returns DEV_NO_WORK, return */
/*                      REC_NO_WORK.                                 */
/*                                                                   */
/*                  7.  An exception may be taken during             */
/*                      receive, in which case DEV_REC_EXCEPTION     */
/*                      will be returned.  In that case, return      */
/*                      REC_FATAL_ERROR.                             */
/*                                                                   */
/*                  8.  Return to caller.                            */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
{
  DEVSTAT dv = DEV_EMPTY;
  int rv = 0;
  Z90C_DEVICE * devPtr;
  Z90C_CALLER * caller_p = NULL;
  CDEVICE_TYPE devType = LEEDS2;
  ica_rsa_modexpo_t * icaMsg_p;
  int temp_rv = 0;     //SCAFFOLDING
  struct list_head *ptr = NULL;
  struct list_head *tptr = NULL;
  struct list_head *hptr = NULL;
#ifdef Z90CRYPT_DEBUG
  int i;
  struct list_head *btr = NULL;
#endif

  memcpy(psmid, NULL_psmid, sizeof(NULL_psmid));

  // If the terminating flag is on, don't try receiving
  if (z90crypt.z90c_terminating == 1)
    return (REC_FATAL_ERROR);

  do {

    PDEBUG("Dequeue called for device %i\n",index);
    devPtr = z90crypt.z90c_device_p[index];
    if (devPtr == (void *)NULL) {
      rv = REC_NO_RESPONSE;
      break;
    }

    if (devPtr->z90c_dev_self_x != index) {
      PRINTK("Corrupt dev ptr in receiveFromAP\n");
      z90crypt.z90c_terminating = 1;
      rv = REC_FATAL_ERROR;
      break;
    }

    if (devPtr->z90c_disabled == TRUE) {
      rv = REC_NO_RESPONSE;
      break;
    }

    devType = devPtr->z90c_dev_type;

    if (!(devPtr->z90c_dev_resp_l) || !(devPtr->z90c_dev_resp_p)) {
      dv = DEV_REC_EXCEPTION;
      PRINTK("Response Length and Pointer at receiveFromAP: %d, %p\n",
             devPtr->z90c_dev_resp_l, devPtr->z90c_dev_resp_p);
    }
    else {
      dv = receiveFromAP(index,
                         z90crypt.z90c_cdx,
                         devPtr->z90c_dev_resp_l,
                         devPtr->z90c_dev_resp_p,
                         psmid);
    }
    if (dv == DEV_REC_EXCEPTION) {
      rv = REC_FATAL_ERROR;
      z90crypt.z90c_terminating = 1;
      PRINTKC("Exception during receive from device %i\n",index);
      break;
    }
    switch (dv) {
      case DEV_ONLINE:
        rv = OK;
        break;
      case DEV_EMPTY:
        rv = REC_EMPTY;
        break;
      case DEV_NO_WORK:
        rv = REC_NO_WORK;
        break;
      case DEV_BAD_MESSAGE:
      case DEV_GONE:
      case REC_HARDWARE_FAILED:
      default:
        rv = REC_NO_RESPONSE;
        break;
    }

    if (rv != OK)
      break;

    if (devPtr->z90c_dev_caller_count == 0) {
      rv = REC_USER_GONE;
      break;
    }

    hptr = &(devPtr->z90c_dev_caller_list);
    list_for_each_safe(ptr,tptr,hptr) {
      caller_p = (Z90C_CALLER *)list_entry(ptr,
                                           struct z90c_caller,
                                           z90c_caller_liste);
      if (!(memcmp(caller_p->z90c_caller_id,
                   psmid,
                   sizeof(caller_p->z90c_caller_id)))) {
        list_del(ptr);
        --(devPtr->z90c_dev_caller_count);
        break;
      }
      else {
        caller_p = NULL;
      }
    }

    if (caller_p == NULL) {
      rv = REC_USER_GONE;
      break;
    }

    PDEBUG("caller_p after successful receive: %p\n",caller_p);
    rv = convertResponse(devPtr->z90c_dev_resp_p,
                         devPtr->z90c_dev_type,
                         caller_p->z90c_caller_buf_l,
                         caller_p->z90c_caller_buf_p,
                         caller_p->z90c_caller_function,
                         caller_p->z90c_pad_rule,
                         buff_len_p,
                         buff);
    switch (rv) {
      case BAD_MESSAGE_USER:
        rv = REC_USER_ERROR;
        break;
      case WRONG_DEVICE_TYPE:
      case REC_HARDWARE_FAILED:
      case BAD_MESSAGE_DRIVER:
        rv = REC_NO_RESPONSE;
        PRINTK("Device %d returned hardware error %d\n", index, rv);
        break;
      case REC_RELEASED:
        break;
      default:
        break;
    }


  } while(0);

  switch(rv) {
    case OK:
      PDEBUG("Successful receive from device %i\n",index);
      icaMsg_p = (ica_rsa_modexpo_t *)(caller_p->z90c_caller_buf_p);
      PDEBUG("icaMsg_p as pointer: %p\n",icaMsg_p);
      PDEBUG("dest_p_p as pointer: %p\n",dest_p_p);
      *dest_p_p = icaMsg_p->outputdata;
      if (*buff_len_p == 0)
        PRINTK("Zero *buff_len_p in receiveFromCryptoDevice\n");
      break;
    case REC_NO_RESPONSE:
      removeDevice(devPtr);
      break;
    default:
      break;
  }

  if (caller_p != NULL) {
    if ((temp_rv = unbuildCaller(devPtr,caller_p))!=OK){
      rv = REC_FATAL_ERROR;
      z90crypt.z90c_terminating = 1;
    }
  }

  return (rv);
} // end receiveFromCryptoDevice

int queryZ90crypt(Z90C_STATUS_STR * z90cryptStatus)
/*********************************************************************/
/*                                                                   */
/*         inputs:  none                                             */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*         output:  z90cryptStatus                                   */
/*                                                                   */
/* The output is the status array imbedded in the z90crypt.          */
/*                                                                   */
/* Not sure why anybody would call this function.  Why not look in   */
/* the z90crypt?                                                     */
/*                                                                   */
/*********************************************************************/
{
  memcpy (z90cryptStatus,
          &(z90crypt.z90c_mask),
          sizeof(struct z90c_status_str));

  return 0;
}

DEVSTAT queryCryptoDevice(int index)
/*********************************************************************/
/*                                                                   */
/*         inputs:  an index                                         */
/*                                                                   */
/* implied inputs:  the z90crypt                                     */
/*                                                                   */
/*         output:  DEV_GONE, DEV_ONLINE, aut ceterum.               */
/*                                                                   */
/*                                                                   */
/* Not sure why anybody would call this function.  Why not look in   */
/* the z90crypt?                                                     */
/*                                                                   */
/*********************************************************************/
{
  DEVSTAT dev_stat;

  dev_stat = z90crypt.z90c_mask.z90c_st_mask[index];

  return (dev_stat);
} // end queryCryptoDevice

int destroyZ90crypt()
/*********************************************************************/
/*                                                                   */
/*  implied input:  the z90crypt                                     */
/*                                                                   */
/*         output:  none                                             */
/*                                                                   */
/* All work will be drained.  All structures pointed to from the     */
/* z90crypt will be unallocated.                                     */
/*                                                                   */
/*********************************************************************/
{
  int i;
  Z90C_DEVICE * dev_ptr;

  for (i=0;i<z90crypt.z90c_max_count;i++) {
    if ((dev_ptr = z90crypt.z90c_device_p[i]) !=
                                                (Z90C_DEVICE *)NULL)
      destroyCryptoDevice(i);
  }

  if (z90crypt.z90c_hdware_info != NULL)
    kfree((void *)z90crypt.z90c_hdware_info);

  memset((void *)&z90crypt,
         0,
         sizeof(struct z90crypt));

  return OK;
}

int selectDeviceType (short func, CDEVICE_TYPE *devType_p)
/*********************************************************************/
/*                                                                   */
/* selectDeviceType:                                                 */
/*                                                                   */
/*   If a leedslite card is present, returns LEEDSLITE               */
/*   If a leeds2 card is present, returns LEEDS2                     */
/*   Otherwise returns -1                                            */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
{
  int rv = 0;

  if (func != PCI_FUNC_KEY_DECRYPT && func != PCI_FUNC_KEY_ENCRYPT){
    PDEBUG("Bad function code top of select\n");
    return (-1);
  }

  if ((*devType_p == LEEDSLITE) &&
      (z90crypt.z90c_hdware_info->
                        z90c_type_mask[LEEDSLITE].z90c_disabled_count >=
        z90crypt.z90c_hdware_info->
                              z90c_type_mask[LEEDSLITE].z90c_st_count)) {
    return (-1);
  }

  if ((*devType_p == LEEDS2) &&
      (z90crypt.z90c_hdware_info->
                        z90c_type_mask[LEEDS2].z90c_disabled_count >=
        z90crypt.z90c_hdware_info->
                              z90c_type_mask[LEEDS2].z90c_st_count)) {
    return (-1);
  }

  if (z90crypt.z90c_hdware_info->
                        z90c_type_mask[LEEDSLITE].z90c_disabled_count <
        z90crypt.z90c_hdware_info->
                              z90c_type_mask[LEEDSLITE].z90c_st_count)
    *devType_p = LEEDSLITE;
  else
    if (z90crypt.z90c_hdware_info->
                        z90c_type_mask[LEEDS2].z90c_disabled_count <
          z90crypt.z90c_hdware_info->
                              z90c_type_mask[LEEDS2].z90c_st_count)
      *devType_p = LEEDS2;
    else
      rv = -1;

  return rv;
} // end selectDeviceType

int selectDevice (short func, CDEVICE_TYPE *devType_p , int *device_nr_p)
/*********************************************************************/
/*                                                                   */
/*           inputs:  func    -- low order two bytes contain a       */
/*                               function code.                      */
/*                                                                   */
/*                    devType -- 0 (LeedsLite), 1 (Leeds2)           */
/*                               or -1 if no type is preferred       */
/*                                                                   */
/*                    device_nr -- a number in [0,...,63] or -1      */
/*                                                                   */
/*    implied input:  z90crypt                                       */
/*                                                                   */
/*          outputs:  *device_nr_p   -- a device number from 0 to 63 */
/*                                      or -1 if no device is        */
/*                                      available.                   */
/*                                                                   */
/*     return value:  same as *device_nr_p.  If no device is         */
/*                    available, returns -1.                         */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*          process:  1.  if device_nr is specified, returns it.     */
/*                                                                   */
/*                    2.  if devType is specified, selects a device  */
/*                        at random from the list associated with    */
/*                        that type.                                 */
/*                                                                   */
/*                    3.  if devType is not specified, selects a     */
/*                        device at random from the master list.     */
/*                                                                   */
/*********************************************************************/
{
  int nr = 0;
  unsigned int ix = 0xff000000;
  int i;
  int indx = -1;
  int devTp;
  struct z90c_device_x *index_p;
  Z90C_DEVICE * dev_ptr = NULL;
  int max;
  int low_count = 0x0000ffff;
  int low_indx = -1;

  PDEBUG("top of selectDevice:  func = %d\n",func);
  PDEBUG("top of selectDevice:  device type = %d\n",*devType_p);
  PDEBUG("top of selectDevice:  device index = %d\n",*device_nr_p);

  max = 2 * z90crypt.z90c_mask.z90c_st_count;

  devTp = *devType_p;
  nr = *device_nr_p;

  if (func != PCI_FUNC_KEY_DECRYPT && func != PCI_FUNC_KEY_ENCRYPT){
    PDEBUG("Bad function code top of select\n");
    return (-1);
  }

  if (nr >= 0 && nr < Z90CRYPT_NUM_DEVS) {
    dev_ptr = z90crypt.z90c_device_p[nr];
  }

  if (dev_ptr != NULL &&
      dev_ptr->z90c_dev_stat != DEV_GONE &&
      dev_ptr->z90c_disabled == FALSE) {
    *devType_p = dev_ptr->z90c_dev_type;
    return (nr);
  }
  else
    nr = -1;

  index_p = &(z90crypt.z90c_hdware_info->z90c_type_x_addr[LEEDSLITE]);
  nr = z90crypt.z90c_hdware_info->
                          z90c_type_mask[LEEDSLITE].z90c_st_count;
  if (nr > z90crypt.z90c_hdware_info->z90c_type_mask[LEEDSLITE].z90c_disabled_count)
  {
    devTp = LEEDSLITE;
  }
  else {
    nr = z90crypt.z90c_hdware_info->
                          z90c_type_mask[LEEDS2].z90c_st_count;
    if (nr > z90crypt.z90c_hdware_info->z90c_type_mask[LEEDS2].z90c_disabled_count)
    {
      devTp = LEEDS2;
      index_p = &(z90crypt.z90c_hdware_info->z90c_type_x_addr[LEEDS2]);
    }
    else {
      devTp = -1;
    }
  }

  PDEBUG("select Device:  number to choose from: %d\n",nr);

  if (devTp == -1) {
    *device_nr_p = -1;
    return (-1);
  }

  for (i=0;i<nr;i++){
    indx = index_p->z90c_device_index[i];
    dev_ptr = z90crypt.z90c_device_p[indx];

    if ((dev_ptr != NULL) &&
        (dev_ptr->z90c_dev_stat != DEV_GONE) &&
        (dev_ptr->z90c_disabled == FALSE) &&
        ((devTp >= LEEDSLITE && devTp == dev_ptr->z90c_dev_type) ||
         (devTp=-1))) {
       low_count = dev_ptr->z90c_dev_caller_count;
       low_indx = indx;
       break;
    }
  }

  if (i == nr - 1) {
    *device_nr_p = low_indx;
    return (low_indx);
  }

  if (i >= nr){
    *device_nr_p = -1;
    return (-1);
  }

  for (ix = i+1;ix<nr;ix++){
    indx = index_p->z90c_device_index[ix];
    dev_ptr = z90crypt.z90c_device_p[indx];

    if ((dev_ptr != NULL) &&
        (dev_ptr->z90c_dev_stat != DEV_GONE) &&
        (dev_ptr->z90c_disabled == FALSE) &&
        ((devTp >= LEEDSLITE && devTp == dev_ptr->z90c_dev_type) ||
         (devTp=-1)))
       if (low_count>dev_ptr->z90c_dev_caller_count) {
         low_count = dev_ptr->z90c_dev_caller_count;
         low_indx = indx;
       }
  }

  *device_nr_p = low_indx;
  return (low_indx);

}; // end selectDevice

int getZ90cryptStatus(ica_z90_status * sp, int sl)
/*********************************************************************/
/*                                                                   */
/*           inputs:  sp -- address of an ica_z90_status (z90crypt.h)*/
/*                                                                   */
/*                    sl -- length of the ica_z90_status             */
/*                                                                   */
/*    implied input:  z90crypt                                       */
/*                                                                   */
/*          outputs:  *sp -- struct is filled                        */
/*                                                                   */
/*     return value:  0 if all goes well                             */
/*                    SEN_USER_ERROR if sl is too small              */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*          process:  1.  checks output length                       */
/*                                                                   */
/*                    2.  stores the total device count              */
/*                                                                   */
/*                    3.  stores the leedslite count                 */
/*                                                                   */
/*                    4.  stores the leeds2 count                    */
/*                                                                   */
/*                    5.  stores the type mask                       */
/*                                                                   */
/*                    6.  for each online device:                    */
/*                          stores the current count of sent         */
/*                          work elements                            */
/*                                                                   */
/*                    7.  returns                                    */
/*                                                                   */
/*********************************************************************/
{
  int ct = 0;
  int i,ix;

  if (sl < sizeof(ica_z90_status))
    return (EBADPARM);

  ct = sp->totalcount =
           z90crypt.z90c_hdware_info->z90c_hdware_mask.z90c_st_count;

  sp->leedslitecount =
   z90crypt.z90c_hdware_info->z90c_type_mask[LEEDSLITE].z90c_st_count;

  sp->leeds2count =
      z90crypt.z90c_hdware_info->z90c_type_mask[LEEDS2].z90c_st_count;

  memcpy(sp->status,z90crypt.z90c_hdware_info->z90c_device_type_array,
         sizeof(sp->status));

  memset(sp->qdepth,0,sizeof(sp->qdepth));

  for (i=0;i<ct;i++) {
    ix = z90crypt.z90c_overall_device_x.z90c_device_index[i];
    sp->qdepth[ix] =
                   z90crypt.z90c_device_p[ix]->z90c_dev_caller_count;
  }

  return (0);

} // end getZ90cryptStatus;

int probeDeviceType(Z90C_DEVICE * devPtr)
/*********************************************************************/
/*                                                                   */
/*           inputs:  devPtr  -- address of a Z90C_DEVICE            */
/*                                                                   */
/*    implied input:  z90crypt                                       */
/*                                                                   */
/*          outputs:  devPtr-> z90c_dev_type                         */
/*                                                                   */
/*     return value:  propagated from low level send/receive         */
/*                    routines.  Otherwise, returns OK               */
/*                                                                   */
/*   implied output:  none                                           */
/*                                                                   */
/*          process:  1.  initialize device type to indeterminate    */
/*                                                                   */
/*                    2.  sendToAP using a canned Leeds2             */
/*                        request.                                   */
/*                                                                   */
/*                    3.  Do 6 times:                                */
/*                                                                   */
/*                        a.  sleep for a little while               */
/*                                                                   */
/*                        b.  receiveFromAP.                         */
/*                                                                   */
/*                        c.  consider the result:                   */
/*                            1)  No response:                       */
/*                                   a) device busy                  */
/*                                         continue                  */
/*                                   b) no work ready                */
/*                                         continue                  */
/*                                   c) anything else                */
/*                                         break                     */
/*                            2)  Type 86 response:                  */
/*                                   deviceType is Leeds2            */
/*                                   break                           */
/*                            3)  Type 82 response                   */
/*                                   deviceType is LeedsLite         */
/*                                   break                           */
/*                                                                   */
/*                    4.  Return                                     */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
{
  int rv = OK;
  int dv = DEV_ONLINE;
  int i;                           // loop counter
  int index;                       // device number
  unsigned char psmid[8];          // returned by receiveFromAP
  int testmsg_len = 384;           // sorry about that; you need that
                                   // much space
  unsigned char dyn_testmsg[testmsg_len];

  index = devPtr->z90c_dev_self_x;

  do {
    getTestMsg(dyn_testmsg, &testmsg_len);
    if ((dv = sendToAP(index,
                       z90crypt.z90c_cdx,
                       testmsg_len-24,    // allow for 'header'
                       dyn_testmsg)) != OK){
      PDEBUG("dv returned by send during probe: %d\n",dv);
      if (dv==DEV_SEN_EXCEPTION) {
        rv = SEN_FATAL_ERROR;
        PRINTKC("Exception during send to device %i\n",index);
        break;
      }
      PDEBUG("return value from sendToAP:  %i\n",rv);
      switch(dv) {
        case DEV_GONE:
          PDEBUG("Device %d not available\n", index);
          rv = SEN_NOT_AVAIL;
          break;
        case DEV_ONLINE:
          rv = OK;
          break;
        case DEV_EMPTY:
          rv = SEN_NOT_AVAIL;
          break;
        case DEV_NO_WORK:
          rv = SEN_FATAL_ERROR;
          break;
        case DEV_BAD_MESSAGE:
          rv = SEN_USER_ERROR;
          break;
        case DEV_QUEUE_FULL:
          rv = SEN_QUEUE_FULL;
          break;
        default:
          break;
      } // end switch (dv from sendToAP)
    } // end if dv != OK

    if (rv != OK)
      break;

    for (i=0;i<6;i++) {
      // sleep(.3)
      mdelay(300);

      dv = receiveFromAP(index,
                         z90crypt.z90c_cdx,
                         devPtr->z90c_dev_resp_l,
                         devPtr->z90c_dev_resp_p,
                         psmid);

      if (dv == DEV_REC_EXCEPTION) {
        rv = REC_FATAL_ERROR;
        PRINTKC("Exception during receive from device %i\n",index);
        break;
      }
      PDEBUG("probeD:  dv returned by DQ = %d\n",dv);
      switch (dv) {
        case DEV_ONLINE:
          rv = OK;
          break;
        case DEV_EMPTY:
          rv = REC_EMPTY;
          break;
        case DEV_NO_WORK:
          rv = REC_NO_WORK;
          break;
        case DEV_BAD_MESSAGE:
        case DEV_GONE:
        default:
          rv = REC_NO_RESPONSE;
          break;
      } // end switch (dv from receiveFromAP)

      if ((rv != OK) && (rv != REC_NO_WORK))
        break;

      if (rv == OK)
         break;

    } // end for (i=0;....

    if (rv != OK)
      break;

    rv = testReply(devPtr->z90c_dev_resp_p);

    if (rv == 0) {
      devPtr->z90c_dev_type = LEEDS2;
      break;
    }

    devPtr->z90c_dev_type = LEEDSLITE;
    rv = 0;

  } while(0);

  // In a general error case, the card is not marked online

  return (rv);

} // end probeDevicetype

int isPKCS1_2padded (unsigned char * buff, int bufflen)
/********************************************************************/
/*                                                                  */
/* isPKCS1_2padded:  Tests whether a string of bytes is (possibly)  */
/*                   padded in accordance with the PKCS#1.2         */
/*                   standard                                       */
/*                                                                  */
/*           input:  buff -- a string of bytes                      */
/*                                                                  */
/*                   bufflen -- length of the string                */
/*                                                                  */
/*         process:  1.  First byte must be 0                       */
/*                   2.  Second byte must be 0x02                   */
/*                   3.  There must be a string of 8 or more        */
/*                       non-zero bytes                             */
/*                   4.  There must be at least one zero byte       */
/*                                                                  */
/*          output:  returns 1 if string could be PKCS1.2 padded    */
/*                   returns 0 if any of the above conditions is    */
/*                             violated.                            */
/*                                                                  */
/********************************************************************/
{
  int rv = 0;
  int i;                              // loop counter

  do {
    if (bufflen < 12)
      break;

    if (buff[0] != 0)
      break;

    if (buff[1] != 0x02)
      break;

    for (i=2;((i<bufflen) && (buff[i]));i++);

    if (i > bufflen - 2)
      break;

    if (i < 9)
      break;

    rv = 1;

  } while(0);

  return (rv);

} // end isPKCS1_2padded
