/*
 *  scsi_scan.c Copyright (C) 2000 Eric Youngdale
 *
 *  major rework 8/2003 Kurt Garloff <garloff@suse.de>
 *  inspired by 2.6 code from Pat Mansfield and others.
 * 
 *  Bus scan logic.
 *
 *  This used to live in scsi.c, but that file was just a laundry basket
 *  full of misc stuff.  This got separated out in order to make things
 *  clearer.
 */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>

#include <linux/blk.h>

#include "scsi.h"
#include "hosts.h"
#include "constants.h"

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

/* 
 * Flags for irregular SCSI devices that need special treatment 
 */
#define BLIST_NOLUN     	0x001	/* Don't scan for LUNs */
#define BLIST_FORCELUN  	0x002	/* Known to have LUNs, force scanning, deprecated */
#define BLIST_BORKEN    	0x004	/* Flag for broken handshaking */
#define BLIST_KEY       	0x008	/* Needs to be unlocked by special command */
#define BLIST_SINGLELUN 	0x010	/* LUNs should better not be used in parallel */
#define BLIST_NOTQ		0x020	/* Buggy Tagged Command Queuing */
#define BLIST_SPARSELUN 	0x040	/* Non consecutive LUN numbering */
#define BLIST_MAX5LUN		0x080	/* Avoid LUNS >= 5 */
#define BLIST_ISDISK    	0x100	/* Treat as (removable) disk */
#define BLIST_ISROM     	0x200	/* Treat as (removable) CD-ROM */
#define BLIST_LARGELUN		0x400	/* LUNs larger than 7 despite reporting as SCSI 2 */
#define BLIST_INQUIRY_36	0x400	/* override additional length field */
#define BLIST_INQUIRY_58	0x800	/* ... for broken inquiry responses */
#if 0	// 2.6 kernel
#define BLIST_NOSTARTONADD	0x1000  /* do not do automatic start on add */
#endif
#define BLIST_NOREPLUN		0x2000	/* do not try to use REPORT_LUNS when scanning */

typedef struct scsi_lun {
	u8 scsi_lun[8];
} ScsiLun;

static void print_inquiry(unsigned char *data);
static int scan_scsis_single(unsigned int channel, unsigned int dev,
		unsigned int lun, int lun0_scsi_level, 
		unsigned int *max_scsi_dev, unsigned int *sparse_lun, 
		Scsi_Device ** SDpnt, char *scsi_result);
static int find_lun0_scsi_level(unsigned int channel, unsigned int dev,
				struct Scsi_Host *shpnt);
static int scsi_send_replun(Scsi_Device *SDpnt, unsigned int channel, 
			    unsigned int target_id,
			    struct scsi_lun* lun_data, int length);

struct dev_info {
	const char *vendor;
	const char *model;
	const char *revision;	/* Latest revision known to be bad.  Not used yet */
	unsigned flags;
};

/*
 * This is what was previously known as the blacklist.  The concept
 * has been expanded so that we can specify other types of things we
 * need to be aware of.
 */
static struct dev_info device_list[] =
{
/* The following devices are known not to tolerate a lun != 0 scan for
 * one reason or another.  Some will respond to all luns, others will
 * lock up.
 */
	{"Aashima", "IMAGERY 2400SP", "1.03", BLIST_NOLUN},	/* Locks up if polled for lun != 0 */
	{"CHINON", "CD-ROM CDS-431", "H42", BLIST_NOLUN},	/* Locks up if polled for lun != 0 */
	{"CHINON", "CD-ROM CDS-535", "Q14", BLIST_NOLUN},	/* Locks up if polled for lun != 0 */
	{"DENON", "DRD-25X", "V", BLIST_NOLUN},			/* Locks up if probed for lun != 0 */
	{"HITACHI", "DK312C", "CM81", BLIST_NOLUN},		/* Responds to all lun - dtg */
	{"HITACHI", "DK314C", "CR21", BLIST_NOLUN},		/* responds to all lun */
	{"IMS", "CDD521/10", "2.06", BLIST_NOLUN},		/* Locks-up when LUN>0 polled. */
	{"MAXTOR", "XT-3280", "PR02", BLIST_NOLUN},		/* Locks-up when LUN>0 polled. */
	{"MAXTOR", "XT-4380S", "B3C", BLIST_NOLUN},		/* Locks-up when LUN>0 polled. */
	{"MAXTOR", "MXT-1240S", "I1.2", BLIST_NOLUN},		/* Locks up when LUN>0 polled */
	{"MAXTOR", "XT-4170S", "B5A", BLIST_NOLUN},		/* Locks-up sometimes when LUN>0 polled. */
	{"MAXTOR", "XT-8760S", "B7B", BLIST_NOLUN},		/* guess what? */
	{"MEDIAVIS", "RENO CD-ROMX2A", "2.03", BLIST_NOLUN},	/*Responds to all lun */
	{"NEC", "CD-ROM DRIVE:841", "1.0", BLIST_NOLUN},	/* Locks-up when LUN>0 polled. */
	{"PHILIPS", "PCA80SC", "V4-2", BLIST_NOLUN},		/* Responds to all lun */
	{"RODIME", "RO3000S", "2.33", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"SANYO", "CRD-250S", "1.20", BLIST_NOLUN},		/* causes failed REQUEST SENSE on lun 1
								 * for aha152x controller, which causes
								 * SCSI code to reset bus.*/
	{"SEAGATE", "ST157N", "\004|j", BLIST_NOLUN},		/* causes failed REQUEST SENSE on lun 1
								 * for aha152x controller, which causes
								 * SCSI code to reset bus.*/
	{"SEAGATE", "ST296", "921", BLIST_NOLUN},		/* Responds to all lun */
	{"SEAGATE", "ST1581", "6538", BLIST_NOLUN},		/* Responds to all lun */
	{"SONY", "CD-ROM CDU-541", "4.3d", BLIST_NOLUN},	
	{"SONY", "CD-ROM CDU-55S", "1.0i", BLIST_NOLUN},
	{"SONY", "CD-ROM CDU-561", "1.7x", BLIST_NOLUN},
	{"SONY", "CD-ROM CDU-8012", "*", BLIST_NOLUN},
	{"TANDBERG", "TDC 3600", "U07", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"TEAC", "CD-R55S", "1.0H", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"TEAC", "CD-ROM", "1.06", BLIST_NOLUN},		/* causes failed REQUEST SENSE on lun 1
								 * for seagate controller, which causes
								 * SCSI code to reset bus.*/
	{"TEAC", "MT-2ST/45S2-27", "RV M", BLIST_NOLUN},	/* Responds to all lun */
	{"TEXEL", "CD-ROM", "1.06", BLIST_NOLUN},		/* causes failed REQUEST SENSE on lun 1
								 * for seagate controller, which causes
								 * SCSI code to reset bus.*/
	{"QUANTUM", "LPS525S", "3110", BLIST_NOLUN},		/* Locks sometimes if polled for lun != 0 */
	{"QUANTUM", "PD1225S", "3110", BLIST_NOLUN},		/* Locks sometimes if polled for lun != 0 */
	{"QUANTUM", "FIREBALL ST4.3S", "0F0C", BLIST_NOLUN},	/* Locks up when polled for lun != 0 */
	{"MEDIAVIS", "CDR-H93MV", "1.31", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"SANKYO", "CP525", "6.64", BLIST_NOLUN},		/* causes failed REQ SENSE, extra reset */
	{"HP", "C1750A", "3226", BLIST_NOLUN},			/* scanjet iic */
	{"HP", "C1790A", "", BLIST_NOLUN},			/* scanjet iip */
	{"HP", "C2500A", "", BLIST_NOLUN},			/* scanjet iicx */
	{"YAMAHA", "CDR100", "1.00", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"YAMAHA", "CDR102", "1.00", BLIST_NOLUN},		/* Locks up if polled for lun != 0  
								 * extra reset */
	{"YAMAHA", "CRW8424S", "1.0", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"YAMAHA", "CRW6416S", "1.0c", BLIST_NOLUN},		/* Locks up if polled for lun != 0 */
	{"MITSUMI", "CD-R CR-2201CS", "6119", BLIST_NOLUN},	/* Locks up if polled for lun != 0 */
	{"RELISYS", "Scorpio", "*", BLIST_NOLUN},		/* responds to all LUN */
	{"RELISYS", "VM3530+", "*", BLIST_NOLUN},		/* responds to all LUN */
	{"ACROSS", "", "*", BLIST_NOLUN},			/* responds to all LUN */
	{"MICROTEK", "ScanMaker II", "5.61", BLIST_NOLUN},	/* responds to all LUN */

/*
 * Other types of devices that have special flags.
 */
	{"SONY", "CD-ROM CDU-8001", "*", BLIST_BORKEN},
	{"TEXEL", "CD-ROM", "1.06", BLIST_BORKEN},
	{"IOMEGA", "Io20S         *F", "*", BLIST_KEY},
	{"iomega", "jaz 1GB", "J.86", BLIST_NOTQ | BLIST_NOLUN},
	{"INSITE", "Floptical   F*8I", "*", BLIST_KEY},
	{"INSITE", "I325VM", "*", BLIST_KEY},
	{"LASOUND","CDX7405","3.10", BLIST_MAX5LUN | BLIST_SINGLELUN},
	{"MICROP", "4110", "*", BLIST_NOTQ},			/* Buggy Tagged Queuing */
	{"NRC", "MBR-7", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"NRC", "MBR-7.4", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"REGAL", "CDC-4X", "*", BLIST_MAX5LUN | BLIST_SINGLELUN},
	{"NAKAMICH", "MJ-4.8S", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"NAKAMICH", "MJ-5.16S", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"PIONEER", "CD-ROM DRM-600", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"PIONEER", "CD-ROM DRM-602X", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"PIONEER", "CD-ROM DRM-604X", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"EMULEX", "MD21/S2     ESDI", "*", BLIST_SINGLELUN},
	{"CANON", "IPUBJD", "*", BLIST_SPARSELUN},
	{"nCipher", "Fastness Crypto", "*", BLIST_FORCELUN},
	/* Note about the following BLIST_FORCELUN entries: Most should probably be
	 * BLIST_SPARSELUN or BLIST_SPARSELUN | BLIST_LARGELUN. 
	 * 2003-03-21, garloff@suse.de */
	{"DEC","HSG80","*", BLIST_FORCELUN | BLIST_LARGELUN},
	{"COMPAQ", "MSA1000", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"COMPAQ","HSV110","*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"HP","HSV100","*", BLIST_SPARSELUN | BLIST_LARGELUN},
	//{"COMPAQ", "MSA1000", "*", BLIST_FORCELUN},
	{"COMPAQ","LOGICAL VOLUME","*", BLIST_FORCELUN},
	{"COMPAQ","CR3500","*", BLIST_FORCELUN},
	{"NEC", "PD-1 ODX654P", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
	{"MATSHITA", "PD-1", "*", BLIST_FORCELUN | BLIST_SINGLELUN},
 	{"TOSHIBA","CDROM","*", BLIST_ISROM},
 	{"TOSHIBA","CD-ROM","*", BLIST_ISROM},
	{"MegaRAID", "LD", "*", BLIST_FORCELUN},
	{"DGC",  "RAID",      "*", BLIST_SPARSELUN | BLIST_LARGELUN}, // Dell PV 650F (tgt @ LUN 0)
	{"DGC",  "DISK",      "*", BLIST_SPARSELUN | BLIST_LARGELUN}, // Dell PV 650F (no tgt @ LUN 0) 
	{"DELL", "PV660F",   "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"DELL", "PV660F   PSEUDO",   "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"DELL", "PSEUDO DEVICE .",   "*", BLIST_SPARSELUN | BLIST_LARGELUN}, // Dell PV 530F
	{"DELL", "PV530F",    "*", BLIST_SPARSELUN | BLIST_LARGELUN}, // Dell PV 530F
	{"DELL", "PV-136T-FC", "*", BLIST_FORCELUN},
	{"DELL", "PERCRAID", "*", BLIST_FORCELUN},
	{"EMC", "SYMMETRIX", "*", BLIST_SPARSELUN | BLIST_LARGELUN /*| BLIST_FORCELUN */},
	{"EUROLOGC", "FC2502", "*", BLIST_LARGELUN | BLIST_SPARSELUN},
	{"HP", "A6188A", "*", BLIST_SPARSELUN | BLIST_LARGELUN},/* HP Va7100 Array */
	{"HP", "A6189A", "*", BLIST_SPARSELUN | BLIST_LARGELUN},/* HP Va7400 Array */
	{"HP", "A6218A", "*", BLIST_SPARSELUN | BLIST_LARGELUN},/* HP Va7410 Array */
	{"HP", "OPEN-", "*", BLIST_SPARSELUN | BLIST_LARGELUN},	/* HP XP Arrays */
	{"HP", "NetRAID-4M", "*", BLIST_FORCELUN | BLIST_LARGELUN},	// HP NetRaid
	{"HP", "C1557A", "*", BLIST_FORCELUN},
	{"CMD", "CRA-7280", "*", BLIST_SPARSELUN | BLIST_LARGELUN},   // CMD RAID Controller
	{"CNSI", "G7324", "*", BLIST_SPARSELUN | BLIST_LARGELUN},     // Chaparral G7324 RAID
	{"CNSi", "G8324", "*", BLIST_SPARSELUN | BLIST_LARGELUN},     // Chaparral G8324 RAID
	{"Zzyzx", "RocketStor 500S", "*", BLIST_SPARSELUN},
	{"Zzyzx", "RocketStor 2000", "*", BLIST_SPARSELUN},
	{"SONY", "TSL",       "*", BLIST_FORCELUN},  // DDS3 & DDS4 autoloaders
	{"ADAPTEC", "AACRAID", "*", BLIST_FORCELUN},
	{"ADAPTEC", "Adaptec 5400S", "*", BLIST_FORCELUN},
	{"IBM", "AuSaV1S2", "*", BLIST_FORCELUN},
	{"IBM", "ProFibre 4000R", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"IBM", "3542", "*", BLIST_FORCELUN},
	{"FSC", "CentricStor", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"DDN", "SAN DataDirector", "*", BLIST_SPARSELUN},
	{"HITACHI", "DF400", "*", BLIST_SPARSELUN},
	{"HITACHI", "DF500", "*", BLIST_SPARSELUN},
	{"HITACHI", "DF600", "*", BLIST_SPARSELUN},
	{"HITACHI", "OPEN-", "*", BLIST_SPARSELUN | BLIST_LARGELUN},	/* HITACHI XP Arrays */
	{"WINSYS","FLASHDISK G6", "*", BLIST_SPARSELUN},	// was: BLIST_FORCELUN
	{"DotHill","SANnet RAID X300", "*", BLIST_SPARSELUN},	
	{"SUN", "T300", "*", BLIST_SPARSELUN},
	{"SUN", "T4", "*", BLIST_SPARSELUN},
	{"SGI", "RAID", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"SGI", "TP", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"MYLEX", "DACARMRB", "*", BLIST_SPARSELUN | BLIST_LARGELUN},
	{"NEC", "iStorage", "*", BLIST_SPARSELUN | BLIST_LARGELUN | BLIST_FORCELUN},
	/*
	 * Must be at end of list...
	 */
	{NULL, NULL, NULL}
};

#define MAX_SCSI_LUNS 0xFFFFFFFF

#ifdef CONFIG_SCSI_MULTI_LUN
static unsigned int max_scsi_luns = MAX_SCSI_LUNS;
#else
static unsigned int max_scsi_luns = 1;
#endif

#define MAX_LLUN_BLKLEN 8
static int llun_blklst[3*MAX_LLUN_BLKLEN] = {
	-1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1 };

static inline int is_on_llun_blklst (int hostno, int channel, int target)
{
	int i;
	for (i = 0; i < MAX_LLUN_BLKLEN; ++i) {
		if (llun_blklst[i*3] == -1)
			break;
		if (llun_blklst[i*3+0] == hostno &&
		    llun_blklst[i*3+1] == channel &&
		    llun_blklst[i*3+2] == target) {
			printk (KERN_DEBUG "Found c%ib%it%i on llun_blklst\n",
				hostno, channel, target);
			return 1;
		}
	}
	return 0;
}

static int scsi_allow_ghost_devices;

static int scsi_sparselun;
static int scsi_largelun;
static int max_scsi_sparseluns;
static int scsi_noreportlun;
static int scsi_reportlun2;
/*
 * max_report_luns: the maximum number of LUNS that will be
 * returned from the REPORT LUNS command. 8 times this value must
 * be allocated. In theory this could be up to an 8 byte value, but
 * in practice, the maximum number of LUNs suppored by any device
 * is about 16k.
 */
static unsigned int max_report_luns = 511;

/* Some AChip ARC765 based DVD-ROM's take 15 or more seconds
 * to reset.  A scan will fail if made right after a reset.
 * It's completely broken device behaviour: SCSI specification
 * says devices need to be able to respond to INQUIRY always
 * (after a selection timeout ... of 250ms).
 */
static unsigned int scsi_inq_timeout;

#ifdef MODULE

MODULE_PARM(max_scsi_luns, "i");
MODULE_PARM_DESC(max_scsi_luns, "last scsi LUN (should be between 1 and 2^32-1)");

MODULE_PARM(llun_blklst, "3-24i");
MODULE_PARM_DESC(llun_blklst, "SCSI-2 devs (C,B,T) that need sparse+large LUNs support");

MODULE_PARM(scsi_sparselun, "i");
MODULE_PARM_DESC(scsi_sparselun, "Assume sparse LUNs for all SCSI devices");

MODULE_PARM(scsi_largelun, "i");
MODULE_PARM_DESC(scsi_largelun, "Assume all SCSI-2 devs support more than 8 LUNs");

MODULE_PARM(max_scsi_sparseluns, "i");
MODULE_PARM_DESC(max_scsi_sparseluns, "Limit LUNs for scanning sparse LUN devices");

MODULE_PARM(scsi_allow_ghost_devices, "i");
MODULE_PARM_DESC(scsi_allow_ghost_devices, "allow devices marked as being offline to be accessed anyway (0 = off, else allow ghosts on lun 0 through allow_ghost_devices - 1");

MODULE_PARM(scsi_noreportlun, "i");
MODULE_PARM_DESC(scsi_noreportlun, "Don't use REPORT_LUNs for scanning SCSI-3 devs");

MODULE_PARM(scsi_reportlun2, "i");
MODULE_PARM_DESC(scsi_reportlun, "Use REPORT_LUNs for scanning SCSI-2 devs as well");

MODULE_PARM(max_report_luns, "i");
MODULE_PARM_DESC(max_report_luns,
		 "REPORT LUNS maximum number of LUNS received (should be"
		 " between 1 and 16383)");

MODULE_PARM(scsi_inq_timeout, "i");
MODULE_PARM_DESC(scsi_inq_timeout, "Timeout (in seconds) waiting for devices to answer INQUIRY. Default is 6. Some non-compliant devices need more.");
#else

static int __init scsi_luns_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		max_scsi_luns = tmp;
		return 1;
	} else {
		printk("scsi_luns_setup : usage max_scsi_luns=n "
		       "(n should be between 1 and 2^32-1)\n");
		return 0;
	}
}

__setup("max_scsi_luns=", scsi_luns_setup);

static int __init llun_blklst_setup(char *str)
{
	unsigned int tmp;
	int i = 0;
	do {
		int rc;
		if (get_option(&str, &tmp) == 2)
			llun_blklst[i++] = tmp;
		else 
			break;
		if (get_option(&str, &tmp) == 2)
			llun_blklst[i++] = tmp;
		else 
			break;
		if ((rc = get_option(&str, &tmp)) >= 1)
			llun_blklst[i++] = tmp;
		if (rc != 2)
			break;
	} while (1);
	return i;
}

__setup("llun_blklst=", llun_blklst_setup);

static int __init scsi_sparselun_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_sparselun = tmp;
		return 1;
	} else {
		scsi_sparselun = 1;
		return 0;
	}
}

__setup("scsi_sparselun=", scsi_sparselun_setup);

static int __init scsi_largelun_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_largelun = tmp;
		return 1;
	} else {
		scsi_largelun = 1;
		return 0;
	}
}

__setup("scsi_largelun=", scsi_largelun_setup);

static int __init scsi_max_sparseluns_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		max_scsi_sparseluns = tmp;
		return 1;
	} else {
		printk("scsi_max_sparseluns_setup : usage max_scsi_sparseluns=n "
		       "(n should be between 1 and 2^32-1)\n");
		return 0;
	}
}

__setup("max_scsi_sparseluns=", scsi_max_sparseluns_setup);

static int __init scsi_allow_ghost_devices_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_allow_ghost_devices = tmp;
		return 1;
	} else {
		printk("scsi_allow_ghost_devices_setup: usage scsi_allow_ghost_devices=n (0: off else\nallow ghost devices (ghost devices are devices that report themselves as\nbeing offline but which we allow access to anyway) on lun 0 through n - 1.\n");
		return 0;
	}
}

__setup("scsi_allow_ghost_devices=", scsi_allow_ghost_devices_setup);

static int __init scsi_noreplun_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_noreportlun = tmp;
		return 1;
	} else {
		scsi_noreportlun = 1;
		return 0;
	}
}

__setup("scsi_noreportlun=", scsi_noreplun_setup);

static int __init scsi_replun2_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_reportlun2 = tmp;
		return 1;
	} else {
		scsi_reportlun2 = 1;
		return 0;
	}
}

__setup("scsi_reportlun2=", scsi_replun2_setup);

static int __init scsi_max_reportluns_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		max_report_luns = tmp;
		return 1;
	} else {
		printk("scsi_max_reportluns_setup : usage max_report_luns=n "
		       "(n should be between 1 and 16383)\n");
		return 0;
	}
}

__setup("max_report_luns=", scsi_max_reportluns_setup);

static int __init scsi_inq_timeout_setup(char *str)
{
	unsigned int tmp;

	if (get_option(&str, &tmp) == 1) {
		scsi_inq_timeout = tmp;
		return 1;
	} else {
		printk("scsi_inq_timeout_setup: usage scsi_inq_timeout=n "
		       "(n is in seconds and should be between 0 and 20)\n");
		return 0;
	}
}

__setup("scsi_inq_timeout=", scsi_inq_timeout_setup);

#endif

static void print_inquiry(unsigned char *data)
{
	int i;

	printk("  Vendor: ");
	for (i = 8; i < 16; i++) {
		if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
	}

	printk("  Model: ");
	for (i = 16; i < 32; i++) {
		if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
	}

	printk("  Rev: ");
	for (i = 32; i < 36; i++) {
		if (data[i] >= 0x20 && i < data[4] + 5)
			printk("%c", data[i]);
		else
			printk(" ");
	}

	printk("\n");

	i = data[0] & 0x1f;

	printk("  Type:   %s ",
	       i < MAX_SCSI_DEVICE_CODE ? scsi_device_types[i] : "Unknown          ");
	printk("                 ANSI SCSI revision: %02x", data[2] & 0x07);
	if ((data[2] & 0x07) == 1 && (data[3] & 0x0f) == 1)
		printk(" CCS\n");
	else
		printk("\n");
}

static int get_device_flags(unsigned char *response_data)
{
	int i = 0;
	unsigned char *pnt;
	for (i = 0; 1; i++) {
		if (device_list[i].vendor == NULL)
			return 0;
		pnt = &response_data[8];
		while (*pnt && *pnt == ' ')
			pnt++;
		if (memcmp(device_list[i].vendor, pnt,
			   strlen(device_list[i].vendor)))
			continue;
		pnt = &response_data[16];
		while (*pnt && *pnt == ' ')
			pnt++;
		if (memcmp(device_list[i].model, pnt,
			   strlen(device_list[i].model)))
			continue;
		return device_list[i].flags;
	}
	return 0;
}

/* 
 * Allocate Scsi_Device structure,
 * set up a blockdev queue, the merge fn and 
 * the waitqueue and chain it into the host adapter
 * device list.
 */
static Scsi_Device *Scsi_Alloc_SD(struct Scsi_Host *shpnt)
{
	Scsi_Device *SDpnt;

	SDpnt = (Scsi_Device *) kmalloc(sizeof(Scsi_Device),
					GFP_ATOMIC);
	if (SDpnt) {
		memset(SDpnt, 0, sizeof(Scsi_Device));
		/*
		 * Register the queue for the device.  All I/O requests will
		 * come in through here.
		 */
		scsi_initialize_queue(SDpnt, shpnt);
	} else
		return NULL;

	/*
	 * We must chain ourself in the host_queue, so commands can time out 
	 */
	SDpnt->queue_depth = 1;
	SDpnt->host = shpnt;
	SDpnt->online = TRUE;

	initialize_merge_fn(SDpnt);

        /*
         * Initialize the object that we will use to wait for command blocks.
         */
	init_waitqueue_head(&SDpnt->scpnt_wait);

	/*
	 * Next, hook the device to the host in question.
	 */
	SDpnt->prev = NULL;
	SDpnt->next = NULL;
	if (shpnt->host_queue != NULL) {
		Scsi_Device *SDtail = shpnt->host_queue;
		while (SDtail->next != NULL)
			SDtail = SDtail->next;

		SDtail->next = SDpnt;
		SDpnt->prev = SDtail;
	} else {
		shpnt->host_queue = SDpnt;
	}
	return SDpnt;
}

/*
 * Reverse Scsi_Alloc_SD(): Free the Scsi_Device:
 * Unlink it from the host adapter list, free the blkdev queue,
 * and free the structure's memory
 */
static void Scsi_Free_SD(Scsi_Device* SDpnt)
{	
	struct Scsi_Host *shpnt = SDpnt->host;
	/* Unchain SRpnt from host_queue */
	Scsi_Device *prev, *next;
	Scsi_Device *dqptr;

	for (dqptr = shpnt->host_queue; dqptr != SDpnt; dqptr = dqptr->next)
		continue;
	if (dqptr) {
		prev = dqptr->prev;
		next = dqptr->next;
		if (prev)
			prev->next = next;
		else
			shpnt->host_queue = next;
		if (next)
			next->prev = prev;
	}
	blk_cleanup_queue(&SDpnt->request_queue);
	kfree((char *) SDpnt);
	
	{
		Scsi_Device *sdev;
		Scsi_Cmnd *scmd;
	
		SCSI_LOG_SCAN_BUS(4, printk("Host status for host %p:\n", shpnt));
		for (sdev = shpnt->host_queue; sdev; sdev = sdev->next) {
			SCSI_LOG_SCAN_BUS(4, printk("Device %d %p: ", sdev->id, sdev));
			for (scmd = sdev->device_queue; scmd; scmd = scmd->next) {
				SCSI_LOG_SCAN_BUS(4, printk("%p ", scmd));
			}
			SCSI_LOG_SCAN_BUS(4, printk("\n"));
		}
	}
}

static void scsi_scan_hardcoded (struct Scsi_Host* shpnt, uint channel, 
				 uint dev, uint lun)
{
	unsigned char scsi_result0[256];
	unsigned char *scsi_result;
	Scsi_Device *SDpnt, *oldSDpnt;
	int lun0_sl = SCSI_3;	/* actually don't care for 0 == lun */
	unsigned int max_dev_lun = (max_scsi_luns < shpnt->max_lun ?
		max_scsi_luns : shpnt->max_lun);
	unsigned int sparse_lun = 0;

	if (channel > shpnt->max_channel ||
	    dev >= shpnt->max_id ||
	    lun >= shpnt->max_lun)
		return;
		
	/* Make sure we have something that is valid for DMA purposes */
	scsi_result = ((!shpnt->unchecked_isa_dma)
		       ? &scsi_result0[0] : kmalloc(512, GFP_DMA));
	if (scsi_result == NULL)
		return;

	SDpnt = Scsi_Alloc_SD(shpnt);
	if (!SDpnt)
		goto Out_Alloc_SD_failed;
	
	oldSDpnt = SDpnt;
	
	if (lun > 0 && lun <= 7)
		lun0_sl = find_lun0_scsi_level(channel, dev, shpnt);
	
	scan_scsis_single(channel, dev, lun, lun0_sl, &max_dev_lun, 
			  &sparse_lun, &SDpnt, scsi_result);
	
	/* 
	 * Did we find something? 
	 * Then select_queue_depths(), init(), attach(), 
	 * build_commandblocks(), resize_dma_pool(), finish()
	 */	
	if (SDpnt != oldSDpnt) {
		struct Scsi_Device_Template *sdtpnt;

		/* it could happen the blockdevice hasn't yet been inited */
		/* queue_depth() moved from scsi_proc_info() so that
		   it is called before scsi_build_commandblocks() */
		if (shpnt->select_queue_depths != NULL)
			(shpnt->select_queue_depths)(shpnt,
						     shpnt->host_queue);

		for (sdtpnt = scsi_devicelist; sdtpnt; sdtpnt = sdtpnt->next)
			if (sdtpnt->init && sdtpnt->dev_noticed)
				(*sdtpnt->init) ();

		for (sdtpnt = scsi_devicelist; sdtpnt; sdtpnt = sdtpnt->next) {
			if (sdtpnt->attach) {
				(*sdtpnt->attach) (oldSDpnt);
				if (oldSDpnt->attached) {
					scsi_build_commandblocks(oldSDpnt);
					if (0 == oldSDpnt->has_cmdblocks) {
						printk("scan_scsis: DANGER, no command blocks\n");
						/* What to do now ?? */
					}
				}
			}
		}
		scsi_resize_dma_pool();

		for (sdtpnt = scsi_devicelist; sdtpnt; sdtpnt = sdtpnt->next)
			if (sdtpnt->finish && sdtpnt->nr_dev)
				(*sdtpnt->finish) ();
	}

	Scsi_Free_SD(SDpnt);
	
    Out_Alloc_SD_failed:
	/* If we allocated a buffer so we could do DMA, free it now */
	if (scsi_result != &scsi_result0[0] && scsi_result != NULL) 
		kfree(scsi_result);
} 


static int scsilun_to_int(struct scsi_lun *scsilun)
{
	int i;
	unsigned int lun;

	lun = 0;
	for (i = 0; i < sizeof(lun); i += 2)
		lun = lun | (((scsilun->scsi_lun[i] << 8) |
			      scsilun->scsi_lun[i + 1]) << (i * 8));
	return lun;
}

static void scsi_scan_target(Scsi_Device **SDpnt2, unsigned char* scsi_result,
			     unsigned int channel, unsigned int dev)
{
	/* Actual ID. PC ordering is 0->n IBM/spec ordering is n->0 */
	int target_id;
	int lun0_sl = SCSI_2, bflags;
	unsigned int lun;
	unsigned int sparse_lun = 0;
	struct Scsi_Host *shpnt = (*SDpnt2)->host;
	/* The minimum */
	unsigned int max_dev_lun = (max_scsi_luns < shpnt->max_lun ?
		max_scsi_luns : shpnt->max_lun);
	
	if (shpnt->reverse_ordering)
		/* Shift to scanning 15,14,13... or 7,6,5,4, */
		target_id = shpnt->max_id - dev - 1;
	else
		target_id = dev;

	/* Don't scan for the own target ID */
	if (shpnt->this_id == target_id) 
		return;

	/* OK, first check whether there is a device */
	if (!scan_scsis_single(channel, target_id, 0, SCSI_2,
			      &max_dev_lun, &sparse_lun, SDpnt2,
			      scsi_result))
		return;
	if (!*SDpnt2)
		return;
	bflags = get_device_flags(scsi_result);
	if (bflags & BLIST_NOLUN)
		return;
	
	lun0_sl = (*SDpnt2)->scsi_level;
	if (!(bflags & BLIST_NOREPLUN)
	    && !(shpnt->no_reportlun)
	    && (  (!scsi_noreportlun && lun0_sl >= SCSI_3) 
	        ||( scsi_reportlun2  && lun0_sl >= SCSI_2 
				     && shpnt->max_lun > 8))) {
		/* Send REPORT_LUNS */
		int replun_buflen = (max_report_luns + 1) * sizeof(struct scsi_lun);
		int num_luns;
		struct scsi_lun *lunp;
		struct scsi_lun *lun_data = kmalloc(replun_buflen, GFP_ATOMIC |
			   ((*SDpnt2)->host->unchecked_isa_dma ? __GFP_DMA : 0));
		unsigned char* data;
		if (!lun_data) {
			printk ("scsi: Allocation of REPORT_LEN buffer failed!\n");
			goto serial_scan;
		}
		memset(lun_data, 0, replun_buflen);
		num_luns = scsi_send_replun(*SDpnt2, channel, target_id, 
					    lun_data, replun_buflen); 
		if (num_luns == -1)
			goto serial_scan_freebuf;
		/* OK, successful, scan for the LUNs reported */
		SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: REPORT LUN scan of"
				  " host %d channel %d id %d\n", shpnt->host_no,
				  channel, target_id));
		
		for (lunp = &lun_data[1]; lunp <= &lun_data[num_luns]; ++lunp) {
			lun = scsilun_to_int(lunp);
			if (memcmp(&lunp->scsi_lun[sizeof(lun)], "\0\0\0\0", 4)) {
				int i;
				printk(KERN_WARNING "scsi: host %i ch %i id %i lun 0x",
				       shpnt->host_no, channel, target_id);
				data = (unsigned char *)lunp->scsi_lun;
				for (i = 0; i < sizeof(struct scsi_lun); ++i)
					printk("%02x", data[i]);
				printk(" has a LUN larger than currently supported.\n");
			} else if (lun == 0) {
				/* LUN 0 has already been scanned. */
			} else if (lun > shpnt->max_lun) {
				printk(KERN_WARNING "scsi: host %i ch %i id %i lun %d has a LUN larger"
				       " than allowed by the host adapter\n",
				       shpnt->host_no, channel, target_id, lun);
			} else {
				int res;

				res = scan_scsis_single(channel, target_id, lun, lun0_sl,
						        &max_dev_lun, &sparse_lun, SDpnt2, 
							scsi_result);
				if (res == 0) {
					/* Got some results, but now none, abort. */
					printk(KERN_ERR "scsi: Unexpected response"
					       " from host %i ch %i id %i lun %d while scanning, scan"
					       " aborted\n", shpnt->host_no, channel, target_id, lun);
					break;
				}
			}
		}
		

		kfree(lun_data);
		return;
    serial_scan_freebuf:
		kfree(lun_data);
	}

    serial_scan:
	if (lun0_sl < SCSI_3 && 
	    (bflags & BLIST_LARGELUN ||
	     is_on_llun_blklst (shpnt->host_no, channel, target_id)))
		lun0_sl = SCSI_3; /* treat as SCSI 3 */

	if (lun0_sl < SCSI_3 && !scsi_largelun && max_dev_lun > 8)
		max_dev_lun = 8;
	
	for (lun = 1; lun < max_dev_lun; ++lun) {

		if (!scan_scsis_single(channel, target_id, lun, lun0_sl,
				       &max_dev_lun, &sparse_lun, SDpnt2, 
				       scsi_result)
		    && !sparse_lun && !scsi_sparselun)
			break;
		if (!*SDpnt2)
			break;
	}	/* for lun ends */
}


/*
 *  Detecting SCSI devices :
 *  We scan all present host adapter's busses,  from ID 0 to ID (max_id).
 *  We use the INQUIRY command, determine device type, and pass the ID /
 *  lun address of all sequential devices to the tape driver, all random
 *  devices to the disk driver.
 */
void scsi_scan_host(struct Scsi_Host* shpnt)
{
	unsigned char scsi_result0[256];
	unsigned char *scsi_result;
	Scsi_Device *SDpnt;
	unsigned int channel, dev;

	/* Make sure we have something that is valid for DMA purposes */
	scsi_result = ((!shpnt->unchecked_isa_dma)
		       ? &scsi_result0[0] : kmalloc(512, GFP_DMA));
	if (scsi_result == NULL)
		return;

	SDpnt = Scsi_Alloc_SD(shpnt);
	if (!SDpnt)
		goto Out_Alloc_SD_failed;
	
	for (channel = 0; channel <= shpnt->max_channel; ++channel)
		for (dev = 0; dev < shpnt->max_id; ++dev) 
		scsi_scan_target(&SDpnt, scsi_result,
				 channel, dev);

	Scsi_Free_SD(SDpnt);

    Out_Alloc_SD_failed:
	/* If we allocated a buffer so we could do DMA, free it now */
	if (scsi_result != &scsi_result0[0] && scsi_result != NULL) 
		kfree(scsi_result);
}

/* Old crufty interface */
void scan_scsis(struct Scsi_Host *shpnt,
		       uint hardcoded, uint hchannel,
		       uint hid, uint hlun)
{
	if (1 == hardcoded)
		scsi_scan_hardcoded(shpnt, hchannel, hid, hlun);
	else 
		scsi_scan_host(shpnt);
}

/*
 * Fill in channel, dev, lun and 
 * allocate request and command blocks
 */
static Scsi_Request* Scsi_Device_Init(Scsi_Device *SDpnt,
		     		      unsigned int channel, 
				      unsigned int dev,
				      unsigned int lun)
{
	Scsi_Request *SRpnt;
	SDpnt->channel = channel;
	SDpnt->id = dev;
	SDpnt->lun = lun;
	SDpnt->online = TRUE;

	scsi_build_commandblocks(SDpnt);
 
	/* Some low level driver could use device->type (DB) */
	SDpnt->type = -1;

	/*
	 * Assume that the device will have handshaking problems, and then fix
	 * this field later if it turns out it doesn't
	 */
	SDpnt->borken = 1;
	SDpnt->was_reset = 0;
	SDpnt->expecting_cc_ua = 0;
	SDpnt->starved = 0;

	SRpnt = scsi_allocate_request(SDpnt);
	if (NULL == SRpnt) {
		printk("scan_scsis_single: no memory for request\n");
		scsi_release_commandblocks(SDpnt);
	}
	return SRpnt;
}

/* Send INQUIRY */
static int send_inquiry(Scsi_Request *SRpnt, char *scsi_result, 
			int inq_lun, int timeout, int alloclen)
{
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	Scsi_Device *SDpnt = SRpnt->sr_device;

	scsi_cmd[0] = INQUIRY;
	scsi_cmd[1] = (inq_lun << 5) & 0xe0;
	scsi_cmd[2] = 0;
	scsi_cmd[3] = 0;
	scsi_cmd[4] = alloclen;
	scsi_cmd[5] = 0;
	SRpnt->sr_cmd_len = 0;
	SRpnt->sr_data_direction = SCSI_DATA_READ;

	scsi_wait_req (SRpnt, (void *) scsi_cmd,
	          (void *) scsi_result,
	          scsi_cmd[4], timeout, 3);

	SCSI_LOG_SCAN_BUS(3, printk("scsi: INQUIRY(%i) to host %i channel %i target %i LUN %i %s with code 0x%x\n",
		scsi_cmd[4], 
		SDpnt->host->host_no, SDpnt->channel,
		SDpnt->id, SDpnt->lun,
		SRpnt->sr_result ? "failed" : "succeeded", 
		SRpnt->sr_result));
	
	return SRpnt->sr_result;
}	



/* Send one or two INQUIRIES, get device_flags and return 0 on success */
static int scsi_probe_lun(Scsi_Request *SRpnt, char* scsi_result,
			  int *bflags, int inq_lun)
{
	int inq_len;
	/*
	 * Build an INQUIRY command block.
	 * Strategy: Send 36 byte INQUIRY first and a second, longer one in 
	 * case the device is SCSI-3 and supports more.
	 */
	memset (scsi_result, 0, 256);

	if (!scsi_inq_timeout) {
#ifdef __powerpc64__
		scsi_inq_timeout = SCSI_TIMEOUT/HZ+25;
#else
		scsi_inq_timeout = SCSI_TIMEOUT/HZ+4;
#endif
	}
	send_inquiry(SRpnt, scsi_result, inq_lun, HZ/2+scsi_inq_timeout*HZ, 36);

	/*
	 * Now that we don't do TEST_UNIT_READY anymore, we must be prepared
	 * for media change conditions here, so cannot require zero result.
	 */
	if (SRpnt->sr_result) {
		if ((driver_byte(SRpnt->sr_result) & DRIVER_SENSE) != 0 &&
		    (SRpnt->sr_sense_buffer[2] & 0xf) == UNIT_ATTENTION &&
		    SRpnt->sr_sense_buffer[12] == 0x28 &&
		    SRpnt->sr_sense_buffer[13] == 0) {
			/* not-ready to ready transition - good */
		} else 
			return 1;
	}

	/* Send a longer INQUIRY again .. SCSI-3 specifies some bits in byte 56 
	 * which might be snooped by the HBA.
	 */
	inq_len = (unsigned char)scsi_result[4] + 5;
	if (BLIST_INQUIRY_36 & *bflags)
		inq_len = 36;
	else if (BLIST_INQUIRY_58 & *bflags)
		inq_len = 58;
	else if (inq_len > 255)
		inq_len = 36;	/* sanity */

	if (inq_len > 36 && ((unsigned char)scsi_result[2] & 0x7f) >= 3)
		send_inquiry(SRpnt, scsi_result, inq_lun, 
			     (1+scsi_inq_timeout)*HZ/2, inq_len);
	/* We don't care about the result, actually, if the first INQUIRY
	 * has been successful. */

	/* TODO: What if we got an answer shorter than 32/36 bytes? */
	
	/*
	 * Check for SPARSELUN before checking the peripheral qualifier,
	 * so sparse lun devices are completely scanned.
	 *
	 * Get any flags for this device.  
	 */
	*bflags = get_device_flags (scsi_result);

	return 0;	
}

static int Scsi_Add_Lun(Scsi_Device *SDpnt, char* scsi_result, int bflags)
{
	char devname[64];
	int type = -1;
	extern devfs_handle_t scsi_devfs_handle;
	
	/*   The Toshiba ROM was "gender-changed" here as an inline hack.
	      This is now much more generic.
	      This is a mess: What we really want is to leave the scsi_result
	      alone, and just change the SDpnt structure. And the SDpnt is what
	      we want print_inquiry to print.  -- REW
	 */
	if (bflags & BLIST_ISDISK) {
		scsi_result[0] = TYPE_DISK;
		scsi_result[1] |= 0x80;     /* removable */
	}

	if (bflags & BLIST_ISROM) {
		scsi_result[0] = TYPE_ROM;
		scsi_result[1] |= 0x80;     /* removable */
	}
    
	memcpy(SDpnt->vendor, scsi_result + 8, 8);
	memcpy(SDpnt->model, scsi_result + 16, 16);
	memcpy(SDpnt->rev, scsi_result + 32, 4);

	SDpnt->removable = (0x80 & scsi_result[1]) >> 7;
	/* Use the peripheral qualifier field to determine online/offline */
 	if ((((scsi_result[0] >> 5) & 7) == 1) &&
 	    (SDpnt->lun >= scsi_allow_ghost_devices))
		SDpnt->online = FALSE;
	else 
		SDpnt->online = TRUE;
	SDpnt->lockable = SDpnt->removable;
	SDpnt->changed = 0;
	SDpnt->access_count = 0;
	SDpnt->busy = 0;
	SDpnt->has_cmdblocks = 0;
	/*
	 * Currently, all sequential devices are assumed to be tapes, all random
	 * devices disk, with the appropriate read only flags set for ROM / WORM
	 * treated as RO.
	 */
	switch (type = (scsi_result[0] & 0x1f)) {
	case TYPE_TAPE:
	case TYPE_DISK:
	case TYPE_PRINTER:
	case TYPE_MOD:
	case TYPE_PROCESSOR:
	case TYPE_SCANNER:
	case TYPE_MEDIUM_CHANGER:
	case TYPE_ENCLOSURE:
	case TYPE_COMM:
		SDpnt->writeable = 1;
		break;
	case TYPE_WORM:
	case TYPE_ROM:
		SDpnt->writeable = 0;
		break;
	default:
		printk("scsi: unknown type %d\n", type);
	}

	SDpnt->device_blocked = FALSE;
	atomic_set(&SDpnt->device_busy,0);
	SDpnt->single_lun = 0;
	SDpnt->soft_reset =
	    (scsi_result[7] & 1) && ((scsi_result[3] & 7) == 2);
	SDpnt->random = (type == TYPE_TAPE) ? 0 : 1;
	SDpnt->type = (type & 0x1f);

	print_inquiry(scsi_result);

        sprintf (devname, "host%d/bus%d/target%d/lun%d",
                 SDpnt->host->host_no, SDpnt->channel, SDpnt->id, SDpnt->lun);

        if (SDpnt->de) 
		printk ("DEBUG: dir: \"%s\" already exists\n", devname);
        else 
		SDpnt->de = devfs_mk_dir (scsi_devfs_handle, devname, NULL);

	SDpnt->scsi_level = scsi_result[2] & 0x07;
	if (SDpnt->scsi_level >= 2 ||
	    (SDpnt->scsi_level == 1 &&
	     (scsi_result[3] & 0x0f) == 1))
		SDpnt->scsi_level++;

	/*
	 * Accommodate drivers that want to sleep when they should be in a
	 * polling loop.
	 */
	SDpnt->disconnect = 0;

	/*
	 * Set the tagged_queue flag for SCSI-II devices that purport to support
	 * tagged queuing in the INQUIRY data.
	 */
	SDpnt->tagged_queue = 0;
	if ((SDpnt->scsi_level >= SCSI_2) &&
	    (scsi_result[7] & 2) &&
	    !(bflags & BLIST_NOTQ)) {
		SDpnt->tagged_supported = 1;
		SDpnt->current_tag = 0;
	}
	/*
	 * Some revisions of the Texel CD ROM drives have handshaking problems
	 * when used with the Seagate controllers.  Before we know what type of
	 * device we're talking to, we assume it's borken and then change it
	 * here if it turns out that it isn't a TEXEL drive.
	 */
	if ((bflags & BLIST_BORKEN) == 0)
		SDpnt->borken = 0;

	/*
	 * If we want to only allow I/O to one of the luns attached to this
	 * device at a time, then we set this flag.
	 */
	if (bflags & BLIST_SINGLELUN)
		SDpnt->single_lun = 1;

	return 0;
}

static int Scsi_Unlock_Floptical(Scsi_Request* SRpnt, char* scsi_result,
				 int inq_lun)
{
	unsigned char scsi_cmd[MAX_COMMAND_SIZE];
	
	printk("Unlock floptical drive.\n");
	scsi_cmd[0] = MODE_SENSE;
	scsi_cmd[1] = (inq_lun << 5) & 0xe0;
	scsi_cmd[2] = 0x2e;
	scsi_cmd[3] = 0;
	scsi_cmd[4] = 0x2a;
	scsi_cmd[5] = 0;
	SRpnt->sr_cmd_len = 0;
	SRpnt->sr_data_direction = SCSI_DATA_READ;
	scsi_wait_req (SRpnt, (void *) scsi_cmd,
		       (void *) scsi_result, 0x2a,
		       SCSI_TIMEOUT, 3);
	return SRpnt->sr_result;
}

#ifndef REPORT_LUNS
# define REPORT_LUNS           0xa0
#endif

/* Returns -1 on failure, num_lun on success */
static int scsi_send_replun(Scsi_Device *SDpnt, unsigned int channel, 
			    unsigned int target_id,
			    struct scsi_lun* lun_data, int length)
{	
	char scsi_cmd[MAX_COMMAND_SIZE];
	int retries, result;
	unsigned int num_luns;
	unsigned char *data = (unsigned char*)lun_data->scsi_lun;
	
	Scsi_Request *SRpnt = Scsi_Device_Init(SDpnt, channel, target_id, 0);
	if (!SRpnt)
		return -1; 
	
	scsi_cmd[0] = REPORT_LUNS;
	/* bytes 1 - 5: reserved, set to zero. */
	memset(scsi_cmd+1, 0, 5);
	/* bytes 6 - 9: length of the command. */
	scsi_cmd[6] = (unsigned char) (length >> 24) & 0xff;
	scsi_cmd[7] = (unsigned char) (length >> 16) & 0xff;
	scsi_cmd[8] = (unsigned char) (length >>  8) & 0xff;
	scsi_cmd[9] = (unsigned char)  length        & 0xff;

	scsi_cmd[10] = 0;	/* reserved */
	scsi_cmd[11] = 0;	/* control */
	SRpnt->sr_cmd_len = 0;
	SRpnt->sr_data_direction = SCSI_DATA_READ;

	/*
	 * We can get a UNIT ATTENTION, for example a power on/reset, so
	 * retry a few times (like sd.c does for TEST UNIT READY).
	 * Experience shows some combinations of adapter/devices get at
	 * least two power on/resets.
	 *
	 * Illegal requests (for devices that do not support REPORT LUNS)
	 * should come through as a check condition, and will not generate
	 * a retry.
	 */
	for (retries = 0; retries < 3; ++retries) {
		SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: Sending"
				  " REPORT LUNS to host %i ch %i id %i (try %d)\n",
				  SDpnt->host->host_no, SDpnt->channel, SDpnt->id,		  
				  retries));
		scsi_wait_req(SRpnt, scsi_cmd, lun_data, length,
			      SCSI_TIMEOUT + 2*HZ, 3);
		SCSI_LOG_SCAN_BUS(3, printk (KERN_INFO "scsi scan: REPORT LUNS"
				  " %s (try %d) result 0x%x\n", SRpnt->sr_result
				  ?  "failed" : "successful", retries,
				  SRpnt->sr_result));
		if ((result = SRpnt->sr_result) == 0 ||
		    SRpnt->sr_sense_buffer[2] != UNIT_ATTENTION)
			break;
	}
	
	scsi_release_request(SRpnt);
	scsi_release_commandblocks(SDpnt);
	if (result)
		return -1;
	
	length = ((data[0] << 24) | (data[1] << 16) |
		  (data[2] << 8) | (data[3] << 0));

	num_luns = (length / sizeof(struct scsi_lun));
	if (num_luns > max_report_luns) {
		printk(KERN_WARNING "scsi: On host %i ch %i id %i only %d (max_report_luns)"
		       " of %d luns reported, try increasing"
		       " max_report_luns.\n",
		       SDpnt->host->host_no, SDpnt->channel, SDpnt->id,		  
		       max_report_luns, num_luns);
		num_luns = max_report_luns;
	}
	
	return num_luns;
}


/*
 * The worker for scan_scsis.
 * TODO: Returning 0 means No response, 
 * 1 means target present, 2 means device attached to LUN.
 * Global variables used : scsi_devices(linked list)
 */
static int scan_scsis_single(unsigned int channel, unsigned int id,
		unsigned int lun, int lun0_scsi_level,
		unsigned int *max_dev_lun, unsigned int *sparse_lun, 
		Scsi_Device ** SDpnt2, char *scsi_result)
{
	struct Scsi_Device_Template *sdtpnt;
	Scsi_Device *SDpnt = *SDpnt2;
	struct Scsi_Host *shpnt = SDpnt->host;
	Scsi_Request *SRpnt;
	int bflags, scsi_level;
	int inq_lun = lun0_scsi_level <= SCSI_2? lun: 0;

	/* Set C,I,L and get cmdblocks and request */
	SRpnt = Scsi_Device_Init(SDpnt, channel, id, lun);
	if (NULL == SRpnt)
		return 0;

	/*
	 * We used to do a TEST_UNIT_READY before the INQUIRY but that was 
	 * not really necessary.  Spec recommends using INQUIRY to scan for
	 * devices (and TEST_UNIT_READY to poll for media change). - Paul G.
	 */

	SCSI_LOG_SCAN_BUS(3, printk("scsi: performing INQUIRY\n"));
	if (scsi_probe_lun(SRpnt, scsi_result, &bflags, inq_lun))
		goto Out_Release;
	
	SDpnt->bflags = bflags;

	if (bflags & BLIST_SPARSELUN ||
	    is_on_llun_blklst (shpnt->host_no, channel, id))
		*sparse_lun = 1;

	/*
	 * Check the peripheral qualifier field - this tells us whether LUNS
	 * are supported here or not.
	 */
	if ((scsi_result[0] >> 5) == 3) 
		goto Out_Release;

	Scsi_Add_Lun(SDpnt, scsi_result, bflags);
	
	for (sdtpnt = scsi_devicelist; sdtpnt;
	     sdtpnt = sdtpnt->next)
		if (sdtpnt->detect)
			SDpnt->attached +=
			    (*sdtpnt->detect) (SDpnt);

	scsi_level = SDpnt->scsi_level;

	/*
	 * These devices need this "key" to unlock the devices so we can use it
	 */
	if ((bflags & BLIST_KEY) != 0) {
		SDpnt->lockable = 0;
		Scsi_Unlock_Floptical(SRpnt, scsi_result, inq_lun);
	}

	scsi_release_request(SRpnt);
	scsi_release_commandblocks(SDpnt);

	/*
	 * Since we just found one device, there had damn well better be one 
	 * in the list already.
	 */
	if (shpnt->host_queue == NULL)
		panic("scan_scsis_single: Host queue == NULL\n");
	/*
	 * This device was already hooked up to the host in question,
	 * so at this point we just let go of it and it should be fine.  
	 * We do need to allocate a new one and attach it to the host 
	 * so that we can further scan the bus.
	 */
	SDpnt = Scsi_Alloc_SD(shpnt);
	
	*SDpnt2 = SDpnt;
	SDpnt->scsi_level = scsi_level;

	/*
	 * Some scsi devices cannot be polled for lun != 0 due to firmware bugs
	 */
	if (bflags & BLIST_NOLUN)
		return 0;	/* break; */

	/*
	 * If this device is known to support sparse multiple units, override
	 * the other settings, and scan all of them.
	 */
	if (bflags & BLIST_SPARSELUN ||
	    is_on_llun_blklst (shpnt->host_no, channel, id)) {
		/* Normally max_dev_lun = min(max_scsi_luns, shpny-max_lun)
		 * For SparseLUN devs, it's increased to shpnt->max_lun */
		*max_dev_lun = (max_scsi_sparseluns > 0 && max_scsi_sparseluns < shpnt->max_lun)?
				max_scsi_sparseluns: shpnt->max_lun;
		*sparse_lun = 1;
		return 1;	/* Not really needed, handled in the caller */
	}
	/*
	 * If this device is known to support multiple units, override the other
	 * settings, and scan all of them. 
	 * Note: This FORCELUN flag should not be needed.
	 * SPARSELUN has a clearer meaning.
	 * TODO: Eliminate
	 */
	if (bflags & BLIST_FORCELUN) {
		/* 
		 * Scanning MAX_SCSI_LUNS units would be a bad idea.
		 * Any better idea?
		 * I think we need REPORT LUNS in future to avoid scanning
		 * of unused LUNs. But, that is another item.
		 */
		/* Increase to shpnt->max_lun (overrides max_scsi_luns) */
		if (*max_dev_lun < shpnt->max_lun)
			*max_dev_lun = shpnt->max_lun;
			/* Add more LUNs in steps of shpnt->max_lun until
			 * we've more than max_scsi_luns/2 (Why?) 
			 * (Why not set directly to max_scsi_luns?) */
		else 	if ((max_scsi_luns >> 1) >= *max_dev_lun)
				*max_dev_lun += shpnt->max_lun;
			/* ... and finally do set to max_scsi_luns */
			else	*max_dev_lun = max_scsi_luns;
		return 1;
	}
	/*
	 * REGAL CDC-4X: avoid hang after LUN 4
	 */
	if (bflags & BLIST_MAX5LUN) {
		*max_dev_lun = 5;
		return 1;
	}

	/*
	 * We assume the device can't handle lun!=0 if: - it reports scsi-0
	 * (ANSI SCSI Revision 0) (old drives like MAXTOR XT-3280) or - it
	 * reports scsi-1 (ANSI SCSI Revision 1) and Response Data Format 0
	 */
	if (((scsi_result[2] & 0x07) == 0) ||
	    ((scsi_result[2] & 0x07) == 1 && (scsi_result[3] & 0x0f) == 0))
		return 0;
	return 1;
	
    Out_Release:
	scsi_release_request(SRpnt);
	scsi_release_commandblocks(SDpnt);
	return 0;	/* assume no peripheral if any sort of error */
}

/*
 * Returns the scsi_level of lun0 on this host, channel and dev (if already
 * known), otherwise returns SCSI_2.
 */
static int find_lun0_scsi_level(unsigned int channel, unsigned int dev,
				struct Scsi_Host *shpnt)
{
	int res = SCSI_2;
	Scsi_Device *SDpnt;

	for (SDpnt = shpnt->host_queue; SDpnt; SDpnt = SDpnt->next) {
		if ((0 == SDpnt->lun) && (dev == SDpnt->id) &&
		    (channel == SDpnt->channel))
			return (int)SDpnt->scsi_level;
	}
	/* haven't found lun0, should send INQUIRY but take easy route */
	return res;
}

/* scan the hosts for new devices using the explicit hardcoded call to 
 * scan_scsis. 
 */
void scsi_scan_new_devices(void)
{
        struct Scsi_Host *shpnt;
	int channel,dev, order_dev;
	
        for (shpnt = scsi_hostlist; shpnt != NULL; shpnt = shpnt->next) {
                for (channel = 0; channel <= shpnt->max_channel; channel++){
			for (dev = 0; dev < shpnt->max_id; ++dev)   {
				if( shpnt->reverse_ordering)
					/* Shift to scanning 15,14,13... or 7,6,5,4, */
					order_dev = shpnt->max_id-dev-1;
				else
					order_dev = dev;				
				if (shpnt->this_id != order_dev) {
					Scsi_Device *sdev;
					/* first check to see if we've already seen this device */
					for(sdev = shpnt->host_queue; sdev; sdev = sdev->next) {
						if(sdev->channel == channel
						   && sdev->id == order_dev
						   && sdev->lun == 0) {
							break;
						}
					}					
					if (sdev) 
						continue;
					scsi_scan_hardcoded(shpnt, channel, order_dev, 0);
					/* now check to see if we found lun 0 */
					for(sdev = shpnt->host_queue; sdev; sdev = sdev->next) {
						if(sdev->channel == channel
						   && sdev->id == order_dev
						   && sdev->lun == 0) {
							break;
						}
					}
					if(sdev && sdev->type == TYPE_DISK &&
					   !(sdev->bflags & BLIST_NOLUN)) {
						int lun, max_lun;
						if(sdev->scsi_level == SCSI_3 ||
						   sdev->bflags & BLIST_FORCELUN)
							max_lun = shpnt->max_lun;
						else
							max_lun = min_t(unsigned, 8, shpnt->max_lun);
						for(lun = 1; lun < max_lun; lun++)
							scsi_scan_hardcoded(shpnt, channel, order_dev, lun);
					}
				} /* this_id  != dev */
			} /* for dev */
		} /* for channel */
	} /* for host */
}
