/* sg_dynalloc.c
 * dynamically allocated structures for the sg driver 
 * derived by Holger Smolinski <holger@kunterbunt.bb.bawue.de>
 * from sd_dynalloc: (c) Kurt Garloff <garloff@suse.de>, 7/2002
 * License: GNU GPL v2
 */


#include "sg_dynalloc.h"
#include <linux/spinlock.h>

#define SG_REUSE_SLOTS 1
//#define SG_DYN_DEBUG 1

# define SGD_PRINTK(ARGS...) printk(KERN_WARNING "sg: " ARGS)
#ifdef SG_DYN_DEBUG
# define SGD_DPRINTK(ARGS...) printk(KERN_DEBUG "sg: " ARGS)
#else
# define SGD_DPRINTK(ARGS...)
#endif

/* Locking:
 * We protect the following variables by a spinlock:
 * sg_majors, sg_per_major_ptrs, sg_major_to_midx, next_sg
 * Actually, we try to only use the locks for writing.
 * This should be possible, as we do build and fill
 * structs before we link them into a list and we never
 * remove any majors.
 * So we only need to protect against two CPUs trying to
 * change the structures.
 */

int sg_majors = 0;
/* Index into the last sg_major struct for next free sg */
int sg_next = SG_DEVICES_PER_MAJOR;

/* Maps majors to indices into sg_major_ptrs */
unsigned char sg_major_to_midx[MAJORS];
/* List of pointers to SG_Major structs */
struct sg_per_major *sg_major_ptrs[MAX_SG_DYN_MAJORS];

inline int sg_midx_to_major (const int midx)
{
	if (midx < 0 || midx >= sg_majors)
		return SG_NO_MAJOR;
	return SG_MAJOR_PTR(midx)->major;
}

/* We should already have the lock in write mode when entering */
int sg_alloc_major (void)
{
	int major,rc;
	SG_Major *sg_major;
	
	/* Initialize sg_major_to_midx array */
	if (sg_majors == 0) {	
	SGD_DPRINTK("initializing sg_major_array ...");
		memset (sg_major_to_midx, SG_NO_MIDX, MAJORS);
	}	
	if (sg_majors >= MAX_SG_DYN_MAJORS) {
	SGD_DPRINTK("too many sg dynamic majors ...");
		return SG_NO_MAJOR;
	}
	if (!sg_majors)
		major = SCSI_GENERIC_MAJOR;
	else {
#ifdef CONFIG_DEVFS_FS
		major = devfs_alloc_major(DEVFS_SPECIAL_CHR);
		if ( major < 0 ) {
			SGD_PRINTK("alloc_major: could not allocate major \n");
			goto alloc_fail;
		}
#else
		major = 0;
#endif
	}
	
	rc = devfs_register_chrdev(major, "sg", &sg_fops);
	if (rc < 0) {
		SGD_PRINTK("alloc_major: could not register major %i, rc=%d\n", 
			   major,rc);
		goto dealloc_major;
	} else {
		if ( major == 0 && rc > 0 ) {
			major = rc;
		} 
	}
	/* Now perform memory allocations */
	sg_major = (SG_Major *)kmalloc (sizeof (SG_Major), GFP_ATOMIC);
	if (!sg_major) {
		SGD_PRINTK("alloc_major: could not allocate sg_major structure\n");
		goto unregister_major;
	}
	memset (sg_major, 0, sizeof (SG_Major));
	
	/* Now, nothing can fail ... except the programmer, of course */
	sg_major->major = major;
	sg_major_ptrs[sg_majors]   = sg_major;
	sg_major_to_midx[major] = sg_majors;
	sg_majors++;
	sg_next = 0;
	/* FIXME: used to be done at finish time, should be OK here. KG. */
	SGD_DPRINTK("allocated major %i\n", major);
	return major;

 unregister_major:
	devfs_unregister_chrdev(major, "sg");
 dealloc_major:
#ifdef CONFIG_DEVFS_FS	
	if (sg_majors)
		devfs_dealloc_major(DEVFS_SPECIAL_CHR, major);
#endif
 alloc_fail:
	return SG_NO_MAJOR;
}
	
/* Deallocate one major */
void sg_dealloc_major (const int midx)
{
	SG_Major *sg_major = SG_MAJOR_PTR(midx);
	const int major = sg_major->major;
	SGD_DPRINTK("about to deallocate idx %i ptr %p major %i\n", 
		   midx, sg_major, major);
	devfs_unregister_chrdev(major, "sg");
	sg_major_ptrs[midx] = NULL;
	sg_major_to_midx[major] = SG_NO_MIDX;
#ifdef CONFIG_DEVFS_FS	
	if (midx)
		devfs_dealloc_major(DEVFS_SPECIAL_CHR, major);
#endif
	kfree (sg_major);
}

/* Deallocate all majors */
void sg_deallocate_all (void)
{
	while (sg_majors--) 
		sg_dealloc_major (sg_majors);
}

/* Currently, sg_find_free_slot() does not use SGp.
 * One could think of clever heuristics, that attaches a detached
 * sg dev to the same slot again. KG. 
 * Find a free slot to attach a SG device 
 * Locking: Caller has to make sure that deviceptr has to be
 * assigned atomically with sg_find_free_slot. HSM. */
int sg_find_free_slot (Scsi_Device *SGp, int *midx, int *didx)
{
	SG_Major *sg_major;
	Sg_device *device;
	SGD_DPRINTK("find_free_slot ...");
#ifdef SG_REUSE_SLOTS
	for (*midx = 0; *midx < sg_majors; ++*midx) {
		sg_major = SG_MAJOR_PTR(*midx);
		for (*didx = 0; *didx < SG_DEVICES_PER_MAJOR; ++*didx) {
			device = *SG_DEVICE_PTR(sg_major, *didx);
			if (!device) 
				goto found;
		}
	}
#else
	if (sg_next < SG_DEVICES_PER_MAJOR) {
		*midx = sg_majors-1;
		*didx = sg_next;
		goto found;
	}
#endif
	/* OK, we need to allocate a new major */
	if (sg_alloc_major () < 0) {
		return SG_NO_MAJOR;
	}
	*midx = sg_majors-1;
	*didx = sg_next;	/* 0 */
	
    found:
	SGD_DPRINTK("... found %02x:%02x\n", 
		    SG_MIDX_TO_MAJOR(*midx), *didx);
	++sg_next;
	return 0;
}
