/*
 * This is a debug register allocation mechanism, to be used by
 * all debuggers, which use debug registers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __ASM_I386_DR_ALLOC_H
#define __ASM_I386_DR_ALLOC_H

/* No of debug registers available on IA32 processors */
#define DR_MAX	4

struct dr_struct {
	unsigned long flag;
	unsigned long use_count;
};

/* dr_struct flags */
#define DR_UNUSED	0
#define DR_LOCAL	1
#define DR_GLOBAL	2
	
/* debug register numbers */
#define DR_ANY		DR_MAX + 1

/* global or local allocation requests */
#define DR_ALLOC_GLOBAL		0
#define DR_ALLOC_LOCAL		1

#define DR7_RW_SET(dr, regnum, rw)                              \
       do {                                                    \
               (dr) &= ~(0x3 << (16 + (4 * (regnum))));         \
               (dr) |= (((rw) & 0x3) << (16 + (4 * (regnum)))); \
       } while (0)

#define DR7_RW_VAL(dr, regnum) \
       (((dr) >> (16 + (4 * (regnum)))) & 0x3)

#define DR7_LEN_SET(dr, regnum, rw)                             \
       do {                                                    \
               (dr) &= ~(0x3 << (18 + (4 * (regnum))));         \
               (dr) |= (((rw) & 0x3) << (18 + (4 * (regnum)))); \
       } while (0)

#define DR7_LEN_VAL(dr, regnum) \
       (((dr) >> (18 + (4 * (regnum)))) & 0x3)

#define DR7_G0(dr)    (((dr)>>1)&0x1)
#define DR7_G0SET(dr) ((dr) |= 0x2)
#define DR7_G0CLR(dr) ((dr) &= ~0x2)
#define DR7_G1(dr)    (((dr)>>3)&0x1)
#define DR7_G1SET(dr) ((dr) |= 0x8)
#define DR7_G1CLR(dr) ((dr) &= ~0x8)
#define DR7_G2(dr)    (((dr)>>5)&0x1)
#define DR7_G2SET(dr) ((dr) |= 0x20)
#define DR7_G2CLR(dr) ((dr) &= ~0x20)
#define DR7_G3(dr)    (((dr)>>7)&0x1)
#define DR7_G3SET(dr) ((dr) |= 0x80)
#define DR7_G3CLR(dr) ((dr) &= ~0x80)

#define DR7_L0(dr)    (((dr))&0x1)
#define DR7_L0SET(dr) ((dr) |= 0x1)
#define DR7_L0CLR(dr) ((dr) &= ~0x1)
#define DR7_L1(dr)    (((dr)>>2)&0x1)
#define DR7_L1SET(dr) ((dr) |= 0x4)
#define DR7_L1CLR(dr) ((dr) &= ~0x4)
#define DR7_L2(dr)    (((dr)>>4)&0x1)
#define DR7_L2SET(dr) ((dr) |= 0x10)
#define DR7_L2CLR(dr) ((dr) &= ~0x10)
#define DR7_L3(dr)    (((dr)>>6)&0x1)
#define DR7_L3SET(dr) ((dr) |= 0x40)
#define DR7_L3CLR(dr) ((dr) &= ~0x40)

#define DR_IS_LOCAL(dr, num) ((dr) & (1UL << (num <<1)))

#define DR7_L0_BIT		0x00000001
#define DR7_L1_BIT		0x00000004
#define DR7_L2_BIT		0x00000010
#define DR7_L3_BIT		0x00000040

/* Set the rw, len and global flag in dr7 for a debug register */
#define SET_DR7(dr, regnum, access, len) \
	DR7_RW_SET(dr, regnum, access); \
	DR7_LEN_SET(dr, regnum, len); \
	dr |= (2UL << regnum*2);

/* Disable a debug register by clearing the global/local flag in dr7 */
#define RESET_DR7(dr, regnum) \
	dr &= ~(3UL << regnum*2); \

#define DR7_DR0_BITS		0x000F0003
#define DR7_DR1_BITS		0x00F0000C
#define DR7_DR2_BITS		0x0F000030
#define DR7_DR3_BITS		0xF00000C0

#define DR_TRAP_MASK 		0xF

#define DR_TYPE_EXECUTE		0x0
#define DR_TYPE_WRITE		0x1
#define DR_TYPE_IO		0x2
#define DR_TYPE_RW		0x3

#define get_dr(regnum, val) \
		__asm__("movl %%db" #regnum ", %0"  \
			:"=r" (val))
static inline unsigned long read_dr(int regnum)
{
	unsigned long val = 0;
	switch(regnum) {
		case 0: get_dr(0, val); break;
		case 1: get_dr(1, val); break;
		case 2: get_dr(2, val); break;
		case 3: get_dr(3, val); break;
		case 7: get_dr(7, val); break;
		default: printk("read_dr: invalid debug register\n"); 
			break;
	}
	return val;
}
#undef get_dr
	
#define set_dr(regnum, val) \
		__asm__("movl %0,%%db" #regnum  \
			: /* no output */ \
			:"r" (val))
static inline void write_dr(int regnum, unsigned long val)
{
	switch(regnum) {
		case 0: set_dr(0, val); break;
		case 1: set_dr(1, val); break;
		case 2: set_dr(2, val); break;
		case 3: set_dr(3, val); break;
		case 7: set_dr(7, val); break;
		default: printk("write_dr: invalid debug register %d\n", regnum); 
			break;
	}
	return;
}
#undef set_dr

/*
 * Given the debug status register, returns the debug register number
 * which caused the debug trap.
 */
static inline int dr_trap(unsigned int condition)
{
	int i, reg_shift = 1UL;
	for (i = 0; i < DR_MAX; i++, reg_shift <<= 1)
		if ((condition & reg_shift))
			return i;	
	return -1;
}

/*
 * Given the debug status register, returns the address due to which
 * the debug trap occured.
 */
static inline unsigned long dr_trap_addr(unsigned int condition)
{
	int regnum = dr_trap(condition);

	if (regnum == -1)
		return -1;

	return read_dr(regnum);
}

/*
 * Given the debug status register, returns the type debug trap -
 * execute, read/write, write or io.
 */
static inline int dr_trap_type(unsigned int condition)
{
	int regnum = dr_trap(condition);

	if (regnum == -1)
		return -1;

	return DR7_RW_VAL(read_dr(7), regnum);
}

/* Function declarations */
extern int dr_trap(unsigned int);
extern unsigned long dr_trap_addr(unsigned int);
extern int dr_trap_type(unsigned int);

extern unsigned long read_dr(int);
extern void write_dr(int, unsigned long);

extern int dr_alloc(int regnum, int flag);
extern int dr_free(int regnum);
extern int dr_inc_use_count(unsigned long mask);
extern int dr_dec_use_count(unsigned long mask);

extern struct dr_struct dr_list[DR_MAX];
extern unsigned long dr7_global_mask;

#endif
