/*
 * IBM PowerPC Virtual I/O Infrastructure Support.
 *
 * Dave Engebretsen engebret@us.ibm.com
 *    Copyright (c) 2003 Dave Engebretsen
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 */

#ifndef _VIO_H
#define _VIO_H

struct vio_dev;
struct vio_driver;

dma_addr_t vio_map_single(struct vio_dev *dev, void *vaddr, 
			  size_t size, int direction);
void vio_unmap_single(struct vio_dev *dev, dma_addr_t dma_handle, 
		      size_t size, int direction);
int vio_map_sg(struct vio_dev *vdev, struct scatterlist *sglist, 
	       int nelems, int direction);
void vio_unmap_sg(struct vio_dev *vdev, struct scatterlist *sglist, 
		  int nelems, int direction);
void *vio_alloc_consistent(struct vio_dev *dev, size_t size, 
			   dma_addr_t *dma_handle);
void vio_free_consistent(struct vio_dev *dev, size_t size, void *vaddr, 
			 dma_addr_t dma_handle);

struct vio_device_id {
	unsigned int vendor, device;	  /* Vendor and device ID or PCI_ANY_ID */
	unsigned int subvendor, subdevice;/* Subsystem ID's or PCI_ANY_ID */
	unsigned int class, class_mask;	  /* (class,subclass,prog-if)triplet */
	unsigned long driver_data;	  /* Data private to the driver */
};

struct vio_driver {
	struct list_head node;
	char *name;
	const struct vio_device_id *id_table;	/* NULL if wants all devices */
	int  (*probe)  (struct vio_dev *dev, const struct vio_device_id *id);	/* New device inserted */
	void (*remove) (struct vio_dev *dev);	/* Device removed (NULL if not a hot-plug capable driver) */
	int  (*save_state) (struct vio_dev *dev, u32 state);    /* Save Device Context */
	int  (*suspend) (struct vio_dev *dev, u32 state);	/* Device suspended */
	int  (*resume) (struct vio_dev *dev);	                /* Device woken up */
	int  (*enable_wake) (struct vio_dev *dev, u32 state, int enable);   /* Enable wake event */

};

struct vio_bus;
/*
 * The vio_dev structure is used to describe virtual I/O devices.
 */
struct vio_dev {
	struct list_head global_list;	/* node in list of all vio devices */
	struct list_head bus_list;	/* node in per-bus list */
	struct vio_bus	*bus;		/* bus this device is on */

	void		*sysdata;	/* hook for sys-specific extension */
	struct proc_dir_entry *procent;	/* device entry in /proc/bus/vio */

	unsigned short	vendor;
	unsigned short	device;
	unsigned short	subsystem_vendor;
	unsigned short	subsystem_device;
	unsigned int	class;		/* 3 bytes: (base,sub,prog-if) */

	struct vio_driver *driver;	/* which driver has allocated this device */
	struct TceTable   *tce_table;
	void		  *driver_data;	/* data private to the driver */
	unsigned long     unit_address; 
	unsigned long     liobn; 
	unsigned long     riobn;
	unsigned long     mac_address; 
	unsigned int      mcastFilterSize;
	unsigned int      isConsole;

	char    name[64];
	char    bus_id[64];

	/*
	 * Instead of touching interrupt line and base address registers
	 * directly, use the values stored here. They might be different!
	 */
	unsigned int	irq;
	struct resource resource[DEVICE_COUNT_RESOURCE]; /* I/O and memory regions + expansion ROMs */
	struct resource dma_resource[DEVICE_COUNT_DMA];
	struct resource irq_resource[DEVICE_COUNT_IRQ];

};

struct vio_bus {
        struct list_head node;          /* node in list of buses */
        struct list_head devices;       /* list of devices on this bus */
        struct  device  * dev;
};

int vio_register_driver(struct vio_driver *drv);

static inline int vio_module_init(struct vio_driver *drv)
{
        int rc = vio_register_driver (drv);

        if (rc > 0)
                return 0;

        /* iff CONFIG_HOTPLUG and built into kernel, we should
         * leave the driver around for future hotplug events.
         * For the module case, a hotplug daemon of some sort
         * should load a module in response to an insert event. */
#if defined(CONFIG_HOTPLUG) && !defined(MODULE)
        if (rc == 0)
                return 0;
#else
        if (rc == 0)
                rc = -ENODEV;
#endif

        /* if we get here, we need to clean up vio driver instance
         * and return some sort of error */

        return rc;
}

#endif /* _PHYP_H */
