/*
 * receive code
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999, 2000, 2001  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/version.h>

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/slab.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct net_device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/skbuff.h>
#include <linux/spinlock.h> /* *lock* */
#include <asm/uaccess.h>

#include <linux/in6.h>
#include <asm/checksum.h>
#include <net/ip.h>

#include <linux/pfkeyv2.h>
#include <linux/pfkey.h>
#include <linux/ipsec.h>
#include <net/sadb.h>
#include <net/spd.h>

#include "ipsec_tunnel.h"
#include "ipsec4.h"

/*
 * Check-replay-window routine, adapted from the original 
 * by J. Hughes, from draft-ietf-ipsec-esp-des-md5-03.txt
 *
 *  This is a routine that implements a 64 packet window. This is intend-
 *  ed on being an implementation sample.
 */
int
ipsec4_check_replaywindow(struct sa_replay_window *win, __u32 seq)
{
	__u32 diff;
	
	if (!win->size || !sysctl_ipsec_replay_window)
		return 1;
	if (seq == 0) 
		return 0;		/* first == 0 or wrapped */

	/* new larger sequence number */
	if (seq > win->last_seq)
		return 1;		/* larger is good */

	diff = win->last_seq - seq;

	/* too old or wrapped */ /* if wrapped, kill off SA? */
	if (diff >= win->size)
		return 0;

	/* this packet already seen */
	if (win->bitmap & (1 << diff))
		return 0;
	/* out of order but good */
	return 1;
}

void
ipsec4_update_replaywindow(struct sa_replay_window *win, __u32 seq)
{
	__u32 diff;
	
	if (!win->size || !sysctl_ipsec_replay_window || !seq)
		return;

	/* new larger sequence number */
	if (seq > win->last_seq) {
		diff = seq - win->last_seq;

		/* In win, set bit for this pkt */
		if (diff < win->size)
			win->bitmap <<= diff;
		else
			win->bitmap = 0;
		win->last_seq = seq;
	}

	diff = win->last_seq - seq;
	if (diff < win->size)
		win->bitmap |= 1 << diff;
	return;
}


/*
 * Check AH header received
 */
static int
ipsec4_rcv_ah(struct sk_buff *skb, struct ipsec_sa **sap, u8 *nexthdr)
{
	struct iphdr *iph = skb->nh.iph;
	struct ip_auth_hdr *ah;
	struct sockaddr_in *sin;
	struct sa_index sa_idx;
	struct ipsec_sa *sa = NULL;
	unsigned int ah_len;
	unsigned char hash[AH_AMAX];
	int res = -EINVAL;

	*sap = NULL;

	ah = (struct ip_auth_hdr *) skb->data;
	if (skb->len < sizeof(*ah)
	 || skb->len < (ah_len = (ah->ah_hl << 2) + 8)) {
		if (net_ratelimit())
			printk(KERN_NOTICE
			       "ipsec4_rcv_ah: "
			       "incomplete AH packet header (skb->len = %d)\n",
			       skb->len);
		goto out;
	}
	if (ah_len < AH_HDRLEN) {
		if (net_ratelimit())
			IPSEC4_ERROR("AH header too short: %d bytes\n", ah_len);
		goto out;
	}

	/* Pull the AH */
	skb_pull(skb, ah_len);

	sa_index_init(&sa_idx);
	sin = (struct sockaddr_in *) &sa_idx.dst;
	sin->sin_addr.s_addr = iph->daddr;
	sin->sin_family = AF_INET;
	sa_idx.prefixlen_d = 32;
	sa_idx.ipsec_proto = SADB_SATYPE_AH;
	sa_idx.spi = ah->ah_spi;

	res = -EINVAL;
	if (!(sa = sadb_find_by_sa_index(&sa_idx)))
		goto out;

	write_lock_bh(&sa->lock);
	if (!ipsec4_check_replaywindow(&sa->replay_window, ah->ah_rpl))
		goto out_unlock;

	/* Now verify the AH */
	res = ipsec4_compute_ah(&sa->auth_algo, skb, 0, hash, sizeof(hash));
	if (res > 0) {
		if (res != ah_len - AH_HDRLEN
		 || memcmp(ah->ah_data, hash, res)) {
			if (net_ratelimit())
				printk(KERN_NOTICE
					"bad AH checksum from %u.%u.%u.%u\n",
					NIPQUAD(iph->saddr));
			res = -EINVAL;
		}
	}

out_unlock:
	write_unlock_bh(&sa->lock);
	if (res >= 0) {
		*nexthdr = ah->ah_nh;
		*sap = sa;
	} else {
		ipsec_sa_put(sa);
	}

out:	return res;
}

static int
ipsec4_rcv_esp(struct sk_buff *skb, struct ipsec_sa **sap, u8 *nexthdr)
{
	struct iphdr *iph = skb->nh.iph;
	struct ip_esp_hdr *esp;
	struct sockaddr_in *sin;
	struct sa_index sa_idx;
	struct auth_algo_info *aalg;
	struct esp_algo_info *ealg;
	struct ipsec_sa *sa = NULL;
	unsigned int len, got;
	int locked = 0;
	int res;

	*sap = NULL;

	res = -EINVAL;
	if (skb->len < sizeof(struct ip_esp_hdr)) 
		goto short_packet;

	esp = (struct ip_esp_hdr *) skb->data;
	/* don't pull the ESP header yet; we may need to checksum it */

	/* XXX is this check actually required? */
	len = ntohs(iph->tot_len);
	got = skb->len + ((char *) skb->data - (char *) iph);
	if (len > got)
		goto short_packet;

	IPSEC4_DEBUG("spi=0x%x seq=0x%x\n",
		ntohl(esp->esp_spi), ntohl(esp->esp_rpl));

	sa_index_init(&sa_idx);
	sin = (struct sockaddr_in *) &sa_idx.dst;
	sin->sin_addr.s_addr = iph->daddr;
	sin->sin_family = AF_INET;
	sa_idx.prefixlen_d = 32;
	sa_idx.ipsec_proto = SADB_SATYPE_ESP;
	sa_idx.spi = esp->esp_spi;

	res = -EINVAL;
	if (!(sa = sadb_find_by_sa_index(&sa_idx)))
		return res;

	write_lock_bh(&sa->lock);
	locked = 1;

	ealg = &sa->esp_algo;
	aalg = &sa->auth_algo;

	if (!ipsec4_check_replaywindow(&sa->replay_window, esp->esp_rpl))
		goto out_unlock;

	/* Verify authenticator */
	if (aalg->algo != SADB_AALG_NONE) {
		unsigned char hash[AH_AMAX];
		int hashsize;

		hashsize = aalg->digest_len;
		if (skb->len < hashsize)
			goto short_packet;
		if (hashsize > sizeof(hash))
			goto out_unlock;

		/* authenticator follows data, cut it off */
		skb_trim(skb, skb->len - hashsize);

		/* Now compute the hash */
		aalg->dx->di->hmac_atomic(aalg->dx, aalg->key, aalg->key_len,
				skb->data, skb->len, hash);

		if (memcmp(skb->data + skb->len, hash, hashsize)) {
			if (net_ratelimit())
				printk(KERN_NOTICE
				       "ipsec4_rcv_esp: checksum failed\n");
			goto out_unlock;
		}
		IPSEC4_DEBUG("checksum okay\n");
	}

	/* Done with the authenticator, pull the ESP header how */
	skb_pull(skb, ESP_HDRLEN);

	/* Decrypt packet */
	if (ealg->algo != SADB_EALG_NONE) {
		unsigned char *iv, *data;
		unsigned int iv_size;
		int pad;

		if (!ealg->cx || !ealg->cx->ci)
			goto out_unlock;

		iv_size = ealg->cx->ci->ivsize;
		if (skb->len < iv_size + 4)
			goto short_packet;
		iv = (char *) skb->data;
		skb_pull(skb, iv_size);

		if (skb->len % ealg->cx->ci->blocksize) {
			if (net_ratelimit())
				printk(KERN_NOTICE
				       "ipsec4_rcv_esp: "
				       "packet with odd content length = %d\n",
				       skb->len);
			goto out_unlock;
		}
		ealg->cx->ci->decrypt_atomic_iv(ealg->cx,
				skb->data, skb->data, skb->len, iv);

		data = (unsigned char *) skb->data;
		*nexthdr = data[skb->len-1];
		pad = data[skb->len-2];
		if (pad + 2 > skb->len)
			goto short_packet;
		skb_trim(skb, skb->len - 2 - pad);

		/* need to update IP header */
		iph->protocol = *nexthdr;
		iph->tot_len = htons(skb->len);
		skb->nh.iph->check = ip_fast_csum((char *) iph, iph->ihl);

		IPSEC4_DEBUG("decrypted payload, nexthdr=%u, len=%d\n",
			*nexthdr, skb->len);

		res = 0;
	}

out_unlock:
	if (locked)
		write_unlock_bh(&sa->lock);

	if (res >= 0) {
		*sap = sa;
	} else {
		ipsec_sa_put(sa);
	}

	return res;

short_packet:
	if (net_ratelimit())
		printk(KERN_NOTICE "ipsec4_rcv_esp: short packet\n");
	goto out_unlock;
}

static int
ipsec4_rcv_comp(struct sk_buff *skb, struct ipsec_sa **sap, u8 *nexthdr)
{
	if (net_ratelimit())
		printk("IPCOMP not yet implemented\n");
	return -EINVAL;
}

/*
 * Check policy vs incoming packet
 */
static int
ipsec4_check_input_policy(struct sk_buff *skb, int nexthdr,
		struct ipsec_sa **present, unsigned int sa_count)
{
	struct ipsec_sa *required[MAX_CHAINED_SAS];
	struct ipsec_sp *policy;
	int num_required = 0, j, k;
	int verdict;
#ifdef CONFIG_IPSEC_DEBUG
	char buf[64];
#endif

	/* Policy check */
	if (!(policy = ipsec4_policy_lookup(skb->nh.iph, nexthdr))) {
		if (net_ratelimit())
			printk(KERN_NOTICE
			       "ipsec4_rcv: no policy for packet\n");
		return IPSEC_POLICY_DROP;
	}

	num_required = ipsec4_policy_to_salist(policy,
				required, MAX_CHAINED_SAS);
	if (num_required < 0)
		return IPSEC_POLICY_DROP;

	/* Get the action from the policy */
	verdict = policy->policy_action;

	/* The list of required SAs is sorted inner to outer,
	 * the list we have is sorted outer to inner */
	j = 0; k = sa_count;
	while (j < num_required) {
		/* Skip any IPIP and COMP SAs specified by the policy;
		 * the packet is not required to have them */
		if (required[j]->ipsec_proto == SADB_X_SATYPE_IPIP
		 || required[j]->ipsec_proto == SADB_X_SATYPE_COMP) {
			j++;
			continue;
		}

		if (k == 0)
			break;

		/* Skip any COMP SAs we found in the packet */
		if (present[k-1]->ipsec_proto == SADB_X_SATYPE_COMP) {
			k--;
			continue;
		}

		if (required[j] != present[k-1]) {
#ifdef CONFIG_IPSEC_DEBUG
			IPSEC4_DEBUG("policy check failed; require %s",
				ipsec4_print_sa(required[j], buf, sizeof(buf)));
			printk("; found %s\n", 
				ipsec4_print_sa(present[k-1], buf, sizeof(buf)));
#endif
			verdict = IPSEC_POLICY_DROP;
			goto done;
		}
		j++, k--;
	}

	if (j < num_required) {
		IPSEC4_DEBUG("policy check failed; missing %s\n",
			ipsec4_print_sa(required[j], buf, sizeof(buf)));
		verdict = IPSEC_POLICY_DROP;
	} else if (k > 0) {
		IPSEC4_DEBUG("policy check failed; unexpected %s\n",
			ipsec4_print_sa(present[k-1], buf, sizeof(buf)));
		verdict = IPSEC_POLICY_DROP;
	}

done:	for (j = 0; j < num_required; j++)
		ipsec_sa_put(required[j]);

	ipsec_sp_put(policy);
	return verdict;
}

/*
 * Receive an IPsec packet.
 * When we get here, the outer IP header has been pulled
 */
int
ipsec4_rcv(struct sk_buff *skb)
{
	struct iphdr *ipp;
	struct ipsec_sa *sa_index[MAX_CHAINED_SAS];
	unsigned int sa_count = 0;
	struct net_device_stats *stats = NULL;
	__u8 nexthdr = 0;
	__u8 proto;
	int res, verdict;

	IPSEC4_DEBUG(KERN_DEBUG "ipsec4_rcv called.\n");
	/* Don't unlink in the middle of a turnaround */
	MOD_INC_USE_COUNT;
	
	if (skb == NULL) {
		IPSEC4_DEBUG("NULL skb passed in.\n");
		goto out;
	}
		
	if (skb->data == NULL) {
		IPSEC4_DEBUG("NULL skb->data passed in, packet is bogus, dropping.\n");
		goto out;
	}

	/* The original FreeSWAN code creates its own copy of the skb
	 * in case it's cloned. We don't because we don't alter
	 * its contents. */

	/* In Linux 2.4.4, we may have to reassemble fragments. They are
	 * not assembled automatically to save TCP from having to copy
	 * twice.
	 */
	if (skb_is_nonlinear(skb)) {
		if (skb_linearize(skb, GFP_ATOMIC) != 0)
			goto out;
	}

	ipp = (struct iphdr *)skb->nh.iph;
	IPSEC4_PRINT_IP(ipp);

	proto = ipp->protocol;
	switch (proto) {
	case IPPROTO_AH:
	case IPPROTO_ESP:
	case IPPROTO_COMP:
		break;
	default:
		if (net_ratelimit())
			printk(KERN_NOTICE
				"ipsec_rcv: unexpected protocol %d\n",
				ipp->protocol);
		goto out;
	}

	if (skb->dev) {
		struct net_device *ipsecdev;
		struct ipsecpriv *priv;

		ipsecdev = ipsec4_tunnel_from_device(skb->dev);
		if (ipsecdev == NULL) {
			if (net_ratelimit())
				printk(KERN_NOTICE
					"No tunnel device found for incoming "
					"IPsec packet.\n");
			goto out;
		}

		/* this shouldn't happen */
		if (!(priv = (struct ipsecpriv *) ipsecdev->priv))
			goto out;
		skb->dev = ipsecdev;
		stats = (struct net_device_stats *) &priv->mystats;
		stats->rx_packets++;
	} else {
		IPSEC4_DEBUG("device supplied with skb is NULL\n");
		stats = NULL;
	}

	/* begin decapsulating loop here */
	do {
		struct ipsec_sa *sa = NULL;
#ifdef CONFIG_IPSEC_DEBUG
		char sabuf[64];
#endif

		skb->h.raw = skb->data;
		switch (proto) {
		case IPPROTO_ESP:
			res = ipsec4_rcv_esp(skb, &sa, &nexthdr);
			break;
		case IPPROTO_AH:
			res = ipsec4_rcv_ah(skb, &sa, &nexthdr);
			break;
		case IPPROTO_COMP:
			res = ipsec4_rcv_comp(skb, &sa, &nexthdr);
			break;
		default:
			if (net_ratelimit())
				printk(KERN_NOTICE
					"Unexpected protocol %d "
					"in ipsec4_rcv\n", proto);
			res = -EINVAL;
		}

		if (res < 0)
			goto rx_error;

		/* Leave room for IPIP SA, if present */
		if (sa_count >= MAX_CHAINED_SAS - 2) {
			if (net_ratelimit())
				printk(KERN_WARNING
				       "Too many nested IPsec headers\n");
			ipsec_sa_put(sa);
			goto rx_error;
		}

		sa_index[sa_count++] = sa;

		/* Update lifetime */
		ipsec_sa_update_lifetime(sa, skb->len);

		proto = nexthdr;

		IPSEC4_DEBUG("Applied transform %s; nexthdr=%u\n",
			ipsec4_print_sa(sa, sabuf, sizeof(sabuf)),
			nexthdr);
	} while (proto == IPPROTO_AH || proto == IPPROTO_ESP || proto == IPPROTO_COMP);

	/* Get the inner IP header in case of a tunnel */
	if (proto == IPPROTO_IPIP) {
		struct iphdr *iph = (struct iphdr *) skb->data;

		if (skb->len < sizeof(struct iphdr)
		 || skb->len < (iph->ihl << 2)) {
			if (net_ratelimit())
				printk(KERN_NOTICE
				       "ipsec4_rcv: short packet (IPIP)\n");
			goto rx_error;
		}

		IPSEC4_DEBUG("innermost header is IPIP\n");
		IPSEC4_PRINT_IP((struct iphdr *) skb->data);
	} else {
		unsigned int ihl = ipp->ihl << 2;

		IPSEC4_DEBUG("transport mode packet, no IPIP header\n");
		/* Push back the IP header for netif_rx */
		skb_push(skb, ihl);
		memmove(skb->data, ipp, ihl);

		/* Update the IP header and redo checksum */
		ipp = (struct iphdr *) skb->data;
		ipp->protocol = proto;
		ipp->tot_len = htons(skb->len);
		ipp->check = 0;
		ipp->check = ip_fast_csum((unsigned char *)ipp, ipp->ihl);

		IPSEC4_PRINT_IP(ipp);
	}

	memset(&(IPCB(skb)->opt), 0, sizeof(struct ip_options));
	skb->mac.raw = skb->data - 14; /* ETH_HLEN */
	skb->nh.raw = skb->data;
	skb->h.raw = NULL;
	skb->protocol = htons(ETH_P_IP);
	skb->ip_summed = 0;

	/* Policy check */
	verdict = ipsec4_check_input_policy(skb, proto,
			sa_index, sa_count);

	switch (verdict) {
	case IPSEC_POLICY_BYPASS:
	case IPSEC_POLICY_APPLY:
		/* okay, nothing */
		break;
	default:
		if (net_ratelimit())
			printk(KERN_NOTICE
				"ipsec4_rcv: incoming packet failed "
				"policy check; dropped\n");
		goto rx_error;
	}

	if(stats)
		stats->rx_bytes += skb->len;
	if (skb->dst) {
		dst_release(skb->dst);
		skb->dst = NULL;
	}
	skb->pkt_type = PACKET_HOST;

#ifdef CONFIG_NETFILTER
        nf_conntrack_put(skb->nfct);
        skb->nfct = NULL;
#endif
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif /* CONFIG_NETFILTER_DEBUG */

	IPSEC4_DEBUG("netif_rx() called.\n");
	netif_rx(skb);
	skb = NULL;

out:	if(skb)
                kfree_skb(skb);

	while (sa_count--)
		ipsec_sa_put(sa_index[sa_count]);

	MOD_DEC_USE_COUNT;
	return(0);

rx_error:
	if (stats)
		stats->rx_errors++;
	goto out;
}
