/*
 * IPSEC Tunneling code. Heavily based on drivers/net/new_tunnel.c
 * Copyright (C) 1996, 1997  John Ioannidis.
 * Copyright (C) 1998, 1999, 2000, 2001  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h> /* printk() */

#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/mm.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/udp.h>         /* struct udphdr */
#include <linux/skbuff.h>
#define MSS_HACK_		/* experimental */
#include <asm/uaccess.h>
#include <linux/in6.h>
#include <net/dst.h>

#include <asm/checksum.h>
#include <net/icmp.h>		/* icmp_send() */
#include <net/ip.h>
#include <linux/netfilter_ipv4.h>

#include <linux/if_arp.h>
#ifdef MSS_HACK
# include <net/tcp.h>		/* TCP options */
#endif	/* MSS_HACK */

#include <linux/ipsec.h>
#include <linux/pfkeyv2.h>
#include <linux/pfkey.h>
#include <net/sadb.h>
#include <net/spd.h>

#include "ipsec_tunnel.h"
#include "ipsec4.h"

#define IPSEC_NUM_IF	4

#ifdef CONFIG_IPSEC_DEBUG_
static void
dmp(char *s, caddr_t bb, int len)
{
	int i;
	unsigned char *b = bb;
  
	if (sysctl_ipsec_debug_ipv4) {
		printk(KERN_INFO "klips_debug:ipsec4_tunnel_:dmp: "
		       "at %s, len=%d:",
		       s,
		       len);
		for (i=0; i < len; i++) {
			if(!(i%16)){
				printk("\nklips_debug:  ");
			}
			printk(" %02x", *b++);
		}
		printk("\n");
	}
}
#else /* CONFIG_IPSEC_DEBUG */
#define dmp(_x, _y, _z) 
#endif /* CONFIG_IPSEC_DEBUG */

static int
ipsec4_tunnel_open(struct net_device *dev)
{
	struct ipsecpriv *prv = dev->priv;
	
	/*
	 * Can't open until attached.
	 */

	IPSEC4_DEBUG("dev = %s, prv->dev = %s\n",
		    dev->name, prv->dev?prv->dev->name:"NONE");

	if (prv->dev == NULL)
		return -ENODEV;
	
	MOD_INC_USE_COUNT;
	return 0;
}

static int
ipsec4_tunnel_close(struct net_device *dev)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

#ifdef MSS_HACK
/*
 * Issues:
 *  1) Fragments arriving in the tunnel should probably be rejected.
 *  2) How does this affect syncookies, mss_cache, dst cache ?
 *  3) Path MTU discovery handling needs to be reviewed.  For example,
 *     if we receive an ICMP 'packet too big' message from an intermediate 
 *     router specifying it's next hop MTU, our stack may process this and
 *     adjust the MSS without taking our AH/ESP overheads into account.
 */

 
/*
 * Recaclulate checksum using differences between changed datum, 
 * borrowed from netfilter.
 */
static u_int16_t 
ipsec_fast_csum(u_int32_t oldvalinv, u_int32_t newval, u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs),
	oldcheck^0xFFFF));
}

/*
 * Determine effective MSS.
 *
 * Note that we assume that there is always an MSS option for our own
 * SYN segments, which is mentioned in tcp_syn_build_options(), kernel 2.2.x.
 * This could change, and we should probably parse TCP options instead.
 *
 */
static void
__ipsec_adjust_mss(struct sk_buff *skb, struct tcphdr *tcph, u_int16_t mtu)
{
	u_int16_t oldmss, newmss;
	u_int32_t *mssp;
	struct sock *sk = skb->sk;
	
	newmss = tcp_sync_mss(sk, mtu);
	printk(KERN_INFO "klips: setting mss to %u\n", newmss);
	mssp = (u_int32_t *)tcph + sizeof(struct tcphdr) / sizeof(u_int32_t);
	oldmss = ntohl(*mssp) & 0x0000FFFF;
	*mssp = htonl((TCPOPT_MSS << 24) | (TCPOLEN_MSS << 16) | newmss);
	tcph->check = ipsec_fast_csum(htons(~oldmss), 
	                              htons(newmss), tcph->check);
}

static inline void
ipsec_adjust_mss(struct sk_buff *skb, struct tcphdr *tcph, u_int16_t mtu)
{
	struct tcphdr *tcph = skb->h.th;

	if (tcph->syn && !tcph->ack)
		__ipsec_adjust_mss(skb, tcph, mtu);
}
#endif	/* MSS_HACK */
                                                        
static inline int ipsec4_tunnel_xmit2(struct sk_buff *skb)
{
	return ip_send(skb);
}

/*
 * Handle a TRAP policy.
 */
static int
ipsec_handle_trap(struct ipsec_sp *sp, struct sk_buff *skb)
{
	int err;

	if ((err = ipsec4_acquire(sp, skb)) < 0) {
		if (net_ratelimit())
			printk(KERN_NOTICE
			      "ipsec_rcv: Failed to send ACQUIRE, errno=%d\n",
			      -err);
	} else {
		/* Successfully sent ACQUIRE message.
		 * Expect key mgmt daemon to set up SAs as required. */
		write_lock_bh(&sp->lock);
		if (sp->policy_action == IPSEC_POLICY_TRAPSUBNET
		 || sp->policy_action == IPSEC_POLICY_TRAP) {
			sp->policy_action = IPSEC_POLICY_HOLD;
			if (sp->hold.last)
				kfree_skb(sp->hold.last);
			sp->hold.last = skb;
			skb_get(skb);
		}
		write_unlock_bh(&sp->lock);
	}

	return err;
}

/*
 * Check if the packet is an IKE packet
 */
static int ipsec4_is_ike(struct sk_buff *skb)
{
	struct iphdr	*iph = skb->nh.iph;
	unsigned int	ihl = iph->ihl << 2;
	struct udphdr	*uh;

	if (iph->protocol != IPPROTO_UDP
	 || inet_addr_type((unsigned long) iph->saddr) != RTN_LOCAL)
		return 0;

	/* We can't deal with fragmented IKE packets */
	if ((iph->frag_off & htons(IP_OFFSET)) != 0)
		return 0;

	if (sizeof(*uh) > (ntohs(iph->tot_len) - ihl))
		return 0;

	uh = (struct udphdr *) ((unsigned char *) iph + ihl);
	return uh->source == htons(500) && uh->dest == htons(500);
}

/*
 * Make sure the SA is valid
 */
static int ipsec_sa_valid(struct ipsec_sa *sa)
{
	int	res = 1;

	read_lock_bh(&sa->lock);

	switch (sa->state) {
	case SADB_SASTATE_MATURE:
	case SADB_SASTATE_DYING:
		break;
	case SADB_SASTATE_LARVAL:
		/* XXX send up ACQUIRE */
		ipsec_sa_acquire(sa);
	default:
		IPSEC4_DEBUG("SA not valid (in state %d)\n", sa->state);
		res = 0;
	}

	/* Prevent replay counter from wrapping */
	if (sa->replay_window.seq_num == -1) {
		IPSEC4_DEBUG("SA not valid: replay window counter wrapped.\n");
		res = 0;
	}

	/* Lifetime check is performed by a timer, no need to mess
	 * with here */

	read_unlock_bh(&sa->lock);
	return res;
}

/*
 * Compute headroom/tailroom
 */
static int ipsec4_headroom(struct ipsec_sa *sa,
			unsigned int *headroom, unsigned int *tailroom)
{
	int res = 0;

	read_lock_bh(&sa->lock);
	switch (sa->ipsec_proto) {
	case SADB_SATYPE_AH:
		if (sa->auth_algo.algo != SADB_AALG_NONE) {
			*headroom += (AH_HDRLEN + sa->auth_algo.digest_len + 7) & ~7;
		}
		break;

	case SADB_SATYPE_ESP:
		if (sa->esp_algo.algo != SADB_EALG_NONE) {
			unsigned int blksize = sa->esp_algo.cx->ci->blocksize;

			*headroom += ESP_HDRLEN + sa->esp_algo.cx->ci->ivsize;
			*tailroom += ((blksize + 3) & ~3) + 2 - 1;
		}
		if (sa->auth_algo.algo != SADB_AALG_NONE)
			*tailroom += (sa->auth_algo.digest_len + 3) & ~3;

		break;

	case SADB_X_SATYPE_IPIP:
		*headroom += sizeof(struct iphdr);
		break;

	case SADB_X_SATYPE_COMP:
		/*
		 * We can't predict how much the packet will
		 * shrink without doing the actual compression.
		 * We could do it here, if we were the first
		 * encapsulation in the chain.  That might save
		 * us a skb_copy_expand, since we might fit
		 * into the existing skb then.  However, this
		 * would be a bit unclean (and this hack has
		 * bit us once), so we better not do it. After
		 * all, the skb_copy_expand is cheap in
		 * comparison to the actual compression.
		 * At least we know the packet will not grow.
		 */
		break;

	default:
		IPSEC4_DEBUG("cannot compute headroom for unknown sa type %d\n",
			sa->ipsec_proto);
		res = -1;
	}

	read_unlock_bh(&sa->lock);
	return res;
}


/*
 * Update the tunnel's MTU if required
 */
static inline int
ipsec4_tunnel_update_mtu(struct ipsecpriv *prv, unsigned int overhead)
{
	int	new_mtu, cur_mtu = prv->mtu;

	if (!prv->dev || (new_mtu = prv->dev->mtu - overhead) < 0)
		return -1;

	if (new_mtu < cur_mtu) {
		IPSEC4_DEBUG("dev %s mtu of %d decreased by %d to %d\n",
			    prv->dev->name,
			    cur_mtu, cur_mtu - new_mtu, new_mtu);
		prv->mtu = new_mtu;
	}

	return 0;
}

/*
 * Build the authentication header, and compute the checksum.
 */
static int
ipsec4_build_ah(struct sk_buff *skb, struct iphdr *iph,
		struct ip_auth_hdr *ahp, size_t hdrlen,
		struct ipsec_sa *sa)
{
	int ret;

	if (sa->auth_algo.algo == SADB_AALG_NONE) {
		if (net_ratelimit())
			IPSEC4_ERROR("AH SA with hash algorithm NONE\n");
		return -EINVAL;
	}

	ahp->ah_hl  = (hdrlen >> 2) - sizeof(__u64)/sizeof(__u32);
	ahp->ah_nh  = iph->protocol;
	ahp->ah_rv  = 0;
	ahp->ah_spi = sa->spi;
	ahp->ah_rpl = htonl(++(sa->replay_window.seq_num));
	iph->protocol = IPPROTO_AH;
	skb->h.raw = (unsigned char *) ahp;

	ret = ipsec4_compute_ah(&sa->auth_algo, skb,
				((unsigned char *) ahp - skb->data) + hdrlen,
				ahp->ah_data, hdrlen - AH_HDRLEN);

	return ret;
}

/*
 * Build the ESP header, and encrypt the packet
 */
static int
ipsec4_build_esp(struct sk_buff *skb, struct iphdr *iph,
		struct ip_esp_hdr *esp, struct ipsec_sa *sa)
{
	struct esp_algo_info *ealg = &sa->esp_algo;
	struct auth_algo_info *aalg = &sa->auth_algo;
	unsigned int offset, length, blksize, i, padding;
	unsigned char *payload;
	unsigned char next_header = iph->protocol;

	iph->protocol = IPPROTO_ESP;
	esp->esp_rpl = htonl(++(sa->replay_window.seq_num));
	esp->esp_spi = sa->spi;

	if (ealg->algo == SADB_EALG_NONE) {
		if (net_ratelimit())
			IPSEC4_ERROR("ESP SA with encryption algo NONE\n");
		return -EINVAL;
	}
	if (!ealg->cx || !ealg->cx->ci) {
		if (net_ratelimit())
			IPSEC4_ERROR("ESP SA not initialized\n");
		return -EINVAL;
	}

	/* Offset of payload that need to be encrypted */
	offset  = ((char *) esp->esp_data - (char *) skb->data) + ealg->cx->ci->ivsize;
	length  = skb->len - offset;
	payload = skb->data + offset;

	blksize = (ealg->cx->ci->blocksize + 3) & ~3;
	padding = blksize - ((length + 2) % blksize);

	/* shouldn't happen */
	if (skb_tailroom(skb) < padding + 2)
		return -EINVAL;
	skb_put(skb, padding + 2);

	/* pad stuff */
	for (i = 0; i < padding; length++, i++)
		payload[length] = i + 1;

	/* pad byte */
	payload[length++] = padding;
	payload[length++] = next_header;

	IPSEC4_DEBUG("Added ESP header (seq=%d, pad=%d, nh=%d)\n",
			htonl(esp->esp_rpl), padding, next_header);

#ifdef CONFIG_IPSEC_DEBUG
	if (sysctl_ipsec_debug_ipv4 > 1) {
		unsigned int i;

		IPSEC4_DEBUG("IV");
		for (i = 0; i < ealg->cx->ci->ivsize; i++)
			printk("%c%02x", i? ':' : '=', ealg->iv[i]);
		printk("\n");
	}
#endif

	ealg->cx->ci->encrypt_atomic_iv(ealg->cx, payload, payload, length, ealg->iv);
	memcpy(esp->esp_data, ealg->iv, ealg->cx->ci->ivsize);

	if (aalg->algo != SADB_AALG_NONE) {
		unsigned char digest[AH_AMAX];
		int hashsize = aalg->digest_len;

		/* shouldn't happen */
		if (skb_tailroom(skb) < hashsize)
			return -EINVAL;

		IPSEC4_DEBUG("adding checksum, len=%d\n", hashsize);

		/* Note we don't copy the digest into the skb directly
		 * because the algorithm's true digest length is usually
		 * more than aalg->digest_len (usu 12). */
		offset  = (char *) esp - (char *) skb->data;
		aalg->dx->di->hmac_atomic(aalg->dx, aalg->key, aalg->key_len,
				skb->data + offset,
				skb->len  - offset,
				digest);
		memcpy(skb->data + skb->len, digest, hashsize);
		skb_put(skb, hashsize);
	}

	skb->h.raw = (unsigned char*) esp;
	return 0;
}

/*
 * Add IPIP encapsulation.
 * The original IP header has already been copied to the outer
 * header, all we need to do is fill in the details.
 */
static int
ipsec4_build_ipip(struct sk_buff *skb, struct iphdr *iph,
		  struct iphdr *inner, struct ipsec_sa *sa)
{
	IPSEC4_DEBUG("Adding IPIP hdr src=%u.%u.%u.%u dst=%u.%u.%u.%u\n",
		NIPQUAD(((struct sockaddr_in *) &sa->src)->sin_addr.s_addr),
		NIPQUAD(((struct sockaddr_in *) &sa->dst)->sin_addr.s_addr));

	iph->version  = 4;
	if (sysctl_ipsec4_tos)
		iph->tos = sysctl_ipsec4_tos;
	else
		iph->tos = inner->tos;
	iph->ttl      = sysctl_ip_default_ttl;
	iph->frag_off = 0;
	iph->saddr    = ((struct sockaddr_in *) &sa->src)->sin_addr.s_addr;
	iph->daddr    = ((struct sockaddr_in *) &sa->dst)->sin_addr.s_addr;
	iph->protocol = IPPROTO_IPIP;
	iph->ihl      = sizeof(struct iphdr) >> 2 /* 5 */;
	skb->h.ipiph  = iph;
	ip_select_ident(iph);

	return 0;
}

/*
 * Build IPCOMP header
 */
static int
ipsec4_build_comp(struct sk_buff *skb, struct iphdr *iph, void *base,
		struct ipsec_sa *sa)
{
#if 1
	return -EINVAL;
#else
	{
		unsigned int flags = 0;
		unsigned int old_tot_len = ntohs(iph->tot_len);
		tdbp->tdb_comp_ratio_dbytes += ntohs(iph->tot_len);

		skb = skb_compress(skb, tdbp, &flags);
		iph = skb->nh.iph;

		tdbp->tdb_comp_ratio_cbytes += ntohs(iph->tot_len);

		if (debug_tunnel & DB_TN_CROUT) {
			if (old_tot_len > ntohs(iph->tot_len))
				IPSEC4_DEBUG("packet shrunk from %d to %d bytes after compression, cpi=%04x (should be from spi=%08x, spi&0xffff=%04x.\n",
					    old_tot_len, ntohs(iph->tot_len),
					    ntohs(((struct ipcomphdr*)(((char*)iph) + ((iph->ihl) << 2)))->ipcomp_cpi),
					    ntohl(tdbp->tdb_said.spi),
					    (__u16)(ntohl(tdbp->tdb_said.spi) & 0x0000ffff));
			else
				IPSEC4_DEBUG("packet did not compress (flags = %d).\n",
					    flags);
		}
	}
#endif
}

/*
 *	This function assumes it is being called from dev_queue_xmit()
 *	and that skb is filled properly by that function.
 */

static int
ipsec4_tunnel_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct ipsecpriv *prv;		/* Our device' private space */
	struct sk_buff *old_skb = NULL;	/* Original skb pointer */
	struct net_device_stats *stats;	/* This device's statistics */
	struct iphdr  *iph;		/* Our new IP header */
	int	headroom;
	int	tailroom;
	__u8	*saved_header = NULL;	/* saved copy of the hard header */
	struct ipsec_sp *policy = NULL;
	struct ipsec_sa *sa_index[MAX_CHAINED_SAS];
	unsigned int j, sa_count = 0;
	unsigned char saved_hard_header[24];
	int hard_header_len = 0;
	struct net_device *physdev;
	struct rtable *rt = NULL;
	int pass = 0;
	int verdict;
	int error = 0;
	struct ipsec_sa *sa;

	/*
	 *	Return if there is nothing to do.  (Does this ever happen?) XXX
	 */
	if (skb == NULL) {
		IPSEC4_DEBUG("Nothing to do!\n" );
		goto cleanup;
	}
	if (dev == NULL) {
		IPSEC4_DEBUG("No device associated with skb!\n" );
		goto cleanup;
	}

	prv = dev->priv;
	if (prv == NULL) {
		IPSEC4_DEBUG("Device has no private structure!\n" );
		goto cleanup;
	}

	physdev = prv->dev;
	if (physdev == NULL) {
		IPSEC4_DEBUG("Device is not attached to physical device!\n" );
		goto cleanup;
	}

	stats = (struct net_device_stats *) &(prv->mystats);

	/* if skb was cloned (most likely due to a packet sniffer such as
	   tcpdump being momentarily attached to the interface), make
	   a copy of our own to modify */
	if(skb_cloned(skb)) {
		if (skb_cow(skb, skb_headroom(skb)) != 0) {
			IPSEC4_DEBUG("skb_cow failed to allocate buffer, dropping.\n" );
			stats->tx_dropped++;
			goto cleanup;
		}
	}

	iph = skb->nh.iph;

	/* sanity check for IP version as we can't handle IPv6 right now */
	if (iph->version != 4) {
		IPSEC4_DEBUG("found IP Version %d but cannot process other IP versions than v4.\n",
			    iph->version); /* XXX */
		goto dropit;
	}
	
#ifdef CONFIG_IPSEC_DEBUG
	if (sysctl_ipsec_debug_ipv4) {
		int i;

		IPSEC4_DEBUG(">>> skb->len=%ld hard_header_len:%d",
		       (unsigned long int)skb->len, hard_header_len);
		for (i=0; i < hard_header_len; i++)
			printk("%c%02x", i? ':' : ' ', skb->data[i]);
		printk(" \n");
	}
#endif /* CONFIG_IPSEC_DEBUG */

	IPSEC4_PRINT_IP(iph);

	/*
	 * Sanity checks
	 */

	if ((iph->ihl << 2) != sizeof (struct iphdr)) {
		IPSEC4_DEBUG("cannot process IP header options yet.  May be mal-formed packet.\n"); /* XXX */
		stats->tx_dropped++;
		goto cleanup;
	}
	
	/*
	 * First things first -- look us up in the policy database.
	 * Always try transport mode first, if that fails try for
	 * a tunnel.
	 */
	if (!(policy = ipsec4_policy_lookup(iph, iph->protocol)))
		policy = ipsec4_policy_lookup(iph, IPPROTO_IPIP);

	if (ipsec4_is_ike(skb)) {
		/* If this is an IKE packet, let it through
		 * without interference.
		 */
		IPSEC4_DEBUG("IKE packet, bypass\n");
		verdict = IPSEC_POLICY_BYPASS;
	} else if (policy) {
		policy->stats.count++;
		policy->stats.last_used = jiffies/HZ;
		verdict = policy->policy_action;
	} else {
		/* default to a %drop eroute */
		verdict = IPSEC_POLICY_DROP;
	}

	switch (verdict) {
	case IPSEC_POLICY_HOLD:
		IPSEC4_DEBUG("shunt SA of HOLD: skb stored in HOLD.\n");
		write_lock_bh(&policy->lock);
		if (policy->hold.last != NULL) {
			kfree_skb(policy->hold.last);
			stats->tx_dropped++;
		}
		policy->hold.last = skb;
		write_unlock_bh(&policy->lock);
		skb = NULL;
		goto cleanup;

	case IPSEC_POLICY_TRAP:
	case IPSEC_POLICY_TRAPSUBNET:
		if (ipsec_handle_trap(policy, skb) < 0)
			stats->tx_dropped++;
		goto cleanup;

	case IPSEC_POLICY_DROP:
		IPSEC4_DEBUG("shunt SA of DROP or no eroute: dropping.\n");
		goto dropit;
	
	case IPSEC_POLICY_REJECT:
		IPSEC4_DEBUG("shunt SA of REJECT: notifying and dropping.\n");
		icmp_send(skb,
			  ICMP_DEST_UNREACH,
			  ICMP_PKT_FILTERED,
			  0);
		goto dropit;

	case IPSEC_POLICY_BYPASS:
		pass = 1;
		IPSEC4_DEBUG("PASS: calling dev_queue_xmit\n");
		goto bypass;

	default:
		break;
	}

	IPSEC4_DEBUG("Original head,tailroom: %d,%d\n",
		    skb_headroom(skb), skb_tailroom(skb));

	if (policy == NULL) {
		if (net_ratelimit())
			printk(KERN_NOTICE "NULL policy in ipsec4_tunnel\n");
		goto dropit;
	}

	/* Extract the list of SAs associated with this policy, sorted
	 * in inner to outer */
	error = ipsec4_policy_to_salist(policy, sa_index, MAX_CHAINED_SAS);
	if (error < 0)
		goto dropit;
	sa_count = error;

	/* Make sure the SAs we have are valid */
	for (j = 0; j < sa_count; j++) {
		if (!ipsec_sa_valid(sa_index[j]))
			goto dropit;
	}

	/* Compute head and tail room required */
	headroom = tailroom = 0;
	for (j = 0; j < sa_count; j++) {
		if (ipsec4_headroom(sa_index[j], &headroom, &tailroom) < 0)
			goto dropit;
	}
	IPSEC4_DEBUG("Required head,tailroom: %d,%d\n",
		    headroom, tailroom);

	if (ipsec4_tunnel_update_mtu(prv, headroom + tailroom) < 0)
		goto dropit;

	/* Make sure the packet doesn't get too big */
	if (skb->len + headroom + tailroom > 0xfff0) {
		if (net_ratelimit())
			printk(KERN_WARNING
				"ipsec4_tunnel_start_xmit: packet "
				"got too big (%d, max 65520)\n",
				skb->len + headroom + tailroom);
		goto dropit;
	}

	/* 
	 * If the sender is doing PMTU discovery, and the
	 * packet doesn't fit within prv->mtu, notify him
	 * (unless it was an ICMP packet, or it was not the
	 * zero-offset packet) and send it anyways.
	 *
	 * Note: buggy firewall configuration may prevent the
	 * ICMP packet from getting back.
	 */
	if(sysctl_ipsec4_icmp
	   && prv->mtu < ntohs(iph->tot_len)
	   && (iph->frag_off & __constant_htons(IP_DF)) ) {
		if (iph->protocol != IPPROTO_ICMP
		 && (iph->frag_off & __constant_htons(IP_OFFSET)) == 0)
			icmp_send(skb,
				  ICMP_DEST_UNREACH,
				  ICMP_FRAG_NEEDED,
				  htonl(prv->mtu));

		/* FreeSWAN quietly passes on the packet, ignoring the
		 * DF bit and having the fragmenter have a go at it.
		 * We do so too. */
	}

#ifdef MSS_HACK
	/*
	 * If this is a transport mode TCP packet with
	 * SYN set, determine an effective MSS based on 
	 * AH/ESP overheads determined above.
	 */
	if (iph->protocol == IPPROTO_TCP 
	 && sa_index[0]->ipsec_proto != IPPROTO_IPIP)
		ipsec_adjust_mss(skb, prv->mtu);
#endif /* MSS_HACK */

	/*
	 * Strip off hard_header if present.
	 *
	 * FIXME? Since we're moving around the ip header anyway, we might as
	 * well move both IP and hard header in a single memmove rather
	 * than wasting stack space. --okir
	 */
	hard_header_len = (unsigned char *) iph - skb->data;
	if (hard_header_len < 0) {
		if (net_ratelimit())
			printk(KERN_NOTICE "ipsec4_tunnel: negative hard_header_len\n");
		goto dropit;
	} else
	if (hard_header_len > sizeof(saved_hard_header)) {
		if (net_ratelimit())
			printk(KERN_NOTICE "ipsec4_tunnel: oversized hard_header_len\n");
		goto dropit;
	} else if (hard_header_len) {
		memcpy(saved_hard_header, skb->data, hard_header_len);
		skb_pull(skb, hard_header_len);
	}
	headroom += (hard_header_len + 15) & ~15;

	/*
	 * Copy data over to new buffer if necessary
	 */
	if (skb_headroom(skb) < headroom
	 || skb_tailroom(skb) < tailroom) {
		old_skb = skb;
		skb = skb_copy_expand(old_skb, headroom, tailroom, GFP_ATOMIC);
		if (!skb) {
			if (net_ratelimit())
				printk(KERN_NOTICE
				       "ipsec4_tunnel_start_xmit: "
				       "failed to allocate skbuff "
				       "(headroom=%d, tailroom=%d)\n",
				       headroom, tailroom);
			goto dropit;
		}
		if (skb && old_skb->sk)
			skb_set_owner_w(skb, old_skb->sk);
	}

	/*
	 * Apply all transforms to packet
	 */
	for (j = 0; j < sa_count; j++) {
		unsigned char *base;
		sa = sa_index[j];

#ifdef CONFIG_IPSEC_DEBUG
		{
			char sabuf[64];
			IPSEC4_DEBUG("Applying transform %s\n",
				ipsec4_print_sa(sa, sabuf, sizeof(sabuf)));
		}
#endif

		read_lock_bh(&sa->lock);

		headroom = tailroom = 0;
		if (ipsec4_headroom(sa, &headroom, &tailroom) < 0)
			goto dropit_and_unlock_sa;

		if (skb_headroom(skb) < headroom
		 || skb_tailroom(skb) < tailroom) {
			if (net_ratelimit())
				printk(KERN_WARNING
					"ipsec4_tunnel_start_xmit: not enough "
					"head/tailroom in packet "
					"(got %d,%d; need %d,%d)\n",
					skb_headroom(skb),
					skb_tailroom(skb),
					headroom, 
					tailroom);
			goto dropit_and_unlock_sa;
		}

		/* Move IP header to make room for transform header */
		base = (unsigned char *) skb_push(skb, headroom);
		/* skb_put(skb, tailroom); xxx */

		switch (sa->ipsec_proto) {
		case SADB_SATYPE_AH:
			memmove(base, base + headroom, iph->ihl << 2);
			skb->nh.iph = iph = (struct iphdr *) base;
			base += (iph->ihl << 2);
			error = ipsec4_build_ah(skb, iph,
					(struct ip_auth_hdr *) base,
					headroom,
					sa);
			break;

		case SADB_SATYPE_ESP:
			memmove(base, base + headroom, iph->ihl << 2);
			skb->nh.iph = iph = (struct iphdr *) base;
			base += (iph->ihl << 2);
			error = ipsec4_build_esp(skb, iph,
					(struct ip_esp_hdr *) base,
					sa);
			break;

		case SADB_X_SATYPE_IPIP:
			skb->nh.iph = iph = (struct iphdr *) base;
			base += sizeof(*iph);
			error = ipsec4_build_ipip(skb, iph,
					(struct iphdr *) base,
					sa);
			break;

		case SADB_X_SATYPE_COMP:
			memmove(base, base + headroom, iph->ihl << 2);
			skb->nh.iph = iph = (struct iphdr *) base;
			base += (iph->ihl << 2);
			error = ipsec4_build_comp(skb, iph, base, sa);
			break;

		default:
			error = -EINVAL;
		}
		read_unlock_bh(&sa->lock);

		if (error < 0)
			goto dropit;

		/* Update use time and counters */
		ipsec_sa_update_lifetime(sa, skb->len);

		/* Update IP total length and checksum */
		iph->tot_len = htons(skb->len);
		iph->check = 0;
		iph->check = ip_fast_csum((unsigned char *)iph, iph->ihl);

		IPSEC4_DEBUG("After applying transform:\n");
		IPSEC4_PRINT_IP(skb->nh.iph);
	}

	/* Old FreeSWAN would do another policy lookup here to see
	 * whether it needed to apply other policies, and continue
	 * if there was. We don't support that at the moment. */

	/* Tack the saved hard header back on */
	if (hard_header_len) {
		if (skb_headroom(skb) < hard_header_len) {
			if (net_ratelimit())
				printk(KERN_NOTICE
					"not enough headroom for hard header "
					"(need %d, have %d)\n",
					hard_header_len, skb_headroom(skb));
			goto dropit;
		}

		skb_push(skb, hard_header_len);
		memcpy(skb->data, saved_hard_header, hard_header_len);
	}

bypass:
	IPSEC4_DEBUG("With hard_header, final head,tailroom: %d,%d\n",
		    skb_headroom(skb), skb_tailroom(skb));
	IPSEC4_PRINT_IP(skb->nh.iph);

	/* new route/dst cache code from James Morris */
	skb->dev = physdev;
	error = ip_route_output(&rt,
			    skb->nh.iph->daddr,
			    pass ? 0 : skb->nh.iph->saddr,
			    RT_TOS(skb->nh.iph->tos),
			    physdev->iflink /* rgb: should this be 0? */);

	if (error) {
		u32 saddr = pass? 0 : skb->nh.iph->saddr;
		const char *dev_name;

		dev_name = (rt && rt->u.dst.dev)?
				rt->u.dst.dev->name : "<none>";
		IPSEC4_DEBUG("ip_route_output(%u.%u.%u.%u -> %u.%u.%u.%u) "
			     "failed: error=%d, rt->u.dst.dev=%s, dropped\n",
			     NIPQUAD(skb->nh.iph->daddr), NIPQUAD(saddr),
			     -error, dev_name);
		goto dropit;
	}

	if (dev == rt->u.dst.dev) {
		ip_rt_put(rt);
		/* This is recursion, drop it. */
		IPSEC4_DEBUG("suspect recursion, dev=rt->u.dst.dev=%s, dropped\n", dev->name);
		goto dropit;
	}

	dst_release(skb->dst);
	skb->dst = &rt->u.dst;
	stats->tx_bytes += skb->len;

	if (skb->len < skb->nh.raw - skb->data) {
		printk(KERN_WARNING
		       "tried to __skb_pull nh-data=%ld, %d available.  This should never happen, please report.\n",
		       skb->nh.raw - skb->data, skb->len);
		goto dropit;
	}
	__skb_pull(skb, skb->nh.raw - skb->data);

#ifdef CONFIG_NETFILTER
	nf_conntrack_put(skb->nfct);
	skb->nfct = NULL;
#endif
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif /* CONFIG_NETFILTER_DEBUG */

	IPSEC4_DEBUG("...done, calling ip_send() on device:%s\n",
		    skb->dev ? skb->dev->name : "NULL");
	IPSEC4_PRINT_IP(skb->nh.iph);

	{
		int err;

		err = NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, rt->u.dst.dev,
			      ipsec4_tunnel_xmit2);

		skb = NULL; /* the skb is gone after this point */
		if(err != NET_XMIT_SUCCESS && err != NET_XMIT_CN) {
			if(net_ratelimit())
				printk(KERN_ERR
				       "ip_send() failed, err=%d\n", 
				       -err);
			stats->tx_aborted_errors++;
			goto dropit;
		}
	}

	stats->tx_packets++;

 cleanup:
	/* Release references to any SAs we have */
	while (sa_count)
		ipsec_sa_put(sa_index[--sa_count]);

#if defined(HAS_NETIF_QUEUE) || defined (HAVE_NETIF_QUEUE)
	netif_wake_queue(dev);
#else /* defined(HAS_NETIF_QUEUE) || defined (HAVE_NETIF_QUEUE) */
	dev->tbusy = 0;
#endif /* defined(HAS_NETIF_QUEUE) || defined (HAVE_NETIF_QUEUE) */
	if (saved_header)
		kfree(saved_header);
	if(skb)
		kfree_skb(skb);
	if (old_skb)
		kfree_skb(old_skb);
	if (policy)
		ipsec_sp_put(policy);
	return 0;

dropit_and_unlock_sa:
	read_unlock_bh(&sa->lock);
dropit:
	stats->tx_dropped++;
	goto cleanup;
}

static struct net_device_stats *
ipsec4_tunnel_get_stats(struct net_device *dev)
{
	return &(((struct ipsecpriv *)(dev->priv))->mystats);
}

/*
 * Revectored calls.
 * For each of these calls, a field exists in our private structure.
 */

static int
ipsec4_tunnel_hard_header(struct sk_buff *skb, struct net_device *dev,
	unsigned short type, void *daddr, void *saddr, unsigned len)
{
	struct ipsecpriv *prv = dev->priv;
	struct net_device *tmp;
	int ret;
	struct net_device_stats *stats;	/* This device's statistics */
	
	if(skb == NULL) {
		IPSEC4_DEBUG("no skb...\n");
		return -ENODATA;
	}

	if(dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return -ENODEV;
	}

	IPSEC4_DEBUG("skb->dev=%s dev=%s.\n",
		    skb->dev ? skb->dev->name : "NULL",
		    dev->name);
	
	if(prv == NULL) {
		IPSEC4_ERROR("no private space associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		return -ENODEV;
	}

	stats = (struct net_device_stats *) &(prv->mystats);

	if(prv->dev == NULL) {
		IPSEC4_ERROR("no physical device associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		stats->tx_dropped++;
		return -ENODEV;
	}

	/* check if we have to send a IPv6 packet. It might be a Router
	   Solicitation, where the building of the packet happens in
	   reverse order:
	   1. ll hdr,
	   2. IPv6 hdr,
	   3. ICMPv6 hdr
	   -> skb->nh.raw is still uninitialized when this function is
	   called!!  If this is no IPv6 packet, we can print debugging
	   messages, otherwise we skip all debugging messages and just
	   build the ll header */
	if(type != ETH_P_IPV6) {
		/* execute this only, if we don't have to build the
		   header for a IPv6 packet */
		if(!prv->hard_header) {
			IPSEC4_DEBUG("physical device has been detached, "
				     "packet dropped 0x%p->0x%p len=%d "
				     "type=%d dev=%s->NULL "
				     "%u.%u.%u.%u -> %u.%u.%u.%u\n",
				    saddr,
				    daddr,
				    len,
				    type,
				    dev->name,
				    NIPQUAD(skb->nh.iph->saddr),
				    NIPQUAD(skb->nh.iph->daddr));
			stats->tx_dropped++;
			return -ENODEV;
		}
		
#define da ((struct net_device *)(prv->dev))->dev_addr
		IPSEC4_DEBUG("Revectored 0x%p->0x%p len=%d type=%d dev=%s->%s "
			     "dev_addr=%02x:%02x:%02x:%02x:%02x:%02x "
			     "ip %u.%u.%u.%u -> %u.%u.%u.%u\n",
			    saddr,
			    daddr,
			    len,
			    type,
			    dev->name,
			    prv->dev->name,
			    da[0], da[1], da[2], da[3], da[4], da[5],
			    NIPQUAD(skb->nh.iph->saddr),
			    NIPQUAD(skb->nh.iph->daddr));
	} else {
		IPSEC4_DEBUG("is IPv6 packet, skip debugging messages, only "
			     "revector and build linklocal header.\n");
	}                                                                       
	tmp = skb->dev;
	skb->dev = prv->dev;
	ret = prv->hard_header(skb, prv->dev, type, (void *)daddr, (void *)saddr, len);
	skb->dev = tmp;
	return ret;
}

static int
ipsec4_tunnel_rebuild_header(struct sk_buff *skb)
{
	struct ipsecpriv *prv = skb->dev->priv;
	struct net_device *tmp;
	int ret;
	struct net_device_stats *stats;	/* This device's statistics */
	
	if(skb->dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return -ENODEV;
	}

	if(prv == NULL) {
		IPSEC4_DEBUG("no private space associated with dev=%s\n",
			    skb->dev->name ? skb->dev->name : "NULL");
		return -ENODEV;
	}

	stats = (struct net_device_stats *) &(prv->mystats);

	if(prv->dev == NULL) {
		IPSEC4_DEBUG("no physical device associated with dev=%s\n",
			    skb->dev->name ? skb->dev->name : "NULL");
		stats->tx_dropped++;
		return -ENODEV;
	}

	if(!prv->rebuild_header) {
		IPSEC4_DEBUG("physical device has been detached, packet dropped "
			     "skb->dev=%s->NULL ip=%u.%u.%u.%u->%u.%u.%u.%u\n",
			    skb->dev->name,
			    NIPQUAD(skb->nh.iph->saddr),
			    NIPQUAD(skb->nh.iph->daddr));
		stats->tx_dropped++;
		return -ENODEV;
	}

	IPSEC4_DEBUG("Revectored rebuild_header dev=%s->%s "
		     "ip=%u.%u.%u.%u->%u.%u.%u.%u\n",
		    skb->dev->name, prv->dev->name,
		    NIPQUAD(skb->nh.iph->saddr),
		    NIPQUAD(skb->nh.iph->daddr));
	tmp = skb->dev;
	skb->dev = prv->dev;
	
	ret = prv->rebuild_header(skb);
	skb->dev = tmp;
	return ret;
}

static int
ipsec4_tunnel_set_mac_address(struct net_device *dev, void *addr)
{
	struct ipsecpriv *prv = dev->priv;
	
	struct net_device_stats *stats;	/* This device's statistics */
	
	if(dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return -ENODEV;
	}

	if(prv == NULL) {
		IPSEC4_DEBUG("no private space associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		return -ENODEV;
	}

	stats = (struct net_device_stats *) &(prv->mystats);

	if(prv->dev == NULL) {
		IPSEC4_DEBUG("no physical device associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		stats->tx_dropped++;
		return -ENODEV;
	}

	if(!prv->set_mac_address) {
		IPSEC4_DEBUG("physical device has been detached, cannot set - skb->dev=%s->NULL\n",
			    dev->name);
		return -ENODEV;
	}

	IPSEC4_DEBUG("Revectored dev=%s->%s addr=%p\n",
		    dev->name, prv->dev->name, addr);
	return prv->set_mac_address(prv->dev, addr);

}

static void
ipsec4_tunnel_cache_update(struct hh_cache *hh, struct net_device *dev, unsigned char *  haddr)
{
	struct ipsecpriv *prv = dev->priv;
	
	struct net_device_stats *stats;	/* This device's statistics */
	
	if(dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return;
	}

	if(prv == NULL) {
		IPSEC4_DEBUG("no private space associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		return;
	}

	stats = (struct net_device_stats *) &(prv->mystats);

	if(prv->dev == NULL) {
		IPSEC4_DEBUG("no physical device associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		stats->tx_dropped++;
		return;
	}

	if(!prv->header_cache_update) {
		IPSEC4_DEBUG("physical device has been detached, cannot set - skb->dev=%s->NULL\n",
			    dev->name);
		return;
	}

	IPSEC4_DEBUG("Revectored cache_update\n");
	prv->header_cache_update(hh, prv->dev, haddr);
	return;
}

static int
ipsec4_tunnel_neigh_setup(struct neighbour *n)
{
	IPSEC4_DEBUG("called\n");
        if (n->nud_state == NUD_NONE) {
                n->ops = &arp_broken_ops;
                n->output = n->ops->output;
        }
        return 0;
}

static int
ipsec4_tunnel_neigh_setup_dev(struct net_device *dev, struct neigh_parms *p)
{
	IPSEC4_DEBUG("setting up %s\n",
		    dev ? dev->name : "NULL");

        if (p->tbl->family == AF_INET) {
                p->neigh_setup = ipsec4_tunnel_neigh_setup;
                p->ucast_probes = 0;
                p->mcast_probes = 0;
        }
        return 0;
}

/*
 * We call the attach routine to attach another device.
 */

static int
ipsec4_tunnel_attach(struct net_device *dev, struct net_device *physdev)
{
        int i;
	struct ipsecpriv *prv = dev->priv;

	if(dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return -ENODEV;
	}

	if(prv == NULL) {
		IPSEC4_DEBUG("no private space associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		return -ENODATA;
	}

	prv->dev = physdev;
	prv->hard_start_xmit = physdev->hard_start_xmit;
	prv->get_stats = physdev->get_stats;

	if (physdev->hard_header) {
		prv->hard_header = physdev->hard_header;
		dev->hard_header = ipsec4_tunnel_hard_header;
	} else
		dev->hard_header = NULL;
	
	if (physdev->rebuild_header) {
		prv->rebuild_header = physdev->rebuild_header;
		dev->rebuild_header = ipsec4_tunnel_rebuild_header;
	} else
		dev->rebuild_header = NULL;
	
	if (physdev->set_mac_address) {
		prv->set_mac_address = physdev->set_mac_address;
		dev->set_mac_address = ipsec4_tunnel_set_mac_address;
	} else
		dev->set_mac_address = NULL;
	
	if (physdev->header_cache_update) {
		prv->header_cache_update = physdev->header_cache_update;
		dev->header_cache_update = ipsec4_tunnel_cache_update;
	} else
		dev->header_cache_update = NULL;

	dev->hard_header_len = physdev->hard_header_len;

/*	prv->neigh_setup        = physdev->neigh_setup; */
	dev->neigh_setup        = ipsec4_tunnel_neigh_setup_dev;
	dev->mtu = 16260; /* 0xfff0; */ /* dev->mtu; */
	prv->mtu = physdev->mtu;

	dev->type = physdev->type /* ARPHRD_TUNNEL */;	/* initially */

	dev->addr_len = physdev->addr_len;
	for (i=0; i<dev->addr_len; i++) {
		dev->dev_addr[i] = physdev->dev_addr[i];
	}
#ifdef CONFIG_IPSEC_DEBUG
	if (sysctl_ipsec_debug_ipv4) {
		IPSEC4_DEBUG("physical device %s being attached has HW address: ",
		       physdev->name);
		for (i=0; i < physdev->addr_len; i++) {
			printk("%c%02x", i? ':' : ' ', physdev->dev_addr[i]);
		}
		printk("\n");
	}
#endif /* CONFIG_IPSEC_DEBUG */

	return 0;
}

/*
 * We call the detach routine to detach the ipsec tunnel from another device.
 */

static int
ipsec4_tunnel_detach(struct net_device *dev)
{
        int i;
	struct ipsecpriv *prv = dev->priv;

	if(dev == NULL) {
		IPSEC4_DEBUG("no device...\n");
		return -ENODEV;
	}

	if(prv == NULL) {
		IPSEC4_DEBUG("no private space associated with dev=%s\n",
			    dev->name ? dev->name : "NULL");
		return -ENODATA;
	}

	IPSEC4_DEBUG("physical device %s being detached from virtual device %s\n",
		    prv->dev ? prv->dev->name : "NULL",
		    dev->name);

	prv->dev = NULL;
	prv->hard_start_xmit = NULL;
	prv->get_stats = NULL;

	prv->hard_header = NULL;
#ifdef DETACH_AND_DOWN
	dev->hard_header = NULL;
#endif /* DETACH_AND_DOWN */
	
	prv->rebuild_header = NULL;
#ifdef DETACH_AND_DOWN
	dev->rebuild_header = NULL;
#endif /* DETACH_AND_DOWN */
	
	prv->set_mac_address = NULL;
#ifdef DETACH_AND_DOWN
	dev->set_mac_address = NULL;
#endif /* DETACH_AND_DOWN */
	
	prv->header_cache_update = NULL;
#ifdef DETACH_AND_DOWN
	dev->header_cache_update = NULL;
#endif /* DETACH_AND_DOWN */

/*	prv->neigh_setup        = NULL; */
#ifdef DETACH_AND_DOWN
	dev->neigh_setup        = NULL;
#endif /* DETACH_AND_DOWN */
	dev->hard_header_len = 0;
#ifdef DETACH_AND_DOWN
	dev->mtu = 0;
#endif /* DETACH_AND_DOWN */
	prv->mtu = 0;
	for (i=0; i<MAX_ADDR_LEN; i++) {
		dev->dev_addr[i] = 0;
	}
	dev->addr_len = 0;
	dev->type = ARPHRD_TUNNEL;

	return 0;
}

/*
 * We call the clear routine to detach all ipsec tunnels from other devices.
 */
static int
ipsec4_tunnel_clear(void)
{
	int i;
	struct net_device *ipsecdev = NULL, *prvdev;
	struct ipsecpriv *prv;
	char name[9];
	int ret;

	IPSEC4_DEBUG("called\n");

	for(i = 0; i < IPSEC_NUM_IF; i++) {
		sprintf(name, "ipsec%d", i);
		if((ipsecdev = __dev_get_by_name(name)) != NULL) {
			if((prv = (struct ipsecpriv *)(ipsecdev->priv))) {
				prvdev = (struct net_device *)(prv->dev);
				if(prvdev) {
					IPSEC4_DEBUG("physical device for device %s is %s\n",
						    name, prvdev->name);
					if((ret = ipsec4_tunnel_detach(ipsecdev))) {
						IPSEC4_DEBUG("error %d detatching device %s from device %s.\n",
							    ret, name, prvdev->name);
						return ret;
					}
				}
			}
		}
	}
	return 0;
}

static int
ipsec4_tunnel_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	struct ipsectunnelconf *cf = (struct ipsectunnelconf *)&ifr->ifr_data;
	struct ipsecpriv *prv = dev->priv;
	struct net_device *them; /* physical device */
	char *colon;
	char realphysname[IFNAMSIZ];
	
	if(dev == NULL) {
		IPSEC4_DEBUG("device not supplied.\n");
		return -ENODEV;
	}

	IPSEC4_DEBUG("tncfg service call #%d for dev=%s\n",
		    cmd,
		    dev->name ? dev->name : "NULL");

	switch (cmd) {
	/* attach a virtual ipsec? device to a physical device */
	case IPSEC_SET_DEV:
		IPSEC4_DEBUG("calling ipsec4_tunnel_attatch...\n");
		/* If this is an IP alias interface, get its real physical name */
		strncpy(realphysname, cf->cf_name, IFNAMSIZ);
		realphysname[IFNAMSIZ-1] = 0;
		colon = strchr(realphysname, ':');
		if (colon) *colon = 0;
		them = __dev_get_by_name(realphysname);

		if (them == NULL) {
			IPSEC4_DEBUG("physical device %s requested is null\n",
				    cf->cf_name);
			return -ENXIO;
		}
		
#if 0
		if (them->flags & IFF_UP) {
			IPSEC4_DEBUG("physical device %s requested is not up.\n",
				    cf->cf_name);
			return -ENXIO;
		}
#endif
		
		if (prv && prv->dev) {
			IPSEC4_DEBUG("virtual device is already connected to %s.\n",
				    prv->dev->name ? prv->dev->name : "NULL");
			return -EBUSY;
		}
		return ipsec4_tunnel_attach(dev, them);

	case IPSEC_DEL_DEV:
		IPSEC4_DEBUG("calling ipsec4_tunnel_detatch.\n");
		if (! prv->dev) {
			IPSEC4_DEBUG("physical device not connected.\n");
			return -ENODEV;
		}
		return ipsec4_tunnel_detach(dev);
	       
	case IPSEC_CLR_DEV:
		IPSEC4_DEBUG("calling ipsec4_tunnel_clear.\n");
		return ipsec4_tunnel_clear();

	default:
		IPSEC4_DEBUG("unknown command %d.\n", cmd);
		return -EOPNOTSUPP;
	}
}

static int
ipsec_device_event(struct notifier_block *unused, unsigned long event, void *ptr)
{
	struct net_device *dev = ptr;
	struct net_device *ipsec_dev;

	if (dev == NULL) {
		IPSEC4_DEBUG("dev=NULL for event type %ld.\n", event);
		return(NOTIFY_DONE);
	}

	/* check for loopback devices */
	if (dev && (dev->flags & IFF_LOOPBACK)) {
		return(NOTIFY_DONE);
	}

	switch (event) {
	case NETDEV_DOWN:
		/* look very carefully at the scope of these compiler
		   directives before changing anything... -- RGB */
	case NETDEV_UNREGISTER:
		switch (event) {
		case NETDEV_DOWN:
			IPSEC4_DEBUG("NETDEV_DOWN dev=%s flags=%x\n",
				    dev->name,
				    dev->flags);
			if(strncmp(dev->name, "ipsec", strlen("ipsec")) == 0) {
				printk(KERN_CRIT "IPSEC EVENT: KLIPS device %s shut down.\n",
				       dev->name);
			}
			break;
		case NETDEV_UNREGISTER:
			IPSEC4_DEBUG("NETDEV_UNREGISTER dev=%s flags=%x\n",
				    dev->name,
				    dev->flags);
			break;
		}
		
		/* find the attached physical device and detach it. */
		ipsec_dev = ipsec4_tunnel_from_device(dev);
		if (ipsec_dev && ipsec_dev != dev) {
			ipsec4_tunnel_detach(ipsec_dev);
			IPSEC4_DEBUG("device '%s' has been detached.\n",
				    ipsec_dev->name);
		}
		break;
	case NETDEV_UP:
		IPSEC4_DEBUG("NETDEV_UP dev=%s\n",
			    dev->name);
		break;
	case NETDEV_REBOOT:
		IPSEC4_DEBUG("NETDEV_REBOOT dev=%s\n",
			    dev->name);
		break;
	case NETDEV_CHANGE:
		IPSEC4_DEBUG("NETDEV_CHANGE dev=%s flags=%x\n",
			    dev->name,
			    dev->flags);
		break;
	case NETDEV_REGISTER:
		IPSEC4_DEBUG("NETDEV_REGISTER dev=%s\n",
			    dev->name);
		break;
	case NETDEV_CHANGEMTU:
		IPSEC4_DEBUG("NETDEV_CHANGEMTU dev=%s to mtu=%d\n",
			    dev->name,
			    dev->mtu);
		break;
	case NETDEV_CHANGEADDR:
		IPSEC4_DEBUG("NETDEV_CHANGEADDR dev=%s\n",
			    dev->name);
		break;
	case NETDEV_GOING_DOWN:
		IPSEC4_DEBUG("NETDEV_GOING_DOWN dev=%s\n",
			    dev->name);
		break;
	case NETDEV_CHANGENAME:
		IPSEC4_DEBUG("NETDEV_CHANGENAME dev=%s\n",
			    dev->name);
		break;
	default:
		IPSEC4_DEBUG("event type %ld unrecognised for dev=%s\n",
			    event,
			    dev->name);
		break;
	}
	return NOTIFY_DONE;
}

/*
 *	Called when an ipsec tunnel device is initialized.
 *	The ipsec tunnel device structure is passed to us.
 */
 
static int
ipsec4_tunnel_init_device(struct net_device *dev)
{
	/* Add our tunnel functions to the device */
	dev->open		= ipsec4_tunnel_open;
	dev->stop		= ipsec4_tunnel_close;
	dev->hard_start_xmit	= ipsec4_tunnel_start_xmit;
	dev->get_stats		= ipsec4_tunnel_get_stats;

	dev->priv = kmalloc(sizeof(struct ipsecpriv), GFP_KERNEL);
	if (dev->priv == NULL)
		return -ENOMEM;
	memset(dev->priv, 0, sizeof(struct ipsecpriv));

	dev->set_multicast_list = NULL;
	dev->do_ioctl		= ipsec4_tunnel_ioctl;
	dev->hard_header	= NULL;
	dev->rebuild_header 	= NULL;
	dev->set_mac_address 	= NULL;
	dev->header_cache_update= NULL;

/*	prv->neigh_setup        = NULL; */
	dev->neigh_setup        = ipsec4_tunnel_neigh_setup_dev;
	dev->hard_header_len 	= 0;
	dev->mtu		= 0;
	dev->addr_len		= 0;
	dev->type		= ARPHRD_TUNNEL; /* 0 */ /* ARPHRD_ETHER; */ /* initially */
	dev->tx_queue_len	= 10;		/* Small queue */
	memset(dev->broadcast,0xFF, ETH_ALEN);	/* what if this is not attached to ethernet? */

	/* New-style flags. */
	dev->flags		= IFF_NOARP /* 0 */ /* Petr Novak */;
	dev_init_buffers(dev);

	/* We're done.  Have I forgotten anything? */
	return 0;
}

/*
 * Get the tunnel device attached to this physical device
 */
struct net_device *
ipsec4_tunnel_from_device(struct net_device *phys_dev)
{
	struct net_device *ipsec_dev;
	struct ipsecpriv *priv;
	char name[IFNAMSIZ];
	int i;

	for(i = 0; i < IPSEC_NUM_IF; i++) {
		sprintf(name, "ipsec%d", i);
		ipsec_dev = __dev_get_by_name(name);
		if (ipsec_dev) {
			if (ipsec_dev == phys_dev)
				return ipsec_dev;
			priv = (struct ipsecpriv *)(ipsec_dev->priv);
			if (priv && (struct net_device *) priv->dev == phys_dev)
				return ipsec_dev;
		}
	}

	return NULL;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*  Module specific interface (but it links with the rest of IPSEC  */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int
ipsec4_tunnel_probe(struct net_device *dev)
{
	ipsec4_tunnel_init_device(dev); 
	return 0;
}

struct net_device dev_ipsec3 = 
{
	"ipsec3\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec4_tunnel_probe	/* setup */
};

struct net_device dev_ipsec2 = 
{
	"ipsec2\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec4_tunnel_probe	/* setup */
};

struct net_device dev_ipsec1 = 
{
	"ipsec1\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec4_tunnel_probe	/* setup */
};

struct net_device dev_ipsec0 = 
{
	"ipsec0\0   ",		/* name */
	0,			/* recv memory end */
	0,			/* recv memory start */
	0,			/* memory end */
	0,			/* memory start */
 	0x0,			/* base I/O address */
	0,			/* IRQ */
	0, 0, 0,		/* flags */
	NULL,			/* next device */
	ipsec4_tunnel_probe	/* setup */
};

static struct notifier_block ipsec_dev_notifier={
	ipsec_device_event,
	NULL,
	0
};

int 
ipsec4_tunnel_init(void)
{
	int err;

	err = register_netdevice_notifier(&ipsec_dev_notifier);
	if (err < 0)
		return err;

	if (register_netdev(&dev_ipsec0) != 0)
		return -EIO;
	if (register_netdev(&dev_ipsec1) != 0)
		return -EIO;
	if (register_netdev(&dev_ipsec2) != 0)
		return -EIO;
	if (register_netdev(&dev_ipsec3) != 0)
		return -EIO;
	return 0;
}

void
ipsec4_tunnel_cleanup(void)
{
	unregister_netdevice_notifier(&ipsec_dev_notifier);
	unregister_netdev(&dev_ipsec0);
	unregister_netdev(&dev_ipsec1);
	unregister_netdev(&dev_ipsec2);
	unregister_netdev(&dev_ipsec3);
	kfree(dev_ipsec0.priv);
	kfree(dev_ipsec1.priv);
	kfree(dev_ipsec2.priv);
	kfree(dev_ipsec3.priv);
	dev_ipsec0.priv=NULL;
	dev_ipsec1.priv=NULL;
	dev_ipsec2.priv=NULL;
	dev_ipsec3.priv=NULL;
}
