/*
 * Debugging functions - dump a PFKEY message
 *
 * Copyright (C) 2003 SuSE Linux AG
 *
 * Written by Olaf Kirch <okir@suse.de>
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/major.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/socket.h>
#include <linux/un.h>
#include <linux/fcntl.h>
#include <linux/termios.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/net.h> /* struct socket */
#include <linux/in.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <net/sock.h> /* struct sock */
#include <net/af_unix.h>
#include <linux/spinlock.h>

#include <linux/types.h>
 
#include <asm/uaccess.h>
#include <linux/in6.h>

#include <linux/pfkeyv2.h>
#include <linux/pfkey.h>
#include <net/spd.h>

#include <linux/ipsec.h>

#include "pfkey_v2_msg.h"
#include "sockaddr_utils.h"

#ifdef CONFIG_IPSEC_DEBUG

static const char *	pfkeyv2_msg_type[SADB_MAX+1] = {
	"RESERVED",
	"GETSPI",
	"UPDATE",
	"ADD",
	"DELETE",
	"GET",
	"ACQUIRE",
	"REGISTER",
	"EXPIRE",
	"FLUSH",
	"DUMP",
	"X_PROMISC",
	"X_PCHANGE",
	"X_GRPSA",
	"X_ADDFLOW",
	"X_DELFLOW",
	"X_DEBUG",
	"X_FLUSH_SP",
};

static const char *pfkeyv2_sa_type[SADB_SATYPE_MAX+1] = {
	"UNSPEC",
	"SATYPE-1",
	"AH",
	"ESP",
	"SATYPE-4",
	"RSVP",
	"OSPFV2",
	"RIPV2",
	"MIP",
	"IPIP",
	"COMP",
	"INT"
};

static void	pfkeyv2_print_ext_sa(struct sadb_ext *);
static void	pfkeyv2_print_ext_satype(struct sadb_ext *);
static void	pfkeyv2_print_ext_lifetime(struct sadb_ext *);
static void	pfkeyv2_print_ext_address(struct sadb_ext *);
static void	pfkeyv2_print_ext_key(struct sadb_ext *);

static struct pfkeyv2_ext_info {
	const char *	name;
	void		(*print)(struct sadb_ext *);
} pfkeyv2_ext[SADB_EXT_MAX+1] = {
{ "RESERVED",		NULL },
{ "SA",			pfkeyv2_print_ext_sa },
{ "LIFETIME_CURRENT",	pfkeyv2_print_ext_lifetime },
{ "LIFETIME_HARD",	pfkeyv2_print_ext_lifetime },
{ "LIFETIME_SOFT",	pfkeyv2_print_ext_lifetime },
{ "ADDRESS_SRC",	pfkeyv2_print_ext_address },
{ "ADDRESS_DST",	pfkeyv2_print_ext_address },
{ "ADDRESS_PROXY",	pfkeyv2_print_ext_address },
{ "KEY_AUTH",		pfkeyv2_print_ext_key },
{ "KEY_ENCRYPT",	pfkeyv2_print_ext_key },
{ "IDENTITY_SRC",	NULL },
{ "IDENTITY_DST",	NULL },
{ "SENSITIVITY",	NULL },
{ "PROPOSAL",		NULL },
{ "SUPPORTED_AUTH",	NULL },
{ "SUPPORTED_ENCRYPT",	NULL },
{ "SPIRANGE",		NULL },
{ "KMPRIVATE",		NULL },
{ "SATYPE2",		pfkeyv2_print_ext_satype },
{ "SA2",		pfkeyv2_print_ext_sa },
{ "ADDRESS_DST2",	pfkeyv2_print_ext_address },
{ "ADDRESS_SRC_FLOW",	pfkeyv2_print_ext_address },
{ "ADDRESS_DST_FLOW",	pfkeyv2_print_ext_address },
{ "ADDRESS_SRC_MASK",	pfkeyv2_print_ext_address },
{ "ADDRESS_DST_MASK",	pfkeyv2_print_ext_address },
{ "DEBUG",		NULL },
};

void
pfkeyv2_print_msg(const char *verb, struct sadb_msg *msg)
{
	if (!sysctl_ipsec_debug_pfkey)
		return;
	printk(KERN_DEBUG
		"%s pfkey msg v=%d type=%d(%s) satype=%d/%s",
		verb,
		msg->sadb_msg_version,
		msg->sadb_msg_type,
		msg->sadb_msg_type <= SADB_MAX
			?  pfkeyv2_msg_type[msg->sadb_msg_type]
			: "???",
		msg->sadb_msg_satype,
		msg->sadb_msg_satype <= SADB_SATYPE_MAX
			? pfkeyv2_sa_type[msg->sadb_msg_satype]
			: "???");
	if (sysctl_ipsec_debug_pfkey > 1)
		printk(" seq=%d pid=%d len=%d",
			msg->sadb_msg_seq,
			msg->sadb_msg_pid,
			msg->sadb_msg_len);
	printk("\n");
}

void
pfkeyv2_print_extensions(struct sadb_ext **ext)
{
	int i;

	if (!sysctl_ipsec_debug_pfkey)
		return;
	for (i = 0; i < SADB_EXT_MAX + 1; i++) {
		const char *ext_name;

		if (!ext[i])
			continue;
		ext_name = pfkeyv2_ext[i].name;

		printk(KERN_DEBUG
		       "  extension %d/%s present, len=%d",
		       i, ext_name? ext_name : "<unknown>",
		       ext[i]->sadb_ext_len);
		if (pfkeyv2_ext[i].print)
			pfkeyv2_ext[i].print(ext[i]);
		printk("\n");
	}
}

void
pfkeyv2_print_ext_sa(struct sadb_ext *ext)
{
	struct sadb_sa *sa = (struct sadb_sa *) ext;
	const char *state_name = NULL, *sep = "";
	int flags;

	switch (sa->sadb_sa_state) {
	case SADB_SASTATE_LARVAL: state_name = "larval"; break;
	case SADB_SASTATE_MATURE: state_name = "mature"; break;
	case SADB_SASTATE_DYING: state_name = "dying"; break;
	case SADB_SASTATE_DEAD: state_name = "dead"; break;
	default: state_name = "unknown";
	}

	printk(": spi=0x%x replay=%d state=%d/%s "
			  "auth=%d encrypt=%d flags=",
		ntohl(sa->sadb_sa_spi),
		sa->sadb_sa_replay,
		sa->sadb_sa_state,
		state_name,
		sa->sadb_sa_auth,
		sa->sadb_sa_encrypt);

	flags = sa->sadb_sa_flags;
	if (flags & SADB_SAFLAGS_PFS) {
		printk("%spfs", sep);
		flags &= ~SADB_SAFLAGS_PFS;
		sep = ",";
	}
	if (flags & SADB_X_SAFLAGS_REPLACEFLOW) {
		printk("%sreplaceflow", sep);
		flags &= ~SADB_X_SAFLAGS_REPLACEFLOW;
		sep = ",";
	}
	if (flags & SADB_X_SAFLAGS_CLEARFLOW) {
		printk("%sclearflow", sep);
		flags &= ~SADB_X_SAFLAGS_CLEARFLOW;
		sep = ",";
	}
	if (flags & SADB_X_SAFLAGS_TUNNEL) {
		printk("%stunnel", sep);
		flags &= ~SADB_X_SAFLAGS_TUNNEL;
		sep = ",";
	}
	if (flags || !*sep)
		printk("%s0x%x", sep, flags);
}

void
pfkeyv2_print_ext_satype(struct sadb_ext *ext)
{
	struct sadb_x_satype *sa = (struct sadb_x_satype *) ext;

	printk(": type=%d/%s",
		sa->sadb_x_satype_satype,
		sa->sadb_x_satype_satype <= SADB_SATYPE_MAX
			? pfkeyv2_sa_type[sa->sadb_x_satype_satype]
			: "???");
}

void
pfkeyv2_print_ext_lifetime(struct sadb_ext *ext)
{
	struct sadb_lifetime *lf = (struct sadb_lifetime *) ext;

	printk(": allocations=%d bytes=%Lu addtime=%Lu usetime=%Lu",
			lf->sadb_lifetime_allocations,
			(unsigned long long) lf->sadb_lifetime_bytes,
			(unsigned long long) lf->sadb_lifetime_addtime,
			(unsigned long long) lf->sadb_lifetime_usetime);
}

void
pfkeyv2_print_ext_address(struct sadb_ext *ext)
{
	struct sadb_address *addr = (struct sadb_address *) ext;
	struct sockaddr_storage ss;

	if (sadb_address_to_sockaddr(addr, (struct sockaddr *) &ss)) {
		printk(": <cannot parse address>");
	} else {
		char buffer[64];

		sockaddrtoa((struct sockaddr *) &ss, buffer, sizeof(buffer));
		printk(": addr=%s", buffer);
		if (addr->sadb_address_prefixlen)
			printk("/%u", addr->sadb_address_prefixlen);
	}
}

void
pfkeyv2_print_ext_key(struct sadb_ext *ext)
{
	struct sadb_key *key = (struct sadb_key *) key;

	printk(": len=%d, keybits=%u",
		key->sadb_key_len,
		key->sadb_key_bits);
}

#endif /* CONFIG_IPSEC_DEBUG */
