/* $USAGI: pfkey_v2_msg_flow.c,v 1.25 2002/12/02 11:09:53 miyazawa Exp $ */
/*
 * Copyright (C)2001 USAGI/WIDE Project
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Mitsuru KANDA <mk@linux-ipv6.org> / USAGI
 *   Kazunori MIYAZAWA <miyazawa@linux-ip.org> / USAGI
 */
/*
 * This is a parse routine for messages SADB_EXT_ADDFLOW and SADB_EXT_DELFLOW.
 * We use FreeS/WAN's user land routine for manipulating SA and Policy.
 * These are FreeS/WAN specific.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/random.h>
#include <linux/ipsec.h>
#include <linux/pfkeyv2.h>
#include <linux/pfkey.h>
#include <net/sadb.h>
#include <net/spd.h>

#include "pfkey_v2_msg.h"
#include "sockaddr_utils.h"

#define BUFSIZE 64

static int sadb_mask_to_prefixlen(const struct sadb_address *ext_msg)
{
	int rtn = 0;

	struct sockaddr* tmp_addr = NULL;

	if (!ext_msg) {
		PFKEY_DEBUG("ext_msg is NULL\n");
		return -EINVAL;
	}

	tmp_addr = (struct sockaddr*)((char*)ext_msg + sizeof(struct sadb_address));

	switch (tmp_addr->sa_family) {

	case AF_INET:
	{
		int i;
		__u32 tmp;
		tmp = ntohl(((struct sockaddr_in*)tmp_addr)->sin_addr.s_addr);
		for (i=0; i<32; i++) {
			if ((tmp>>i)&0x01)
				break;
		}
		rtn = 32 - i;
	}
	break;

	case AF_INET6:
	{
		int i,j;
		__u16 tmp;
		for (i=0; i<8; i++ ) {
			tmp = ntohs(((struct sockaddr_in6*)tmp_addr)->sin6_addr.s6_addr16[7-i]);
			for (j=0; j<16; j++) {
				if ((tmp>>j)&0x01) {
					rtn = 128 - i*16 - j;
					PFKEY_DEBUG("result i=%d, j=%d\n", i, j);
					goto err;
				}
			}
		}
	}
	break;
	default:
		PFKEY_DEBUG("address family is unknown\n");
		rtn = -EINVAL;
		goto err;
	}

err:
#ifdef CONFIG_IPSEC_DEBUG
	if (rtn)
		PFKEY_DEBUG("prefixlen=%d\n", rtn);
#endif
	return rtn;
}

static int
update_policy(struct ipsec_sp *policy, struct sa_index *sa_idx, int flags)
{
	struct sa_index **dest = NULL;

	switch (ntohl(sa_idx->spi)) {
	case IPSEC_SPI_DROP:
		policy->policy_action = IPSEC_POLICY_DROP;
		break;
	case IPSEC_SPI_BYPASS:
		policy->policy_action = IPSEC_POLICY_BYPASS;
		break;
	case IPSEC_SPI_REJECT:
		policy->policy_action = IPSEC_POLICY_REJECT;
		break;
	case IPSEC_SPI_HOLD:
		policy->policy_action = IPSEC_POLICY_HOLD;
		break;
	case IPSEC_SPI_TRAP:
		policy->policy_action = IPSEC_POLICY_TRAP;
		break;
	case IPSEC_SPI_TRAPSUBNET:
		policy->policy_action = IPSEC_POLICY_TRAPSUBNET;
		break;
	default:
		policy->policy_action = IPSEC_POLICY_APPLY;
	}

	if (policy->policy_action != IPSEC_POLICY_APPLY) {
		if (policy->hold.last)
			kfree_skb(policy->hold.last);
		policy->hold.last = NULL;
		sa_index_kfree(policy->auth_sa_idx);
		policy->auth_sa_idx = NULL;
		sa_index_kfree(policy->esp_sa_idx);
		policy->esp_sa_idx = NULL;
		sa_index_kfree(policy->comp_sa_idx);
		policy->comp_sa_idx = NULL;
		sa_index_kfree(policy->ipip_sa_idx);
		policy->ipip_sa_idx = NULL;
		return 0;
	}

	switch(sa_idx->ipsec_proto) {
	case SADB_SATYPE_AH:
		dest = &policy->auth_sa_idx;
		break;
	case SADB_SATYPE_ESP:
		dest = &policy->esp_sa_idx;
		break;
	case SADB_X_SATYPE_COMP:
		dest = &policy->comp_sa_idx;
		break;
	case SADB_X_SATYPE_IPIP:
		dest = &policy->ipip_sa_idx;
		break;
	default:
		return -EINVAL;
	}

	if (*dest) {
		if (!(flags & SADB_X_SAFLAGS_REPLACEFLOW))
			return -EEXIST;
		sa_index_kfree(*dest);
	}

	if (!(*dest = sa_index_kmalloc()))
		return -ENOMEM;

	return sa_index_copy(*dest, sa_idx);
}

int sadb_msg_addflow_parse(struct sock *sk, struct sadb_msg* msg, struct sadb_msg **reply)
{
	int error = 0;
	int tmp = 0;
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1], *reply_ext_msgs[SADB_EXT_MAX+1];
	struct sadb_sa *sa = NULL;
	struct sadb_address *dst = NULL;
	struct sadb_address *sflow = NULL;
	struct sadb_address *dflow = NULL;
	struct ipsec_sp *policy = NULL;
	struct sa_index sa_idx;
	struct selector selector;


	if (!msg) {
		PFKEY_DEBUG("msg is NULL\n");
		error = -EINVAL;
		goto err;
	}

	memset(ext_msgs, 0, sizeof(ext_msgs));
	memset(reply_ext_msgs, 0, sizeof(reply_ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	memset(&selector, 0, sizeof(selector));
	if (ext_msgs[SADB_EXT_SA] &&
		ext_msgs[SADB_EXT_ADDRESS_DST] &&
		ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] &&
		ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW])
	{
		sa  = (struct sadb_sa*)ext_msgs[SADB_EXT_SA];
		dst = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_DST];
		sflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
		dflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];

		selector.mode = sa->sadb_sa_flags & SADB_X_SAFLAGS_TUNNEL
				? IPSEC_MODE_TUNNEL
				: IPSEC_MODE_TRANSPORT;

		if (selector.mode == IPSEC_MODE_TRANSPORT) {
			if (sflow->sadb_address_proto == dflow->sadb_address_proto) {
				selector.proto = sflow->sadb_address_proto;
			} else {
				error = -EINVAL;
				goto err;
			}
		}

		sa_index_init(&sa_idx);

		sa_idx.ipsec_proto = msg->sadb_msg_satype;
		sa_idx.spi = sa->sadb_sa_spi;

		sadb_address_to_sockaddr(dst, (struct sockaddr*)&sa_idx.dst);
		sa_idx.prefixlen_d = dst->sadb_address_prefixlen;

		sadb_address_to_sockaddr(sflow, (struct sockaddr *) &selector.src);
		selector.prefixlen_s = sflow->sadb_address_prefixlen;
		sadb_address_to_sockaddr(dflow, (struct sockaddr *) &selector.dst);
		selector.prefixlen_d = dflow->sadb_address_prefixlen;

		if (ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK]) {
			tmp = sadb_mask_to_prefixlen((struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK]);
			if (tmp<0) {
				error = tmp;
				goto err;
			} else {
				selector.prefixlen_s = tmp;
			}
		}

		if (ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK]) {
			tmp = sadb_mask_to_prefixlen((struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK]);
			if (tmp<0) {
				error = tmp;
				goto err;
			} else {
				selector.prefixlen_d = tmp;
			}
		}

		/* This one creates the policy if it doesn't exist. */
		error = spd_find_by_selector(&selector, &policy, 1);

		if (error == -ESRCH || error == -EEXIST) {
			int flags = 0, is_new = 0;

			if (error == -ESRCH)
				is_new = 1;
			else
				flags = sa->sadb_sa_flags;

			write_lock_bh(&policy->lock);
			error = update_policy(policy, &sa_idx, flags);
			/* spd_find_by_selector may have matched a policy
			 * with a different mode; update it */
			policy->selector.mode = selector.mode;
			write_unlock_bh(&policy->lock);

			if (error < 0) {
				PFKEY_DEBUG("Failed to initialize policy, error=%d\n",
					-error);
				if (is_new)
					spd_remove(&selector);
			}
			ipsec_sp_put(policy);
		} else {
			/* Something is wrong */
			PFKEY_DEBUG("spd_find_by_selector failed, error=%d\n",
				-error);
			goto err;
		}

	}
	
err:
	reply_ext_msgs[0] = (struct sadb_ext*) msg;
	reply_ext_msgs[SADB_EXT_SA] = ext_msgs[SADB_EXT_SA];
	reply_ext_msgs[SADB_EXT_ADDRESS_SRC] = ext_msgs[SADB_EXT_ADDRESS_SRC];
	reply_ext_msgs[SADB_EXT_ADDRESS_DST] = ext_msgs[SADB_EXT_ADDRESS_DST];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] = ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW] = ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK] = ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK] = ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK];
	error = pfkey_msg_build(reply, reply_ext_msgs, EXT_BITS_OUT);

	return error;

}

int sadb_msg_delflow_parse(struct sock *sk, struct sadb_msg* msg, struct sadb_msg **reply)
{
	int error = 0;
	int tmp = 0;
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1], *reply_ext_msgs[SADB_EXT_MAX+1];
	struct sadb_sa *sa = NULL;
	struct sadb_address *sflow = NULL;
	struct sadb_address *dflow = NULL;
	struct selector selector;


	if (!msg) {
		PFKEY_DEBUG("msg is NULL\n");
		error = -EINVAL;
		goto err;
	}

	memset(ext_msgs, 0, sizeof(ext_msgs));
	memset(reply_ext_msgs, 0, sizeof(reply_ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	memset(&selector, 0, sizeof(selector));

	if (ext_msgs[SADB_EXT_SA] &&
		ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] &&
		ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW])
	{
		sa  = (struct sadb_sa*)ext_msgs[SADB_EXT_SA];

		sflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
		dflow = (struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];
#ifdef CONFIG_IPSEC_TUNNEL
		selector.mode = sa->sadb_sa_flags & SADB_X_SAFLAGS_TUNNEL
				? IPSEC_MODE_TUNNEL
				: IPSEC_MODE_TRANSPORT;

		if (selector.mode == IPSEC_MODE_TRANSPORT) {
#endif
			if (sflow->sadb_address_proto == dflow->sadb_address_proto) {
				selector.proto = sflow->sadb_address_proto;
			} else {
				error = -EINVAL;
				goto err;
			}	
#ifdef CONFIG_IPSEC_TUNNEL
		}
#endif
		sadb_address_to_sockaddr( sflow, (struct sockaddr*)&selector.src);
		sadb_address_to_sockaddr( dflow, (struct sockaddr*)&selector.dst);
		selector.prefixlen_s = sflow->sadb_address_prefixlen;
		selector.prefixlen_d = dflow->sadb_address_prefixlen;

		if (ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK]) {
			tmp = sadb_mask_to_prefixlen(
				(struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK]);
			if (tmp<0) {
				error = tmp;
				goto err;
			} else {
				selector.prefixlen_s = tmp;
			}
		}

		if (ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK]) {
			tmp = sadb_mask_to_prefixlen(
				(struct sadb_address*)ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK]);
			if (tmp<0) {
				error = tmp;
				goto err;
			} else {
				selector.prefixlen_d = tmp;
			}
		}

		error = spd_remove(&selector);

		if (error) {
			PFKEY_DEBUG("spd_remove failed\n");
			goto err;
		}

	}

	reply_ext_msgs[0] = (struct sadb_ext*) msg;
	reply_ext_msgs[SADB_EXT_SA] = ext_msgs[SADB_EXT_SA];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW] = ext_msgs[SADB_X_EXT_ADDRESS_SRC_FLOW];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW] = ext_msgs[SADB_X_EXT_ADDRESS_DST_FLOW];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK] = ext_msgs[SADB_X_EXT_ADDRESS_SRC_MASK];
	reply_ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK] = ext_msgs[SADB_X_EXT_ADDRESS_DST_MASK];
	error = pfkey_msg_build(reply, reply_ext_msgs, EXT_BITS_OUT);

err:
	return error;

}



