/* $USAGI: sockaddr_utils.c,v 1.6 2002/10/04 11:31:28 mk Exp $ */
/*
 * Copyright (C)2001 USAGI/WIDE Project
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Kazunori MIYAZAWA <miyazawa@linux-ip.org>
 *   Mitsuru KANDA <mk@linux-ipv6.org>
 */
/*
 * sadb_utils.c include utility functions for SADB handling.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <net/ipv6.h>
#include <linux/inet.h>
#include <linux/ipsec.h>
#include <linux/pfkeyv2.h>
#include <net/sadb.h>

#define BUFSIZE 64

int
sockaddrtoa(struct sockaddr *addr, char *buf, size_t buflen)
{
	int ret = 0;

	if (addr == NULL) {
		snprintf(buf, buflen, "<null>");
		return 0;
	}

	switch (addr->sa_family) {
	case AF_UNSPEC:
		snprintf(buf, buflen, "UNSPEC");
		break;
	case AF_INET:
		snprintf(buf, buflen, "%d.%d.%d.%d", NIPQUAD((((struct sockaddr_in *)addr)->sin_addr)));
		break;
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	case AF_INET6:
		in6_ntop(&(((struct sockaddr_in6*)addr)->sin6_addr), buf, buflen);
		break;
#endif		
	default:
		snprintf(buf, buflen, "<addr family %d>", addr->sa_family);
		ret = -EAFNOSUPPORT;
		break;
	}

	return ret;
}

int
sockporttoa(struct sockaddr *addr, char *buf, size_t buflen)
{
	int ret = 0;

	switch (addr->sa_family) {
	case AF_UNSPEC:
		snprintf(buf, buflen, "UNSPEC");
	case AF_INET:
		snprintf(buf, buflen, "%hd", ntohs(((struct sockaddr_in *)addr)->sin_port));
		break;
	case AF_INET6:
		snprintf(buf, buflen, "%hd", ntohs(((struct sockaddr_in6*)addr)->sin6_port));
		break;
	default:
		snprintf(buf, buflen, "<addr family %d>", addr->sa_family);
		ret = -EAFNOSUPPORT;
		break;
	}

	return ret;
}

/*
 * addr1, prefixlen1 : packet(must set 128 or 32 befor call this) 
 * addr2, prefixlen2 : sa/sp
 */
int
sockaddr_prefix_compare(struct sockaddr *addr1, __u8 prefixlen1,
			struct sockaddr *addr2, __u8 prefixlen2)
{
	__u8 prefixlen;

	if (!addr1 || !addr2) {
		SADB_DEBUG("addr1 or add2 is NULL\n");
		return -EINVAL;
	}

	/* AF_UNSPEC addresses always match */
	if (addr1->sa_family == AF_UNSPEC
	 || addr2->sa_family == AF_UNSPEC)
	 	return 0;

	if (addr1->sa_family != addr2->sa_family) {
		SADB_DEBUG("sa_family not match\n");
		return 1;
	}

	if (prefixlen1 < prefixlen2) 
		prefixlen = prefixlen1;
	else
		prefixlen = prefixlen2;
	SADB_DEBUG("prefixlen: %d, prefixlen1: %d, prefixlen2: %d\n", prefixlen, prefixlen1, prefixlen2);

	switch (addr1->sa_family) {
	case AF_INET:
		if (prefixlen > 32 )
			return 1;
		/* The compiler doesn't really do what you expect if
		 * you ask it to shift by 32bits */
		if (prefixlen == 0)
			return 0;
		return (((struct sockaddr_in *)addr1)->sin_addr.s_addr ^
			  ((struct sockaddr_in *)addr2)->sin_addr.s_addr) &
			 htonl((0xffffffff << (32 - prefixlen)));
	case AF_INET6:
		if (prefixlen > 128)
			return 1;

		return ipv6_prefix_cmp(&((struct sockaddr_in6 *)addr1)->sin6_addr,
				       &((struct sockaddr_in6 *)addr2)->sin6_addr,
				       prefixlen);
	default:
		SADB_DEBUG("unknown sa_family\n");
		return 1;
	}
}

int 
sockaddr_compare_ports(struct sockaddr *addr1, struct sockaddr *addr2)
{
	struct sockaddr_in *sin1, *sin2;
	struct sockaddr_in6 *six1, *six2;

	/* AF_UNSPEC addresses always match */
	if (addr1->sa_family == AF_UNSPEC
	 || addr2->sa_family == AF_UNSPEC)
	 	return 0;

	if (addr1->sa_family != addr2->sa_family)
		return -EINVAL;

	switch (addr1->sa_family) {
	case AF_INET:
		sin1 = (struct sockaddr_in *) addr1;
		sin2 = (struct sockaddr_in *) addr2;
		if (sin1->sin_port && sin2->sin_port)
			return !(sin1->sin_port == sin1->sin_port);
		break;
	case AF_INET6:
		six1 = (struct sockaddr_in6 *) addr1;
		six2 = (struct sockaddr_in6 *) addr2;
		if (six1->sin6_port && six2->sin6_port)
			return !(six1->sin6_port == six2->sin6_port);
		break;
	default:
		SADB_DEBUG("unsupported address family: %d\n", 
			  addr1->sa_family);
		return -EINVAL;
	}

	return 0; /* should never reach here */
}

